/* FILE:        CppViewSimpleLin.cpp
 * DATE:        9/29/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Very simple demonstration of the ImageView library, using the
 *              BitFlow Ci library for GNU Linux.
 */


//// Include files.

// BitFlow includes for acquisition.
#include <BFciLib.h>

// BitFlow ImageView library includes.
#include <ImageView/ImageViewCApi.h>

// Resource files.
#include "CView64_xpm.h"

// C library includes.
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

// Threading support.
#include <pthread.h>


//// Minimal implementation of the ImageView acquisition interface methods.

// Global variables.
typedef struct UserDataStruct
{
    IVApp app;
    IVWindow window;
    IVRawImage rawImage;

    tCIp sCIp;

    IVbool acqStarted;
    IVbool acqThreadQuit;

    pthread_t acqThread;
} UserData;

// Methods.
IVbool acqCleanup (IVUserData ud)
{
    UserData *const userData = (UserData*)ud;

    if (userData->sCIp)
    {
        CiAqSWreset(userData->sCIp);
        CiUnmapFrameBuffers(userData->sCIp);
        CiVFGclose(userData->sCIp);

        userData->sCIp = IV_NULL;
    }

    userData->acqThreadQuit = IV_TRUE;
    if (userData->acqThread)
    {
        pthread_join(userData->acqThread, IV_NULL);
        userData->acqThread = IV_NULL;
    }

    return IV_TRUE;
}

IVbool acqStart (IVUserData ud)
{
    UserData *const userData = (UserData*)ud;

    if (userData->sCIp)
    {
        if (CiAqStart(userData->sCIp, 0) == kCIEnoErr)
        {
            userData->acqStarted = IV_TRUE;
            return IV_TRUE;
        }
    }
    return IV_FALSE;
}

IVbool acqStop (IVUserData ud)
{
    UserData *const userData = (UserData*)ud;

    if (userData->sCIp)
        return CiAqStop(userData->sCIp) == kCIEnoErr;
    return IV_FALSE;
}

IVbool acqCanStart (IVUserData ud)
{
    UserData *const userData = (UserData*)ud;

    return !userData->acqThreadQuit && !userData->acqStarted;
}

IVbool acqCanStop (IVUserData ud)
{
    UserData *const userData = (UserData*)ud;

    return !userData->acqThreadQuit && userData->acqStarted;
}

// Returns IV_TRUE/IV_FALSE, whether or not the interface is setup for acquisition.
IVbool acqIsSetup (IVUserData ud)
{
    UserData *const userData = (UserData*)ud;

    return !userData->acqThreadQuit;
}

// Returns IV_TRUE/IV_FALSE, whether or not acquisition is running.
IVbool acqIsRunning (IVUserData ud)
{
    UserData *const userData = (UserData*)ud;

    return !userData->acqThreadQuit && userData->acqStarted;
}

// Acquisition thread function.
void* acqThreadFunc (void *ud)
{
    UserData *const userData = (UserData*)ud;

    tCIRC circ;

    tCIU32 frameID;
    tCIU8 *frameP;

    tCIU32 state, frames, errors;

    // Loop until acquisition quit.
    while (!userData->acqThreadQuit)
    {
        // Get the next frame.
        circ = CiGetOldestNotDeliveredFrame(userData->sCIp, &frameID, &frameP);
        switch (circ)
        {
        case kCIEnoErr:
            // We have a frame.
            userData->rawImage.buffer = frameP;
            IVWindowSetNextFrame(userData->window, &userData->rawImage);
            break;

        case kCIEnoNewData:
            break;

        default:
            fprintf(stderr, "CiGetOldestNotDeliveredFrame failed with \"%s\".\n", CiErrStr(circ));
            userData->acqThreadQuit = IV_TRUE;
            IVSendSignal(userData->window, IV_WINDOW_ACQUISITION_STATE_CHANGE);
            return IV_NULL;
        }

        // Check the acquisition status.
        circ = CiAqGetStatus(userData->sCIp, &state, &frames, &errors);
        if (kCIEnoErr != circ)
        {
            fprintf(stderr, "CiAqGetStatus failed with \"%s\".\n", CiErrStr(circ));
            userData->acqThreadQuit = IV_TRUE;
            IVSendSignal(userData->window, IV_WINDOW_ACQUISITION_STATE_CHANGE);
            return IV_NULL;
        }

        switch (state & kCIAQSstateMask)
        {
        default:
            // For non-process valid states, clear the userData->acqStarted flag.
            if (userData->acqStarted)
            {
                userData->acqStarted = IV_FALSE;
                IVSendSignal(userData->window, IV_WINDOW_ACQUISITION_STATE_CHANGE);
            }

            // Fall through...
        case kCIAQSsnapShotInProcess:
        case kCIAQScontinuousInProcess:
            // Wait for another frame.
            circ = CiWaitNextUndeliveredFrame(userData->sCIp, 17);
            if (userData->acqThreadQuit)
                return IV_NULL;
            else if (kCIEnoErr != circ && kCIEtimeoutErr != circ)
            {
                fprintf(stderr, "CiWaitNextUndeliveredFrame failed with \"%s\".\n", CiErrStr(circ));
                userData->acqThreadQuit = IV_TRUE;
                IVSendSignal(userData->window, IV_WINDOW_ACQUISITION_STATE_CHANGE);
                return IV_NULL;
            }
            break;

        case kCIAQSinvalidState:
            fprintf(stderr, "CiAqGetStatus failed with \"%s\".\n", CiErrStr(circ));
            userData->acqThreadQuit = IV_TRUE;
            IVSendSignal(userData->window, IV_WINDOW_ACQUISITION_STATE_CHANGE);
            return IV_NULL;
        }
    }

    return IV_NULL;
}


//// Local function definitions

// The application main function.
int main (const int argc, const char *const argv[])
{
    UserData userData;

    tCIRC circ;
    tCIU32 nPtrs;
    tCIU8 **uPtrs=IV_NULL;
    tCIU32 nFrames, hROIoffset, vROIoffset, stride;

    tCIU32 width, height, bitDepth;

    tCIU32 brdCnt;
    tCIU32 brdNumber;

    IVRC rc;

    // Set userData to its initial null state.
    memset(&userData, IV_NULL, sizeof(userData));

    // Ask the user which board to open.
    if (kCIEnoErr != CiSysVFGenum(&brdCnt))
    {
        fprintf(stderr, "Unable to determine the system board count.\n");
        return EXIT_FAILURE;
    }

    if (!brdCnt)
    {
        fprintf(stderr, "No BitFlow boards installed on this system.\n");
        return EXIT_FAILURE;
    }

    do
    {
        printf("Which board to open (0-%u)? ", brdCnt - 1);
        if (scanf("%u", &brdNumber) != 1)
            brdNumber = brdCnt;
        while(getchar() != '\n');
    }
    while (brdNumber >= brdCnt);

    // Open the frame grabber with writer permission.
    circ = CiVFGopen(brdNumber, kCIBO_writeAccess, &userData.sCIp);
    if (kCIEnoErr != circ)
    {
        acqCleanup(&userData);
        fprintf(stderr, "CiVFGopen failed with \"%s\".\n", CiErrStr(circ));
        return EXIT_FAILURE;
    }

    // Initialize with the default config file.
    circ = CiVFGinitialize(userData.sCIp, kBFDEFAULTCFGFILE);
    if (kCIEnoErr != circ)
    {
        acqCleanup(&userData);
        fprintf(stderr, "CiVFGinitialize failed with \"%s\".\n", CiErrStr(circ));
        return EXIT_FAILURE;
    }

    // Configure the VFG for 10 frame buffers.
    circ = CiDrvrBuffConfigure(userData.sCIp, 10, 0, 0, 0, 0);
    if (kCIEnoErr != circ)
    {
        acqCleanup(&userData);
        fprintf(stderr, "CiDrvrBuffConfigure failed with \"%s\".\n", CiErrStr(circ));
        return EXIT_FAILURE;
    }

    // Determine the frame parameters.
    circ = CiBufferInterrogate(userData.sCIp, &nFrames, &bitDepth, &hROIoffset, &width, &vROIoffset, &height, &stride);
    if (kCIEnoErr != circ)
    {
        acqCleanup(&userData);
        fprintf(stderr, "CiBufferInterrogate failed with \"%s\".\n", CiErrStr(circ));
        return EXIT_FAILURE;
    }

    userData.rawImage.width = width;
    userData.rawImage.height = height;
    userData.rawImage.format = (IVFormat)bitDepth;
    userData.rawImage.userData = &userData;

    // Get the buffer pointers for read/write access to buffers.
    circ = CiMapFrameBuffers(userData.sCIp,1,&nPtrs,&uPtrs);
    if (kCIEnoErr != circ)
    {
        acqCleanup(&userData);
        fprintf(stderr, "CiMapFrameBuffers failed with \"%s\".\n", CiErrStr(circ));
        return EXIT_FAILURE;
    }

    // Reset acquisition and clear all error conditions.
    circ = CiAqSWreset(userData.sCIp);
    if (kCIEnoErr != circ)
    {
        acqCleanup(&userData);
        fprintf(stderr, "CiAqSWreset failed with \"%s\".\n", CiErrStr(circ));
        return EXIT_FAILURE;
    }

    // Start the acquisition thread.
    if (pthread_create(&userData.acqThread, IV_NULL, &acqThreadFunc, &userData))
    {
        acqCleanup(&userData);
        fprintf(stderr, "pthread_create failed.\n");
        return EXIT_FAILURE;
    }

    // Create the ImageView Application.
    rc = IVAppCreate(argc, argv, &userData.app);
    if (IV_OK != rc)
    {
        acqCleanup(&userData);
        fprintf(stderr, "Failed to create the ImageView Application.\n");
        return EXIT_FAILURE;
    }

    // Create an ImageView Window.
    rc = IVWindowCreate(userData.app, &userData.window, &userData);
    if (IV_OK != rc)
    {
        IVAppDestroy(userData.app);
        acqCleanup(&userData);

        fprintf(stderr, "Failed to create the ImageView Window.\n");
        return EXIT_FAILURE;
    }

    // Setup the ImageView Window features.
    IVSetFeatureBV(userData.window, IV_WINDOW_BV_START, &acqStart);
    IVSetFeatureBV(userData.window, IV_WINDOW_BV_STOP, &acqStop);

    IVSetFeatureBV(userData.window, IV_WINDOW_BV_CAN_START, &acqCanStart);
    IVSetFeatureBV(userData.window, IV_WINDOW_BV_CAN_STOP, &acqCanStop);

    IVSetFeatureBV(userData.window, IV_WINDOW_BV_IS_SETUP, &acqIsSetup);
    IVSetFeatureBV(userData.window, IV_WINDOW_BV_IS_RUNNING, &acqIsRunning);

    IVSendSignal(userData.window, IV_WINDOW_FEATURE_SET_CHANGE);

    // Setup the ImageView Window properties.
    IVSetPropData(userData.window, IV_WINDOW_APPLICATION_ABOUT_ICON_XPM, CView64_xpm, sizeof(CView64_xpm));
    IVSetPropCStr(userData.window, IV_WINDOW_APPLICATION_ABOUT_TITLE, "About CViewSimpleLin");
    IVSetPropCStr(userData.window, IV_WINDOW_APPLICATION_ABOUT_DESCRIPTION, "<html><h3>CViewSimpleLin</h3><p><i>Live preview of acquisition, for the BitFlow Linux SDK.</i></p></html>");

    // Buffers will remain allocated, and do not need to be explicitly locked.
    IVSetPropBool(userData.window, IV_WINDOW_ACQUISITION_FRAMES_LOCKED, IV_TRUE);

    // Show the window.
    IVSetPropBool(userData.window, IV_WINDOW_VISIBLE, IV_TRUE);

    // Asynchronously start acquisition.
    IVSendSignal(userData.window, IV_WINDOW_SET_RUNNING);

    // Execute the Application.
    rc = IVAppExec(userData.app, IV_NULL);

    // Cleanup after acquisition.
    acqCleanup(&userData);

    IVWindowDestroy(userData.window);
    IVAppDestroy(userData.app);

    return EXIT_SUCCESS;
}
