/* FILE:        CppViewSimpleLin.cpp
 * DATE:        9/29/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Very simple demonstration of the ImageView library, using the
 *              BitFlow Ci library for GNU Linux.
 */


//// Include files.

// BitFlow includes for acquisition.
#include <BFciLib.h>

// BitFlow ImageView library includes.
#include <ImageView/ImageViewApplication.h>
#include <ImageView/ImageViewWindow.h>
#include <ImageView/ImperativeAcquisitionInterface.h>

// Resource files.
#include "CppView64_xpm.h"

// C++ std library includes.
#include <cstddef>
#include <cstdint>

#include <iostream>
#include <sstream>

#include <string>

#include <atomic>
#include <condition_variable>
#include <thread>


//// Minimal implementation of the ImageView::ImperativeAcquisitionInterface.

class MinimalAcqIf final : public ImageView::ImperativeAcquisitionInterface
{
private:
    inline void throwInitError (std::string const& context, const tCIRC circ);

public:
    inline MinimalAcqIf (const unsigned int brdNumber, const unsigned int numBuffers);
    inline ~MinimalAcqIf (void) override;

    // Acquisition interface methods.
    inline bool start (void) override;
    inline bool stop (void) override;
    inline bool cleanup (void) override;

    inline bool canStart (void) const override;
    inline bool canStop (void) const override;

    inline bool isSetup (void) const override;
    inline bool isRunning (void) const override;

private:
    // Acquisition thread function.
    inline void acqThreadFunc (void);

    // Member variables.
    tCIp m_sCIp;

    tCIU32 m_width, m_height, m_bitDepth;

    bool m_acqStarted;

    bool m_acqThreadQuit;
    std::thread m_acqThread;

    // Illegal.
    MinimalAcqIf (void);
    MinimalAcqIf (MinimalAcqIf const&);
    MinimalAcqIf& operator= (MinimalAcqIf const&);
};

void MinimalAcqIf::throwInitError (std::string const& context, const tCIRC circ)
{
    cleanup();

    std::stringstream errStrm;
    errStrm << context << " failed with \"" << CiErrStr(circ) << "\".\n";
    throw std::runtime_error(errStrm.str());
}

// Construct an acquisition interface for the given board number.
MinimalAcqIf::MinimalAcqIf (const unsigned int brdNumber, const unsigned int numBuffers)
    : m_sCIp (NULL)
    , m_acqStarted (false)
    , m_acqThreadQuit (false)
{
    tCIRC circ;
    tCIU32 nPtrs;
    tCIU8 **uPtrs=NULL;
    tCIU32 nFrames, hROIoffset, vROIoffset, stride;

    // Open the frame grabber with writer permission.
    circ = CiVFGopen(brdNumber, kCIBO_writeAccess, &m_sCIp);
    if (kCIEnoErr != circ)
        throwInitError("CiVFGopen", circ);

    // Initialize with the default config file.
    circ = CiVFGinitialize(m_sCIp, kBFDEFAULTCFGFILE);
    if (kCIEnoErr != circ)
        throwInitError("CiVFGinitialize", circ);

    // Configure the VFG for 10 frame buffers.
    circ = CiDrvrBuffConfigure(m_sCIp, numBuffers, 0, 0, 0, 0);
    if (kCIEnoErr != circ)
        throwInitError("CiDrvrBuffConfigure", circ);

    // Determine the frame parameters.
    circ = CiBufferInterrogate(m_sCIp, &nFrames, &m_bitDepth, &hROIoffset, &m_width, &vROIoffset, &m_height, &stride);
    if (kCIEnoErr != circ)
        throwInitError("CiBufferInterrogate", circ);

    // Get the buffer pointers for read/write access to buffers.
    circ = CiMapFrameBuffers(m_sCIp,1,&nPtrs,&uPtrs);
    if (kCIEnoErr != circ)
        throwInitError("CiMapFrameBuffers", circ);

    // Reset acquisition and clear all error conditions.
    circ = CiAqSWreset(m_sCIp);
    if (kCIEnoErr != circ)
        throwInitError("CiAqSWreset", circ);

    // Enable the acquisition interface start and stop features.
    setHasStart(true);
    setHasStop(true);

    // Indicate that acquired frames will remain allocated throughout
    // the lifetime of the ImperativeAcquisitionInterface.
    setFramesLocked(true);

    // Start the acquisition thread.
    m_acqThread = std::thread(&MinimalAcqIf::acqThreadFunc, this);
}

// Destructor.
MinimalAcqIf::~MinimalAcqIf (void)
{
    // Abort acquisition, if still running.
    m_acqThreadQuit = true;

    // Join the acquisition thread.
    if (m_acqThread.joinable())
        m_acqThread.join();

    // Board cleanup.
    cleanup();
}

bool MinimalAcqIf::start (void)
{
    if (m_sCIp)
    {
        if (CiAqStart(m_sCIp, 0) == kCIEnoErr)
        {
            m_acqStarted = true;
            return true;
        }
    }
    return false;
}

bool MinimalAcqIf::stop (void)
{
    if (m_sCIp)
        return CiAqStop(m_sCIp) == kCIEnoErr;
    return false;
}

bool MinimalAcqIf::cleanup (void)
{
    if (m_sCIp)
    {
        CiAqSWreset(m_sCIp);
        CiUnmapFrameBuffers(m_sCIp);
        CiVFGclose(m_sCIp);

        m_sCIp = NULL;
    }

    return true;
}

bool MinimalAcqIf::canStart (void) const
{
    return !m_acqThreadQuit && !m_acqStarted;
}

bool MinimalAcqIf::canStop (void) const
{
    return isRunning();
}

// Returns true/false, whether or not the interface is setup for acquisition.
bool MinimalAcqIf::isSetup (void) const
{
    return !m_acqThreadQuit;
}

// Returns true/false, whether or not acquisition is running.
bool MinimalAcqIf::isRunning (void) const
{
    return !m_acqThreadQuit && m_acqStarted;
}

// Acquisition thread function.
void MinimalAcqIf::acqThreadFunc (void)
{
    tCIRC circ;

    // Loop until acquisition quit.
    while (!m_acqThreadQuit)
    {
        tCIU32 frameID;
        tCIU8 *frameP;

        tCIU32 state, frames, errors;

        // Get the next frame.
        circ = CiGetOldestNotDeliveredFrame(m_sCIp, &frameID, &frameP);
        switch (circ)
        {
        case kCIEnoErr:
            // We have a frame.
            setNextFrame(frameP, m_width, m_height, (ImageView::Format)m_bitDepth);
            break;

        case kCIEnoNewData:
            break;

        default:
            std::cerr << "CiGetOldestNotDeliveredFrame failed with \"" << CiErrStr(circ) << "\".\n";
            m_acqThreadQuit = true;
            signalAcquisitionStateChange();
            return;
        }

        // Check the acquisition status.
        circ = CiAqGetStatus(m_sCIp, &state, &frames, &errors);
        if (kCIEnoErr != circ)
        {
            std::cerr << "CiAqGetStatus failed with \"" << CiErrStr(circ) << "\".\n";
            m_acqThreadQuit = true;
            signalAcquisitionStateChange();
            return;
        }

        switch (state & kCIAQSstateMask)
        {
        default:
            // For non-process valid states, clear the acqStarted flag.
            if (m_acqStarted)
            {
                m_acqStarted = false;
                signalAcquisitionStateChange();
            }

            // Fall through...
        case kCIAQSsnapShotInProcess:
        case kCIAQScontinuousInProcess:
            // Wait for another frame.
            circ = CiWaitNextUndeliveredFrame(m_sCIp, 17);
            if (kCIEnoErr != circ && kCIEtimeoutErr != circ)
            {
                std::cerr << "CiWaitNextUndeliveredFrame failed with \"" << CiErrStr(circ) << "\".\n";
                m_acqThreadQuit = true;
                signalAcquisitionStateChange();
                return;
            }
            break;

        case kCIAQSinvalidState:
            std::cerr << "CiAqGetStatus state invalid (state == " << state << ").\n";
            m_acqThreadQuit = true;
            signalAcquisitionStateChange();
            return;
        }
    }
}


//// Local function definitions

// The application main function.
int main (int argc, char *argv[])
{
    // Ask the user which board to open.
    tCIU32 brdCnt;
    if (kCIEnoErr != CiSysVFGenum(&brdCnt))
    {
        std::cerr << "Unable to determine the system board count.\n";
        return EXIT_FAILURE;
    }

    if (!brdCnt)
    {
        std::cerr << "No BitFlow boards installed on this system.\n";
        return EXIT_FAILURE;
    }

    tCIU32 brdNumber = brdCnt;
    do
    {
        std::cout << "Which board to open (0-" << (brdCnt - 1) << ")? ";
        std::cin >> brdNumber;
    }
    while (brdNumber >= brdCnt);

    try
    {
        // Initialize the ImageView::Application.
        ImageView::Application app (argc, argv);

        // Attempt to open the MinimalAcqIf.
        auto acqIfPtr = ImageView::make_scoped(new MinimalAcqIf(brdNumber, 10));

        // Create the ImageView::Window.
        ImageView::Window window (std::move(acqIfPtr));

        window.setApplicationAbout("About CppViewSimpleLin", "<html><h3>CppViewSimpleLin</h3><p><i>Live preview of acquisition, for the BitFlow Linux SDK.</i></p></html>", CppView64_xpm, CppView64_xpm, CppView64_xpm);

        // Display the window.
        window.show();

        // Start acquisition asynchronously.
        window.setRunning();

        // Enter the application main loop.
        return app.exec();
    }
    catch (std::exception const& err)
    {
        std::cerr << "Caught an exception: " << err.what() << "\n";
    }
    catch (...)
    {
        std::cerr << "Caught an unknown exception.\n";
    }

    return EXIT_FAILURE;
}
