#ifndef INCLUDED__IMAGE__VIEW__ABSTRACT__ACQUISITION__INTERFACE__H
#define INCLUDED__IMAGE__VIEW__ABSTRACT__ACQUISITION__INTERFACE__H

/* FILE:        AbstractAcquisitionInterface.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Abstract interface for ImageView acquisition.
 */

#include "ImageView.h"
#include "ImageViewBaseClass.h"

#include "ImageViewRawImage.h"
#include "ImageViewAbstractLock.h"

namespace ImageView
{
    // The acquisition interface for the ImageView::Window class.
    class IMAGEVIEW_EXPORT AbstractAcquisitionInterface : public BaseClass
    {
    private:
        struct PrivateData;
        PrivateData &m_pd;

        // Illegal.
        AbstractAcquisitionInterface (AbstractAcquisitionInterface const&);
        AbstractAcquisitionInterface& operator= (AbstractAcquisitionInterface const&);

    protected:
        AbstractAcquisitionInterface (void);

        // Update the GUI for all features that the interface has.
        void signalFeatureSetChange (void);

        // Update the GUI to reflect the current acquisition state.
        void signalAcquisitionStateChange (void);

        // Initiate a call to getNextFrame as soon as possible.
        void signalGetFrameNow (void);

    public:
        virtual ~AbstractAcquisitionInterface (void);

        // Method executed when the user interface is about to close. Default does nothing.
        virtual void uxClosed (void);

        // Retrieve the AcquisitionSignalHandler.
        AcquisitionSignalHandler* signalHandler (void);
        const AcquisitionSignalHandler* signalHandler (void) const;

        //// Non-dynamic features. signalFeatureSetChange propagates to the Window.

        // Returns true/false, whether or not the interface supports acquisition setup.
        bool hasSetup (void) const;
        void setHasSetup (const bool enabled);

        // Returns true/false, whether or not the interface supports acquisition setup with a mode.
        bool hasSetupMode (void) const;
        void setHasSetupMode (const bool enabled);

        // Returns true/false, whether or not the interface supports acquisition cleanup.
        bool hasCleanup (void) const;
        void setHasCleanup (const bool enabled);

        // Returns true/false, whether or not the interface supports acquisition triggering.
        bool hasTrigger (void) const;
        void setHasTrigger (const bool enabled);

        // Returns true/false, whether or not the interface supports reverse acquisition start of a single frame.
        bool hasStartSingleReverse (void) const;
        void setHasStartSingleReverse (const bool enabled);

        // Returns true/false, whether or not the interface supports reverse acquisition start.
        bool hasStartReverse (void) const;
        void setHasStartReverse (const bool enabled);

        // Returns true/false, whether or not the interface supports acquisition start.
        bool hasStart (void) const;
        void setHasStart (const bool enabled);

        // Returns true/false, whether or not the interface supports acquisition of a fixed count of frames.
        bool hasStartFixed (void) const;
        void setHasStartFixed (const bool enabled);

        // Returns true/false, whether or not the interface supports grabbing a single frame.
        bool hasStartSingle (void) const;
        void setHasStartSingle (const bool enabled);

        // Returns true/false, whether or not the interface supports pausing an active acquisition.
        bool hasPause (void) const;
        void setHasPause (const bool enabled);

        // Returns true/false, whether or not the interface supports arbitrary acquisition stop.
        bool hasStop (void) const;
        void setHasStop (const bool enabled);

        // Returns true/false, whether or not the interface supports frame-rate querying.
        bool hasFrameRate (void) const;
        void setHasFrameRate (const bool enabled);

        // Returns true/false, whether or not the interface supports a queued frame count.
        bool hasFramesQueued (void) const;
        void setHasFramesQueued (const bool enabled);

        // Returns true/false, whether or not the interface supports a filled (waiting) frame count.
        bool hasFramesFilled (void) const;
        void setHasFramesFilled (const bool enabled);

        // Returns true/false, whether or not the interface supports a dropped frame count.
        bool hasFramesDropped (void) const;
        void setHasFramesDropped (const bool enabled);

        // Returns true/false, whether or not the interface supports a delivered frame count.
        bool hasFramesDelivered (void) const;
        void setHasFramesDelivered (const bool enabled);

        // The minimum startFixed count. Default is 1.
        std::int32_t startFixedMin (void) const;
        void setStartFixedMin (const std::int32_t val);

        // The maximum startFixed count. Default is 65536.
        std::int32_t startFixedMax (void) const;
        void setStartFixedMax (const std::int32_t val);

        // The startFixed count increment, from startFixedMin. Default is 1.
        std::int32_t startFixedInc (void) const;
        void setStartFixedInc (const std::int32_t val);

        //// Acquisition setup and control.

        // Setup the interface for acquisition. Base implementation simply returns false.
        virtual bool setup (void);

        // Setup the interface for acquisition using a specified mode. Base implementatino simply returns false.
        virtual bool setupMode (const char *const mode);

        // Cleanup the acquisition interface. Base implementation simply returns false.
        virtual bool cleanup (void);

        // Trigger the acquisition interface. Base implementation simply returns false.
        virtual bool trigger (void);

        // Start reverse acquisition of a single frame. Base implementation simply returns false.
        virtual bool startSingleReverse (void);

        // Start a reverse acquisition. Base implementation simply returns false.
        virtual bool startReverse (void);

        // Start acquisition. Base implementation simply returns false.
        virtual bool start (void);

        // Start a fixed-length acquisition. Base implementation simply returns false.
        virtual bool startFixed (const size_t cnt);

        // Start acquisition of a single frame. Base implementation simply returns false.
        virtual bool startSingle (void);

        // Pause an acquisition. Base implementation simply returns false.
        virtual bool pause (void);

        // Stop acquisition. Base implementation simply returns false.
        virtual bool stop (void);

        //// Dynamic GUI feature states.

        // Returns true/false, whether or not the interface may be setup for acquisition.
        virtual bool canSetup (void) const;

        // Returns true/false, whether or not the interface may be setup for acquisition with a mode.
        virtual bool canSetupMode (void) const;

        // Returns true/false, whether or not the interface may be cleaned up.
        virtual bool canCleanup (void) const;

        // Returns true/false, whether or not the interface may be triggered.
        virtual bool canTrigger (void) const;

        // Returns true/false, whether or not reverse acquisition of a single frame can be started.
        virtual bool canStartSingleReverse (void) const;

        // Returns true/false, whether or not a continuous reverse acquisition can be started.
        virtual bool canStartReverse (void) const;

        // Returns true/false, whether or not a continuous acquisition can be started.
        virtual bool canStart (void) const;

        // Returns true/false, whether or not a fixed-frame acquisition can be started.
        virtual bool canStartFixed (void) const;

        // Returns true/false, whether or not a single-frame acquisition can be started.
        virtual bool canStartSingle (void) const;

        // Returns true/false, whether or not acquisition may be paused.
        virtual bool canPause (void) const;

        // Returns true/false, whether or not acquisition may be stopped.
        virtual bool canStop (void) const;

        //// Acquisition and setup status.

        // Returns true/false, whether or not the interface is setup for acquisition.
        virtual bool isSetup (void) const = 0;

        // Returns true/false, whether or not reverse acquisition of a single frame is running. Default returns false.
        virtual bool isRunningSingleReverse (void) const;

        // Returns true/false, whether or not a reverse acquisition is running. Default returns false.
        virtual bool isRunningReverse (void) const;

        // Returns true/false, whether or not acquisition is running.
        virtual bool isRunning (void) const = 0;

        // Returns true/false, whether or not acquisition is grabbing a fixed-length sequence of frames.
        virtual bool isRunningFixed (void) const;

        // Returns true/false, whether or not acquisition is grabbing a single frame.
        virtual bool isRunningSingle (void) const;

        // Returns true/false, whether or not acquisition is running. Default returns false.
        virtual bool isPaused (void) const;

        // Returns true/false, whether or not acquisition is running. Default returns !isRunning().
        virtual bool isStopped (void) const;

        //// Setup mode.

        // Add a mode for acquisition setup, with optional description.
        bool addSetupMode (const char *const mode);
        bool addSetupMode (const char *const mode, const char *const description);

        // Clear all previously added setup modes.
        void clearSetupModes (void);

        // The number of added setup modes.
        size_t setupModeCount (void) const;

        // Return the setup mode name for the given index.
        bool setupModeName (const size_t index, char *const pModeStr, size_t *const pSize) const;

        // Return the setup mode description for the given index.
        bool setupModeDescription (const size_t index, char *const pDescStr, size_t *const pSize) const;

        // Return the name of the setup mode, if any. Default simply returns false.
        virtual bool activeSetupMode (char *const pModeStr, size_t *const pSize) const;

        //// Frame status query.

        // Return the acquisition frame rate. Base implementation simply returns 0.
        virtual double frameRate (void) const;

        // Return the frames queued count. Base implementations simply return 0.
        virtual size_t framesQueued (void) const;

        // Return the frames filled count. Base implementations simply return 0.
        virtual size_t framesFilled (void) const;

        // Return the frames dropped count. Base implementations simply return 0.
        virtual size_t framesDropped (void) const;

        // Return the frames delivered count. Base implementations simply return 0.
        virtual size_t framesDelivered (void) const;

        //// Frame retrieval and access.

        // Retrieve the next frame to display. ImageView::Window will only retrieve frames at
        // Window::refreshRate() frames-per-second, so it often makes sense to drop frames if
        // the are received at a higher rate than that. If false is returned, the ImageView::Window
        // will continue displaying whatever frame was last returned.
        virtual bool getNextFrame (RawImage *const rawImage, ScopedLock *const imageLock) = 0;

        // Retrieve a critical section to protect the allocation of buffers returned by getNextFrame.
        // Return a NULL ScopedLock, if allocation cannot be guaranteed.
        virtual ScopedLock getFrameLock (RawImage const& rawImage) = 0;
    };

    class IMAGEVIEW_EXPORT ScopedAcquisitionInterface
    {
    private:
        AbstractAcquisitionInterface *m_acqIf;

        // Illegal.
        ScopedAcquisitionInterface (int);
        ScopedAcquisitionInterface (std::nullptr_t);
        ScopedAcquisitionInterface (ScopedAcquisitionInterface const&);
        ScopedAcquisitionInterface& operator= (ScopedAcquisitionInterface const&);

    public:
        explicit ScopedAcquisitionInterface (void);
        explicit ScopedAcquisitionInterface (AbstractAcquisitionInterface *const acqIf);
        ScopedAcquisitionInterface (ScopedAcquisitionInterface &&to_take);

        ~ScopedAcquisitionInterface (void);

        operator bool (void) const;
        bool operator! (void) const;

        void take (ScopedAcquisitionInterface &&to_take);
        void reset (AbstractAcquisitionInterface *const acqIf = nullptr);

        AbstractAcquisitionInterface* get (void) const;
        AbstractAcquisitionInterface* operator-> (void) const;
        AbstractAcquisitionInterface& operator* (void) const;
    };

    inline static ScopedAcquisitionInterface make_scoped (AbstractAcquisitionInterface *const acqIf)
    {
        return ScopedAcquisitionInterface(acqIf);
    }
}

#endif // INCLUDED__IMAGE__VIEW__ABSTRACT__ACQUISITION__INTERFACE__H
