/* FILE:        AbstractFileInterface.cpp
* DATE:        1/7/2016
* AUTHOR:      Jeremy Greene
* COMPANY:     BitFlow, Inc.
* COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
* DESCRIPTION: Implementation of AbstractFileInterface.
*/

#include "AbstractFileInterface_Private.h"

#include "ImageViewStringHelpers.h"

// Compare two extensions, ignoring case and the presence or absence of a leading period. If either
// or both are invalid, return std::numeric_limits<int>::min().
int ImageView::AbstractFileInterface::compareExtensions (const char *const extLeft, const char *const extRight)
{
    const std::string left = PrivateData::comparableFileExtension(extLeft);
    const std::string right = PrivateData::comparableFileExtension(extRight);

    if (left.empty() || right.empty())
        return std::numeric_limits<int>::min();

    return left.compare(right);
}

ImageView::AbstractFileInterface::AbstractFileInterface (void)
    : m_pd (*new PrivateData(*this))
{ }

ImageView::AbstractFileInterface::~AbstractFileInterface (void)
{
    delete &m_pd;
}

// Add a file save format with the given primary extension, optional alternate extension(s) (space
// separated), and optional description. The primary and alternate extensions must be unique,
// vis-a-vis all other formats. Extensions may optionally start with a period, and should otherwise
// contain only uppercase and lowercase letters, and decimal digits.
bool ImageView::AbstractFileInterface::addFormat (const char *const primaryExtension, const char *const alternatExtensions, const char *const description)
{
    const auto comparablePrimary = PrivateData::comparableFileExtension(primaryExtension);
    if (comparablePrimary.empty())
        return false;

    // Parse the file format.
    PrivateData::FileFormat format;
    if (!PrivateData::parseFileFormat(primaryExtension, alternatExtensions, description, &format))
        return false;

    // Verify that the primary and alternate extensions are all unique.
    if (hasFormat(primaryExtension))
        return false;

    for (auto const& altExt : format.alternateExtensions)
    {
        if (hasFormat(altExt.first.c_str()))
            return false;
    }

    // Lock the file format map.
    const std::lock_guard<std::mutex> fileSaveFormatsLock (m_pd.m_fileFormatsMutex);

    // Insert the new file format.
    m_pd.m_fileFormats.insert(std::make_pair(comparablePrimary, format));

    return true;
}

// Clear the list of supported file formats.
void ImageView::AbstractFileInterface::clearFormats (void)
{
    // Lock the file format map.
    const std::lock_guard<std::mutex> fileSaveFormatsLock (m_pd.m_fileFormatsMutex);

    // Clear all file formats.
    m_pd.m_fileFormats.clear();
}

// Return true or false, whether or not the extension is supported. This searches both primary and
// alternate extension formats.
bool ImageView::AbstractFileInterface::hasFormat (const char *const extension) const
{
    const auto comparable = PrivateData::comparableFileExtension(extension);
    if (comparable.empty())
        return false;

    // Lock the file format map.
    const std::lock_guard<std::mutex> fileSaveFormatsLock (m_pd.m_fileFormatsMutex);

    // Check the primaries first.
    const auto extantIter = m_pd.m_fileFormats.find(comparable);
    if (m_pd.m_fileFormats.end() != extantIter)
        return true;

    // Check the alternates.
    for (auto const& format : m_pd.m_fileFormats)
    {
        const auto altExtIter = format.second.alternateExtensions.find(comparable);
        if (format.second.alternateExtensions.end() != altExtIter)
            return true;
    }

    return false;
}

// Query the count of file save formats, and the space-separated list of primary extensions.
// Any parameter may be nullptr, but pOutStrSize is required for pOutStr.
bool ImageView::AbstractFileInterface::queryFormats (size_t *const formatCount, char *const pOutStr, size_t *const pOutStrSize) const
{
    if (pOutStr && !pOutStrSize)
        return false;

    size_t fileFormatCount;
    std::string primaryExtensionsList;
    {
        // Lock the file format map.
        const std::lock_guard<std::mutex> fileSaveFormatsLock (m_pd.m_fileFormatsMutex);

        // Get the format count.
        fileFormatCount = m_pd.m_fileFormats.size();

        // If necessary, generate the primaryExtensionsList.
        if (pOutStrSize)
        {
            // Generate the 
            for (auto const& format : m_pd.m_fileFormats)
            {
                if (primaryExtensionsList.size())
                    primaryExtensionsList.push_back(' ');
                primaryExtensionsList.append(format.second.primaryExtension);
            }
        }
    }

    const bool outStrSuccess = !pOutStrSize || String::set(primaryExtensionsList, pOutStr, pOutStrSize);
    if (!outStrSuccess)
        return false;

    if (formatCount)
        *formatCount = fileFormatCount;

    return true;
}

// Query the file save format for the given primary extension.
bool ImageView::AbstractFileInterface::queryFormatPattern (const char *const primaryExtension, char *const pOutStr, size_t *const pOutStrSize) const
{
    const auto comparable = PrivateData::comparableFileExtension(primaryExtension);
    if (comparable.empty())
        return false;

    std::string formatPattern;
    {
        // Lock the file format map.
        const std::lock_guard<std::mutex> userSaveFormatsLock (m_pd.m_fileFormatsMutex);

        const auto formatIter = m_pd.m_fileFormats.find(comparable);
        if (m_pd.m_fileFormats.end() == formatIter)
            return false;

        formatPattern = formatIter->second.pattern;
    }

    return String::set(formatPattern, pOutStr, pOutStrSize);
}

//// ScopedFileInterface
ImageView::ScopedFileInterface::ScopedFileInterface (void)
    : m_fileIf (nullptr)
{ }
ImageView::ScopedFileInterface::ScopedFileInterface (AbstractFileInterface *const fileIf)
    : m_fileIf (fileIf)
{ }

ImageView::ScopedFileInterface::ScopedFileInterface (ScopedFileInterface &&to_take)
    : m_fileIf (to_take.m_fileIf)
{
    to_take.m_fileIf = nullptr;
}

ImageView::ScopedFileInterface::~ScopedFileInterface (void)
{
    delete m_fileIf;
}

ImageView::ScopedFileInterface::operator bool (void) const
{
    return !!m_fileIf;
}
bool ImageView::ScopedFileInterface::operator! (void) const
{
    return !m_fileIf;
}

void ImageView::ScopedFileInterface::take (ScopedFileInterface &&to_take)
{
    delete m_fileIf;
    m_fileIf = to_take.m_fileIf;
    to_take.m_fileIf = nullptr;
}
void ImageView::ScopedFileInterface::reset (AbstractFileInterface *const fileIf)
{
    delete m_fileIf;
    m_fileIf = fileIf;
}

ImageView::AbstractFileInterface* ImageView::ScopedFileInterface::get (void) const
{
    return m_fileIf;
}
ImageView::AbstractFileInterface* ImageView::ScopedFileInterface::operator-> (void) const
{
    return m_fileIf;
}
ImageView::AbstractFileInterface& ImageView::ScopedFileInterface::operator* (void) const
{
    return *m_fileIf;
}
