#ifndef INCLUDED__BIT_FLOW__IMAGE__VIEW__ABSTRACT__FILE__INTERFACE__H
#define INCLUDED__BIT_FLOW__IMAGE__VIEW__ABSTRACT__FILE__INTERFACE__H

/* FILE:        AbstractFileInterface.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Abstract interface for ImageView file I/O.
 */

#include "ImageView.h"
#include "ImageViewBaseClass.h"

#include "ImageViewRawImage.h"

namespace ImageView
{
    class IMAGEVIEW_EXPORT AbstractFileInterface : public BaseClass
    {
    private:
        struct PrivateData;
        PrivateData &m_pd;

        // Illegal.
        AbstractFileInterface (AbstractFileInterface const&);
        AbstractFileInterface& operator= (AbstractFileInterface const&);

    protected:
        AbstractFileInterface (void);

    public:
        // Compare two extensions, ignoring case and the presence or absence of a leading period. If either
        // or both extensions are invalid, return std::numeric_limits<int>::min().
        static int compareExtensions (const char *const extLeft, const char *const extRight);

        // Constructor.
        virtual ~AbstractFileInterface (void);

        // Add a file save format with the given primary extension, optional alternate extension(s) (space
        // separated), and optional description. The primary and alternate extensions must be unique,
        // vis-a-vis all other formats. Extensions may optionally start with a period, and should otherwise
        // contain only uppercase and lowercase letters, and decimal digits. Extension case is preserved
        // internally, but comparisons are case insensitive.
        bool addFormat (const char *const primaryExtension, const char *const alternatExtensions = nullptr, const char *const description = nullptr);

        // Clear the list of supported file formats.
        void clearFormats (void);

        // Return true or false, whether or not the extension is supported. This searches both primary and
        // alternate extension formats.
        bool hasFormat (const char *const extension) const;

        // Query the count of file save formats, and the space-separated list of primary extensions.
        // Any parameter may be nullptr, but pOutStrSize is required for pOutStr.
        bool queryFormats (size_t *const formatCount, char *const pOutStr = nullptr, size_t *const pOutStrSize = nullptr) const;

        // Query the file save format for the given primary extension.
        bool queryFormatPattern (const char *const primaryExtension, char *const pOutStr, size_t *const pOutStrSize) const;

        // Save an image to the specified file path, with format deduced from the file extension. The
        // caller is responsible for guaranteeing image immutability during the save operation.
        virtual bool save (RawImage const& rawImage, const char *const filePath) const = 0;
    };

    class IMAGEVIEW_EXPORT ScopedFileInterface
    {
    private:
        AbstractFileInterface *m_fileIf;

        // Illegal.
        ScopedFileInterface (int);
        ScopedFileInterface (std::nullptr_t);
        ScopedFileInterface (ScopedFileInterface const&);
        ScopedFileInterface& operator= (ScopedFileInterface const&);

    public:
        explicit ScopedFileInterface (void);
        explicit ScopedFileInterface (AbstractFileInterface *const fileIf);
        ScopedFileInterface (ScopedFileInterface &&to_take);

        ~ScopedFileInterface (void);

        operator bool (void) const;
        bool operator! (void) const;

        void take (ScopedFileInterface &&to_take);
        void reset (AbstractFileInterface *const fileIf);

        AbstractFileInterface* get (void) const;
        AbstractFileInterface* operator-> (void) const;
        AbstractFileInterface& operator* (void) const;
    };

    inline static ScopedFileInterface make_scoped (AbstractFileInterface *const acqIf)
    {
        return ScopedFileInterface(acqIf);
    }
}

#endif // INCLUDED__BIT_FLOW__IMAGE__VIEW__ABSTRACT__FILE__INTERFACE__H
