/* FILE:        AbstractFileInterface_Private.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::AbstractFileInterface::PrivateData.
 */

#include "AbstractFileInterface_Private.h"

#include <QString>
#include <QStringList>

#include <algorithm>

const std::string ImageView::AbstractFileInterface::PrivateData::valid_extension_chars = [](void)
{
    std::string vec;
    for (char c = '0'; c <= '9'; c++)
        vec.push_back(c);
    for (char c = 'a'; c <= 'z'; c++)
        vec.push_back(c);
    for (char c = 'A'; c <= 'Z'; c++)
        vec.push_back(c);
    return vec;
}();

ImageView::AbstractFileInterface::PrivateData::PrivateData (AbstractFileInterface &fileIf)
    : m_fileIf (fileIf)
{ }

//// Helpers.

std::string ImageView::AbstractFileInterface::PrivateData::cleanupFileExtension (const char *const extension)
{
    std::string cleanedExt;

    if (!extension)
        return std::string();

    cleanedExt = extension;
    if (cleanedExt.empty())
        return std::string();

    if (cleanedExt.front() != '.')
        cleanedExt.insert(0, 1, '.');

    if (cleanedExt.find_first_not_of(valid_extension_chars, 1) != std::string::npos)
        return std::string();

    if ("." == cleanedExt)
        return std::string();

    return cleanedExt;
}

std::string ImageView::AbstractFileInterface::PrivateData::comparableFileExtension (const char *const extension)
{
    std::string comparableExt;

    if (!extension)
        return std::string();

    comparableExt = extension;
    if (comparableExt.empty())
        return std::string();

    std::transform(comparableExt.begin(), comparableExt.end(), comparableExt.begin(), ::tolower);
    if (comparableExt.front() != '.')
        comparableExt.insert(0, 1, '.');

    if (comparableExt.find_first_not_of(valid_extension_chars, 1) != std::string::npos)
        return std::string();

    if ("." == comparableExt)
        return std::string();

    return comparableExt;
}

bool ImageView::AbstractFileInterface::PrivateData::parseFileFormat
    (const char *const primaryExtension, const char *const alternateExtensions, const char *const description, FileFormat *const format)
{
    FileFormat tmpFormat;
    tmpFormat.primaryExtension = cleanupFileExtension(primaryExtension);
    if (tmpFormat.primaryExtension.empty())
        return false;

    if (alternateExtensions)
    {
        const QString altExts = alternateExtensions;
        const QStringList altExtList = altExts.split(' ', QString::SkipEmptyParts);
        for (auto const& altExt : altExtList)
        {
            const auto extData = altExt.toLocal8Bit();
            const auto compExt = comparableFileExtension(extData);
            if (compExt.empty())
                return false;

            const auto cleanExt = cleanupFileExtension(extData);
            if (cleanExt.empty())
                return false;

            const auto altExtPair = std::make_pair(compExt, cleanExt);
            tmpFormat.alternateExtensions.insert(altExtPair);
        }
    }

    if (description)
        tmpFormat.description = description;

    // Generate the pattern.
    if (tmpFormat.description.size())
        tmpFormat.pattern = tmpFormat.description + " (";
    else
        tmpFormat.pattern = tmpFormat.primaryExtension + " image file (";

    tmpFormat.pattern += "*" + tmpFormat.primaryExtension;
    for (auto const& altExt : tmpFormat.alternateExtensions)
        tmpFormat.pattern += " *" + altExt.second;
    tmpFormat.pattern += ")";

    *format = tmpFormat;

    return true;
}
