#ifndef INCLUDED__BIT_FLOW__IMAGE__VIEW__BAYER__DEMOSAIC__H
#define INCLUDED__BIT_FLOW__IMAGE__VIEW__BAYER__DEMOSAIC__H

/* FILE:        BayerDemosaic.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Bilinear Bayer demosaic library for grayscale images (8-, 10-, 12-, 14- and 16-bit).
 */

#include "ImageConversionHelpers.h"
#include "ImageViewRawImage.h"

namespace ImageView
{
    // Demosaic buffer-to-QImage conversion.
    struct BayerPatterns
    {
        template <size_t XT, size_t YT>
        struct Pos
        {
            enum : size_t
            {
                x = XT,
                y = YT
            };
        };

        template <typename PosT> struct DestPos : public Pos<1 - PosT::x, 1 - PosT::y> { };
        template <typename PosT> struct PatternPos : public Pos<2 - PosT::x, 2 - PosT::y> { };

        template <typename PosT> struct Up : public Pos<PosT::x, PosT::y - 1> { };
        template <typename PosT> struct Left : public Pos<PosT::x - 1, PosT::y> { };
        template <typename PosT> struct Down : public Pos<PosT::x, PosT::y + 1> { };
        template <typename PosT> struct Right : public Pos<PosT::x + 1, PosT::y> { };

        template <typename PosT> struct UpLeft : public Pos<PosT::x - 1, PosT::y - 1> { };
        template <typename PosT> struct UpRight : public Pos<PosT::x + 1, PosT::y - 1> { };
        template <typename PosT> struct DownLeft : public Pos<PosT::x - 1, PosT::y + 1> { };
        template <typename PosT> struct DownRight : public Pos<PosT::x + 1, PosT::y + 1> { };

        template <typename PosT>
        struct ROffsets
        {
            typedef DestPos<PosT> BR;
            typedef PatternPos<PosT> R;

            typedef Up<R> G0;
            typedef Left<R> G1;
            typedef Down<R> G2;
            typedef Right<R> G3;

            typedef UpLeft<R> B0;
            typedef UpRight<R> B1;
            typedef DownLeft<R> B2;
            typedef DownRight<R> B3;
        };

        enum class GreenRed
        {
            Vertical,
            Horizontal
        };

        template <GreenRed GREEN_RED_ORIENTATION, typename PosT>
        struct GOffsets;

        template <typename PosT>
        struct GOffsets<GreenRed::Vertical, PosT>
        {
            typedef DestPos<PosT> BG;
            typedef PatternPos<PosT> G;

            typedef Up<G> R0;
            typedef Down<G> R1;

            typedef Left<G> B0;
            typedef Right<G> B1;
        };

        template <typename PosT>
        struct GOffsets<GreenRed::Horizontal, PosT>
        {
            typedef DestPos<PosT> BG;
            typedef PatternPos<PosT> G;

            typedef Left<G> R0;
            typedef Right<G> R1;

            typedef Up<G> B0;
            typedef Down<G> B1;
        };

        template <typename PosT>
        struct BOffsets
        {
            typedef DestPos<PosT> BB;
            typedef PatternPos<PosT> B;

            typedef Up<B> G0;
            typedef Left<B> G1;
            typedef Down<B> G2;
            typedef Right<B> G3;

            typedef UpLeft<B> R0;
            typedef UpRight<B> R1;
            typedef DownLeft<B> R2;
            typedef DownRight<B> R3;
        };

        typedef Pos<0, 0> Pos0;
        typedef Pos<1, 0> Pos1;
        typedef Pos<0, 1> Pos2;
        typedef Pos<1, 1> Pos3;

        struct RGGB
        {
            typedef ROffsets<Pos0> R;
            typedef GOffsets<GreenRed::Horizontal, Pos1> G0;
            typedef GOffsets<GreenRed::Vertical, Pos2> G1;
            typedef BOffsets<Pos3> B;
        };

        struct GRBG
        {
            typedef GOffsets<GreenRed::Horizontal, Pos0> G0;
            typedef ROffsets<Pos1> R;
            typedef BOffsets<Pos2> B;
            typedef GOffsets<GreenRed::Vertical, Pos3> G1;
        };

        struct GBRG
        {
            typedef GOffsets<GreenRed::Vertical, Pos0> G0;
            typedef BOffsets<Pos1> B;
            typedef ROffsets<Pos2> R;
            typedef GOffsets<GreenRed::Horizontal, Pos3> G1;
        };

        struct BGGR
        {
            typedef BOffsets<Pos0> B;
            typedef GOffsets<GreenRed::Vertical, Pos1> G0;
            typedef GOffsets<GreenRed::Horizontal, Pos2> G1;
            typedef ROffsets<Pos3> R;
        };
    };

    struct BayerDecoder
    {
    private:

        template <typename SrcT, typename SrcIteratorT>
        struct EdgeCopierT
        {
            SrcT srcQuad0[4], srcQuad1[4], srcQuad2[4], srcQuad3[4];
            SrcT *srcQuads[4];
            SrcIteratorT srcQuadIter0, srcQuadIter1, srcQuadIter2, srcQuadIter3;
            SrcIteratorT *srcHexConst[4];

            EdgeCopierT (void)
                : srcQuadIter0 (srcQuad0)
                , srcQuadIter1 (srcQuad1)
                , srcQuadIter2 (srcQuad2)
                , srcQuadIter3 (srcQuad3)
            {
                srcQuads[0] = srcQuad0;
                srcQuads[1] = srcQuad1;
                srcQuads[2] = srcQuad2;
                srcQuads[3] = srcQuad3;

                srcHexConst[0] = &srcQuadIter0;
                srcHexConst[1] = &srcQuadIter1;
                srcHexConst[2] = &srcQuadIter2;
                srcHexConst[3] = &srcQuadIter3;
            }

            inline void setupFirstPair (SrcIteratorT *srcLine[4])
            {
                for (size_t y = 0; y < 4; y++)
                {
                    srcQuads[y][1] = (*srcLine[y])[0];
                    srcQuads[y][0] = srcQuads[y][2] = (*srcLine[y])[1];
                    srcQuads[y][3] = (*srcLine[y])[2];
                }
            }

            inline void setupFinalPair (SrcIteratorT *srcLine[4])
            {
                for (size_t y = 0; y < 4; y++)
                {
                    srcQuads[y][0] = (*srcLine[y])[0];
                    srcQuads[y][3] = srcQuads[y][1] = (*srcLine[y])[1];
                    srcQuads[y][2] = (*srcLine[y])[2];
                }
            }

            inline void setupOverflowColumn (SrcIteratorT *srcLine[4])
            {
                for (size_t y = 0; y < 4; y++)
                {
                    srcQuads[y][2] = srcQuads[y][0] = (*srcLine[y])[0];
                    srcQuads[y][3] = srcQuads[y][1] = *((*srcLine[y]) + -1);
                }
            }
        };

        template <typename SrcT, Format PackedType>
        struct EdgeCopierT<SrcT, Gray10PackedIterator<PackedType>>
        {
            typedef Gray10PackedIterator<PackedType> SrcIterator;

            typename SrcIterator::ValUnion srcPair0[2], srcPair1[2], srcPair2[2], srcPair3[2];
            typename SrcIterator::ValUnion *srcPairs[4];

            SrcIterator srcQuadIter0, srcQuadIter1, srcQuadIter2, srcQuadIter3;
            SrcIterator *srcHexConst[4];

            EdgeCopierT (void)
                : srcQuadIter0 (&srcPairs[0][0].val64)
                , srcQuadIter1 (&srcPairs[1][0].val64)
                , srcQuadIter2 (&srcPairs[2][0].val64)
                , srcQuadIter3 (&srcPairs[3][0].val64)
            {
                srcPairs[0] = srcPair0;
                srcPairs[1] = srcPair1;
                srcPairs[2] = srcPair2;
                srcPairs[3] = srcPair3;

                srcHexConst[0] = &srcQuadIter0;
                srcHexConst[1] = &srcQuadIter1;
                srcHexConst[2] = &srcQuadIter2;
                srcHexConst[3] = &srcQuadIter3;
            }

            inline void setupEndPair (SrcIterator *srcLine[4])
            {
                for (size_t y = 0; y < 4; y++)
                {
                    srcPairs[y][0].val64 = srcLine[y]->val64();
                    if (PackedType != PackedNoSwap)
                        srcPairs[y][0].swap();
                    memcpy(srcPairs[y][0].val8 + 5, srcPairs[y][0].val8, 5);
                    (*srcHexConst[y] = SrcIterator(&srcPairs[y][0].val64))++;
                }
            }

            inline void setupFirstPair (SrcIterator *srcLine[4])
            {
                setupEndPair(srcLine);
            }

            inline void setupFinalPair (SrcIterator *srcLine[4])
            {
                setupEndPair(srcLine);
            }

            inline void setupOverflowColumn (SrcIterator *srcLine[4])
            {
                setupEndPair(srcLine);
                for (size_t y = 0; y < 4; y++)
                    (*srcHexConst[y])++;
            }
        };

        template <typename DstIterator, typename Filter, typename SrcT, typename SrcIteratorT, size_t shift>
        struct Decoder
        {
            typedef EdgeCopierT<SrcT, SrcIteratorT> EdgeCopier;

            typedef typename Filter::R R;
            typedef typename Filter::G0 G0;
            typedef typename Filter::G1 G1;
            typedef typename Filter::B B;

            inline static void rAverage (DstIterator dst, SrcIteratorT *src[4])
            {
                dst.set
                    ( ((*src[R::R::y])[R::R::x]) >> shift
                    , ((*src[R::G0::y])[R::G0::x] + (*src[R::G1::y])[R::G1::x] + (*src[R::G2::y])[R::G2::x] + (*src[R::G3::y])[R::G3::x]) / 4 >> shift
                    , ((*src[R::B0::y])[R::B0::x] + (*src[R::B1::y])[R::B1::x] + (*src[R::B2::y])[R::B2::x] + (*src[R::B3::y])[R::B3::x]) / 4 >> shift );
            }

            inline static void g0Average (DstIterator dst, SrcIteratorT *src[4])
            {
                dst.set
                    ( ((*src[G0::R0::y])[G0::R0::x] + (*src[G0::R1::y])[G0::R1::x]) / 2 >> shift
                    , ((*src[G0::G::y])[G0::G::x]) >> shift
                    , ((*src[G0::B0::y])[G0::B0::x] + (*src[G0::B1::y])[G0::B1::x]) / 2 >> shift );
            }

            inline static void g1Average (DstIterator dst, SrcIteratorT *src[4])
            {
                dst.set
                    ( ((*src[G1::R0::y])[G1::R0::x] + (*src[G1::R1::y])[G1::R1::x]) / 2 >> shift
                    , ((*src[G1::G::y])[G1::G::x]) >> shift
                    , ((*src[G1::B0::y])[G1::B0::x] + (*src[G1::B1::y])[G1::B1::x]) / 2 >> shift );
            }

            inline static void bAverage (DstIterator dst, SrcIteratorT *src[4])
            {
                dst.set
                    ( ((*src[B::R0::y])[B::R0::x] + (*src[B::R1::y])[B::R1::x] + (*src[B::R2::y])[B::R2::x] + (*src[B::R3::y])[B::R3::x]) / 4 >> shift
                    , ((*src[B::G0::y])[B::G0::x] + (*src[B::G1::y])[B::G1::x] + (*src[B::G2::y])[B::G2::x] + (*src[B::G3::y])[B::G3::x]) / 4 >> shift
                    , ((*src[B::B::y])[B::B::x]) >> shift );
            }

            inline static void decodeQuad (DstIterator *const dstLine, SrcIteratorT *src[4])
            {
                rAverage(dstLine[R::BR::y][R::BR::x], src);
                g0Average(dstLine[G0::BG::y][G0::BG::x], src);
                g1Average(dstLine[G1::BG::y][G1::BG::x], src);
                bAverage(dstLine[B::BB::y][B::BB::x], src);
            }

            inline static void decodeLinePair (uchar *const dst, const size_t dstLineSize, const SrcT *const srcPtr, const size_t width, const size_t height, const size_t line)
            {
                DstIterator dstLine [2] =
                    {
                        dst + dstLineSize * line,
                        dst + dstLineSize * (line + 1 < height ? line + 1 : line)
                    };

                SrcIteratorT srcLine1 = srcPtr + width * line;
                SrcIteratorT srcLine2 = srcPtr + width * (line + 1 < height ? line + 1 : (line > 0 ? line - 1 : line));
                SrcIteratorT srcLine3 = srcPtr + width * (line + 2 < height ? line + 2 : line);
                SrcIteratorT srcLine0 = line > 0 ? srcPtr + width * (line - 1) : srcLine2;
                SrcIteratorT *srcLine[4] = { &srcLine0, &srcLine1, &srcLine2, &srcLine3 };

                const size_t safeWidth = (width & ~size_t(1)) - 2;
                const size_t endWidth = safeWidth - 1;
                SrcIteratorT srcLine0PairsEnd = srcLine0 + endWidth;
                SrcIteratorT srcLine0PairsUnrolled = srcLine0 + (endWidth - endWidth % (2 * 4));

                // Decode the first column using special handling.
                EdgeCopier EdgeCopier;
                EdgeCopier.setupFirstPair(srcLine);

                decodeQuad(dstLine, EdgeCopier.srcHexConst);
                dstLine[0] += 2; dstLine[1] += 2;
                srcLine0++; srcLine1++; srcLine2++; srcLine3++;

                // Decode the main part.
                    // Unroll a few entries.
                while (srcLine0 < srcLine0PairsUnrolled)
                {
                    decodeQuad(dstLine, srcLine);
                    dstLine[0] += 2; dstLine[1] += 2;
                    srcLine0 += 2; srcLine1 += 2; srcLine2 += 2; srcLine3 += 2;

                    decodeQuad(dstLine, srcLine);
                    dstLine[0] += 2; dstLine[1] += 2;
                    srcLine0 += 2; srcLine1 += 2; srcLine2 += 2; srcLine3 += 2;

                    decodeQuad(dstLine, srcLine);
                    dstLine[0] += 2; dstLine[1] += 2;
                    srcLine0 += 2; srcLine1 += 2; srcLine2 += 2; srcLine3 += 2;

                    decodeQuad(dstLine, srcLine);
                    dstLine[0] += 2; dstLine[1] += 2;
                    srcLine0 += 2; srcLine1 += 2; srcLine2 += 2; srcLine3 += 2;
                }
                // Do the rest.
                while (srcLine0 < srcLine0PairsEnd)
                {
                    decodeQuad(dstLine, srcLine);
                    dstLine[0] += 2; dstLine[1] += 2;
                    srcLine0 += 2; srcLine1 += 2; srcLine2 += 2; srcLine3 += 2;
                }

                // Decode the final whole column pair.
                EdgeCopier.setupFinalPair(srcLine);

                decodeQuad(dstLine, EdgeCopier.srcHexConst);
                dstLine[0] += 2; dstLine[1] += 2;
                srcLine0 += 2; srcLine1 += 2; srcLine2 += 2; srcLine3 += 2;

                // If any, decode the final column.
                if (width & 1u)
                {
                    EdgeCopier.setupOverflowColumn(srcLine);

                    typename DstIterator::Pix tmpDstPair0 [2], tmpDstPair1 [2];
                    DstIterator tmpDstQuad [2] = { tmpDstPair0, tmpDstPair1 };
                    decodeQuad(tmpDstQuad, EdgeCopier.srcHexConst);

                    dstLine[0].set(tmpDstQuad[0].get());
                    dstLine[1].set(tmpDstQuad[1].get());
                }
            }
        };

        ImageConversion &m_imageConverter;

        template <typename Filter, size_t shift, typename SrcT, typename SrcIteratorT>
        inline void multithreadedDecode (uchar *const dst, const size_t dstLineSize, const bool dstAlpha, RawImage const& rawImage)
        {
            typedef Decoder<RgbxIterator, Filter, SrcT, SrcIteratorT, shift> DecoderImpRgbx;
            typedef Decoder<RgbIterator, Filter, SrcT, SrcIteratorT, shift> DecoderImpRgb;

            std::atomic<size_t> lineAtomic (0);

            const auto *const srcPtr = reinterpret_cast<const SrcT*>(rawImage.m_buffer);

            if (dstAlpha)
                m_imageConverter.logicalMultithreadedExe([&](const size_t id, const size_t cnt)
                    {
                        for (size_t line = lineAtomic.fetch_add(2); line < rawImage.m_height; line = lineAtomic.fetch_add(2))
                            DecoderImpRgbx::decodeLinePair(dst, dstLineSize, srcPtr, rawImage.m_width, rawImage.m_height, line);
                    });
            else
                m_imageConverter.logicalMultithreadedExe([&](const size_t id, const size_t cnt)
                    {
                        for (size_t line = lineAtomic.fetch_add(2); line < rawImage.m_height; line = lineAtomic.fetch_add(2))
                            DecoderImpRgb::decodeLinePair(dst, dstLineSize, srcPtr, rawImage.m_width, rawImage.m_height, line);
                    });
        }

        template <int FORMAT_BASE, typename SrcT, typename SrcIteratorT>
        inline bool decodeBase (uchar *const dst, const size_t dstLineSize, const bool dstAlpha, RawImage const& rawImage)
        {
            enum : size_t { downShift = (FORMAT_BASE & ImageView::PackedFormatMask) ? 0 : (FORMAT_BASE & ImageView::BaseFormatMask) - 8 };

            switch (rawImage.m_format & ImageView::BayerFormatMask)
            {
            case ImageView::BayerRGGB:
                multithreadedDecode<BayerPatterns::RGGB, downShift, SrcT, SrcIteratorT>(dst, dstLineSize, dstAlpha, rawImage);
                return true;
            case ImageView::BayerGRBG:
                multithreadedDecode<BayerPatterns::GRBG, downShift, SrcT, SrcIteratorT>(dst, dstLineSize, dstAlpha, rawImage);
                return true;
            case ImageView::BayerGBRG:
                multithreadedDecode<BayerPatterns::GBRG, downShift, SrcT, SrcIteratorT>(dst, dstLineSize, dstAlpha, rawImage);
                return true;
            case ImageView::BayerBGGR:
                multithreadedDecode<BayerPatterns::BGGR, downShift, SrcT, SrcIteratorT>(dst, dstLineSize, dstAlpha, rawImage);
                return true;
            }

            return false;
        }

    public:
        BayerDecoder (ImageConversion &imageConverter)
            : m_imageConverter (imageConverter)
        { }

        inline bool decode8 (uchar *const dst, const size_t dstLineSize, const bool dstAlpha, RawImage const& rawImage)
        {
            switch (rawImage.m_format & ImageView::BaseFormatMask)
            {
            case Gray8:
                decodeBase<Gray8, std::uint8_t, const std::uint8_t*>(dst, dstLineSize, dstAlpha, rawImage);
                return true;
            }
            return false;
        }

        inline bool decode16 (uchar *const dst, const size_t dstLineSize, const bool dstAlpha, RawImage const& rawImage)
        {
            switch (rawImage.m_format & BaseFormatMask)
            {
            case Gray10:
                switch (rawImage.m_format & PackedFormatMask)
                {
                case Packed:
                    decodeBase<Gray10|Packed, std::uint16_t, Gray10PackedIterator<Packed>>(dst, dstLineSize, dstAlpha, rawImage);
                    break;
                case PackedNoSwap:
                    decodeBase<Gray10|PackedNoSwap, std::uint16_t, Gray10PackedIterator<PackedNoSwap>>(dst, dstLineSize, dstAlpha, rawImage);
                    break;
                default:
                    decodeBase<Gray10, std::uint16_t, const std::uint16_t*>(dst, dstLineSize, dstAlpha, rawImage);
                }
                return true;
            case Gray12:
                decodeBase<Gray12, std::uint16_t, const std::uint16_t*>(dst, dstLineSize, dstAlpha, rawImage);
                return true;
            case Gray14:
                decodeBase<Gray14, std::uint16_t, const std::uint16_t*>(dst, dstLineSize, dstAlpha, rawImage);
                return true;
            case Gray16:
                decodeBase<Gray16, std::uint16_t, const std::uint16_t*>(dst, dstLineSize, dstAlpha, rawImage);
                return true;
            }
            return false;
        }
    };
}

#endif // INCLUDED__BIT_FLOW__IMAGE__VIEW__BAYER__DEMOSAIC__H
