/* FILE:        ImageConversionHelpers.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::ImageConversion.
 */

#include "ImageConversionHelpers_Private.h"
#include "ImageViewThread.h"

#include <algorithm>

namespace ImageView
{
    template <>
    void Gray10PackedIterator<ImageView::Packed>::ValUnion::swap (void)
    {
        std::swap(val8[0], val8[7]);
        std::swap(val8[1], val8[6]);
        std::swap(val8[2], val8[5]);
        std::swap(val8[3], val8[4]);
    }

    template <>
    void Gray10PackedIterator<ImageView::PackedNoSwap>::ValUnion::swap (void)
    {
    }
}

// Private implementation.
ImageView::ImageConversion::PrivateData::PrivateData (void)
    : m_logicalThreadCount (std::max(ImageView::thread::hardware_concurrency(), 1u))
    , m_extraThreadCount (1 >= m_logicalThreadCount ? 0 : m_logicalThreadCount - 1)
    , m_keepRunning (true)
    , m_hasActiveFunc (false)
    , m_threadsAwake (0)
    , m_funcsStarted (0)
{ }

ImageView::ImageConversion::PrivateData::~PrivateData (void)
{
    {
        std::unique_lock<std::mutex> activeFuncLock (m_activeFuncMutex);

        m_keepRunning.store(false);
        m_activeFuncCondition.notify_all();

        while (m_threadsAwake > 0)
            m_threadAwokeCondition.wait(activeFuncLock);
    }

    for (auto &t : m_threads)
    {
        if (t.joinable())
            t.join();
    }
}

void ImageView::ImageConversion::PrivateData::checkThreads (void)
{
    if (m_extraThreadCount && !m_threads.size())
    {
        m_threads.resize(m_extraThreadCount);

        size_t threadId = 0;

        for (auto &t : m_threads)
        {
            std::promise<void> launchPromise;
            t = ImageView::thread(&PrivateData::logicalThreadExe, this, ImageView::ref(launchPromise), ++threadId);
            launchPromise.get_future().wait();
        }
    }
}

void ImageView::ImageConversion::PrivateData::logicalThreadExe (std::promise<void> &launched, const size_t id)
{
    std::unique_lock<std::mutex> activeFuncLock (m_activeFuncMutex);
    launched.set_value();

    while (m_keepRunning.load())
    {
        m_activeFuncCondition.wait(activeFuncLock);
        if (!m_hasActiveFunc)
            continue;

        const Func func = m_activeFunc;

        m_funcsStarted++;
        m_threadsAwake++;
        activeFuncLock.unlock();
        try
        {
            func(id, m_logicalThreadCount);
        }
        catch (...)
        { }
        activeFuncLock.lock();
        m_threadsAwake--;

        m_threadAwokeCondition.notify_one();
    }
}

void ImageView::ImageConversion::PrivateData::logicalMultithreadedExe (Func const& func)
{
    const std::lock_guard<std::mutex> logicalMultithreadedExeLock (m_logicalMultithreadedExeMutex);

    checkThreads();

    size_t funcsStartedAtFinish;
    {
        std::unique_lock<std::mutex> activeFuncLock (m_activeFuncMutex);

        funcsStartedAtFinish = m_funcsStarted + m_extraThreadCount;
        m_activeFunc = func;
        m_hasActiveFunc = true;

        m_activeFuncCondition.notify_all();
    }

    func(0, m_logicalThreadCount);

    {
        std::unique_lock<std::mutex> activeFuncLock (m_activeFuncMutex);

        while (m_threadsAwake > 0 || funcsStartedAtFinish > m_funcsStarted)
            m_threadAwokeCondition.wait(activeFuncLock);

        m_hasActiveFunc = false;
    }
}

ImageView::ImageConversion::ImageConversion (void)
    : m_pd (*new PrivateData)
{ }

ImageView::ImageConversion::~ImageConversion (void)
{
    delete &m_pd;
}

// Public features.
void ImageView::ImageConversion::logicalMultithreadedExe (Func const& func)
{
    m_pd.logicalMultithreadedExe(func);
}
