#ifndef INCLUDED__BIT_FLOW__IMAGE__CONVERSION__HELPERS__H
#define INCLUDED__BIT_FLOW__IMAGE__CONVERSION__HELPERS__H

/* FILE:        ImageConversionHelpers.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Image conversion for the library ImageView, supporting conversion
 *              from any BitFlow format to 8-bit indexed, and 24-bit and 32-bit RGB formats.
 */

#include "ImageView.h"
#include "ImageViewRawImage.h"

#include <cstdint>
#include <functional>
#include <atomic>

namespace ImageView
{
    struct ImageConversion
    {
        typedef std::function<void(const size_t, const size_t)> Func;

    private:
        struct PrivateData;
        PrivateData &m_pd;

        ImageConversion (ImageConversion const&);
        ImageConversion& operator= (ImageConversion const&);

    public:
        ImageConversion (void);
        ~ImageConversion (void);

        // Execute the function in as many threads as their are logical cores on the system. This is a
        // synchronous function, and only one call may execute at a time.
        void logicalMultithreadedExe (Func const& func);
    };

    template <typename SrcT, size_t ShiftBy = 0, typename SrcIteratorT = const SrcT*>
    struct ThreadableCopyGray
    {
    private:
        std::atomic<size_t> m_lineAtomic;
        const RawImage m_rawImage;
        const SrcT *const m_src;
        std::uint8_t *const m_dst;
        const size_t m_dstLineSize;
        const size_t m_srcLineSize;
        const size_t m_unrolledLineWidth;

        inline void copyFunc (const size_t id, const size_t cnt)
        {
            for (size_t line = m_lineAtomic.fetch_add(1); line < m_rawImage.m_height; line = m_lineAtomic.fetch_add(1))
            {
                std::uint8_t *dstIter = m_dst + line * m_dstLineSize;
                const std::uint8_t *const dstUnrolledEnd = dstIter + m_unrolledLineWidth;
                const std::uint8_t *const dstEnd = dstIter + m_rawImage.m_width;
                SrcIteratorT srcIter = m_src + line * m_srcLineSize;

                while (dstIter < dstUnrolledEnd)
                {
                    *dstIter = std::uint8_t(*srcIter >> ShiftBy); dstIter++; srcIter++;
                    *dstIter = std::uint8_t(*srcIter >> ShiftBy); dstIter++; srcIter++;
                    *dstIter = std::uint8_t(*srcIter >> ShiftBy); dstIter++; srcIter++;
                    *dstIter = std::uint8_t(*srcIter >> ShiftBy); dstIter++; srcIter++;
                    *dstIter = std::uint8_t(*srcIter >> ShiftBy); dstIter++; srcIter++;
                    *dstIter = std::uint8_t(*srcIter >> ShiftBy); dstIter++; srcIter++;
                    *dstIter = std::uint8_t(*srcIter >> ShiftBy); dstIter++; srcIter++;
                    *dstIter = std::uint8_t(*srcIter >> ShiftBy); dstIter++; srcIter++;
                }

                while (dstIter < dstEnd)
                    *dstIter = std::uint8_t(*srcIter >> ShiftBy); dstIter++; srcIter++;
            }
        }

    public:
        ThreadableCopyGray (RawImage const& rawImage, std::uint8_t *const dst, const size_t dstLineSize)
            : m_lineAtomic (0)
            , m_rawImage (rawImage)
            , m_src (reinterpret_cast<const SrcT*>(rawImage.m_buffer))
            , m_dst (dst)
            , m_dstLineSize (dstLineSize)
            , m_srcLineSize (rawImage.m_width)
            , m_unrolledLineWidth (rawImage.m_width & ~size_t(8 - 1))
        { }

        inline operator ImageConversion::Func (void)
        {
            return std::bind(&ThreadableCopyGray::copyFunc, this, std::placeholders::_1, std::placeholders::_2);
        }
    };

    template<>
    struct ThreadableCopyGray<std::uint8_t>
    {
    private:
        const RawImage m_rawImage;
        const std::uint8_t *const m_src;
        std::uint8_t *const m_dst;
        const size_t m_imageSize;

        inline void copyFunc (const size_t id, const size_t cnt)
        {
            const size_t blockSize = m_imageSize / cnt;
            const size_t blockOffset = blockSize * id;

            if (id + 1 < cnt)
                memcpy(m_dst + blockOffset, m_src + blockOffset, blockSize);
            else
            {
                const size_t blockSizeRemainder = m_imageSize % cnt;
                memcpy(m_dst + blockOffset, m_src + blockOffset, blockSize + blockSizeRemainder);
            }
        }

    public:
        ThreadableCopyGray (RawImage const& rawImage, std::uint8_t *const dst)
            : m_rawImage (rawImage)
            , m_src (reinterpret_cast<const std::uint8_t*>(rawImage.m_buffer))
            , m_dst (dst)
            , m_imageSize (rawImage.m_width * rawImage.m_height)
        { }

        inline operator ImageConversion::Func (void)
        {
            return std::bind(&ThreadableCopyGray::copyFunc, this, std::placeholders::_1, std::placeholders::_2);
        }
    };

    template <typename SrcT, size_t SrcPixSize, size_t ShiftBy = 0, typename SrcIteratorT = const SrcT*>
    struct ThreadableCopyColor
    {
    private:
        std::atomic<size_t> m_lineAtomic;
        const RawImage m_rawImage;
        const SrcT *const m_src;
        std::uint8_t *const m_dst;
        const size_t m_dstLineSize;
        const bool m_dstAlpha;
        const size_t m_srcLineSize;
        const size_t m_unrolledLineWidth;

        struct ColorOrder
        {
            struct RGB
            {
                enum : size_t
                {
                    R = 0,
                    G = 1,
                    B = 2
                };
            };

            struct BGR
            {
                enum : size_t
                {
                    R = 2,
                    G = 1,
                    B = 0
                };
            };
        };

        template <typename DstIterator, typename CO>
        inline void copyFunc (const size_t id, const size_t cnt)
        {
            for (size_t line = m_lineAtomic.fetch_add(1); line < m_rawImage.m_height; line = m_lineAtomic.fetch_add(1))
            {
                DstIterator dstIter = m_dst + line * m_dstLineSize;
                const DstIterator dstUnrolledEnd = dstIter + m_unrolledLineWidth;
                const DstIterator dstEnd = dstIter + m_rawImage.m_width;
                SrcIteratorT srcIter = m_src + line * m_srcLineSize;

                while (dstIter < dstUnrolledEnd)
                {
                    dstIter.set(srcIter[CO::R] >> ShiftBy, srcIter[CO::G] >> ShiftBy, srcIter[CO::B] >> ShiftBy);
                    ++dstIter;
                    srcIter += SrcPixSize;

                    dstIter.set(srcIter[CO::R] >> ShiftBy, srcIter[CO::G] >> ShiftBy, srcIter[CO::B] >> ShiftBy);
                    ++dstIter;
                    srcIter += SrcPixSize;

                    dstIter.set(srcIter[CO::R] >> ShiftBy, srcIter[CO::G] >> ShiftBy, srcIter[CO::B] >> ShiftBy);
                    ++dstIter;
                    srcIter += SrcPixSize;

                    dstIter.set(srcIter[CO::R] >> ShiftBy, srcIter[CO::G] >> ShiftBy, srcIter[CO::B] >> ShiftBy);
                    ++dstIter;
                    srcIter += SrcPixSize;

                    dstIter.set(srcIter[CO::R] >> ShiftBy, srcIter[CO::G] >> ShiftBy, srcIter[CO::B] >> ShiftBy);
                    ++dstIter;
                    srcIter += SrcPixSize;

                    dstIter.set(srcIter[CO::R] >> ShiftBy, srcIter[CO::G] >> ShiftBy, srcIter[CO::B] >> ShiftBy);
                    ++dstIter;
                    srcIter += SrcPixSize;

                    dstIter.set(srcIter[CO::R] >> ShiftBy, srcIter[CO::G] >> ShiftBy, srcIter[CO::B] >> ShiftBy);
                    ++dstIter;
                    srcIter += SrcPixSize;

                    dstIter.set(srcIter[CO::R] >> ShiftBy, srcIter[CO::G] >> ShiftBy, srcIter[CO::B] >> ShiftBy);
                    ++dstIter;
                    srcIter += SrcPixSize;
                }

                while (dstIter < dstEnd)
                {
                    dstIter.set(srcIter[CO::R] >> ShiftBy, srcIter[CO::G] >> ShiftBy, srcIter[CO::B] >> ShiftBy);
                    ++dstIter;
                    srcIter += SrcPixSize;
                }
            }
        }

    public:
        ThreadableCopyColor (RawImage const& rawImage, std::uint8_t *const dst, const size_t dstLineSize, const bool dstAlpha)
            : m_lineAtomic (0)
            , m_rawImage (rawImage)
            , m_src (reinterpret_cast<const SrcT*>(rawImage.m_buffer))
            , m_dst (dst)
            , m_dstLineSize (dstLineSize)
            , m_dstAlpha (dstAlpha)
            , m_srcLineSize (rawImage.m_width * SrcPixSize)
            , m_unrolledLineWidth (rawImage.m_width & ~size_t(8 - 1))
        { }

        inline operator ImageConversion::Func (void);
    };

    struct RgbxIterator
    {
        typedef std::uint8_t Pix[4];

        std::uint32_t *m_rgbx;

        template<typename DstT>
        RgbxIterator (DstT *const rgbx)
            : m_rgbx (reinterpret_cast<std::uint32_t*>(rgbx))
        { }

        inline void set (const std::uint32_t r, const std::uint32_t g, const std::uint32_t b)
        {
            *m_rgbx = (0xFF << 24) | ((r & 0xFF) << 16) | ((g & 0xFF) << 8) | (b & 0xFF);
        }

        inline void set (const std::uint32_t rgb)
        {
            *m_rgbx = (0xFF << 24) | rgb;
        }

        inline std::uint32_t get (void) const
        {
            return *m_rgbx;
        }

        inline RgbxIterator& operator++ (void)
        {
            ++m_rgbx;
            return *this;
        }

        template<typename OffsetT>
        inline RgbxIterator& operator+= (const OffsetT offset)
        {
            m_rgbx += offset;
            return *this;
        }

        template<typename OffsetT>
        inline RgbxIterator operator+ (const OffsetT offset) const
        {
            return RgbxIterator(m_rgbx + offset);
        }

        template<typename OffsetT>
        inline RgbxIterator operator[] (const OffsetT offset) const
        {
            return RgbxIterator (m_rgbx + offset);
        }

        inline bool operator== (RgbxIterator const& other) const
        {
            return m_rgbx == other.m_rgbx;
        }
        inline bool operator< (RgbxIterator const& other) const
        {
            return m_rgbx < other.m_rgbx;
        }
    };

    struct RgbIterator
    {
        typedef std::uint8_t Pix[3];

        std::uint8_t *m_rgb;

        template<typename DstT>
        RgbIterator (DstT *const rgb)
            : m_rgb (reinterpret_cast<std::uint8_t*>(rgb))
        { }

        inline void set (const std::uint32_t r, const std::uint32_t g, const std::uint32_t b)
        {
            m_rgb[2] = (std::uint8_t)r;
            m_rgb[1] = (std::uint8_t)g;
            m_rgb[0] = (std::uint8_t)b;
        }

        inline void set (const std::uint32_t rgb)
        {
            m_rgb[2] = (std::uint8_t)((rgb >> 16) & 0xFF);
            m_rgb[1] = (std::uint8_t)((rgb >> 8) & 0xFF);
            m_rgb[0] = (std::uint8_t)((rgb >> 0) & 0xFF);
        }

        inline std::uint32_t get (void) const
        {
            return (0xFF << 24)
                | ((std::uint32_t(m_rgb[2]) << 16) & 0xFF)
                | ((std::uint32_t(m_rgb[1]) << 8) & 0xFF)
                | (std::uint32_t(m_rgb[0]) & 0xFF);
        }

        inline RgbIterator& operator++ (void)
        {
            m_rgb += 3;
            return *this;
        }

        template<typename OffsetT>
        inline RgbIterator& operator+= (const OffsetT offset)
        {
            m_rgb += 3 * offset;
            return *this;
        }

        template<typename OffsetT>
        inline RgbIterator operator+ (const OffsetT offset) const
        {
            return RgbIterator(m_rgb + 3 * offset);
        }

        template<typename OffsetT>
        inline RgbIterator operator[] (const OffsetT offset) const
        {
            return RgbIterator (m_rgb + 3 * offset);
        }

        inline bool operator== (RgbIterator const& other) const
        {
            return m_rgb == other.m_rgb;
        }
        inline bool operator< (RgbIterator const& other) const
        {
            return m_rgb < other.m_rgb;
        }
    };

    template <Format PackedFormat>
    struct Gray10PackedIterator
    {
    public:
        union ValUnion
        {
            std::uint8_t val8[8];
            std::uint64_t val64;

            void swap (void);
        };

    private:
        ptrdiff_t m_cnt;
        const std::uint8_t *m_src;

        mutable ValUnion m_valUnion;
        mutable bool m_needsUpdate;

        inline void update (void) const
        {
            m_valUnion.val64 = *reinterpret_cast<const std::uint64_t*>(m_src);
            m_valUnion.swap();
            m_needsUpdate = false;
        }

    public:
        template <typename SrcT>
        Gray10PackedIterator (const SrcT *const src)
            : m_cnt (0)
            , m_src (reinterpret_cast<const std::uint8_t*>(src))
            , m_needsUpdate (true)
        { }

        inline std::uint64_t val64 (void) const
        {
            if (m_needsUpdate)
                update();
            return m_valUnion.val64;
        }

        inline std::uint8_t get (void) const
        {
            switch (m_cnt & 3u)
            {
            case 0: return val64() >> 56;
            case 1: return val64() >> 46;
            case 2: return val64() >> 36;
            }
            return val64() >> 26;
        }

        inline std::uint16_t get_all (void) const
        {
            switch (m_cnt & 3u)
            {
            case 0: return (val64() >> 54) & 0x3FF;
            case 1: return (val64() >> 44) & 0x3FF;
            case 2: return (val64() >> 34) & 0x3FF;
            }
            return (val64() >> 24) & 0x3FF;
        }

        inline void advance (const ptrdiff_t by)
        {
            const ptrdiff_t oldQuarter = m_cnt >> 2;
            m_cnt += by;
            const ptrdiff_t newQuarter = m_cnt >> 2;

            m_src += 5 * (newQuarter - oldQuarter);
            m_needsUpdate = m_needsUpdate || oldQuarter != newQuarter;
        }

        inline void next (void)
        {
            const ptrdiff_t cntLsb = ++m_cnt & 3u;
            m_src += cntLsb ? 0 : 5;
            m_needsUpdate = m_needsUpdate || !cntLsb;
        }

        inline Gray10PackedIterator& operator++ (int)
        {
            next();
            return *this;
        }

        inline Gray10PackedIterator& operator+= (const ptrdiff_t by)
        {
            advance(by);
            return *this;
        }

        inline Gray10PackedIterator operator+ (const ptrdiff_t by) const
        {
            Gray10PackedIterator sum (*this);
            sum.advance(by);
            return sum;
        }
        inline Gray10PackedIterator operator- (const ptrdiff_t by) const
        {
            Gray10PackedIterator sum (*this);
            sum.advance(-by);
            return sum;
        }

        inline std::uint8_t operator* (void) const
        {
            return get();
        }

        inline std::uint8_t operator[] (const ptrdiff_t at) const
        {
            if (0 == at)
                return get();

            Gray10PackedIterator offset (*this);
            offset.advance(at);
            return offset.get();
        }

        inline bool operator< (Gray10PackedIterator const& other) const
        {
            return m_cnt < other.m_cnt;
        }
    };

    struct Color30Iterator
    {
        const std::uint32_t *m_data;

        Color30Iterator (const std::uint32_t *const data)
            : m_data (data)
        { }

        inline std::uint32_t operator[] (const ptrdiff_t index) const
        {
            switch (index)
            {
            case 0: return (*m_data >> 20) & 0x03FF;
            case 1: return (*m_data >> 10) & 0x03FF;
            }
            return (*m_data) & 0x03FF;
        }

        inline Color30Iterator& operator+= (const ptrdiff_t offset)
        {
            m_data += offset;
            return *this;
        }
    };

    typedef ThreadableCopyGray<std::uint8_t>
        ThreadableCopyGray8;

    typedef ThreadableCopyGray<std::uint16_t, Gray10 - 8>
        ThreadableCopyGray10;

    typedef ThreadableCopyGray<std::uint16_t, 0, Gray10PackedIterator<Packed>>
        ThreadableCopyGray10Packed;

    typedef ThreadableCopyGray<std::uint16_t, 0, Gray10PackedIterator<PackedNoSwap>>
        ThreadableCopyGray10PackedNoSwap;

    typedef ThreadableCopyGray<std::uint16_t, Gray12 - 8>
        ThreadableCopyGray12;

    typedef ThreadableCopyGray<std::uint16_t, Gray14 - 8>
        ThreadableCopyGray14;

    typedef ThreadableCopyGray<std::uint16_t, Gray16 - 8>
        ThreadableCopyGray16;

    typedef ThreadableCopyColor<std::uint8_t, 3>
        ThreadableCopyColor24;

    typedef ThreadableCopyColor<std::uint32_t, 1, 2, Color30Iterator>
        ThreadableCopyColor30;

    typedef ThreadableCopyColor<std::uint8_t, 4>
        ThreadableCopyColor32;

    typedef ThreadableCopyColor<std::uint16_t, 3, HiColor36 / 3 - 8>
        ThreadableCopyHiColor36;

    typedef ThreadableCopyColor<std::uint16_t, 3, HiColor42 / 3 - 8>
        ThreadableCopyHiColor42;

    typedef ThreadableCopyColor<std::uint16_t, 3, HiColor48 / 3 - 8>
        ThreadableCopyHiColor48;

    // Inline functions.
    template <typename SrcT, size_t SrcPixSize, size_t ShiftBy, typename SrcIteratorT>
    ThreadableCopyColor<SrcT, SrcPixSize, ShiftBy, SrcIteratorT>::operator ImageConversion::Func (void)
    {
        switch (m_rawImage.m_format & ColorOrderFormatMask)
        {
        default:
        case ColorOrderRGB:
            if (m_dstAlpha)
                return std::bind(&ThreadableCopyColor::copyFunc<RgbxIterator, typename ColorOrder::RGB>, this, std::placeholders::_1, std::placeholders::_2);
            return std::bind(&ThreadableCopyColor::copyFunc<RgbIterator, typename ColorOrder::RGB>, this, std::placeholders::_1, std::placeholders::_2);
        case ColorOrderBGR:
            if (m_dstAlpha)
                return std::bind(&ThreadableCopyColor::copyFunc<RgbxIterator, typename ColorOrder::BGR>, this, std::placeholders::_1, std::placeholders::_2);
            return std::bind(&ThreadableCopyColor::copyFunc<RgbIterator, typename ColorOrder::BGR>, this, std::placeholders::_1, std::placeholders::_2);
        }
    }
}

#endif // INCLUDED__BIT_FLOW__IMAGE__CONVERSION__HELPERS__H
