#ifndef INCLUDED__BIT_FLOW__IMAGE__VIEW__H
#define INCLUDED__BIT_FLOW__IMAGE__VIEW__H

/* FILE:        ImageView.h
 * DATE:        12/15/2015
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2015, BitFlow, Inc.
 * DESCRIPTION: Primary declarations of ImageView library C++ globals.
 */

#if defined(_WIN32)
#   if defined(IMAGEVIEW_LIB)
#       define IMAGEVIEW_EXPORT __declspec(dllexport)
#   else
#       define IMAGEVIEW_EXPORT __declspec(dllimport)
#   endif
#elif defined(__GNUC__)
#   if defined(IMAGEVIEW_LIB)
#       define IMAGEVIEW_EXPORT [[gnu::visibility("default")]]
#   else
#       define IMAGEVIEW_EXPORT [[gnu::visibility("default")]]
#   endif
#else
#   error No import/export definitions for this platform.
#endif

#include <cstddef>
#include <cstdint>
#include <cstring>

namespace ImageView
{
    // ImageView Base class.
    class BaseClass;

    // Acquisition classes.
    class AcquisitionSignalHandler;
    class AbstractAcquisitionInterface;

    // File handler classes.
    class AbstractFileInterface;

    // GUI and GUI helper classes.
    class Application;
    class Surface;
    class SurfaceStream;
    class SurfaceScrollArea;
    class Window;
    class Action;

    class AbstractActionInterface;

    // Helper classes.
    struct RawImage;
    struct PixPos;
    struct PixVal;
    class AbstractLock;

    // Smart pointers.
    class ScopedAcquisitionInterface;
    class ScopedFileInterface;
    class ScopedActionInterface;
    class ScopedLock;

    // make_scoped exists for each of the Scoped* classes, but is explicitly a no-op for int & nullptr_t.
    inline static void make_scoped (int) { }
    inline static void make_scoped (std::nullptr_t) { }

    // Constants.
    enum Format
    {
        UnknownFormat = 0,

        // Base formats.
        Gray8 = 8,
        Gray10 = 10,
        Gray12 = 12,
        Gray14 = 14,
        Gray16 = 16,

        Color24 = 24,
        Color30 = 30,
        Color32 = 32,

        HiColor36 = 36,
        HiColor42 = 42,
        HiColor48 = 48,

        // Encoding flags.
        Unencoded = 0,
        BayerGBRG = 0x00000100,
        BayerGRBG = 0x00000200,
        BayerRGGB = 0x00000300,
        BayerBGGR = 0x00000400,

        ColorOrderRGB = 0x00001000, // Default
        ColorOrderBGR = 0x00002000,

        Packed = 0x00004000,
        PackedNoSwap = 0x00008000,

        // Masks.
        BaseFormatMask = 0x000000FF,
        BayerFormatMask = 0x00000F00,
        ColorOrderFormatMask = 0x00003000,
        PackedFormatMask = 0x0000C000,
        StructuralFormatMask = BaseFormatMask | PackedFormatMask,
        EncodingFormatMask = ~BaseFormatMask
    };

    // Format helpers.
    inline static bool isGrayscale (const Format format)
    {
        switch (format & BaseFormatMask)
        {
        case Gray8: case Gray10: case Gray12: case Gray14: case Gray16:
            return true;
        }
        return false;
    }
    inline static bool isColor (const Format format)
    {
        switch (format & BaseFormatMask)
        {
        case Color24: case Color30: case Color32: case HiColor36: case HiColor42: case HiColor48:
            return true;
        }
        return false;
    }

    // Grayscale color table formats.
    enum class ColorTable
    {
        Gray = 0,
        Gray2x = 1,
        GrayInverse = 2,
        Sine = 3,
        Cosine = 4,
        Rainbow = 5,
        Random = 6,
        Contours = 7
    };

    // Standard user interface buttons.
    enum class DialogButton : std::uint64_t
    {
        NoButton = 0,
        Ok = 1 << 0,
        Open = 1 << 1,
        Save = 1 << 2,
        Cancel = 1 << 3,
        Close = 1 << 4,
        Discard = 1 << 5,
        Apply = 1 << 6,
        Reset = 1 << 7,
        RestoreDefaults = 1 << 8,
        Help = 1 << 9,
        SaveAll = 1 << 10,
        Yes = 1 << 11,
        YesToAll = 1 << 12,
        No = 1 << 13,
        NoToAll = 1 << 14,
        Abort = 1 << 15,
        Retry = 1 << 16,
        Ignore = 1 << 17
    };

    union DialogButtons
    {
        DialogButton b;
        std::uint64_t v;

        inline DialogButtons (DialogButton const& a_b = DialogButton::NoButton) : b (a_b) { }
        inline DialogButtons (DialogButtons const& a_b) : b (a_b.b) { }
        inline explicit DialogButtons (const std::uint64_t a_v) : v (a_v) { }

        inline operator bool (void) const
        {
            return !!v;
        }

        inline bool operator! (void) const
        {
            return !v;
        }

        inline DialogButtons operator| (DialogButtons const& other) const
        {
            return DialogButtons(v | other.v);
        }

        inline DialogButtons& operator|= (DialogButtons const& other)
        {
            v |= other.v;
            return *this;
        }
    };
}

inline static ImageView::DialogButtons operator| (const ImageView::DialogButton l, const ImageView::DialogButton r)
{
    return ImageView::DialogButtons(l | r);
}

inline static ImageView::DialogButtons operator| (const ImageView::DialogButton l, ImageView::DialogButtons const& r)
{
    return r | l;
}

inline static ImageView::DialogButtons operator| (ImageView::DialogButtons const& l, ImageView::DialogButtons const& r)
{
    return ImageView::DialogButtons(l.v | r.v);
}

inline static ImageView::DialogButtons operator& (ImageView::DialogButtons const& l, ImageView::DialogButtons const& r)
{
    return ImageView::DialogButtons(l.v & r.v);
}

#endif // INCLUDED__BIT_FLOW__IMAGE__VIEW__H
