/* FILE:        ImageViewAction.cpp
 * DATE:        8/11/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Public display Action class for the ImageView library. Used to interface
 *              with custom ImageView::Window menu and tool-bar entries.
 */

#include "ImageViewAction_Private.h"

#include "ImageViewStringHelpers.h"

ImageView::Action::PrivateData::PrivateData (Action *const a_parent, Window *const a_window, QAction *const a_actionQt, ScopedActionInterface &&a_actionInterface)
    : m_parent (a_parent)
    , m_window (a_window)
    , m_actionQt (a_actionQt)
    , m_actionInterface (std::move(a_actionInterface))
{
    m_actionInterface->setAction(m_parent);

    QObject::connect(m_actionQt, SIGNAL(destroyed()), this, SLOT(actionQtDestroyed()));
    QObject::connect(m_actionQt, SIGNAL(triggered(bool)), this, SLOT(actionQtTriggered(bool)));
}

void ImageView::Action::PrivateData::actionQtDestroyed (void)
{
    auto *const parent = m_parent;

    m_actionInterface->uxDestroyed();

    m_parent = nullptr;
    m_window = nullptr;
    m_actionQt = nullptr;
    m_actionInterface->setAction(nullptr);

    delete parent;
}
void ImageView::Action::PrivateData::actionQtTriggered (const bool checked)
{
    m_actionInterface->triggered(checked);
}

// Construct an action. Generally, ImageView::Actions are not created explcitly, but by one
// of the Window::addAction methods.
ImageView::Action::Action (ImageView::Window *const a_window, QAction *const a_action, ImageView::ScopedActionInterface &&a_scopedActionInterface)
    : m_pd (*new PrivateData(this, a_window, a_action, std::move(a_scopedActionInterface)))
{
}

// Destructor. The action may be deleted from user code, but will otherwise be destroyed
// whenever the interface action is destroyed.
ImageView::Action::~Action (void)
{
    if (m_pd.m_actionQt)
        delete m_pd.m_actionQt;
    delete &m_pd;
}

// Retrieve the containing window handle.
ImageView::Window* ImageView::Action::window (void)
{
    return m_pd.m_window;
}
const ImageView::Window* ImageView::Action::window (void) const
{
    return m_pd.m_window;
}

// Retrieve the Qt action handle for the underlying QAction.
QAction* ImageView::Action::qt_handle (void)
{
    return m_pd.m_actionQt;
}
const QAction* ImageView::Action::qt_handle (void) const
{
    return m_pd.m_actionQt;
}

// Retrieve the AbstractActionInterface.
ImageView::AbstractActionInterface* ImageView::Action::actionInterface (void)
{
    return m_pd.m_actionInterface.get();
}
const ImageView::AbstractActionInterface* ImageView::Action::actionInterface (void) const
{
    return m_pd.m_actionInterface.get();
}

// Action properties. These should only be accessed/modified from the ImageView main thread.
bool ImageView::Action::text (char *const pOutStr, size_t *const pOutStrSize) const
{
    return ImageView::String::set(m_pd.m_actionQt->text().toLocal8Bit().data(), pOutStr, pOutStrSize);
}
void ImageView::Action::setText (const char *const str)
{
    m_pd.m_actionQt->setText(str);
}

bool ImageView::Action::toolTip (char *const pOutStr, size_t *const pOutStrSize) const
{
    return ImageView::String::set(m_pd.m_actionQt->toolTip().toLocal8Bit().data(), pOutStr, pOutStrSize);
}
void ImageView::Action::setToolTip (const char *const str)
{
    m_pd.m_actionQt->setToolTip(str);
}

bool ImageView::Action::isCheckable (void) const
{
    return m_pd.m_actionQt->isCheckable();
}
void ImageView::Action::setCheckable (const bool enabled)
{
    m_pd.m_actionQt->setCheckable(enabled);
}

bool ImageView::Action::isChecked (void) const
{
    return m_pd.m_actionQt->isChecked();
}
void ImageView::Action::setChecked (const bool enabled)
{
    m_pd.m_actionQt->setChecked(enabled);
}

bool ImageView::Action::isEnabled (void) const
{
    return m_pd.m_actionQt->isEnabled();
}
void ImageView::Action::setEnabled (const bool enabled)
{
    m_pd.m_actionQt->setEnabled(enabled);
}
