/* FILE:        ImageViewAppplication.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::Appplication.
 */

#include "ImageViewApplication_Private.h"
#include "ImageViewNative_Private.h"

#include <stdexcept>

#include <string>

#include <QFileInfo>

std::atomic<ImageView::Application*> ImageView::Application::PrivateData::globalApplication (nullptr);

ImageView::Application::PrivateData::PrivateData (const int argc, const char *const argv[])
    : m_idleTimer (nullptr)
{
    // Record our arguments.
    m_argc = argc;
    m_argvVals.resize(argc);
    m_argv.resize(argc);

    std::transform(argv, argv + argc, m_argvVals.begin(), [](const char *const arg) { return ArgvVal(arg, arg + strlen(arg) + 1); });
    std::transform(m_argvVals.begin(), m_argvVals.end(), m_argv.begin(), [](ArgvVal &s) { return s.data(); });

#if defined(_WIN32)

    // Use the location of QtXCore as the Qt plugins path.
#   if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0) && QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
#       if defined(_DEBUG)
    const auto qtCoreDll = GetModuleHandle(TEXT("Qt5Cored.dll"));
#       else
    const auto qtCoreDll = GetModuleHandle(TEXT("Qt5Core.dll"));
#       endif
#   else
#       error Unsupported Qt build version. Please specify the QtXCore DLL name.
#   endif

    if (qtCoreDll)
    {
        CHAR qtCorePath[_MAX_PATH];
        if (GetModuleFileNameA(qtCoreDll, qtCorePath, _MAX_PATH))
            QCoreApplication::addLibraryPath(QFileInfo(qtCorePath).path());
    }

#endif

    // Create the Qt application.
    m_qApplication.reset(new QApplication (m_argc, m_argv.data()));
}

ImageView::Application::PrivateData::~PrivateData (void)
{
    ImageView::Native::Cleanup();
    m_qApplication.reset();
}

// Retrieve the global Application handle. The handle must already have been constructed.
ImageView::Application* ImageView::Application::global (void)
{
    return PrivateData::globalApplication.load();
}

// Construct the application. Throw a runtime_error, if the application already exists.
ImageView::Application::Application (const int argc, const char *const argv[])
{
    Application *expected = nullptr;
    if (!PrivateData::globalApplication.compare_exchange_strong(expected, this))
        throw std::runtime_error(std::string(__FUNCTION__) + ": The Applicaion global has already been allocated.");

    m_pd = new PrivateData(argc, argv);
}

// Deconstructor
ImageView::Application::~Application (void)
{
    PrivateData::globalApplication.store(nullptr);
    delete m_pd;
}

// Add a method to be called whenever the event loop is idle. This should only
// be executed from within the exec thread.
void ImageView::Application::addIdleMethod (void(*idleMethod)(void*), void *const userData)
{
    m_pd->m_idleFunctionPairs.push_back(std::make_pair(idleMethod, userData));

    if (!m_pd->m_idleTimer)
    {
        m_pd->m_idleTimer = new QTimer(m_pd->m_qApplication.get());

        QObject::connect(m_pd->m_idleTimer, &QTimer::timeout, [this]
            {
                for (auto const& idleFuncPair : m_pd->m_idleFunctionPairs)
                    idleFuncPair.first(idleFuncPair.second);
            });
        
        m_pd->m_idleTimer->setSingleShot(false);
        m_pd->m_idleTimer->start(0);
    }
}

// Whether or not we should exit exec when the last window is closed.
bool ImageView::Application::quitOnLastWindowClosed (void) const
{
    return m_pd->m_qApplication->quitOnLastWindowClosed();
}
void ImageView::Application::setQuitOnLastWindowClosed (const bool enabled)
{
    m_pd->m_qApplication->setQuitOnLastWindowClosed(enabled);
}

// Execute the application main loop. This will not return until the application
// is ready to exit. This should generally only be executed in the main thread.
int ImageView::Application::exec (void)
{
    return m_pd->m_qApplication->exec();
}

// Exit the application main loop (the exec method), returning the given return code.
void ImageView::Application::exit (const int returnCode)
{
    m_pd->m_qApplication->exit(returnCode);
}
