/* FILE:        ImageViewCApi.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of the ImageView C API.
 */

#include "ImageViewCApi_Private.h"

#include "ImageViewApplication.h"
#include "ImageViewWindow.h"
#include "ImageViewAction.h"
#include "ImageViewMessageBoxDialog.h"

#include "ImageViewStringHelpers.h"

ImageView::Application *g_application = nullptr;

//// ImageView Application methods.
IVRC IVAppCreate (const IVint32 argc, const IVchar *const argv[], IVApp *const pApp)
{
    if (g_application)
        return IV_ERR_APPLICATION_ALREADY_CREATED;

    if (!argv)
        return IV_ERR_NULL_ARGUMENT;

    if (!pApp)
        return IV_ERR_NULL_HANDLE_REF;

    try
    {
        g_application = new ImageView::Application(argc, argv);
        *pApp = reinterpret_cast<IVApp>(g_application);
    }
    catch (...)
    {
        g_application = nullptr;
        return IV_ERR_APPLICATION_CREATION_FAILED;
    }

    return IV_OK;
}
IVRC IVAppDestroy (IVApp app)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (reinterpret_cast<ImageView::Application*>(app) != g_application)
        return IV_ERR_BAD_HANDLE;

    try
    {
        delete g_application;
    }
    catch (...)
    {
    }
    g_application = nullptr;

    return IV_OK;
}

// Retrieve the global ImageView Application handle.
IVRC IVAppGetGlobal (IVApp *const pApp)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!pApp)
        return IV_ERR_NULL_HANDLE_REF;

    *pApp = reinterpret_cast<IVApp>(g_application);

    return IV_OK;
}

IVRC IVAppAddIdleMethod (IVApp app, const IVAppIdleMethod method, IVUserData userData)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (reinterpret_cast<ImageView::Application*>(app) != g_application)
        return IV_ERR_BAD_HANDLE;

    try
    {
        g_application->addIdleMethod(method, userData);
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Enter the application main loop. Does not return until the application exits. pReturnCode may be NULL.
IVRC IVAppExec (IVApp app, IVint32 *const pReturnCode)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (reinterpret_cast<ImageView::Application*>(app) != g_application)
        return IV_ERR_BAD_HANDLE;

    int returnCode;
    try
    {
        returnCode = g_application->exec();
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    if (pReturnCode)
        *pReturnCode = returnCode;

    return IV_OK;
}

// Exit the application on call, returning the specified return code.
IVRC IVAppExit (IVApp app, const IVint32 returnCode)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (reinterpret_cast<ImageView::Application*>(app) != g_application)
        return IV_ERR_BAD_HANDLE;

    try
    {
        g_application->exit(returnCode);
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

//// ImageView Window methods.
IVRC IVWindowCreate (IVApp app, IVWindow *const pWindow, IVUserData windowUserData)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (reinterpret_cast<ImageView::Application*>(app) != g_application)
        return IV_ERR_BAD_HANDLE;

    if (!pWindow)
        return IV_ERR_NULL_HANDLE_REF;

    try
    {
        auto acqIf = ImageView::make_scoped(new IVAcqIfClass(windowUserData));
        auto fileIf = ImageView::make_scoped(new IVFileIfClass(windowUserData));

        *pWindow = IVWindowWrapper::Make(std::move(acqIf), std::move(fileIf));
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}
IVRC IVWindowDestroy (IVWindow window)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!window)
        return IV_ERR_NULL_HANDLE;

    try
    {
        IVWindowWrapper::Destroy(window);
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

IVbool IVWindowIsGood (IVWindow window)
{
    if (!g_application)
        return IV_FALSE;

    if (!window)
        return IV_FALSE;

    try
    {
        return IVWindowWrapper::Get(window).good();
    }
    catch (...)
    {
    }

    return IV_FALSE;
}

// Retrieve the native window handle.
IVRC IVWindowGetNativeHandle (IVWindow window, IMAGE_VIEW_NATIVE_WINDOW *const pNativeHandle)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!window)
        return IV_ERR_NULL_HANDLE;

    if (!pNativeHandle)
        return IV_ERR_NULL_ARGUMENT;

    try
    {
        *pNativeHandle = IVWindowWrapper::Get(window).window().native_handle();
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Retrieve the window User Data handle.
IVRC IVWindowGetUserData (IVWindow window, IVUserData *const pWindowUserData)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!window)
        return IV_ERR_NULL_HANDLE;

    if (!pWindowUserData)
        return IV_ERR_NULL_ARGUMENT;

    try
    {
        *pWindowUserData = IVWindowWrapper::Get(window).acqIf().windowUserData();
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Set the next image frame to display. If nextFrame is IV_NULL, any pending frame buffer will be cleared.
IVRC IVWindowSetNextFrame (IVWindow window, IVRawImage *const nextFrame)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!window)
        return IV_ERR_NULL_HANDLE;

    try
    {
        if (nextFrame)
        {
            const ImageView::RawImage rawImage =
            {
                nextFrame->buffer,
                nextFrame->width,
                nextFrame->height,
                (ImageView::Format)nextFrame->format,
                nextFrame->userData
            };

            IVWindowWrapper::Get(window).acqIf().setNextFrame(rawImage);
        }
        else
            IVWindowWrapper::Get(window).acqIf().clearNextFrame();
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Add a setup mode. The description may be any valid string, but the name must be a unique string.
IVRC IVWindowAddSetupMode (IVWindow window, const char *const name, const char *const description)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!window)
        return IV_ERR_NULL_HANDLE;

    try
    {
        if (!IVWindowWrapper::Get(window).acqIf().addSetupMode(name, description))
            return IV_ERR_WINDOW_BAD_SETUP_MODE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Clear the list of setup modes.
IVRC IVWindowClearSetupModes (IVWindow window)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!window)
        return IV_ERR_NULL_HANDLE;

    try
    {
        IVWindowWrapper::Get(window).acqIf().clearSetupModes();
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

//// ImageView File Interface methods.

// Add a file save format with the given primary extension, optional alternate extension(s) (space
// separated), and optional description. The primary and alternate extensions must be unique,
// vis-a-vis all other formats. Extensions may optionally start with a period, and should otherwise
// contain only uppercase and lowercase letters, and decimal digits. Extension case is preserved
// internally, but comparisons are case insensitive. alternateExts and desc may be NULL.
IVRC IVWindowAddFileFormat (IVWindow window, const IVchar *const primaryExt, const IVchar *const alternateExts, const IVchar *const desc)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!window)
        return IV_ERR_NULL_HANDLE;

    try
    {
        if (!IVWindowWrapper::Get(window).fileIf().addFormat(primaryExt, alternateExts, desc))
            return IV_ERR_WINDOW_BAD_FILE_FORMAT;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Clear the list of custom file formats.
IVRC IVWindowClearFileFormats (IVWindow window)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!window)
        return IV_ERR_NULL_HANDLE;

    try
    {
        IVWindowWrapper::Get(window).fileIf().clearFormats();
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Add a new GUI action to the window at the given anchor point. parentMenuName should specify a
// top-level menu entry name for IV_WINDOW_ACTION_MAIN_MENU, but IV_NULL for any other anchor point.
IVRC IVWindowAddAction (IVWindow window, const IVWindowActionAnchor anchor, const IVchar *const parentMenuName, IVAction *const pAction, IVUserData actionUserData)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!window)
        return IV_ERR_NULL_HANDLE;
    if (!pAction)
        return IV_ERR_NULL_HANDLE_REF;

    ImageView::Action *newAction;
    try
    {
        auto accessor = IVWindowWrapper::Get(window);
        auto &window = accessor.window();
        const auto windowUserData = accessor.acqIf().windowUserData();

        switch (anchor)
        {
        case IV_WINDOW_ACTION_MAIN_MENU:
            if (!parentMenuName)
                return IV_ERR_WINDOW_ACTION_NULL_MENU_NAME;

            newAction = window.addMainMenuAction(ImageView::make_scoped(new IVActionIfClass(windowUserData)), parentMenuName);
            break;

        case IV_WINDOW_ACTION_TOOL_BAR:
            if (parentMenuName)
                return IV_ERR_WINDOW_ACTION_UNEXPECTED_MENU_NAME;

            newAction = window.addToolBarAction(ImageView::make_scoped(new IVActionIfClass(windowUserData)));
            break;

        case IV_WINDOW_ACTION_IMAGE_STATS:
            if (parentMenuName)
                return IV_ERR_WINDOW_ACTION_UNEXPECTED_MENU_NAME;

            newAction = window.addImageStatsAction(ImageView::make_scoped(new IVActionIfClass(windowUserData)));
            break;

        default:
            return IV_ERR_WINDOW_ACTION_UNKNOWN_ANCHOR;
        }
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    *pAction = reinterpret_cast<IVAction>(newAction);

    return IV_OK;
}


//// ImageView Action methods.

// Retrieve the action User Data handle.
IVRC IVActionGetUserData (IVAction action, IVUserData *const pActionUserData)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!action)
        return IV_ERR_NULL_HANDLE;

    if (!pActionUserData)
        return IV_ERR_NULL_ARGUMENT;

    const auto *const base = reinterpret_cast<const ImageView::BaseClass*>(action);

    try
    {
        if (const auto *const action = dynamic_cast<const ImageView::Action*>(base))
        {
            const auto *const actionIf = static_cast<const IVActionIfClass*>(action->actionInterface());
            *pActionUserData = actionIf->userData();
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Destroy the action. Actions will be automatically destroyed when the parent window closes, if
// not explicitly destroyed.
IVRC IVActionDestroy (IVAction action)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!action)
        return IV_ERR_NULL_HANDLE;

    const auto *const base = reinterpret_cast<const ImageView::BaseClass*>(action);

    try
    {
        if (const auto *const action = dynamic_cast<const ImageView::Action*>(base))
            delete action;
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}


//// ImageView MessageBox methods.

// Show a message box of the given type as a modal dialog of the given parent window. On success,
// pUserButton returns the IVButton pressed by the user. pUserButton may be IV_NULL.
IVRC IVMessageBoxDialogShow (const IVMessageBoxDialogType type, IVWindow parentWindow, const IVchar *const title, const IVchar *const text, const IVint32 buttons, const IVDialogButton defaultButton, IVDialogButton *const pUserButton)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!parentWindow)
        return IV_ERR_NULL_HANDLE;

    if (!title || !text)
        return IV_ERR_NULL_ARGUMENT;

    if (buttons > IV_DIALOG_BUTTON_IGNORE)
        return IV_ERR_DIALOG_BUTTON_UNKNOWN_BUTTONS;

    switch (defaultButton)
    {
    default:
        return IV_ERR_DIALOG_BUTTON_BAD_DEFAULT;
    case IV_DIALOG_BUTTON_NO_BUTTON:
    case IV_DIALOG_BUTTON_OK:
    case IV_DIALOG_BUTTON_OPEN:
    case IV_DIALOG_BUTTON_SAVE:
    case IV_DIALOG_BUTTON_CANCEL:
    case IV_DIALOG_BUTTON_CLOSE:
    case IV_DIALOG_BUTTON_DISCARD:
    case IV_DIALOG_BUTTON_APPLY:
    case IV_DIALOG_BUTTON_RESET:
    case IV_DIALOG_BUTTON_RESTORE_DEFAULTS:
    case IV_DIALOG_BUTTON_HELP:
    case IV_DIALOG_BUTTON_SAVE_ALL:
    case IV_DIALOG_BUTTON_YES:
    case IV_DIALOG_BUTTON_YES_TO_ALL:
    case IV_DIALOG_BUTTON_NO:
    case IV_DIALOG_BUTTON_NO_TO_ALL:
    case IV_DIALOG_BUTTON_ABORT:
    case IV_DIALOG_BUTTON_RETRY:
    case IV_DIALOG_BUTTON_IGNORE:
        break;
    }

    try
    {
        auto &window = IVWindowWrapper::Get(parentWindow).window();

        ImageView::DialogButton userButton;
        switch (type)
        {
        default:
            return IV_ERR_MESSAGE_BOX_DIALOG_TYPE_UNKNOWN;

        case IV_MESSAGE_TYPE_CRITICAL:
            userButton = ImageView::MessageBoxDialog::critical(&window, title, text, static_cast<ImageView::DialogButtons>(buttons), static_cast<ImageView::DialogButton>(defaultButton));
            break;
        case IV_MESSAGE_TYPE_WARNING:
            userButton = ImageView::MessageBoxDialog::warning(&window, title, text, static_cast<ImageView::DialogButtons>(buttons), static_cast<ImageView::DialogButton>(defaultButton));
            break;
        case IV_MESSAGE_TYPE_INFORMATION:
            userButton = ImageView::MessageBoxDialog::information(&window, title, text, static_cast<ImageView::DialogButtons>(buttons), static_cast<ImageView::DialogButton>(defaultButton));
            break;
        case IV_MESSAGE_TYPE_QUESTION:
            userButton = ImageView::MessageBoxDialog::question(&window, title, text, static_cast<ImageView::DialogButtons>(buttons), static_cast<ImageView::DialogButton>(defaultButton));
            break;
        }

        if (pUserButton)
            *pUserButton = static_cast<IVDialogButton>(userButton);
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}


//// Generic signaling methods.

// Issue a signal asynchronously.
IVRC IVSendSignal (IVHandle hStruct, const IVint32 signalCode)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (signalCode)
            {
            default:
                return IV_ERR_WINDOW_SIGNAL_CODE_UNKNOWN;

            case IV_WINDOW_RAISE:
                accessor.window().raise();
                break;
            case IV_WINDOW_SET_RUNNING:
                accessor.window().setRunning();
                break;
            case IV_WINDOW_SET_CONFIGURED:
                accessor.window().setConfigured();
                break;
            case IV_WINDOW_SET_UNCONFIGURED:
                accessor.window().setUnconfigured();
                break;

            case IV_WINDOW_FEATURE_SET_CHANGE:
                accessor.acqIf().signalFeatureSetChange();
                break;
            case IV_WINDOW_ACQUISITION_STATE_CHANGE:
                accessor.acqIf().signalAcquisitionStateChange();
                break;
            case IV_WINDOW_GET_FRAME_NOW:
                accessor.acqIf().signalGetFrameNow();
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}


//// Generic property methods.

// Get a boolean property from the structure.
IVRC IVGetPropBool (IVHandle hStruct, const IVint32 prop, IVbool *const pEnabled)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    if (!pEnabled)
        return IV_ERR_NULL_ARGUMENT;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        IVbool bEnabled;

        if (base == g_application)
        {
            switch (prop)
            {
            default:
                return IV_ERR_APPLICATION_PROPERTY_CODE_UNKNOWN;

            case IV_APP_QUIT_ON_LAST_WINDOW_CLOSED:
                bEnabled = g_application->quitOnLastWindowClosed();
                break;
            }
        }
        else if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (prop)
            {
            default:
                return IV_ERR_WINDOW_PROPERTY_CODE_UNKNOWN;

            case IV_WINDOW_PREVIEW_ZOOM_LEVEL:
                bEnabled = accessor.window().previewZoomLevel();
                break;
            case IV_WINDOW_PREVIEW_IMAGE_DIMENSIONS:
                bEnabled = accessor.window().previewImageDimensions();
                break;
            case IV_WINDOW_PREVIEW_IMAGE_FORMAT:
                bEnabled = accessor.window().previewImageFormat();
                break;
            case IV_WINDOW_PREVIEW_PIXEL_VALUE:
                bEnabled = accessor.window().previewPixelValue();
                break;
            case IV_WINDOW_PREVIEW_FRAME_RATE:
                bEnabled = accessor.window().previewFrameRate();
                break;
            case IV_WINDOW_PREVIEW_BUFFERS_FILLED:
                bEnabled = accessor.window().previewBuffersFilled();
                break;
            case IV_WINDOW_PREVIEW_BUFFERS_DROPPED:
                bEnabled = accessor.window().previewBuffersDropped();
                break;
            case IV_WINDOW_PREVIEW_BUFFERS_DELIVERED:
                bEnabled = accessor.window().previewBuffersDelivered();
                break;
            case IV_WINDOW_PREVIEW_BUFFERS_QUEUED:
                bEnabled = accessor.window().previewBuffersQueued();
                break;
            case IV_WINDOW_PREVIEW_SETUP_MODE:
                bEnabled = accessor.window().previewSetupMode();
                break;
            case IV_WINDOW_KEEP_IMAGE_INFO_VISIBLE:
                bEnabled = accessor.window().keepImageInfoVisible();
                break;
            case IV_WINDOW_BAYER_DEMOSAIC_ENABLED:
                bEnabled = accessor.window().bayerDemosaicEnabled();
                break;
            case IV_WINDOW_RGB_ORDER_REVERSED:
                bEnabled = accessor.window().rgbOrderReversed();
                break;
            case IV_WINDOW_PIXEL_CROSSHAIR_VISIBLE:
                bEnabled = accessor.window().pixelCrosshairVisible();
                break;
            case IV_WINDOW_VISIBLE:
                bEnabled = accessor.window().isVisible();
                break;
            case IV_WINDOW_CLOSEABLE:
                bEnabled = accessor.window().closeable();
                break;
            case IV_WINDOW_FULL_SCREEN:
                bEnabled = accessor.window().isFullScreen();
                break;
            case IV_WINDOW_MAXIMIZED:
                bEnabled = accessor.window().isMaximized();
                break;
            case IV_WINDOW_MINIMIZED:
                bEnabled = accessor.window().isMinimized();
                break;
            case IV_WINDOW_ZOOM_TO_FIT:
                bEnabled = accessor.window().zoomToFit();
                break;
            case IV_WINDOW_REFRESH_ENABLED:
                bEnabled = accessor.window().refreshEnabled();
                break;
            case IV_WINDOW_REFRESH_RATE_AUTO:
                bEnabled = accessor.window().refreshRateAuto();
                break;
            case IV_WINDOW_ACQUISITION_FRAMES_LOCKED:
                bEnabled = accessor.acqIf().framesLocked();
                break;
            }
        }
        else if (const auto *const action = dynamic_cast<ImageView::Action*>(base))
        {
            switch (prop)
            {
            default:
                return IV_ERR_ACTION_PROPERTY_CODE_UNKNOWN;

            case IV_ACTION_CHECKABLE:
                bEnabled = action->isCheckable();
                break;
            case IV_ACTION_CHECKED:
                bEnabled = action->isChecked();
                break;
            case IV_ACTION_ENABLED:
                bEnabled = action->isEnabled();
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;

        *pEnabled = bEnabled;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Get an int32 property from the structure.
IVRC IVGetPropInt32 (IVHandle hStruct, const IVint32 prop, IVint32 *const pValue)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    if (!pValue)
        return IV_ERR_NULL_ARGUMENT;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        IVint32 iValue;

        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (prop)
            {
            default:
                return IV_ERR_WINDOW_PROPERTY_CODE_UNKNOWN;

            case IV_WINDOW_POS_X:
                iValue = accessor.window().x();
                break;
            case IV_WINDOW_POS_Y:
                iValue = accessor.window().y();
                break;
            case IV_WINDOW_SIZE_WIDTH:
                iValue = accessor.window().width();
                break;
            case IV_WINDOW_SIZE_HEIGHT:
                iValue = accessor.window().height();
                break;
            case IV_WINDOW_BAYER_FILTER_OVERRIDE:
                iValue = (IVint32)accessor.window().bayerFilterOverride();
                break;
            case IV_WINDOW_COLOR_TABLE:
                iValue = (IVint32)accessor.window().colorTable();
                break;
            case IV_WINDOW_REFRESH_RATE:
                iValue = accessor.window().refreshRate();
                break;
            case IV_WINDOW_IMAGE_OFFSET_X:
                iValue = accessor.window().imageOffsetX();
                break;
            case IV_WINDOW_IMAGE_OFFSET_Y:
                iValue = accessor.window().imageOffsetY();
                break;
#if defined(_WIN32)
            case IV_WINDOW_APPLICATION_ABOUT_ICON_RC:
                iValue = accessor.acqIf().windowApplicationAboutIconRc();
                break;
#endif
            case IV_WINDOW_START_FIXED_MIN:
                iValue = accessor.acqIf().startFixedMin();
                break;
            case IV_WINDOW_START_FIXED_MAX:
                iValue = accessor.acqIf().startFixedMax();
                break;
            case IV_WINDOW_START_FIXED_INC:
                iValue = accessor.acqIf().startFixedInc();
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;

        *pValue = iValue;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Get a double floating point property from the structure.
IVRC IVGetPropDouble (IVHandle hStruct, const IVint32 prop, IVdouble *const pValue)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    if (!pValue)
        return IV_ERR_NULL_ARGUMENT;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        IVdouble iValue;

        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (prop)
            {
            default:
                return IV_ERR_WINDOW_PROPERTY_CODE_UNKNOWN;

            case IV_WINDOW_ZOOM_LEVEL:
                iValue = accessor.window().zoomLevel();
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;

        *pValue = iValue;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Retrieve a c-string property from the structure.
//   pStrBuf    [OUT] The output string buffer. This may be IV_NULL, to retrieve just the string length. If
//                    insufficiently long, filled with as much data as will fit.
//   pSize      [IN]  The length of pStrBuf, if that is not IV_NULL.
//              [OUT] The length of the complete string, including null-terminator.
IVRC IVGetPropCStr (IVHandle hStruct, const IVint32 prop, IVchar *const pStrBuf, IVsize_t *const pSize)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    if (!pSize)
        return IV_ERR_NULL_ARGUMENT;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (prop)
            {
            default:
                return IV_ERR_WINDOW_PROPERTY_CODE_UNKNOWN;

            case IV_WINDOW_TITLE:
                if (!accessor.window().title(pStrBuf, pSize))
                    return IV_ERR_NULL_ARGUMENT;
                break;
            case IV_WINDOW_APPLICATION_ABOUT_TITLE:
                if (!accessor.window().applicationAboutTitle(pStrBuf, pSize))
                    return IV_ERR_NULL_ARGUMENT;
                break;
            case IV_WINDOW_APPLICATION_ABOUT_DESCRIPTION:
                if (!accessor.window().applicationAboutDescription(pStrBuf, pSize))
                    return IV_ERR_NULL_ARGUMENT;
                break;
            case IV_WINDOW_RESTORE_GEOMETRY:
                if (!accessor.window().saveGeometry(pStrBuf, pSize))
                    return IV_ERR_NULL_ARGUMENT;
                break;
            }
        }
        else if (const auto *const action = dynamic_cast<ImageView::Action*>(base))
        {
            switch (prop)
            {
            default:
                return IV_ERR_ACTION_PROPERTY_CODE_UNKNOWN;

            case IV_ACTION_TEXT:
                if (!action->text(pStrBuf, pSize))
                    return IV_ERR_NULL_ARGUMENT;
                break;
            case IV_ACTION_TOOL_TIP:
                if (!action->toolTip(pStrBuf, pSize))
                    return IV_ERR_NULL_ARGUMENT;
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Retrieve a data property from the structure.
//   pDataBuf   [OUT] The output data buffer. This may be IV_NULL, to retrieve just the data size. No data is
//                    returned, if the buffer size is inadequate.
//   pDataSize  [IN]  The length of pDataBuf in bytes, if that is not IV_NULL.
//              [OUT] The exact data size in bytes.
IVRC IVGetPropData (IVHandle hStruct, const IVint32 prop, IVvoid *const pDataBuf, IVsize_t *const pDataSize)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    if (!pDataSize)
        return IV_ERR_NULL_ARGUMENT;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (prop)
            {
            default:
                return IV_ERR_WINDOW_PROPERTY_CODE_UNKNOWN;

            case IV_WINDOW_APPLICATION_ABOUT_ICON_XPM:
                if (!accessor.acqIf().getWindowApplicationAboutIconXPMData(pDataBuf, pDataSize))
                    return IV_ERR_BUFFER_TOO_SMALL;
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Set a boolean property on the structure.
IVRC IVSetPropBool (IVHandle hStruct, const IVint32 prop, const IVbool enabled)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        const bool bEnabled = !!enabled;

        if (base == g_application)
        {
            switch (prop)
            {
            default:
                return IV_ERR_APPLICATION_PROPERTY_CODE_UNKNOWN;

            case IV_APP_QUIT_ON_LAST_WINDOW_CLOSED:
                g_application->setQuitOnLastWindowClosed(bEnabled);
                break;
            }
        }
        else if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (prop)
            {
            default:
                return IV_ERR_WINDOW_PROPERTY_CODE_UNKNOWN;

            case IV_WINDOW_PREVIEW_ZOOM_LEVEL:
                accessor.window().setPreviewZoomLevel(bEnabled);
                break;
            case IV_WINDOW_PREVIEW_IMAGE_DIMENSIONS:
                accessor.window().setPreviewImageDimensions(bEnabled);
                break;
            case IV_WINDOW_PREVIEW_IMAGE_FORMAT:
                accessor.window().setPreviewImageFormat(bEnabled);
                break;
            case IV_WINDOW_PREVIEW_PIXEL_VALUE:
                accessor.window().setPreviewPixelValue(bEnabled);
                break;
            case IV_WINDOW_PREVIEW_FRAME_RATE:
                accessor.window().setPreviewFrameRate(bEnabled);
                break;
            case IV_WINDOW_PREVIEW_BUFFERS_FILLED:
                accessor.window().setPreviewBuffersFilled(bEnabled);
                break;
            case IV_WINDOW_PREVIEW_BUFFERS_DROPPED:
                accessor.window().setPreviewBuffersDropped(bEnabled);
                break;
            case IV_WINDOW_PREVIEW_BUFFERS_DELIVERED:
                accessor.window().setPreviewBuffersDelivered(bEnabled);
                break;
            case IV_WINDOW_PREVIEW_BUFFERS_QUEUED:
                accessor.window().setPreviewBuffersQueued(bEnabled);
                break;
            case IV_WINDOW_PREVIEW_SETUP_MODE:
                accessor.window().setPreviewSetupMode(bEnabled);
                break;
            case IV_WINDOW_KEEP_IMAGE_INFO_VISIBLE:
                accessor.window().setKeepImageInfoVisible(bEnabled);
                break;
            case IV_WINDOW_BAYER_DEMOSAIC_ENABLED:
                accessor.window().setBayerDemosaicEnabled(bEnabled);
                break;
            case IV_WINDOW_RGB_ORDER_REVERSED:
                accessor.window().setRgbOrderReversed(bEnabled);
                break;
            case IV_WINDOW_PIXEL_CROSSHAIR_VISIBLE:
                accessor.window().setPixelCrosshairVisible(bEnabled);
                break;
            case IV_WINDOW_VISIBLE:
                accessor.window().setVisible(bEnabled);
                break;
            case IV_WINDOW_CLOSEABLE:
                accessor.window().setCloseable(bEnabled);
                break;
            case IV_WINDOW_FULL_SCREEN:
                accessor.window().setFullScreen(bEnabled);
                break;
            case IV_WINDOW_MAXIMIZED:
                if (bEnabled)
                    accessor.window().maximize();
                else
                    accessor.window().restore();
                break;
            case IV_WINDOW_MINIMIZED:
                if (bEnabled)
                    accessor.window().minimize();
                else
                    accessor.window().restore();
                break;
            case IV_WINDOW_ZOOM_TO_FIT:
                accessor.window().setZoomToFit(bEnabled);
                break;
            case IV_WINDOW_REFRESH_ENABLED:
                accessor.window().setRefreshEnabled(bEnabled);
                break;
            case IV_WINDOW_REFRESH_RATE_AUTO:
                accessor.window().setRefreshRateAuto(bEnabled);
                break;
            case IV_WINDOW_ACQUISITION_FRAMES_LOCKED:
                accessor.acqIf().setFramesLocked(bEnabled);
                break;
            }
        }
        else if (auto *const action = dynamic_cast<ImageView::Action*>(base))
        {
            switch (prop)
            {
            default:
                return IV_ERR_ACTION_PROPERTY_CODE_UNKNOWN;

            case IV_ACTION_CHECKABLE:
                action->setCheckable(bEnabled);
                break;
            case IV_ACTION_CHECKED:
                action->setChecked(bEnabled);
                break;
            case IV_ACTION_ENABLED:
                action->setEnabled(bEnabled);
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Set an int32 property on the structure.
IVRC IVSetPropInt32 (IVHandle hStruct, const IVint32 prop, const IVint32 value)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (prop)
            {
            default:
                return IV_ERR_WINDOW_PROPERTY_CODE_UNKNOWN;

            case IV_WINDOW_POS_X:
                accessor.window().move(value, accessor.window().y());
                break;
            case IV_WINDOW_POS_Y:
                accessor.window().move(accessor.window().x(), value);
                break;
            case IV_WINDOW_SIZE_WIDTH:
                accessor.window().resize(value, accessor.window().height());
                break;
            case IV_WINDOW_SIZE_HEIGHT:
                accessor.window().resize(accessor.window().width(), value);
                break;
            case IV_WINDOW_BAYER_FILTER_OVERRIDE:
                switch (value)
                {
                default:
                    return IV_ERR_WINDOW_BAYER_FILTER_UNKNOWN;
                case ImageView::Unencoded:
                case ImageView::BayerRGGB:
                case ImageView::BayerGRBG:
                case ImageView::BayerBGGR:
                case ImageView::BayerGBRG:
                    break;
                }
                accessor.window().setBayerFilterOverride((ImageView::Format)value);
                break;
            case IV_WINDOW_COLOR_TABLE:
                switch ((ImageView::ColorTable)value)
                {
                default:
                    return IV_ERR_WINDOW_COLOR_TABLE_UNKNOWN;
                case ImageView::ColorTable::Gray:
                case ImageView::ColorTable::Gray2x:
                case ImageView::ColorTable::GrayInverse:
                case ImageView::ColorTable::Sine:
                case ImageView::ColorTable::Cosine:
                case ImageView::ColorTable::Rainbow:
                case ImageView::ColorTable::Random:
                case ImageView::ColorTable::Contours:
                    break;
                }
                accessor.window().setColorTable((ImageView::ColorTable)value);
                break;
            case IV_WINDOW_REFRESH_RATE:
                accessor.window().setRefreshRate(value);
                break;
            case IV_WINDOW_IMAGE_OFFSET_X:
                accessor.window().setImageOffset(value, accessor.window().imageOffsetY());
                break;
            case IV_WINDOW_IMAGE_OFFSET_Y:
                accessor.window().setImageOffset(accessor.window().imageOffsetX(), value);
                break;
#if defined(_WIN32)
            case IV_WINDOW_APPLICATION_ABOUT_ICON_RC:
                accessor.acqIf().setWindowApplicationAboutIconRc(value);
                break;
#endif
            case IV_WINDOW_START_FIXED_MIN:
                accessor.acqIf().setStartFixedMin(value);
                break;
            case IV_WINDOW_START_FIXED_MAX:
                accessor.acqIf().setStartFixedMax(value);
                break;
            case IV_WINDOW_START_FIXED_INC:
                accessor.acqIf().setStartFixedInc(value);
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Set a double floating point property on the structure.
IVRC IVSetPropDouble (IVHandle hStruct, const IVint32 prop, const IVdouble value)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (prop)
            {
            default:
                return IV_ERR_WINDOW_PROPERTY_CODE_UNKNOWN;

            case IV_WINDOW_ZOOM_LEVEL:
                accessor.window().setZoomLevel(value);
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Set a string property on the structure.
IVRC IVSetPropCStr (IVHandle hStruct, const IVint32 prop, const IVchar *const strBuf)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    if (!strBuf)
        return IV_ERR_NULL_ARGUMENT;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (prop)
            {
            default:
                return IV_ERR_WINDOW_PROPERTY_CODE_UNKNOWN;

            case IV_WINDOW_TITLE:
                accessor.window().setTitle(strBuf);
                break;

            case IV_WINDOW_APPLICATION_ABOUT_TITLE:
            case IV_WINDOW_APPLICATION_ABOUT_DESCRIPTION:
                {
                    ImageView::Window &window = accessor.window();

#if defined(_WIN32)
                    const std::int32_t iconRc = accessor.acqIf().windowApplicationAboutIconRc();
                    HICON hIconBig = (HICON)LoadImage(GetModuleHandle(nullptr), MAKEINTRESOURCE(iconRc), IMAGE_ICON, 256, 256, LR_DEFAULTSIZE | LR_LOADTRANSPARENT);
                    HICON hIconMed = (HICON)LoadImage(GetModuleHandle(nullptr), MAKEINTRESOURCE(iconRc), IMAGE_ICON, 48, 48, LR_DEFAULTSIZE | LR_LOADTRANSPARENT);
                    HICON hIconSmall = (HICON)LoadImage(GetModuleHandle(nullptr), MAKEINTRESOURCE(iconRc), IMAGE_ICON, 16, 16, LR_DEFAULTSIZE | LR_LOADTRANSPARENT);

                    if (hIconBig || hIconMed || hIconSmall)
                    {
                        try
                        {
                            if (IV_WINDOW_APPLICATION_ABOUT_TITLE == prop)
                            {
                                const auto descStr = ImageView::String::read(&ImageView::Window::applicationAboutDescription, &window);
                                window.setApplicationAbout(strBuf, descStr.c_str(), hIconBig, hIconMed, hIconSmall);
                            }
                            else
                            {
                                const auto titleStr = ImageView::String::read(&ImageView::Window::applicationAboutTitle, &window);
                                window.setApplicationAbout(titleStr.c_str(), strBuf, hIconBig, hIconMed, hIconSmall);
                            }
                        }
                        catch (...)
                        {
                            DestroyIcon(hIconBig);
                            DestroyIcon(hIconMed);
                            DestroyIcon(hIconSmall);
                            throw;
                        }

                        DestroyIcon(hIconBig);
                        DestroyIcon(hIconMed);
                        DestroyIcon(hIconSmall);

                        break;
                    }
#endif
                    const auto xpmIcon = accessor.acqIf().windowApplicationAboutIconXPM();

                    if (IV_WINDOW_APPLICATION_ABOUT_TITLE == prop)
                    {
                        const auto descStr = ImageView::String::read(&ImageView::Window::applicationAboutDescription, &window);
                        window.setApplicationAbout(strBuf, descStr.c_str(), xpmIcon, xpmIcon, xpmIcon);
                    }
                    else
                    {
                        const auto titleStr = ImageView::String::read(&ImageView::Window::applicationAboutTitle, &window);
                        window.setApplicationAbout(titleStr.c_str(), strBuf, xpmIcon, xpmIcon, xpmIcon);
                    }
                }
                break;

            case IV_WINDOW_RESTORE_GEOMETRY:
                if (!accessor.window().restoreGeometry(strBuf))
                    return IV_ERR_WINDOW_BAD_GEOMETRY;
                break;
            }
        }
        else if (auto *const action = dynamic_cast<ImageView::Action*>(base))
        {
            switch (prop)
            {
            default:
                return IV_ERR_ACTION_PROPERTY_CODE_UNKNOWN;

            case IV_ACTION_TEXT:
                action->setText(strBuf);
                break;
            case IV_ACTION_TOOL_TIP:
                action->setToolTip(strBuf);
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Set a data property on the structure.
IVRC IVSetPropData (IVHandle hStruct, const IVint32 prop, const IVvoid *const dataBuf, const IVsize_t dataSize)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            switch (prop)
            {
            default:
                return IV_ERR_WINDOW_PROPERTY_CODE_UNKNOWN;

            case IV_WINDOW_APPLICATION_ABOUT_ICON_XPM:
                if (!accessor.acqIf().setWindowApplicationAboutIconXPMData(dataBuf, dataSize))
                    return IV_ERR_NULL_ARGUMENT;
                break;
            }
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

//// Generic feature methods.
template <typename FeatureEnumT, typename FeatureMethodT>
IVRC IVSetFeature_AcqIf_Private (IVHandle hStruct, const IVint32 feature, FeatureMethodT method)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            if (!accessor.acqIf().setFeature((FeatureEnumT)feature, method))
                return IV_ERR_WINDOW_FEATURE_CODE_UNKNOWN;
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

template <typename FeatureEnumT, typename FeatureMethodT>
IVRC IVSetFeature_FileIf_Private (IVHandle hStruct, const IVint32 feature, FeatureMethodT method)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            if (!accessor.fileIf().setFeature((FeatureEnumT)feature, method))
                return IV_ERR_WINDOW_FEATURE_CODE_UNKNOWN;
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    {
        return IV_ERR_WINDOW_NOT_GOOD;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

template <typename FeatureEnumT, typename FeatureMethodT>
IVRC IVSetFeature_ActionIf_Private (IVHandle hStruct, const IVint32 feature, FeatureMethodT method)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (auto *const action = dynamic_cast<ImageView::Action*>(base))
        {
            auto *const actionIf = static_cast<IVActionIfClass*>(action->actionInterface());
            if (!actionIf->setFeature((FeatureEnumT)feature, method))
                return IV_ERR_ACTION_FEATURE_CODE_UNKNOWN;
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

template <typename AcqIfFeatureEnumT, typename ActionIfFeatureEnumT, typename FeatureMethodT>
IVRC IVSetFeature_AcqIf_ActionIf_Private (IVHandle hStruct, const IVint32 feature, FeatureMethodT method)
{
    if (!g_application)
        return IV_ERR_APPLICATION_NOT_CREATED;

    if (!hStruct)
        return IV_ERR_NULL_HANDLE;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            if (!accessor.acqIf().setFeature((AcqIfFeatureEnumT)feature, method))
                return IV_ERR_WINDOW_FEATURE_CODE_UNKNOWN;
        }
        else if (auto *const action = dynamic_cast<ImageView::Action*>(base))
        {
            auto *const actionIf = static_cast<IVActionIfClass*>(action->actionInterface());
            if (!actionIf->setFeature((ActionIfFeatureEnumT)feature, method))
                return IV_ERR_ACTION_FEATURE_CODE_UNKNOWN;
        }
        else
            return IV_ERR_BAD_HANDLE;
    }
    catch (...)
    {
        return IV_ERR_UNKNOWN_ERROR;
    }

    return IV_OK;
}

// Return IV_TRUE if the feature has been enabled upon the given handle, IV_FALSE, otherwise.
IVbool IVIsFeatureSet (IVHandle hStruct, const IVint32 feature)
{
    if (!g_application)
        return IV_FALSE;

    if (!hStruct)
        return IV_FALSE;

    auto *const base = reinterpret_cast<ImageView::BaseClass*>(hStruct);

    try
    {
        if (IVWindowWrapper::Check(base))
        {
            auto accessor = IVWindowWrapper::Get(hStruct);

            if (accessor.acqIf().isFeatureSet(feature) || accessor.fileIf().isFeatureSet(feature))
                return IV_TRUE;
        }
        else if (const auto *const action = dynamic_cast<ImageView::Action*>(base))
        {
            const auto *const actionIf = static_cast<const IVActionIfClass*>(action->actionInterface());
            if (actionIf->isFeatureSet(feature))
                return IV_TRUE;
        }
    }
    catch (IVWindowWrapper::WindowExpiredException const& err)
    { }
    catch (...)
    { }

    return IV_FALSE;
}

// Enable a feature that returns IVvoid after taking IVvoid (ie., no return, no args). Set IV_NULL, to disable a feature.
IVRC IVSetFeatureVV (IVHandle hStruct, const IVint32 feature, IVFeature_void_void_Method method)
{
    return IVSetFeature_AcqIf_ActionIf_Private<IVWindowFeatures_void_void, IVActionFeatures_void_void>(hStruct, feature, method);
}

// Enable a feature that returns IVvoid (ie., no return) after taking IVbool. Set IV_NULL, to disable a feature.
IVRC IVSetFeatureVB (IVHandle hStruct, const IVint32 feature, IVFeature_void_bool_Method method)
{
    return IVSetFeature_ActionIf_Private<IVActionFeatures_void_bool>(hStruct, feature, method);
}

// Enable a feature that returns IVbool after taking IVvoid (ie., no args). Set IV_NULL, to disable a feature.
IVRC IVSetFeatureBV (IVHandle hStruct, const IVint32 feature, IVFeature_bool_void_Method method)
{
    return IVSetFeature_AcqIf_Private<IVWindowFeatures_bool_void>(hStruct, feature, method);
}

// Enable a feature that returns IVbool after taking C-string. Set IV_NULL, to disable a feature.
IVRC IVSetFeatureBCStr (IVHandle hStruct, const IVint32 feature, IVFeature_bool_cCStr_Method method)
{
    return IVSetFeature_AcqIf_Private<IVWindowFeatures_bool_cCStr>(hStruct, feature, method);
}

// Enable a feature that returns IVbool after taking IVsize_t. Set IV_NULL, to disable a feature.
IVRC IVSetFeatureBSt (IVHandle hStruct, const IVint32 feature, IVFeature_bool_size_t_Method method)
{
    return IVSetFeature_AcqIf_Private<IVWindowFeatures_bool_size_t>(hStruct, feature, method);
}

// Enable a feature that returns IVdouble after taking IVvoid (ie., no args). Set IV_NULL, to disable a feature.
IVRC IVSetFeatureDV (IVHandle hStruct, const IVint32 feature, IVFeature_double_void_Method method)
{
    return IVSetFeature_AcqIf_Private<IVWindowFeatures_double_void>(hStruct, feature, method);
}

// Enable a feature that returns IVsize_t after taking IVvoid (ie., no args). Set IV_NULL, to disable a feature.
IVRC IVSetFeatureStV (IVHandle hStruct, const IVint32 feature, IVFeature_size_t_void_Method method)
{
    return IVSetFeature_AcqIf_Private<IVWindowFeatures_size_t_void>(hStruct, feature, method);
}

// Enable a feature that returns IVbool after taking a mutable C-string buffer and a size_t pointer. Set IV_NULL, to disable a feature.
IVRC IVSetFeatureBCStrPSt (IVHandle hStruct, const IVint32 feature, IVFeature_bool_CStr_psize_t_Method method)
{
    return IVSetFeature_AcqIf_Private<IVWindowFeatures_bool_CStr_psize_t>(hStruct, feature, method);
}

// Enable a feature that returns IVbool after taking pointers to an IVRawImage and IVImageLock. Set IV_NULL, to disable a feature.
IVRC IVSetFeatureBRiIl (IVHandle hStruct, const IVint32 feature, IVFeature_bool_pIVRawImage_pIVImageLock_Method method)
{
    return IVSetFeature_AcqIf_Private<IVWindowFeatures_bool_pIVRawImage_pIVImageLock>(hStruct, feature, method);
}

// Enable a feature that returns IVbool after taking a pointer to an IVImageLock. Set IV_NULL, to disable a feature.
IVRC IVSetFeatureBIl (IVHandle hStruct, const IVint32 feature, IVFeature_bool_pIVImageLock_Method method)
{
    return IVSetFeature_AcqIf_Private<IVWindowFeatures_bool_pIVImageLock>(hStruct, feature, method);
}

// Enable a feature that returns IVbool after taking pointers to an IVRawImage and C-string. Set IV_NULL, to disable a feature.
IVRC IVSetFeatureBRiCStr (IVHandle hStruct, const IVint32 feature, IVFeature_bool_pIVRawImage_cCStr_Method method)
{
    return IVSetFeature_FileIf_Private<IVWindowFeatures_bool_pIVRawImage_cCStr>(hStruct, feature, method);
}
