#ifndef INCLUDED__BIT_FLOW__IMAGE__VIEW__C__API__H
#define INCLUDED__BIT_FLOW__IMAGE__VIEW__C__API__H

/* FILE:        ImageViewCApi.h
 * DATE:        11/23/2015
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2015, BitFlow, Inc.
 * DESCRIPTION: Pure C API interface to the ImageView library. API functions are defined here.
 */

#include "ImageViewCDef.h"

#ifdef __cplusplus
extern "C" {
#endif  /* __cplusplus */

//// ImageView Application methods.

// Create the global ImageView Application instance. This should be called exactly once, before any
// other ImageView methods are used. Standard application argument parameters should be provided.
IV_IMPORT_EXPORT IVRC IV_API IVAppCreate (const IVint32 argc, const IVchar *const argv[], IVApp *const pApp);

// Destroy the global ImageView Application instance created by IVAppCreate. If IVAppCreate succeeds,
// this should be called once, after all other ImageView method calls have been made.
IV_IMPORT_EXPORT IVRC IV_API IVAppDestroy (IVApp app);

// Retrieve the global ImageView Application handle. IVAppCreate must first have been called.
IV_IMPORT_EXPORT IVRC IV_API IVAppGetGlobal (IVApp *const pApp);

// Add a method to be executed periodically when the application main thread is idle.
IV_IMPORT_EXPORT IVRC IV_API IVAppAddIdleMethod (IVApp app, const IVAppIdleMethod method, IVUserData userData);

// Enter the application main loop. Does not return until the application exits. pReturnCode may be IV_NULL.
IV_IMPORT_EXPORT IVRC IV_API IVAppExec (IVApp app, IVint32 *const pReturnCode);

// Exit the application on call, returning the specified return code.
IV_IMPORT_EXPORT IVRC IV_API IVAppExit (IVApp app, const IVint32 returnCode);


//// ImageView Window methods.

// Create a new window with the given user data.
IV_IMPORT_EXPORT IVRC IV_API IVWindowCreate (IVApp app, IVWindow *const pWindow, IVUserData windowUserData);

// Destroy the window. This should be called for every Window created with IVWindowCreate.
IV_IMPORT_EXPORT IVRC IV_API IVWindowDestroy (IVWindow window);

// Returns IV_TRUE, if the handle points to a Window that still exists, IV_FALSE otherwise.
IV_IMPORT_EXPORT IVbool IV_API IVWindowIsGood (IVWindow window);

// Retrieve the native window handle.
IV_IMPORT_EXPORT IVRC IV_API IVWindowGetNativeHandle (IVWindow window, IMAGE_VIEW_NATIVE_WINDOW *const pNativeHandle);

// Retrieve the window User Data handle.
IV_IMPORT_EXPORT IVRC IV_API IVWindowGetUserData (IVWindow window, IVUserData *const pWindowUserData);

// Set the next image frame to display. If nextFrame is IV_NULL, any pending frame buffer will be cleared.
IV_IMPORT_EXPORT IVRC IV_API IVWindowSetNextFrame (IVWindow window, IVRawImage *const nextFrame);

// Add a setup mode. The description may be any valid string, but the name must be a unique string.
IV_IMPORT_EXPORT IVRC IV_API IVWindowAddSetupMode (IVWindow window, const char *const name, const char *const description);

// Clear the list of setup modes.
IV_IMPORT_EXPORT IVRC IV_API IVWindowClearSetupModes (IVWindow window);

// Add a file save format with the given primary extension, optional alternate extension(s) (space
// separated), and optional description. The primary and alternate extensions must be unique,
// vis-a-vis all other formats. Extensions may optionally start with a period, and should otherwise
// contain only uppercase and lowercase letters, and decimal digits. Extension case is preserved
// internally, but comparisons are case insensitive. alternateExts and desc may be IV_NULL.
IV_IMPORT_EXPORT IVRC IV_API IVWindowAddFileFormat (IVWindow window, const IVchar *const primaryExt, const IVchar *const alternateExts, const IVchar *const desc);

// Clear the list of custom file formats.
IV_IMPORT_EXPORT IVRC IV_API IVWindowClearFileFormats (IVWindow window);

// Add a new GUI action to the window at the given anchor point. parentMenuName should specify a
// top-level menu entry name for IV_WINDOW_ACTION_MAIN_MENU, but IV_NULL for any other anchor point.
IV_IMPORT_EXPORT IVRC IV_API IVWindowAddAction (IVWindow window, const IVWindowActionAnchor anchor, const IVchar *const parentMenuName, IVAction *const pAction, IVUserData actionUserData);


//// ImageView Action methods.

// Destroy the action. Actions will be automatically destroyed when the parent window closes, if
// not explicitly destroyed prior to that time.
IV_IMPORT_EXPORT IVRC IV_API IVActionDestroy (IVAction action);

// Retrieve the action User Data handle.
IV_IMPORT_EXPORT IVRC IV_API IVActionGetUserData (IVAction action, IVUserData *const pActionUserData);


//// ImageView MessageBox methods.

// Show a message box of the given type as a modal dialog of the given parent window. On success,
// pUserButton returns the IVButton pressed by the user. pUserButton may be IV_NULL. The method is
// thread safe and blocking, only returning on error or user input.
IV_IMPORT_EXPORT IVRC IV_API IVMessageBoxDialogShow (const IVMessageBoxDialogType type, IVWindow parentWindow, const IVchar *const title, const IVchar *const text, const IVint32 buttons, const IVDialogButton defaultButton, IVDialogButton *const pUserButton);

//// Generic signaling methods.

// Issue a signal asynchronously.
IV_IMPORT_EXPORT IVRC IV_API IVSendSignal (IVHandle hStruct, const IVint32 signalCode);


//// Generic property methods.

// Get a boolean property from the structure.
IV_IMPORT_EXPORT IVRC IV_API IVGetPropBool (IVHandle hStruct, const IVint32 prop, IVbool *const pEnabled);

// Get an int32 property from the structure.
IV_IMPORT_EXPORT IVRC IV_API IVGetPropInt32 (IVHandle hStruct, const IVint32 prop, IVint32 *const pValue);

// Get a double floating point property from the structure.
IV_IMPORT_EXPORT IVRC IV_API IVGetPropDouble (IVHandle hStruct, const IVint32 prop, IVdouble *const pValue);

// Retrieve a c-string property from the structure.
//   pStrBuf    [OUT] The output string buffer. This may be IV_NULL, to retrieve just the string length. If
//                    insufficiently long, filled with as much data as will fit.
//   pSize      [IN]  The length of pStrBuf, if that is not IV_NULL.
//              [OUT] The length of the complete string, including null-terminator.
IV_IMPORT_EXPORT IVRC IV_API IVGetPropCStr (IVHandle hStruct, const IVint32 prop, IVchar *const pStrBuf, IVsize_t *const pSize);

// Retrieve a data property from the structure.
//   pDataBuf   [OUT] The output data buffer. This may be IV_NULL, to retrieve just the data size. No data is
//                    returned, if the buffer size is inadequate (IV_ERR_BUFFER_TOO_SMALL).
//   pDataSize  [IN]  The length of pDataBuf in bytes, if that is not IV_NULL.
//              [OUT] The exact data size in bytes.
IV_IMPORT_EXPORT IVRC IV_API IVGetPropData (IVHandle hStruct, const IVint32 prop, IVvoid *const pDataBuf, IVsize_t *const pDataSize);

// Set a boolean property on the structure.
IV_IMPORT_EXPORT IVRC IV_API IVSetPropBool (IVHandle hStruct, const IVint32 prop, const IVbool enabled);

// Set an int32 property on the structure.
IV_IMPORT_EXPORT IVRC IV_API IVSetPropInt32 (IVHandle hStruct, const IVint32 prop, const IVint32 value);

// Set a double floating point property on the structure.
IV_IMPORT_EXPORT IVRC IV_API IVSetPropDouble (IVHandle hStruct, const IVint32 prop, const IVdouble value);

// Set a string property on the structure.
IV_IMPORT_EXPORT IVRC IV_API IVSetPropCStr (IVHandle hStruct, const IVint32 prop, const IVchar *const strBuf);

// Set a data property on the structure.
IV_IMPORT_EXPORT IVRC IV_API IVSetPropData (IVHandle hStruct, const IVint32 prop, const IVvoid *const dataBuf, const IVsize_t dataSize);


//// Generic feature methods.

// Return IV_TRUE if the feature has been enabled upon the given handle, IV_FALSE, otherwise.
IV_IMPORT_EXPORT IVbool IV_API IVIsFeatureSet (IVHandle hStruct, const IVint32 feature);

// Enable a feature that returns IVvoid after taking IVvoid (ie., no return, no args). Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureVV (IVHandle hStruct, const IVint32 feature, IVFeature_void_void_Method method);

// Enable a feature that returns IVvoid (ie., no return) after taking IVbool. Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureVB (IVHandle hStruct, const IVint32 feature, IVFeature_void_bool_Method method);

// Enable a feature that returns IVbool after taking IVvoid (ie., no args). Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureBV (IVHandle hStruct, const IVint32 feature, IVFeature_bool_void_Method method);

// Enable a feature that returns IVbool after taking C-string. Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureBCStr (IVHandle hStruct, const IVint32 feature, IVFeature_bool_cCStr_Method method);

// Enable a feature that returns IVbool after taking IVsize_t. Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureBSt (IVHandle hStruct, const IVint32 feature, IVFeature_bool_size_t_Method method);

// Enable a feature that returns IVdouble after taking IVvoid (ie., no args). Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureDV (IVHandle hStruct, const IVint32 feature, IVFeature_double_void_Method method);

// Enable a feature that returns IVsize_t after taking IVvoid (ie., no args). Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureStV (IVHandle hStruct, const IVint32 feature, IVFeature_size_t_void_Method method);

// Enable a feature that returns IVbool after taking a mutable C-string buffer and a size_t pointer. Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureBCStrPSt (IVHandle hStruct, const IVint32 feature, IVFeature_bool_CStr_psize_t_Method method);

// Enable a feature that returns IVbool after taking pointers to an IVRawImage and IVImageLock. Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureBRiIl (IVHandle hStruct, const IVint32 feature, IVFeature_bool_pIVRawImage_pIVImageLock_Method method);

// Enable a feature that returns IVbool after taking a pointer to an IVImageLock. Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureBIl (IVHandle hStruct, const IVint32 feature, IVFeature_bool_pIVImageLock_Method method);

// Enable a feature that returns IVbool after taking pointers to an IVRawImage and C-string. Set IV_NULL, to disable a feature.
IV_IMPORT_EXPORT IVRC IV_API IVSetFeatureBRiCStr (IVHandle hStruct, const IVint32 feature, IVFeature_bool_pIVRawImage_cCStr_Method method);

#ifdef __cplusplus
}
#endif  /* __cplusplus */

#endif // INCLUDED__BIT_FLOW__IMAGE__VIEW__C__API__H
