/* FILE:        ImageViewCApi_Private.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView C API private features.
 */

#include "ImageViewCApi_Private.h"

//// IVAcqIfClass

IVAcqIfClass::CMethodImageLock::CMethodImageLock (const IVUserData windowUserData, IVImageLock const& imageLock, const IVFeature_bool_pIVImageLock_Method unlockImageMethod)
    : m_windowUserData (windowUserData)
    , m_imageLock (imageLock)
    , m_unlockImageMethod (unlockImageMethod)
{ }

IVAcqIfClass::CMethodImageLock::CMethodImageLock (CMethodImageLock &&to_take)
    : m_windowUserData (to_take.m_windowUserData)
    , m_imageLock (to_take.m_imageLock)
    , m_unlockImageMethod (to_take.m_unlockImageMethod)
{
    to_take.m_windowUserData = nullptr;
    to_take.m_imageLock.userData = nullptr;
    to_take.m_unlockImageMethod = nullptr;
}

IVAcqIfClass::CMethodImageLock::~CMethodImageLock (void)
{
    if (m_unlockImageMethod)
        m_unlockImageMethod(m_windowUserData, &m_imageLock);
}

ImageView::ScopedLock IVAcqIfClass::CMethodImageLock::to_scoped (void)
{
    return ImageView::make_scoped(new CMethodImageLock(std::move(*this)));
}

IVAcqIfClass::IVAcqIfClass (IVUserData windowUserData)
    : m_windowUserData (windowUserData)

    , m_closedMethod (nullptr)
    , m_readyNextFrameMethod (nullptr)

    , m_setupMethod (nullptr)
    , m_cleanupMethod (nullptr)
    , m_triggerMethod (nullptr)
    , m_startSingleReverseMethod (nullptr)
    , m_startReverseMethod (nullptr)
    , m_startMethod (nullptr)
    , m_startSingleMethod (nullptr)
    , m_pauseMethod (nullptr)
    , m_stopMethod (nullptr)

    , m_canSetupMethod (nullptr)
    , m_canSetupModeMethod (nullptr)
    , m_canCleanupMethod (nullptr)
    , m_canTriggerMethod (nullptr)
    , m_canStartSingleReverseMethod (nullptr)
    , m_canStartReverseMethod (nullptr)
    , m_canStartMethod (nullptr)
    , m_canStartSingleMethod (nullptr)
    , m_canStartFixedMethod (nullptr)
    , m_canPauseMethod (nullptr)
    , m_canStopMethod (nullptr)

    , m_isSetupMethod (nullptr)
    , m_isRunningSingleReverseMethod (nullptr)
    , m_isRunningReverseMethod (nullptr)
    , m_isRunningMethod (nullptr)
    , m_isRunningSingleMethod (nullptr)
    , m_isRunningFixedMethod (nullptr)
    , m_isPausedMethod (nullptr)
    , m_isStoppedMethod (nullptr)

    , m_setupModeMethod (nullptr)

    , m_startFixedMethod (nullptr)

    , m_frameRateMethod (nullptr)

    , m_framesQueuedMethod (nullptr)
    , m_framesFilledMethod (nullptr)
    , m_framesDroppedMethod (nullptr)
    , m_framesDeliveredMethod (nullptr)

    , m_activeSetupMode (nullptr)

    , m_lockImageMethod (nullptr)
    , m_unlockImageMethod (nullptr)

#if defined(_WIN32)
    , m_WindowApplicationAboutIconRc (-1)
#endif
{ }

IVAcqIfClass::~IVAcqIfClass (void)
{
    // Make sure that we're no good, and that nothing else can use us.
    m_windowMasterRef.kill_all_refs();
}

IVUserData IVAcqIfClass::windowUserData (void) const
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);
    return m_windowUserData;
}

SlaveRef IVAcqIfClass::windowSlaveRef (void) const
{
    return m_windowMasterRef;
}

#if defined(_WIN32)
std::int32_t IVAcqIfClass::windowApplicationAboutIconRc (void) const
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);
    return m_WindowApplicationAboutIconRc;
}
void IVAcqIfClass::setWindowApplicationAboutIconRc (const std::int32_t rc)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);
    m_WindowApplicationAboutIconRc = rc;
}
#endif

IVAcqIfClass::XpmIcon IVAcqIfClass::windowApplicationAboutIconXPM (void) const
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);
    if (!m_WindowApplicationAboutIconXPM.size())
        return nullptr;

    return (XpmIcon)m_WindowApplicationAboutIconXPM.data();
}
bool IVAcqIfClass::getWindowApplicationAboutIconXPMData (void *const a_data, size_t *const a_pSize) const
{
    if (!a_pSize)
        return false;

    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    const size_t bufSize = *a_pSize;
    const size_t dataSize = m_WindowApplicationAboutIconXPM.size();

    *a_pSize = dataSize;

    if (a_data)
    {
        if (bufSize < dataSize)
            return false;
        memcpy(a_data, m_WindowApplicationAboutIconXPM.data(), dataSize);
    }

    return true;
}
bool IVAcqIfClass::setWindowApplicationAboutIconXPMData (const void *const a_data, const size_t a_size)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    if (a_data && a_size)
    {
        const auto *const data8 = reinterpret_cast<const std::uint8_t*>(a_data);
        m_WindowApplicationAboutIconXPM.assign(data8, data8 + a_size);
        m_WindowApplicationAboutIconXPM.shrink_to_fit();
    }
    else if (!a_data && !a_size)
        m_WindowApplicationAboutIconXPM.clear();
    else
        return false;

    return true;
}

bool IVAcqIfClass::isFeatureSet (const IVint32 feature) const
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_VV_CLOSED:
        return !!m_closedMethod;
    case IV_WINDOW_VV_READY_NEXT_FRAME:
        return !!m_readyNextFrameMethod;
    case IV_WINDOW_BV_SETUP:
        return hasSetup();
    case IV_WINDOW_BV_CLEANUP:
        return hasCleanup();
    case IV_WINDOW_BV_TRIGGER:
        return hasTrigger();
    case IV_WINDOW_BV_START_SINGLE_REVERSE:
        return hasStartSingleReverse();
    case IV_WINDOW_BV_START_REVERSE:
        return hasStartReverse();
    case IV_WINDOW_BV_START:
        return hasStart();
    case IV_WINDOW_BV_START_SINGLE:
        return hasStartSingle();
    case IV_WINDOW_BV_PAUSE:
        return hasPause();
    case IV_WINDOW_BV_STOP:
        return hasStop();
    case IV_WINDOW_BV_CAN_SETUP:
        return !!m_canSetupMethod;
    case IV_WINDOW_BV_CAN_SETUP_MODE:
        return !!m_canSetupModeMethod;
    case IV_WINDOW_BV_CAN_CLEANUP:
        return !!m_canCleanupMethod;
    case IV_WINDOW_BV_CAN_TRIGGER:
        return !!m_canTriggerMethod;
    case IV_WINDOW_BV_CAN_START_SINGLE_REVERSE:
        return !!m_canStartSingleReverseMethod;
    case IV_WINDOW_BV_CAN_START_REVERSE:
        return !!m_canStartReverseMethod;
    case IV_WINDOW_BV_CAN_START:
        return !!m_canStartMethod;
    case IV_WINDOW_BV_CAN_START_SINGLE:
        return !!m_canStartSingleMethod;
    case IV_WINDOW_BV_CAN_START_FIXED:
        return !!m_canStartFixedMethod;
    case IV_WINDOW_BV_CAN_PAUSE:
        return !!m_canPauseMethod;
    case IV_WINDOW_BV_CAN_STOP:
        return !!m_canStopMethod;
    case IV_WINDOW_BV_IS_SETUP:
        return !!m_isSetupMethod;
    case IV_WINDOW_BV_IS_RUNNING_SINGLE_REVERSE:
        return !!m_isRunningSingleReverseMethod;
    case IV_WINDOW_BV_IS_RUNNING_REVERSE:
        return !!m_isRunningReverseMethod;
    case IV_WINDOW_BV_IS_RUNNING:
        return !!m_isRunningMethod;
    case IV_WINDOW_BV_IS_RUNNING_SINGLE:
        return !!m_isRunningSingleMethod;
    case IV_WINDOW_BV_IS_RUNNING_FIXED:
        return !!m_isRunningFixedMethod;
    case IV_WINDOW_BV_IS_PAUSED:
        return !!m_isPausedMethod;
    case IV_WINDOW_BV_IS_STOPPED:
        return !!m_isStoppedMethod;
    case IV_WINDOW_BCSTR_SETUP_MODE:
        return hasSetupMode();
    case IV_WINDOW_BST_START_FIXED:
        return hasStartFixed();
    case IV_WINDOW_DV_FRAME_RATE:
        return hasFrameRate();
    case IV_WINDOW_STV_FRAMES_QUEUED:
        return hasFramesQueued();
    case IV_WINDOW_STV_FRAMES_FILLED:
        return hasFramesFilled();
    case IV_WINDOW_STV_FRAMES_DROPPED:
        return hasFramesDropped();
    case IV_WINDOW_STV_FRAMES_DELIVERED:
        return hasFramesDelivered();
    case IV_WINDOW_BCSTRPST_ACTIVE_SETUP_MODE:
        return !!m_activeSetupMode;
    case IV_WINDOW_BRIIL_LOCK_IMAGE:
        return !!m_lockImageMethod;
    case IV_WINDOW_BIL_UNLOCK_IMAGE:
        return !!m_unlockImageMethod;
    }

    return false;
}

bool IVAcqIfClass::setFeature (const IVWindowFeatures_void_void feature, IVFeature_void_void_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_VV_CLOSED:
        m_closedMethod = method;
        return true;
    case IV_WINDOW_VV_READY_NEXT_FRAME:
        m_readyNextFrameMethod = method;
        return true;
    }

    return false;
}
bool IVAcqIfClass::setFeature (const IVWindowFeatures_bool_void feature, IVFeature_bool_void_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_BV_SETUP:
        m_setupMethod = method;
        setHasSetup(!!method);
        return true;
    case IV_WINDOW_BV_CLEANUP:
        m_cleanupMethod = method;
        setHasCleanup(!!method);
        return true;
    case IV_WINDOW_BV_TRIGGER:
        m_triggerMethod = method;
        setHasTrigger(!!method);
        return true;
    case IV_WINDOW_BV_START_SINGLE_REVERSE:
        m_startSingleReverseMethod = method;
        setHasStartSingleReverse(!!method);
        return true;
    case IV_WINDOW_BV_START_REVERSE:
        m_startReverseMethod = method;
        setHasStartReverse(!!method);
        return true;
    case IV_WINDOW_BV_START:
        m_startMethod = method;
        setHasStart(!!method);
        return true;
    case IV_WINDOW_BV_START_SINGLE:
        m_startSingleMethod = method;
        setHasStartSingle(!!method);
        return true;
    case IV_WINDOW_BV_PAUSE:
        m_pauseMethod = method;
        setHasPause(!!method);
        return true;
    case IV_WINDOW_BV_STOP:
        m_stopMethod = method;
        setHasStop(!!method);
        return true;

    case IV_WINDOW_BV_CAN_SETUP:
        m_canSetupMethod = method;
        return true;
    case IV_WINDOW_BV_CAN_SETUP_MODE:
        m_canSetupModeMethod = method;
        return true;
    case IV_WINDOW_BV_CAN_CLEANUP:
        m_canCleanupMethod = method;
        return true;
    case IV_WINDOW_BV_CAN_TRIGGER:
        m_canTriggerMethod = method;
        return true;
    case IV_WINDOW_BV_CAN_START_SINGLE_REVERSE:
        m_canStartSingleReverseMethod = method;
        return true;
    case IV_WINDOW_BV_CAN_START_REVERSE:
        m_canStartReverseMethod = method;
        return true;
    case IV_WINDOW_BV_CAN_START:
        m_canStartMethod = method;
        return true;
    case IV_WINDOW_BV_CAN_START_SINGLE:
        m_canStartSingleMethod = method;
        return true;
    case IV_WINDOW_BV_CAN_START_FIXED:
        m_canStartFixedMethod = method;
        return true;
    case IV_WINDOW_BV_CAN_PAUSE:
        m_canPauseMethod = method;
        return true;
    case IV_WINDOW_BV_CAN_STOP:
        m_canStopMethod = method;
        return true;

    case IV_WINDOW_BV_IS_SETUP:
        m_isSetupMethod = method;
        return true;
    case IV_WINDOW_BV_IS_RUNNING_SINGLE_REVERSE:
        m_isRunningSingleReverseMethod = method;
        return true;
    case IV_WINDOW_BV_IS_RUNNING_REVERSE:
        m_isRunningReverseMethod = method;
        return true;
    case IV_WINDOW_BV_IS_RUNNING:
        m_isRunningMethod = method;
        return true;
    case IV_WINDOW_BV_IS_RUNNING_SINGLE:
        m_isRunningSingleMethod = method;
        return true;
    case IV_WINDOW_BV_IS_RUNNING_FIXED:
        m_isRunningFixedMethod = method;
        return true;
    case IV_WINDOW_BV_IS_PAUSED:
        m_isPausedMethod = method;
        return true;
    case IV_WINDOW_BV_IS_STOPPED:
        m_isStoppedMethod = method;
        return true;
    }

    return false;
}
bool IVAcqIfClass::setFeature (const IVWindowFeatures_bool_cCStr feature, IVFeature_bool_cCStr_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_BCSTR_SETUP_MODE:
        m_setupModeMethod = method;
        setHasSetupMode(!!method);
        return true;
    }

    return false;
}
bool IVAcqIfClass::setFeature (const IVWindowFeatures_bool_size_t feature, IVFeature_bool_size_t_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_BST_START_FIXED:
        m_startFixedMethod = method;
        setHasStartFixed(!!method);
        return true;
    }

    return false;
}
bool IVAcqIfClass::setFeature (const IVWindowFeatures_double_void feature, IVFeature_double_void_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_DV_FRAME_RATE:
        m_frameRateMethod = method;
        setHasFrameRate(!!method);
        return true;
    }

    return false;
}
bool IVAcqIfClass::setFeature (const IVWindowFeatures_size_t_void feature, IVFeature_size_t_void_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_STV_FRAMES_QUEUED:
        m_framesQueuedMethod = method;
        setHasFramesQueued(!!method);
        return true;
    case IV_WINDOW_STV_FRAMES_FILLED:
        m_framesFilledMethod = method;
        setHasFramesFilled(!!method);
        return true;
    case IV_WINDOW_STV_FRAMES_DROPPED:
        m_framesDroppedMethod = method;
        setHasFramesDropped(!!method);
        return true;
    case IV_WINDOW_STV_FRAMES_DELIVERED:
        m_framesDeliveredMethod = method;
        setHasFramesDelivered(!!method);
        return true;
    }

    return false;
}

bool IVAcqIfClass::setFeature (const IVWindowFeatures_bool_CStr_psize_t feature, IVFeature_bool_CStr_psize_t_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_BCSTRPST_ACTIVE_SETUP_MODE:
        m_activeSetupMode = method;
        return true;
    }

    return false;
}

bool IVAcqIfClass::setFeature (const IVWindowFeatures_bool_pIVRawImage_pIVImageLock feature, IVFeature_bool_pIVRawImage_pIVImageLock_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_BRIIL_LOCK_IMAGE:
        m_lockImageMethod = method;
        return true;
    }

    return false;
}
bool IVAcqIfClass::setFeature (const IVWindowFeatures_bool_pIVImageLock feature, IVFeature_bool_pIVImageLock_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_BIL_UNLOCK_IMAGE:
        m_unlockImageMethod = method;
        return true;
    }

    return false;
}

// Execute a member method.
bool IVAcqIfClass::exe (IVFeature_void_void_Method const& method) const
{
    IVUserData exeWinUsrData;
    IVFeature_void_void_Method exeMethod;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);

        exeWinUsrData = m_windowUserData;
        exeMethod = method;
    }

    if (exeMethod)
    {
        exeMethod(exeWinUsrData);
        return true;
    }

    return false;
}
bool IVAcqIfClass::exe (IVFeature_bool_void_Method const& method) const
{
    IVUserData exeWinUsrData;
    IVFeature_bool_void_Method exeMethod;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);

        exeWinUsrData = m_windowUserData;
        exeMethod = method;
    }

    if (exeMethod)
        return !!exeMethod(exeWinUsrData);

    return false;
}
bool IVAcqIfClass::exe (IVFeature_bool_cCStr_Method const& method, const char *const arg0) const
{
    IVUserData exeWinUsrData;
    IVFeature_bool_cCStr_Method exeMethod;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);

        exeWinUsrData = m_windowUserData;
        exeMethod = method;
    }

    if (exeMethod)
        return !!exeMethod(exeWinUsrData, arg0);

    return false;
}
bool IVAcqIfClass::exe (IVFeature_bool_size_t_Method const& method, const size_t arg0) const
{
    IVUserData exeWinUsrData;
    IVFeature_bool_size_t_Method exeMethod;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);

        exeWinUsrData = m_windowUserData;
        exeMethod = method;
    }

    if (exeMethod)
        return !!exeMethod(exeWinUsrData, arg0);

    return false;
}
double IVAcqIfClass::exe (IVFeature_double_void_Method const& method) const
{
    IVUserData exeWinUsrData;
    IVFeature_double_void_Method exeMethod;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);

        exeWinUsrData = m_windowUserData;
        exeMethod = method;
    }

    if (exeMethod)
        return exeMethod(exeWinUsrData);

    return false;
}
size_t IVAcqIfClass::exe (IVFeature_size_t_void_Method const& method) const
{
    IVUserData exeWinUsrData;
    IVFeature_size_t_void_Method exeMethod;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);

        exeWinUsrData = m_windowUserData;
        exeMethod = method;
    }

    if (exeMethod)
        return exeMethod(exeWinUsrData);

    return false;
}
bool IVAcqIfClass::exe (IVFeature_bool_CStr_psize_t_Method const& method, char *const pModeStr, size_t *const pSize) const
{
    IVUserData exeWinUsrData;
    IVFeature_bool_CStr_psize_t_Method exeMethod;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);

        exeWinUsrData = m_windowUserData;
        exeMethod = method;
    }

    if (exeMethod)
        return exeMethod(exeWinUsrData, pModeStr, pSize);

    return false;
}

// Method executed when the user interface is about to close.
void IVAcqIfClass::uxClosed (void)
{
    if (!exe(m_closedMethod))
    {
        // If setup, than automatically cleanup.
        if (isSetup())
            cleanup();
    }
}

//// Acquisition setup and control.

// Setup the interface for acquisition. Base implementation simply returns false.
bool IVAcqIfClass::setup (void)
{
    return exe(m_setupMethod);
}

// Setup the interface for acquisition using the given mode. Base implementation simply returns false.
bool IVAcqIfClass::setupMode (const char *const modeName)
{
    return exe(m_setupModeMethod, modeName);
}

// Cleanup the acquisition interface. Base implementation simply returns false.
bool IVAcqIfClass::cleanup (void)
{
    return exe(m_cleanupMethod);
}

// Trigger the acquisition interface. Base implementation simply returns false.
bool IVAcqIfClass::trigger (void)
{
    return exe(m_triggerMethod);
}

// Start reverse acquisition of a single frame. Base implementation simply returns false.
bool IVAcqIfClass::startSingleReverse (void)
{
    return exe(m_startSingleReverseMethod);
}

// Start a reverse acquisition. Base implementation simply returns false.
bool IVAcqIfClass::startReverse (void)
{
    return exe(m_startReverseMethod);
}

// Start acquisition. Base implementation simply returns false.
bool IVAcqIfClass::start (void)
{
    return exe(m_startMethod);
}

// Start a fixed-length acquisition. Base implementation simply returns false.
bool IVAcqIfClass::startFixed (const size_t cnt)
{
    return exe(m_startFixedMethod, cnt);
}

// Start acquisition of a single frame. Base implementation simply returns false.
bool IVAcqIfClass::startSingle (void)
{
    return exe(m_startSingleMethod);
}

// Pause an acquisition. Base implementation simply returns false.
bool IVAcqIfClass::pause (void)
{
    return exe(m_pauseMethod);
}

// Stop acquisition. Base implementation simply returns false.
bool IVAcqIfClass::stop (void)
{
    return exe(m_stopMethod);
}

//// Dynamic GUI feature states.

// Returns true/false, whether or not the interface may be setup for acquisition.
bool IVAcqIfClass::canSetup (void) const
{
    return exe(m_canSetupMethod);
}

// Returns true/false, whether or not the interface may be setup for acquisition with a specific mode.
bool IVAcqIfClass::canSetupMode (void) const
{
    return exe(m_canSetupModeMethod);
}

// Returns true/false, whether or not the interface may be cleaned up.
bool IVAcqIfClass::canCleanup (void) const
{
    return exe(m_canCleanupMethod);
}

// Returns true/false, whether or not the interface may be triggered.
bool IVAcqIfClass::canTrigger (void) const
{
    return exe(m_canTriggerMethod);
}

// Returns true/false, whether or not reverse acquisition of a single frame can be started.
bool IVAcqIfClass::canStartSingleReverse (void) const
{
    return exe(m_canStartSingleReverseMethod);
}

// Returns true/false, whether or not a continuous reverse acquisition can be started.
bool IVAcqIfClass::canStartReverse (void) const
{
    return exe(m_canStartReverseMethod);
}

// Returns true/false, whether or not a continuous acquisition can be started.
bool IVAcqIfClass::canStart (void) const
{
    return exe(m_canStartMethod);
}

// Returns true/false, whether or not a fixed-frame acquisition can be started.
bool IVAcqIfClass::canStartFixed (void) const
{
    return exe(m_canStartFixedMethod);
}

// Returns true/false, whether or not a single-frame acquisition can be started.
bool IVAcqIfClass::canStartSingle (void) const
{
    return exe(m_canStartSingleMethod);
}

// Returns true/false, whether or not acquisition may be paused.
bool IVAcqIfClass::canPause (void) const
{
    return exe(m_canPauseMethod);
}

// Returns true/false, whether or not acquisition may be stopped.
bool IVAcqIfClass::canStop (void) const
{
    return exe(m_canStopMethod);
}

//// Acquisition and setup status.

// Returns true/false, whether or not the interface is setup for acquisition.
bool IVAcqIfClass::isSetup (void) const
{
    return exe(m_isSetupMethod);
}

// Returns true/false, whether or not reverse acquisition of a single frame is running. Default returns false.
bool IVAcqIfClass::isRunningSingleReverse (void) const
{
    return exe(m_isRunningSingleReverseMethod);
}

// Returns true/false, whether or not a reverse acquisition is running. Default returns false.
bool IVAcqIfClass::isRunningReverse (void) const
{
    return exe(m_isRunningReverseMethod);
}

// Returns true/false, whether or not acquisition is running.
bool IVAcqIfClass::isRunning (void) const
{
    return exe(m_isRunningMethod);
}

// Returns true/false, whether or not acquisition is grabbing a fixed-length sequence of frames.
bool IVAcqIfClass::isRunningFixed (void) const
{
    return exe(m_isRunningFixedMethod);
}

// Returns true/false, whether or not acquisition is grabbing a single frame.
bool IVAcqIfClass::isRunningSingle (void) const
{
    return exe(m_isRunningSingleMethod);
}

// Returns true/false, whether or not acquisition is running. Default returns false.
bool IVAcqIfClass::isPaused (void) const
{
    return exe(m_isPausedMethod);
}

// Returns true/false, whether or not acquisition is running. Default returns !isRunning().
bool IVAcqIfClass::isStopped (void) const
{
    return exe(m_isStoppedMethod);
}

//// Setup mode.

bool IVAcqIfClass::activeSetupMode (char *const pModeStr, size_t *const pSize) const
{
    return exe(m_activeSetupMode, pModeStr, pSize);
}

//// Frame status query.

// Return the acquisition frame rate. Base implementation simply returns 0.
double IVAcqIfClass::frameRate (void) const
{
    return exe(m_frameRateMethod);
}

// Return the frames queued count. Base implementations simply return 0.
size_t IVAcqIfClass::framesQueued (void) const
{
    return exe(m_framesQueuedMethod);
}

// Return the frames filled count. Base implementations simply return 0.
size_t IVAcqIfClass::framesFilled (void) const
{
    return exe(m_framesFilledMethod);
}

// Return the frames dropped count. Base implementations simply return 0.
size_t IVAcqIfClass::framesDropped (void) const
{
    return exe(m_framesDroppedMethod);
}

// Return the frames delivered count. Base implementations simply return 0.
size_t IVAcqIfClass::framesDelivered (void) const
{
    return exe(m_framesDeliveredMethod);
}

//// Frame retrieval and access.

// Method called by the Imperative getNextFrame method before attempting to return
// the next frame, allowing clients to modify states declaratively. The default
// implementation does nothing.
void IVAcqIfClass::readyNextFrame (void)
{
    IVUserData windowUserData;
    IVFeature_void_void_Method readyNextFrame;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);

        windowUserData = m_windowUserData;
        readyNextFrame = m_readyNextFrameMethod;
    }

    if (!windowUserData || !readyNextFrame)
        return;

    try
    {
        readyNextFrame(windowUserData);
    }
    catch (...)
    {
    }
}

// Retrieve a critical section to protect the allocation of buffers returned by getNextFrame.
// Return a NULL ScopedLock, if allocation cannot be guaranteed.
ImageView::ScopedLock IVAcqIfClass::getFrameLock (ImageView::RawImage const& rawImage)
{
    // Retrieve the lock methods and user data.
    IVUserData windowUserData;
    IVFeature_bool_pIVRawImage_pIVImageLock_Method lockImage;
    IVFeature_bool_pIVImageLock_Method unlockImage;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);

        windowUserData = m_windowUserData;
        lockImage = m_lockImageMethod;
        unlockImage = m_unlockImageMethod;
    }

    // If either lock method is nullptr, defer to the Imperative implementation.
    if (!lockImage || !unlockImage)
        return ImperativeAcquisitionInterface::getFrameLock(rawImage);

    // Attempt to use the callback lock.
    try
    {
        IVRawImage ivRawImage =
        {
            rawImage.m_buffer,
            rawImage.m_width,
            rawImage.m_height,
            (IVFormat)rawImage.m_format,
            rawImage.m_userData
        };

        IVImageLock imageLock;
        if (lockImage(windowUserData, &ivRawImage, &imageLock))
            return CMethodImageLock(windowUserData, imageLock, unlockImage).to_scoped();
    }
    catch (...)
    {
    }

    return ImageView::ScopedLock();
}

//// IVFileIfClass

IVFileIfClass::IVFileIfClass (const IVUserData windowUserData)
    : m_windowUserData (windowUserData)

    , m_saveImageFileMethod (nullptr)
{
}
IVFileIfClass::~IVFileIfClass (void)
{
}

bool IVFileIfClass::isFeatureSet (const IVint32 feature) const
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_BRICSTR_IMAGE_FILE_SAVE:
        return !!m_saveImageFileMethod;
    }

    return false;
}

// Set a file interface feature method.
bool IVFileIfClass::setFeature (const IVWindowFeatures_bool_pIVRawImage_cCStr feature, IVFeature_bool_pIVRawImage_cCStr_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_WINDOW_BRICSTR_IMAGE_FILE_SAVE:
        m_saveImageFileMethod = method;
        return true;
    }

    return false;
}

// Save an image to the specified file path, with format deduced from the file extension. The
// caller is responsible for guaranteeing image immutability during the save operation.
bool IVFileIfClass::save (ImageView::RawImage const& rawImage, const char *const filePath) const
{
    IVUserData windowUserData;
    IVFeature_bool_pIVRawImage_cCStr_Method saveImageFile;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);
        windowUserData = m_windowUserData;
        saveImageFile = m_saveImageFileMethod;
    }

    if (!saveImageFile)
        return false;

    IVRawImage ivRawImage;
    ivRawImage.buffer = rawImage.m_buffer;
    ivRawImage.width = rawImage.m_width;
    ivRawImage.height = rawImage.m_height;
    ivRawImage.format = (IVFormat)rawImage.m_format;
    ivRawImage.userData = rawImage.m_userData;

    return saveImageFile(windowUserData, &ivRawImage, filePath);
}

//// IVActionIfClass.

IVActionIfClass::IVActionIfClass (const IVUserData actionUserData)
    : m_actionUserData (actionUserData)

    , m_destroyedMethod (nullptr)
    , m_triggeredMethod (nullptr)
{
}
IVActionIfClass::~IVActionIfClass (void)
{
}

IVUserData IVActionIfClass::userData (void) const
{
    return m_actionUserData;
}

bool IVActionIfClass::isFeatureSet (const IVint32 feature) const
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_ACTION_VV_DESTROYED:
        return !!m_destroyedMethod;
    case IV_ACTION_VB_TRIGGERED:
        return !!m_triggeredMethod;
    }

    return false;
}

// Set a file interface feature method.
bool IVActionIfClass::setFeature (const IVActionFeatures_void_void feature, IVFeature_void_void_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_ACTION_VV_DESTROYED:
        m_destroyedMethod = method;
        return true;
    }

    return false;
}
bool IVActionIfClass::setFeature (const IVActionFeatures_void_bool feature, IVFeature_void_bool_Method method)
{
    const std::lock_guard<std::mutex> featureLock (m_featureMutex);

    switch (feature)
    {
    case IV_ACTION_VB_TRIGGERED:
        m_triggeredMethod = method;
        return true;
    }

    return false;
}

// Method implementation.
void IVActionIfClass::uxDestroyed (void)
{
    IVUserData actionUserData;
    IVFeature_void_void_Method destroyedMethod;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);
        actionUserData = m_actionUserData;
        destroyedMethod = m_destroyedMethod;
    }

    if (!destroyedMethod)
        return;

    destroyedMethod(actionUserData);
}
void IVActionIfClass::triggered (const bool checked)
{
    IVUserData actionUserData;
    IVFeature_void_bool_Method triggeredMethod;
    {
        const std::lock_guard<std::mutex> featureLock (m_featureMutex);
        actionUserData = m_actionUserData;
        triggeredMethod = m_triggeredMethod;
    }

    if (!triggeredMethod)
        return;

    triggeredMethod(actionUserData, checked);
}
