#ifndef INCLUDED__BIT_FLOW__IMAGE__VIEW__C__API__PRIVATE__H
#define INCLUDED__BIT_FLOW__IMAGE__VIEW__C__API__PRIVATE__H

/* FILE:        ImageViewCApi_Private.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: PrivateData for the ImageView C API.
 */

#include "ImageViewCApi.h"

#include "ImperativeAcquisitionInterface.h"
#include "AbstractFileInterface.h"
#include "AbstractActionInterface.h"

#include "ImageViewWindow.h"

#include "ImageViewThread.h"

#include <mutex>
#include <memory>
#include <vector>
#include <stdexcept>

// MasterRef -- A class that cannot close while it has extant dependents.
class MasterRef
{
private:
    std::shared_ptr<bool> m_sharedPtr;

    friend class SlaveRef;
    friend class DependentRef;

    // Illegal.
    MasterRef (MasterRef const&);
    MasterRef& operator= (MasterRef const&);

public:
    MasterRef (void)
        : m_sharedPtr (new bool(true))
    { }

    ~MasterRef (void)
    {
        kill_all_refs();
    }

    inline void kill_all_refs (void)
    {
        if (m_sharedPtr)
        {
            std::weak_ptr<bool> weakPtr = m_sharedPtr;
            *m_sharedPtr = false;
            m_sharedPtr.reset();
            while (!weakPtr.expired())
                ImageView::this_thread::yield();
        }
    }
};

// SlaveRef -- A class to reference a MasterRef, but isn't a dependent.
class SlaveRef
{
private:
    std::weak_ptr<bool> m_weakPtr;

    friend class DependentRef;

public:
    SlaveRef (void)
    { }

    SlaveRef (SlaveRef const& to_copy)
        : m_weakPtr (to_copy.m_weakPtr)
    { }

    SlaveRef (MasterRef const& master)
        : m_weakPtr (master.m_sharedPtr)
    { }

    SlaveRef& operator= (SlaveRef const& to_copy)
    {
        if (this != &to_copy)
            m_weakPtr = to_copy.m_weakPtr;
        return *this;
    }
};

// DependentRef -- A class that may be a dependent of a MasterRef.
class DependentRef
{
private:
    std::shared_ptr<bool> m_sharedPtr;
    bool m_good;

    // Illegal.
    DependentRef (void);
    DependentRef (DependentRef const&);
    DependentRef& operator= (DependentRef const&);

    // Helpers.
    inline void validate (void)
    {
        if (m_sharedPtr)
        {
            if (!(*m_sharedPtr))
                m_sharedPtr.reset();
        }
    }

public:
    DependentRef (DependentRef &&to_take)
        : m_sharedPtr (std::move(to_take.m_sharedPtr))
    {
        validate();
    }

    explicit DependentRef (SlaveRef const& slave)
        : m_sharedPtr (slave.m_weakPtr.lock())
    {
        validate();
    }

    explicit DependentRef (MasterRef const& master)
        : m_sharedPtr (master.m_sharedPtr)
    {
        validate();
    }

    inline bool good (void) const
    {
        return !!m_sharedPtr;
    }

    inline operator bool (void) const
    {
        return good();
    }
    inline bool operator! (void) const
    {
        return !good();
    }
};

// The acquisition interface for the ImageView::Window class.
class IVAcqIfClass : public ImageView::ImperativeAcquisitionInterface
{
public:
    typedef const char** XpmIcon;
    typedef std::vector<std::uint8_t> DataVector;

private:
    // Illegal.
    IVAcqIfClass (IVAcqIfClass const&);
    IVAcqIfClass& operator= (IVAcqIfClass const&);

    // Members.
    const IVUserData m_windowUserData;

    IVFeature_void_void_Method m_closedMethod;
    IVFeature_void_void_Method m_readyNextFrameMethod;

    IVFeature_bool_void_Method m_setupMethod;
    IVFeature_bool_void_Method m_cleanupMethod;
    IVFeature_bool_void_Method m_triggerMethod;
    IVFeature_bool_void_Method m_startSingleReverseMethod;
    IVFeature_bool_void_Method m_startReverseMethod;
    IVFeature_bool_void_Method m_startMethod;
    IVFeature_bool_void_Method m_startSingleMethod;
    IVFeature_bool_void_Method m_pauseMethod;
    IVFeature_bool_void_Method m_stopMethod;

    IVFeature_bool_void_Method m_canSetupMethod;
    IVFeature_bool_void_Method m_canSetupModeMethod;
    IVFeature_bool_void_Method m_canCleanupMethod;
    IVFeature_bool_void_Method m_canTriggerMethod;
    IVFeature_bool_void_Method m_canStartSingleReverseMethod;
    IVFeature_bool_void_Method m_canStartReverseMethod;
    IVFeature_bool_void_Method m_canStartMethod;
    IVFeature_bool_void_Method m_canStartSingleMethod;
    IVFeature_bool_void_Method m_canStartFixedMethod;
    IVFeature_bool_void_Method m_canPauseMethod;
    IVFeature_bool_void_Method m_canStopMethod;

    IVFeature_bool_void_Method m_isSetupMethod;
    IVFeature_bool_void_Method m_isRunningSingleReverseMethod;
    IVFeature_bool_void_Method m_isRunningReverseMethod;
    IVFeature_bool_void_Method m_isRunningMethod;
    IVFeature_bool_void_Method m_isRunningSingleMethod;
    IVFeature_bool_void_Method m_isRunningFixedMethod;
    IVFeature_bool_void_Method m_isPausedMethod;
    IVFeature_bool_void_Method m_isStoppedMethod;

    IVFeature_bool_cCStr_Method m_setupModeMethod;

    IVFeature_bool_size_t_Method m_startFixedMethod;

    IVFeature_double_void_Method m_frameRateMethod;

    IVFeature_size_t_void_Method m_framesQueuedMethod;
    IVFeature_size_t_void_Method m_framesFilledMethod;
    IVFeature_size_t_void_Method m_framesDroppedMethod;
    IVFeature_size_t_void_Method m_framesDeliveredMethod;

    IVFeature_bool_CStr_psize_t_Method m_activeSetupMode;

    IVFeature_bool_pIVRawImage_pIVImageLock_Method m_lockImageMethod;
    IVFeature_bool_pIVImageLock_Method m_unlockImageMethod;

#if defined(WIN32)
    std::int32_t m_WindowApplicationAboutIconRc;
#endif

    DataVector m_WindowApplicationAboutIconXPM;

    mutable std::mutex m_featureMutex;

    mutable MasterRef m_windowMasterRef;

    struct CMethodImageLock : ImageView::AbstractLock
    {
    private:
        // Members.
        IVUserData m_windowUserData;
        IVImageLock m_imageLock;
        IVFeature_bool_pIVImageLock_Method m_unlockImageMethod;

        // Illegal.
        CMethodImageLock (void);
        CMethodImageLock (CMethodImageLock const&);
        CMethodImageLock& operator= (CMethodImageLock const&);

    public:
        CMethodImageLock (const IVUserData windowUserData, IVImageLock const& imageLock, const IVFeature_bool_pIVImageLock_Method unlockImageMethod);
        CMethodImageLock (CMethodImageLock &&to_take);
        ~CMethodImageLock (void);

        ImageView::ScopedLock to_scoped (void);
    };

public:
    IVAcqIfClass (IVUserData windowUserData = nullptr);
    ~IVAcqIfClass (void);

    IVUserData windowUserData (void) const;

    SlaveRef windowSlaveRef (void) const;

#if defined(_WIN32)
    std::int32_t windowApplicationAboutIconRc (void) const;
    void setWindowApplicationAboutIconRc (const std::int32_t rc);
#endif

    XpmIcon windowApplicationAboutIconXPM (void) const;
    bool getWindowApplicationAboutIconXPMData (void *const a_data, size_t *const a_pSize) const;
    bool setWindowApplicationAboutIconXPMData (const void *const a_data, const size_t a_size);

    bool isFeatureSet (const IVint32 feature) const;

    bool setFeature (const IVWindowFeatures_void_void feature, IVFeature_void_void_Method method = nullptr);
    bool setFeature (const IVWindowFeatures_bool_void feature, IVFeature_bool_void_Method method = nullptr);
    bool setFeature (const IVWindowFeatures_bool_cCStr feature, IVFeature_bool_cCStr_Method method = nullptr);
    bool setFeature (const IVWindowFeatures_bool_size_t feature, IVFeature_bool_size_t_Method method = nullptr);
    bool setFeature (const IVWindowFeatures_double_void feature, IVFeature_double_void_Method method = nullptr);
    bool setFeature (const IVWindowFeatures_size_t_void feature, IVFeature_size_t_void_Method method = nullptr);
    bool setFeature (const IVWindowFeatures_bool_CStr_psize_t feature, IVFeature_bool_CStr_psize_t_Method method = nullptr);

    bool setFeature (const IVWindowFeatures_bool_pIVRawImage_pIVImageLock feature, IVFeature_bool_pIVRawImage_pIVImageLock_Method method = nullptr);
    bool setFeature (const IVWindowFeatures_bool_pIVImageLock feature, IVFeature_bool_pIVImageLock_Method method = nullptr);

    // Make the signal handlers public.
    using AbstractAcquisitionInterface::signalFeatureSetChange;
    using AbstractAcquisitionInterface::signalAcquisitionStateChange;
    using AbstractAcquisitionInterface::signalGetFrameNow;

    // Execute a member method.
    bool exe (IVFeature_void_void_Method const& method) const;
    bool exe (IVFeature_bool_void_Method const& method) const;
    bool exe (IVFeature_bool_cCStr_Method const& method, const char *const arg0) const;
    bool exe (IVFeature_bool_size_t_Method const& method, const size_t arg0) const;
    double exe (IVFeature_double_void_Method const& method) const;
    size_t exe (IVFeature_size_t_void_Method const& method) const;
    bool exe (IVFeature_bool_CStr_psize_t_Method const& method, char *const pModeStr, size_t *const pSize) const;

    // Method executed when the user interface is about to close.
    void uxClosed (void) override;

    //// Acquisition setup and control.

    // Setup the interface for acquisition. Base implementation simply returns false.
    bool setup (void) override;

    // Setup the interface for acquisition using the given mode. Base implementation simply returns false.
    bool setupMode (const char *const modeName) override;

    // Cleanup the acquisition interface. Base implementation simply returns false.
    bool cleanup (void) override;

    // Trigger the acquisition interface. Base implementation simply returns false.
    bool trigger (void) override;

    // Start reverse acquisition of a single frame. Base implementation simply returns false.
    bool startSingleReverse (void) override;

    // Start a reverse acquisition. Base implementation simply returns false.
    bool startReverse (void) override;

    // Start acquisition. Base implementation simply returns false.
    bool start (void) override;

    // Start a fixed-length acquisition. Base implementation simply returns false.
    bool startFixed (const size_t cnt) override;

    // Start acquisition of a single frame. Base implementation simply returns false.
    bool startSingle (void) override;

    // Pause an acquisition. Base implementation simply returns false.
    bool pause (void) override;

    // Stop acquisition. Base implementation simply returns false.
    bool stop (void) override;

    //// Dynamic GUI feature states.

    // Returns true/false, whether or not the interface may be setup for acquisition.
    bool canSetup (void) const override;

    // Returns true/false, whether or not the interface may be setup for acquisition with a specific mode.
    bool canSetupMode (void) const override;

    // Returns true/false, whether or not the interface may be cleaned up.
    bool canCleanup (void) const override;

    // Returns true/false, whether or not the interface may be triggered.
    bool canTrigger (void) const override;

    // Returns true/false, whether or not reverse acquisition of a single frame can be started.
    bool canStartSingleReverse (void) const override;

    // Returns true/false, whether or not a continuous reverse acquisition can be started.
    bool canStartReverse (void) const override;

    // Returns true/false, whether or not a continuous acquisition can be started.
    bool canStart (void) const override;

    // Returns true/false, whether or not a fixed-frame acquisition can be started.
    bool canStartFixed (void) const override;

    // Returns true/false, whether or not a single-frame acquisition can be started.
    bool canStartSingle (void) const override;

    // Returns true/false, whether or not acquisition may be paused.
    bool canPause (void) const override;

    // Returns true/false, whether or not acquisition may be stopped.
    bool canStop (void) const override;

    //// Acquisition and setup status.

    // Returns true/false, whether or not the interface is setup for acquisition.
    bool isSetup (void) const override;

    // Returns true/false, whether or not reverse acquisition of a single frame is running. Default returns false.
    bool isRunningSingleReverse (void) const override;

    // Returns true/false, whether or not a reverse acquisition is running. Default returns false.
    bool isRunningReverse (void) const override;

    // Returns true/false, whether or not acquisition is running.
    bool isRunning (void) const override;

    // Returns true/false, whether or not acquisition is grabbing a fixed-length sequence of frames.
    bool isRunningFixed (void) const override;

    // Returns true/false, whether or not acquisition is grabbing a single frame.
    bool isRunningSingle (void) const override;

    // Returns true/false, whether or not acquisition is running. Default returns false.
    bool isPaused (void) const override;

    // Returns true/false, whether or not acquisition is running. Default returns !isRunning().
    bool isStopped (void) const override;

    //// Setup mode.

    // Return the name of the setup mode, if any. Default simply returns false.
    bool activeSetupMode (char *const pModeStr, size_t *const pSize) const override;

    //// Frame status query.

    // Return the acquisition frame rate. Base implementation simply returns 0.
    double frameRate (void) const override;

    // Return the frames queued count. Base implementations simply return 0.
    size_t framesQueued (void) const override;

    // Return the frames filled count. Base implementations simply return 0.
    size_t framesFilled (void) const override;

    // Return the frames dropped count. Base implementations simply return 0.
    size_t framesDropped (void) const override;

    // Return the frames delivered count. Base implementations simply return 0.
    size_t framesDelivered (void) const override;

    //// Frame retrieval and access.

    // Method called by the Imperative getNextFrame method before attempting to return
    // the next frame, allowing clients to modify states declaratively. The default
    // implementation does nothing.
    void readyNextFrame (void) override;

    // Retrieve a critical section to protect the allocation of buffers returned by getNextFrame.
    // Return a NULL ScopedLock, if allocation cannot be guaranteed.
    ImageView::ScopedLock getFrameLock (ImageView::RawImage const& rawImage) override;
};

class IVFileIfClass : public ImageView::AbstractFileInterface
{
private:
    // Illegal.
    IVFileIfClass (void);
    IVFileIfClass (IVFileIfClass const&);
    IVFileIfClass& operator= (IVFileIfClass const&);

    // Members.
    const IVUserData m_windowUserData;

    IVFeature_bool_pIVRawImage_cCStr_Method m_saveImageFileMethod;

    mutable std::mutex m_featureMutex;

public:
    IVFileIfClass (const IVUserData windowUserData);
    ~IVFileIfClass (void);

    bool isFeatureSet (const IVint32 feature) const;

    // Set a file interface feature method.
    bool setFeature (const IVWindowFeatures_bool_pIVRawImage_cCStr feature, IVFeature_bool_pIVRawImage_cCStr_Method method = nullptr);

    // Save an image to the specified file path, with format deduced from the file extension. The
    // caller is responsible for guaranteeing image immutability during the save operation.
    bool save (ImageView::RawImage const& rawImage, const char *const filePath) const override;
};

class IVActionIfClass : public ImageView::AbstractActionInterface
{
private:
    // Illegal.
    IVActionIfClass (void);
    IVActionIfClass (IVActionIfClass const&);
    IVActionIfClass& operator= (IVActionIfClass const&);

    // Members.
    const IVUserData m_actionUserData;

    IVFeature_void_void_Method m_destroyedMethod;
    IVFeature_void_bool_Method m_triggeredMethod;

    mutable std::mutex m_featureMutex;

public:
    IVActionIfClass (const IVUserData actionUserData);
    ~IVActionIfClass (void);

    IVUserData userData (void) const;

    bool isFeatureSet (const IVint32 feature) const;

    // Set a file interface feature method.
    bool setFeature (const IVActionFeatures_void_void feature, IVFeature_void_void_Method method = nullptr);
    bool setFeature (const IVActionFeatures_void_bool feature, IVFeature_void_bool_Method method = nullptr);

    // Method implementation.
    void uxDestroyed (void) override;
    void triggered (const bool checked) override;
};

class IVWindowWrapper : public ImageView::BaseClass
{
private:
    ImageView::Window &m_window;

    SlaveRef m_windowSlaveRef;

public:
    class WindowExpiredException : public std::exception
    {
        inline const char* what (void) const throw() override
        {
            return "The ImageView Window handle has expired.";
        }
    };

    class Accessor
    {
    private:
        IVWindowWrapper &m_wrapper;
        DependentRef m_windowDependentRef;

    public:
        Accessor (IVWindowWrapper &wrapper)
            : m_wrapper (wrapper)
            , m_windowDependentRef (wrapper.m_windowSlaveRef)
        { }

        Accessor (Accessor &&to_take)
            : m_wrapper (to_take.m_wrapper)
            , m_windowDependentRef (std::move(to_take.m_windowDependentRef))
        { }

        inline bool good (void) const
        {
            return m_windowDependentRef && m_wrapper.m_window.good();
        }

        inline operator bool (void) const
        {
            return good();
        }

        inline bool operator! (void) const
        {
            return !good();
        }

        inline ImageView::Window& window (void) const
        {
            if (!good())
                throw WindowExpiredException();
            return m_wrapper.m_window;
        }

        inline IVAcqIfClass& acqIf (void) const
        {
            return *static_cast<IVAcqIfClass*>(window().acquisitionInterface());
        }

        inline IVFileIfClass& fileIf (void) const
        {
            return *static_cast<IVFileIfClass*>(window().fileInterface());
        }
    };

private:
    IVWindowWrapper (ImageView::ScopedAcquisitionInterface &&acqIfPtr, ImageView::ScopedFileInterface &&fileIfPtr)
        : m_window (*new ImageView::Window(std::move(acqIfPtr), std::move(fileIfPtr)))
    {
        auto &acqIf = *static_cast<IVAcqIfClass*>(m_window.acquisitionInterface());
        m_windowSlaveRef = acqIf.windowSlaveRef();
    }

    ~IVWindowWrapper (void)
    {
        delete &m_window;
    }

public:
    inline static IVWindowWrapper& Make (ImageView::ScopedAcquisitionInterface &&acqIf, ImageView::ScopedFileInterface &&fileIf)
    {
        return *new IVWindowWrapper(std::move(acqIf), std::move(fileIf));
    }

    inline static void Destroy (IVWindow window)
    {
        delete reinterpret_cast<IVWindowWrapper*>(window);
    }

    inline static bool Check (const ImageView::BaseClass *const base)
    {
        return !!dynamic_cast<const IVWindowWrapper*>(base);
    }

    inline static Accessor Get (IVWindow window)
    {
        return Accessor(*reinterpret_cast<IVWindowWrapper*>(window));
    }

    inline static Accessor Get (IVHandle handle)
    {
        return Accessor(*reinterpret_cast<IVWindowWrapper*>(handle));
    }

    inline operator IVWindow (void)
    {
        return reinterpret_cast<const IVWindow>(this);
    }
};

#endif // INCLUDED__BIT_FLOW__IMAGE__VIEW__C__API__PRIVATE__H
