/* FILE:        ImageViewImageAccessor.cpp
 * DATE:        4/4/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Methods for image data conversion and access.
 */

#include "ImageViewImageAccessor_Private.h"

ImageView::ImageAccessor::ImageAccessor (void)
    : m_pd (new PrivateData)
{
}
ImageView::ImageAccessor::~ImageAccessor (void)
{
    delete m_pd;
}

// Generate a QImage that is a suitable destination for the RawImage, matching all parameters.
bool ImageView::ImageAccessor::AllocateQImage (QImage &qImage, std::vector<uchar> &qImageData, const int width, const int height, const Format format) const
{
    if (isColor(format))
        qImage = m_pd->makeQImageColor(qImageData, width, height, (format & BaseFormatMask) == Color32);
    else
        qImage = m_pd->makeQImageGray(qImageData, width, height);

    return !qImage.isNull();
}

// Convert the QImage t a format that ImageView can use. RawImage (if provided) will be populated appropriately.
bool ImageView::ImageAccessor::NormalizeQImage (QImage &a_qImage, std::vector<uchar> &a_qImageData, RawImage *const a_pRawImage) const
{
    QImage cImage = a_qImage;

    // Make sure its' a valid format.
    switch (cImage.format())
    {
    case QImage::Format_Indexed8:
    case QImage::Format_RGB888:
    case QImage::Format_RGBX8888:
        break;

    default:
        if (cImage.isGrayscale())
            cImage = cImage.convertToFormat(QImage::Format_Indexed8);
        else
            cImage = cImage.convertToFormat(QImage::Format_RGBX8888);
    }

    // Determine the respective bytes-per-line values.
    int bytesPerLineIV = cImage.width();
    Format formatIV;
    switch (cImage.format())
    {
    case QImage::Format_Indexed8:
        formatIV = Gray8;
        break;
    case QImage::Format_RGB888:
        formatIV = Color24;
        bytesPerLineIV *= 3;
        break;
    case QImage::Format_RGBX8888:
        formatIV = Color32;
        bytesPerLineIV *= 4;
        break;

    default:
        return false;
    }

    // If necessary, convert to the ImageView bytes-per-line standard.
    const int bytesPerLineQt = cImage.bytesPerLine();
    if (bytesPerLineIV != bytesPerLineQt)
    {
        const int lineWidth = cImage.width();
        const int lineCnt = cImage.height();

        a_qImageData.resize(bytesPerLineIV * lineCnt);
        std::uint8_t *const imageDataPtr = a_qImageData.data();

        std::atomic<int> lineIndexAtomic (0);
        m_pd->m_imageConverter.logicalMultithreadedExe([&](size_t, size_t)
            {
                for (int lineIndex = lineIndexAtomic.fetch_add(1); lineIndex < lineCnt; lineIndex = lineIndexAtomic.fetch_add(1))
                {
                    const size_t lineOffsetIV = lineIndex * bytesPerLineIV;
                    std::uint8_t *const scanLineIV = imageDataPtr + lineOffsetIV;
                    const std::uint8_t *const scanLineQt = cImage.constScanLine(lineIndex);

                    memcpy(scanLineIV, scanLineQt, bytesPerLineIV);
                }
            });

        cImage = QImage (imageDataPtr, lineWidth, lineCnt, bytesPerLineIV, cImage.format());
    }

    if (cImage.isNull())
        return false;

    if (a_pRawImage)
    {
        a_pRawImage->m_buffer = cImage.constBits();
        a_pRawImage->m_width = cImage.width();
        a_pRawImage->m_height = cImage.height();
        a_pRawImage->m_format = formatIV;
    }

    a_qImage = cImage;

    return true;
}

// Convert the input image buffer to a QImage, using qImageData as the qImage data buffer. This
// performs a deep copy, and so the external buffer may be deallocated after calling.
bool ImageView::ImageAccessor::QImageFromBuffer (QImage &qImage, std::vector<uchar> &qImageData, RawImage const& rawImage) const
{
    Format allocFormat;
    if (isColor(rawImage.m_format) || (rawImage.m_format & BayerFormatMask))
        allocFormat = Color32;
    else
        allocFormat = Gray8;

    if (!AllocateQImage(qImage, qImageData, rawImage.m_width, rawImage.m_height, allocFormat))
        return false;

    return QImageFromBufferConst(qImage, qImageData, rawImage);
}
// Equivalent to QImageFromBuffer, but qImageData will not be reallocated.
bool ImageView::ImageAccessor::QImageFromBufferConst (QImage &qImage, std::vector<uchar> &qImageData, RawImage const& rawImage) const
{
    switch (rawImage.m_format & BaseFormatMask)
    {
    case Format::Gray8:
        if (rawImage.m_format & BayerFormatMask)
            return m_pd->imageDecodeBayer8(qImage, qImageData, rawImage);
        else
            return m_pd->imageCopy8(qImage, qImageData, rawImage);
    case Format::Gray10: case Format::Gray12: case Format::Gray14: case Format::Gray16:
        if (rawImage.m_format & BayerFormatMask)
            return m_pd->imageDecodeBayer16(qImage, qImageData, rawImage);
        else
            return m_pd->imageCopy16(qImage, qImageData, rawImage);
    case Format::Color24:
        return m_pd->imageCopy24(qImage, qImageData, rawImage);
    case Format::Color30:
        return m_pd->imageCopy30(qImage, qImageData, rawImage);
    case Format::Color32:
        return m_pd->imageCopy32(qImage, qImageData, rawImage);
    case Format::HiColor36: case Format::HiColor42: case Format::HiColor48:
        return m_pd->imageCopy48(qImage, qImageData, rawImage);
    }

    return false;
}
// Retrieve the buffer pixel value specified.
bool ImageView::ImageAccessor::BufferPixelValue (RawImage const& rawImage, PixPos const& pos, PixVal &val) const
{
    // Default values.
    val.isGrayscale = false;
    val.isColor = false;
    val.gray = -1;
    val.r = -1;
    val.g = -1;
    val.b = -1;

    // Validate our indices.
    if (pos.x >= rawImage.m_width || pos.y >= rawImage.m_height)
        return false;

    const size_t pixOffset = pos.x + rawImage.m_width * pos.y;
    const auto *const imageBuffer8 = reinterpret_cast<const std::uint8_t*>(rawImage.m_buffer);
    const auto *const imageBuffer16 = reinterpret_cast<const std::uint16_t*>(rawImage.m_buffer);
    const auto *const imageBuffer32 = reinterpret_cast<const std::uint32_t*>(rawImage.m_buffer);

    const int maskedFormat = rawImage.m_format & BaseFormatMask;
    switch (maskedFormat)
    {
    default:
        return false;
    case Format::Gray8:
        val.isGrayscale = true;
        val.gray = imageBuffer8[pixOffset];
        break;
    case Format::Gray10:
        val.isGrayscale = true;
        if ((rawImage.m_format & PackedFormatMask) == Packed)
        {
            Gray10PackedIterator<Packed> gray10iter (imageBuffer16 + pos.y * rawImage.m_width);
            gray10iter.advance(pos.x);
            val.gray = gray10iter.get_all();
            break;
        }
        else if ((rawImage.m_format & PackedFormatMask) == PackedNoSwap)
        {
            Gray10PackedIterator<PackedNoSwap> gray10iter (imageBuffer16 + pos.y * rawImage.m_width);
            gray10iter.advance(pos.x);
            val.gray = gray10iter.get_all();
            break;
        }
    case Format::Gray12: case Format::Gray14: case Format::Gray16:
        val.isGrayscale = true;
        val.gray = imageBuffer16[pixOffset];
        break;
    case Format::Color24:
        val.isColor = true;
        val.r = imageBuffer8[3 * pixOffset + 0];
        val.g = imageBuffer8[3 * pixOffset + 1];
        val.b = imageBuffer8[3 * pixOffset + 2];
        break;
    case Format::Color30:
        val.isColor = true;
        val.r = (imageBuffer32[pixOffset] >> 0) & 0x03FF;
        val.g = (imageBuffer32[pixOffset] >> 10) & 0x03FF;
        val.b = (imageBuffer32[pixOffset] >> 20) & 0x03FF;
        break;
    case Format::Color32:
        val.isColor = true;
        val.r = imageBuffer8[4 * pixOffset + 0];
        val.g = imageBuffer8[4 * pixOffset + 1];
        val.b = imageBuffer8[4 * pixOffset + 2];
        break;
    case Format::HiColor36: case Format::HiColor42: case Format::HiColor48:
        val.isColor = true;
        val.r = imageBuffer16[3 * pixOffset + 0] & ((1 << (maskedFormat / 3)) - 1);
        val.g = imageBuffer16[3 * pixOffset + 1] & ((1 << (maskedFormat / 3)) - 1);
        val.b = imageBuffer16[3 * pixOffset + 2] & ((1 << (maskedFormat / 3)) - 1);
        break;
    }

    if ((rawImage.m_format & ColorOrderFormatMask) == ColorOrderBGR)
        std::swap(val.r, val.b);

    return true;
}
