/* FILE:        ImageViewImageAccessor_Private.cpp
 * DATE:        4/4/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Methods for image data conversion and access.
 */

#include "ImageViewImageAccessor_Private.h"

#include "BayerDemosaic.h"

 // Direct buffer-to-QImage conversion.
QImage ImageView::ImageAccessor::PrivateData::makeQImageGray (std::vector<uchar> &qImageData, const int width, const int height)
{
    const size_t imgLineSize = width;
    const size_t imgByteSize = imgLineSize * height;

    if (qImageData.size() != imgByteSize)
    {
        try
        {
            qImageData.resize(imgByteSize);
        }
        catch (...)
        {
            return QImage();
        }
    }

    return QImage(qImageData.data(), (int)width, (int)height, (int)imgLineSize, QImage::Format_Indexed8);
}
QImage ImageView::ImageAccessor::PrivateData::makeQImageColor (std::vector<uchar> &qImageData, const int width, const int height, const bool alpha)
{
    const size_t imgLineSize = width * (alpha ? 4 : 3);
    const size_t imgByteSize = imgLineSize * height;
    const QImage::Format format = alpha ? QImage::Format_RGBX8888 : QImage::Format_RGB888;

    if (qImageData.size() != imgByteSize)
    {
        try
        {
            qImageData.resize(imgByteSize);
        }
        catch (...)
        {
            return QImage();
        }
    }

    return QImage(qImageData.data(), (int)width, (int)height, (int)imgLineSize, format);
}

bool ImageView::ImageAccessor::PrivateData::validateQImageGray (QImage &image, const int width, const int height)
{
    return !image.isNull()
        && image.width() == width
        && image.height() == height
        && image.bytesPerLine() == width
        && image.format() == QImage::Format_Indexed8;
}
bool ImageView::ImageAccessor::PrivateData::validateQImageColor (QImage &image, const int width, const int height)
{
    if (image.isNull() || image.width() != width || image.height() != height)
        return false;

    switch (image.format())
    {
    case QImage::Format_RGB888:
        return image.bytesPerLine() == width * 3;
    case QImage::Format_RGBX8888:
        return image.bytesPerLine() == width * 4;
    default:
        break;
    }

    return false;
}

bool ImageView::ImageAccessor::PrivateData::imageCopy8 (QImage &image, std::vector<uchar> &qImageData, RawImage const& rawImage)
{
    if (!validateQImageGray(image, rawImage.m_width, rawImage.m_height))
        return false;

    memcpy(qImageData.data(), rawImage.m_buffer, qImageData.size());

    return true;
}
bool ImageView::ImageAccessor::PrivateData::imageCopy16 (QImage &image, std::vector<uchar> &qImageData, RawImage const& rawImage)
{
    if (!validateQImageGray(image, rawImage.m_width, rawImage.m_height))
        return false;

    switch (rawImage.m_format & BaseFormatMask)
    {
    case Gray10:
        switch (rawImage.m_format & PackedFormatMask)
        {
        case Packed:
            {
                ThreadableCopyGray10Packed copier (rawImage, qImageData.data(), image.bytesPerLine());
                m_imageConverter.logicalMultithreadedExe(copier);
            }
            break;
        case PackedNoSwap:
            {
                ThreadableCopyGray10PackedNoSwap copier (rawImage, qImageData.data(), image.bytesPerLine());
                m_imageConverter.logicalMultithreadedExe(copier);
            }
            break;
        default:
            {
                ThreadableCopyGray10 copier (rawImage, qImageData.data(), image.bytesPerLine());
                m_imageConverter.logicalMultithreadedExe(copier);
            }
        }
        return true;
    case Gray12:
        {
            ThreadableCopyGray12 copier (rawImage, qImageData.data(), image.bytesPerLine());
            m_imageConverter.logicalMultithreadedExe(copier);
        }
        return true;
    case Gray14:
        {
            ThreadableCopyGray14 copier (rawImage, qImageData.data(), image.bytesPerLine());
            m_imageConverter.logicalMultithreadedExe(copier);
        }
        return true;
    case Gray16:
        {
            ThreadableCopyGray16 copier (rawImage, qImageData.data(), image.bytesPerLine());
            m_imageConverter.logicalMultithreadedExe(copier);
        }
        return true;
    }

    return false;
}

bool ImageView::ImageAccessor::PrivateData::imageCopy24 (QImage &image, std::vector<uchar> &qImageData, RawImage const& rawImage)
{
    if (!validateQImageColor(image, rawImage.m_width, rawImage.m_height))
        return false;

    const bool isRgbx = image.format() == QImage::Format_RGBX8888;

    ThreadableCopyColor24 copier (rawImage, qImageData.data(), image.bytesPerLine(), isRgbx);
    m_imageConverter.logicalMultithreadedExe(copier);

    return true;
}
bool ImageView::ImageAccessor::PrivateData::imageCopy30 (QImage &image, std::vector<uchar> &qImageData, RawImage const& rawImage)
{
    if (!validateQImageColor(image, rawImage.m_width, rawImage.m_height))
        return false;

    const bool isRgbx = image.format() == QImage::Format_RGBX8888;

    ThreadableCopyColor30 copier (rawImage, qImageData.data(), image.bytesPerLine(), isRgbx);
    m_imageConverter.logicalMultithreadedExe(copier);

    return true;
}
bool ImageView::ImageAccessor::PrivateData::imageCopy32 (QImage &image, std::vector<uchar> &qImageData, RawImage const& rawImage)
{
    if (!validateQImageColor(image, rawImage.m_width, rawImage.m_height))
        return false;

    const bool isRgbx = image.format() == QImage::Format_RGBX8888;

    ThreadableCopyColor32 copier (rawImage, qImageData.data(), image.bytesPerLine(), isRgbx);
    m_imageConverter.logicalMultithreadedExe(copier);

    return true;
}
bool ImageView::ImageAccessor::PrivateData::imageCopy48 (QImage &image, std::vector<uchar> &qImageData, RawImage const& rawImage)
{
    if (!validateQImageColor(image, rawImage.m_width, rawImage.m_height))
        return false;

    const bool isRgbx = image.format() == QImage::Format_RGBX8888;

    switch (rawImage.m_format & BaseFormatMask)
    {
    case HiColor36:
    {
        ThreadableCopyHiColor36 copier (rawImage, qImageData.data(), image.bytesPerLine(), isRgbx);
        m_imageConverter.logicalMultithreadedExe(copier);
    }
    return true;

    case HiColor42:
    {
        ThreadableCopyHiColor42 copier (rawImage, qImageData.data(), image.bytesPerLine(), isRgbx);
        m_imageConverter.logicalMultithreadedExe(copier);
    }
    return true;

    case HiColor48:
    {
        ThreadableCopyHiColor48 copier (rawImage, qImageData.data(), image.bytesPerLine(), isRgbx);
        m_imageConverter.logicalMultithreadedExe(copier);
    }
    return true;
    }

    return false;
}

// Demosaic buffer-to-QImage conversion.
bool ImageView::ImageAccessor::PrivateData::imageDecodeBayer8 (QImage &image, std::vector<uchar> &qImageData, RawImage const& rawImage)
{
    if (rawImage.m_width < 4 || rawImage.m_height < 4)
        return false;

    if (!validateQImageColor(image, rawImage.m_width, rawImage.m_height))
        return false;

    const bool isRgbx = image.format() == QImage::Format_RGBX8888;

    return BayerDecoder(m_imageConverter).decode8(qImageData.data(), image.bytesPerLine(), isRgbx, rawImage);
}
bool ImageView::ImageAccessor::PrivateData::imageDecodeBayer16 (QImage &image, std::vector<uchar> &qImageData, RawImage const& rawImage)
{
    if (rawImage.m_width < 4 || rawImage.m_height < 4)
        return false;

    if (!validateQImageColor(image, rawImage.m_width, rawImage.m_height))
        return false;

    const bool isRgbx = image.format() == QImage::Format_RGBX8888;

    return BayerDecoder(m_imageConverter).decode16(qImageData.data(), image.bytesPerLine(), isRgbx, rawImage);
}
