/* FILE:        ImageViewMessageBoxDialog.cpp
 * DATE:        7/6/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Public display MessageBox class for the ImageView library. May only be
 *              used after construction of the ImageView::Application global object.
 *              Use Qt's QMessageBox, for greater versatility.
 */

#include <QApplication>
#include <QThread>

#include "ImageViewMessageBoxDialog_Private.h"

#include "ImageViewWindow.h"

const std::pair<ImageView::DialogButton, QMessageBox::StandardButton> ImageView::MessageBoxDialog::PrivateData::buttonTypePairs [] =
{
    std::make_pair(ImageView::DialogButton::NoButton, QMessageBox::NoButton),
    std::make_pair(ImageView::DialogButton::Ok, QMessageBox::Ok),
    std::make_pair(ImageView::DialogButton::Open, QMessageBox::Open),
    std::make_pair(ImageView::DialogButton::Save, QMessageBox::Save),
    std::make_pair(ImageView::DialogButton::Cancel, QMessageBox::Cancel),
    std::make_pair(ImageView::DialogButton::Close, QMessageBox::Close),
    std::make_pair(ImageView::DialogButton::Discard, QMessageBox::Discard),
    std::make_pair(ImageView::DialogButton::Apply, QMessageBox::Apply),
    std::make_pair(ImageView::DialogButton::Reset, QMessageBox::Reset),
    std::make_pair(ImageView::DialogButton::RestoreDefaults, QMessageBox::RestoreDefaults),
    std::make_pair(ImageView::DialogButton::Help, QMessageBox::Help),
    std::make_pair(ImageView::DialogButton::SaveAll, QMessageBox::SaveAll),
    std::make_pair(ImageView::DialogButton::Yes, QMessageBox::Yes),
    std::make_pair(ImageView::DialogButton::YesToAll, QMessageBox::YesToAll),
    std::make_pair(ImageView::DialogButton::No, QMessageBox::No),
    std::make_pair(ImageView::DialogButton::NoToAll, QMessageBox::NoToAll),
    std::make_pair(ImageView::DialogButton::Abort, QMessageBox::Abort),
    std::make_pair(ImageView::DialogButton::Retry, QMessageBox::Retry),
    std::make_pair(ImageView::DialogButton::Ignore, QMessageBox::Ignore)
};

ImageView::DialogButton ImageView::MessageBoxDialog::PrivateData::qt2iv (const QMessageBox::StandardButton qtButton)
{
    for (auto const& bp : buttonTypePairs)
    {
        if (bp.second == qtButton)
            return bp.first;
    }

    return ImageView::DialogButton::NoButton;
}
QMessageBox::StandardButton ImageView::MessageBoxDialog::PrivateData::iv2qt (const ImageView::DialogButton ivButton)
{
    for (auto const& bp : buttonTypePairs)
    {
        if (bp.first == ivButton)
            return bp.second;
    }

    return QMessageBox::NoButton;
}
QMessageBox::StandardButtons ImageView::MessageBoxDialog::PrivateData::iv2qt (ImageView::DialogButtons const& ivButtons)
{
    QMessageBox::StandardButtons qtButtons;

    for (auto const& bp : buttonTypePairs)
    {
        if (bp.first & ivButtons)
            qtButtons |= bp.second;
    }

    return qtButtons;
}

bool ImageView::MessageBoxDialog::PrivateData::currentIsGuiThread (void)
{
    return QApplication::instance()->thread() == QThread::currentThread();
}

ImageView::MessageBoxDialog::PrivateData::PrivateData (Window *const parentWindow, const char *const title, const char *const text, DialogButtons const& buttons, const DialogButton defaultButton)
    : m_parent (parentWindow->qt_handle())
    , m_title (title)
    , m_text (text)
    , m_buttons (iv2qt(buttons))
    , m_defaultButton (iv2qt(defaultButton))
    , m_result (QMessageBox::NoButton)
{
    moveToThread(QApplication::instance()->thread());

    connect(&m_signalBox, SIGNAL(critical(void)), this, SLOT(doCritical(void)), Qt::BlockingQueuedConnection);
    connect(&m_signalBox, SIGNAL(warning(void)), this, SLOT(doWarning(void)), Qt::BlockingQueuedConnection);
    connect(&m_signalBox, SIGNAL(information(void)), this, SLOT(doInformation(void)), Qt::BlockingQueuedConnection);
    connect(&m_signalBox, SIGNAL(question(void)), this, SLOT(doQuestion(void)), Qt::BlockingQueuedConnection);
}

ImageView::DialogButton ImageView::MessageBoxDialog::PrivateData::critical (void)
{
    if (currentIsGuiThread())
        doCritical();
    else
        emit m_signalBox.critical();

    return qt2iv(m_result);
}
ImageView::DialogButton ImageView::MessageBoxDialog::PrivateData::warning (void)
{
    if (currentIsGuiThread())
        doWarning();
    else
        emit m_signalBox.warning();

    return qt2iv(m_result);
}
ImageView::DialogButton ImageView::MessageBoxDialog::PrivateData::information (void)
{
    if (currentIsGuiThread())
        doInformation();
    else
        emit m_signalBox.information();

    return qt2iv(m_result);
}
ImageView::DialogButton ImageView::MessageBoxDialog::PrivateData::question (void)
{
    if (currentIsGuiThread())
        doQuestion();
    else
        emit m_signalBox.question();

    return qt2iv(m_result);
}

void ImageView::MessageBoxDialog::PrivateData::doCritical (void)
{
    m_result = QMessageBox::critical(m_parent, m_title, m_text, m_buttons, m_defaultButton);
}
void ImageView::MessageBoxDialog::PrivateData::doWarning (void)
{
    m_result = QMessageBox::warning(m_parent, m_title, m_text, m_buttons, m_defaultButton);
}
void ImageView::MessageBoxDialog::PrivateData::doInformation (void)
{
    m_result = QMessageBox::information(m_parent, m_title, m_text, m_buttons, m_defaultButton);
}
void ImageView::MessageBoxDialog::PrivateData::doQuestion (void)
{
    m_result = QMessageBox::question(m_parent, m_title, m_text, m_buttons, m_defaultButton);
}

// Create an error message box, returning the user button selection. This will be
// a modal dialog of the parent window. The text may contain basic HTML elements.
ImageView::DialogButton ImageView::MessageBoxDialog::critical (Window *const parentWindow, const char *const title, const char *const text, DialogButtons const& buttons, const DialogButton defaultButton)
{
    return PrivateData(parentWindow, title, text, buttons, defaultButton).critical();
}

// Create a warning message box, returning the user button selection. This will be
// a modal dialog of the parent window. The text may contain basic HTML elements.
ImageView::DialogButton ImageView::MessageBoxDialog::warning (Window *const parentWindow, const char *const title, const char *const text, DialogButtons const& buttons, const DialogButton defaultButton)
{
    return PrivateData(parentWindow, title, text, buttons, defaultButton).warning();
}

// Create a notification message box, returning the user button selection. This will be
// a modal dialog of the parent window. The text may contain basic HTML elements.
ImageView::DialogButton ImageView::MessageBoxDialog::information (Window *const parentWindow, const char *const title, const char *const text, DialogButtons const& buttons, const DialogButton defaultButton)
{
    return PrivateData(parentWindow, title, text, buttons, defaultButton).information();
}

// Create an inquiry message box, returning the user button selection. This will be
// a modal dialog of the parent window. The text may contain basic HTML elements.
ImageView::DialogButton ImageView::MessageBoxDialog::question (Window *const parentWindow, const char *const title, const char *const text, DialogButtons const& buttons, const DialogButton defaultButton)
{
    return PrivateData(parentWindow, title, text, buttons, defaultButton).question();
}
