/* FILE:        ImageViewMiniMap.cpp
 * DATE:        12/23/2015
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2015, BitFlow, Inc.
 * DESCRIPTION: A MiniMap style widget for the ImageView library, allowing an image to
 *              be panned very quickly.
 */

#include "ImageViewMiniMap_Private.h"

#include <QPainter>

#include <QShortcut>

#include <QPaintEvent>
#include <QShowEvent>
#include <QMouseEvent>
#include <QWheelEvent>

ImageView::MiniMap::PrivateData::PrivateData (MiniMap &miniMap)
    : QObject (&miniMap)
    , m_miniMap (miniMap)
    , m_boundarySize (256, 256)
    , m_scaling (1.)
    , m_mousePress (false)
    , m_mouseDrag (false)
    , m_mouseOutlineHover (false)
{
    m_miniMap.setMouseTracking(true);
    m_miniMap.setCursor(Qt::PointingHandCursor);

    new QShortcut(QKeySequence::ZoomIn, &m_miniMap, SIGNAL(scrollIn()));
    new QShortcut(QKeySequence::ZoomOut, &m_miniMap, SIGNAL(scrollOut()));

    auto *const shortcutUp = new QShortcut(Qt::Key_Up, &m_miniMap, SLOT(moveOffsetUp()));
    auto *const shortcutDown = new QShortcut(Qt::Key_Down, &m_miniMap, SLOT(moveOffsetDown()));
    auto *const shortcutLeft = new QShortcut(Qt::Key_Left, &m_miniMap, SLOT(moveOffsetLeft()));
    auto *const shortcutRight = new QShortcut(Qt::Key_Right, &m_miniMap, SLOT(moveOffsetRight()));

    connect(shortcutUp, SIGNAL(activated()), this, SLOT(emitOffsetChanged()));
    connect(shortcutDown, SIGNAL(activated()), this, SLOT(emitOffsetChanged()));
    connect(shortcutLeft, SIGNAL(activated()), this, SLOT(emitOffsetChanged()));
    connect(shortcutRight, SIGNAL(activated()), this, SLOT(emitOffsetChanged()));

    auto *const shortcutPageUp = new QShortcut(Qt::CTRL + Qt::Key_Up, &m_miniMap, SLOT(pageOffsetUp()));
    auto *const shortcutPageDown = new QShortcut(Qt::CTRL + Qt::Key_Down, &m_miniMap, SLOT(pageOffsetDown()));
    auto *const shortcutPageLeft = new QShortcut(Qt::CTRL + Qt::Key_Left, &m_miniMap, SLOT(pageOffsetLeft()));
    auto *const shortcutPageRight = new QShortcut(Qt::CTRL + Qt::Key_Right, &m_miniMap, SLOT(pageOffsetRight()));

    connect(shortcutPageUp, SIGNAL(activated()), this, SLOT(emitOffsetChanged()));
    connect(shortcutPageDown, SIGNAL(activated()), this, SLOT(emitOffsetChanged()));
    connect(shortcutPageLeft, SIGNAL(activated()), this, SLOT(emitOffsetChanged()));
    connect(shortcutPageRight, SIGNAL(activated()), this, SLOT(emitOffsetChanged()));
}

ImageView::MiniMap::MiniMap (QWidget *const parent)
    : QWidget (parent)
    , m_pd (*new PrivateData(*this))
{ }

ImageView::MiniMap::~MiniMap (void)
{
    delete &m_pd;
}

// Helpers.
void ImageView::MiniMap::PrivateData::localUpdate (void)
{
    const QSizeF pixSize = m_pixmap.size();
    const QSizeF boundarySizeF = m_boundarySize;

    const double widthRatio = boundarySizeF.width() / pixSize.width();
    const double heightRatio = boundarySizeF.height() / pixSize.height();

    if (widthRatio < heightRatio)
    {
        m_miniMap.setFixedSize(boundarySizeF.width(), boundarySizeF.height() * pixSize.height() / pixSize.width());
        m_localScaling = widthRatio;
    }
    else
    {
        m_miniMap.setFixedSize(boundarySizeF.width() * pixSize.width() / pixSize.height(), boundarySizeF.height());
        m_localScaling = heightRatio;
    }

    const QSizeF offsetRange = pixSize - m_viewRect.size();
    m_viewRect.moveTo
    (
        std::max(0., std::min(offsetRange.width(), m_viewRect.left())),
        std::max(0., std::min(offsetRange.height(), m_viewRect.top()))
    );

    cursorUpdate();

    m_miniMap.update();
}
void ImageView::MiniMap::PrivateData::cursorUpdate (QPoint const& pos)
{
    const auto viewRect = scaledViewRect();
    const bool outlineVisible = viewRect.isValid();

    const bool oldMouseOutlineHover = m_mouseOutlineHover;
    m_mouseOutlineHover = outlineVisible && viewRect.contains(m_miniMap.mapFromGlobal(pos));

    if (!outlineVisible)
        m_miniMap.setCursor(Qt::ArrowCursor);
    else if (m_mouseDrag || m_mousePress)
        m_miniMap.setCursor(Qt::ClosedHandCursor);
    else if (m_mouseOutlineHover)
        m_miniMap.setCursor(Qt::OpenHandCursor);
    else
        m_miniMap.setCursor(Qt::PointingHandCursor);

    if (oldMouseOutlineHover != m_mouseOutlineHover)
        m_miniMap.update();
}

QRectF ImageView::MiniMap::PrivateData::scaledViewRect (void) const
{
    const auto localRect = m_miniMap.rect();

    const auto viewPos = m_viewRect.topLeft() * m_localScaling;
    const auto viewSize = m_viewRect.size() * m_localScaling;
    const QRectF viewRectMin (viewPos + QPointF(viewSize.width() - 5, viewSize.height() - 5) / 2, QSizeF(5, 5));
    const auto viewRect = QRectF(viewPos, viewSize).united(viewRectMin);

    if (localRect.width() <= viewRect.width() && localRect.height() <= viewRect.height())
        return QRectF();

    return viewRect;
}

QPixmap ImageView::MiniMap::pixmap (void) const
{
    return m_pd.m_pixmap;
}
void ImageView::MiniMap::setPixmap (QPixmap const& value)
{
    m_pd.m_pixmap = value;
    m_pd.localUpdate();
}

QSizeF ImageView::MiniMap::boundarySize (void) const
{
    return m_pd.m_boundarySize;
}
void ImageView::MiniMap::setBoundarySize (QSizeF const& value)
{
    m_pd.m_boundarySize = value;
    m_pd.localUpdate();
}
void ImageView::MiniMap::setBoundarySize (const double width, const double height)
{
    setBoundarySize(QSizeF(width, height));
}

QRectF ImageView::MiniMap::viewRect (void) const
{
    return m_pd.m_viewRect;
}
void ImageView::MiniMap::setViewRect (QRectF const& value)
{
    m_pd.m_viewRect = value;
    m_pd.localUpdate();
}
void ImageView::MiniMap::setViewRect (const double x, const double y, const double width, const double height)
{
    setViewRect(QRectF(x, y, width, height));
}

QSizeF ImageView::MiniMap::viewSize (void) const
{
    return m_pd.m_viewRect.size();
}
void ImageView::MiniMap::setViewSize (QSizeF const& value)
{
    m_pd.m_viewRect.setSize(value);
    m_pd.localUpdate();
}
void ImageView::MiniMap::setViewSize (const double width, const double height)
{
    setViewSize(QSizeF(width, height));
}

QPointF ImageView::MiniMap::offset (void) const
{
    return m_pd.m_viewRect.topLeft();
}
void ImageView::MiniMap::setOffset (QPointF const& value)
{
    m_pd.m_viewRect.moveTo(value);
    m_pd.localUpdate();
}
void ImageView::MiniMap::setOffset (const double x, const double y)
{
    m_pd.m_viewRect.moveTo(x, y);
    m_pd.localUpdate();
}

void ImageView::MiniMap::moveOffsetOver (QPointF const& pos)
{
    m_pd.m_viewRect.moveCenter(pos);
    m_pd.localUpdate();
}
void ImageView::MiniMap::moveOffsetOver (const double x, const double y)
{
    moveOffsetOver(QPointF(x, y));
}

void ImageView::MiniMap::moveOffsetUp (void)
{
    setOffset(offset() + QPointF(0, -1));
}
void ImageView::MiniMap::moveOffsetDown (void)
{
    setOffset(offset() + QPointF(0, 1));
}
void ImageView::MiniMap::moveOffsetLeft (void)
{
    setOffset(offset() + QPointF(-1, 0));
}
void ImageView::MiniMap::moveOffsetRight (void)
{
    setOffset(offset() + QPointF(1, 0));
}

void ImageView::MiniMap::pageOffsetUp (void)
{
    setOffset(offset() + QPointF(0, -m_pd.m_viewRect.height()));
}
void ImageView::MiniMap::pageOffsetDown (void)
{
    setOffset(offset() + QPointF(0, m_pd.m_viewRect.height()));
}
void ImageView::MiniMap::pageOffsetLeft (void)
{
    setOffset(offset() + QPointF(-m_pd.m_viewRect.width(), 0));
}
void ImageView::MiniMap::pageOffsetRight (void)
{
    setOffset(offset() + QPointF(m_pd.m_viewRect.width(), 0));
}

double ImageView::MiniMap::scaling (void) const
{
    return m_pd.m_scaling;
}
void ImageView::MiniMap::setScaling (const double value)
{
    m_pd.m_scaling = value;
    m_pd.cursorUpdate();
}

void ImageView::MiniMap::PrivateData::emitOffsetChanged (void)
{
    emit m_miniMap.offsetChanged(m_miniMap.offset() * m_scaling);
}

// Event overrides.
void ImageView::MiniMap::paintEvent (QPaintEvent *const evt)
{
    QPainter painter (this);

    auto viewRect = m_pd.scaledViewRect();

    painter.drawPixmap(rect(), m_pd.m_pixmap);

    if (viewRect.isValid())
    {
        const bool tiny = viewRect.width() <= 5 || viewRect.height() <= 5;
        const bool mouseActive = m_pd.m_mouseOutlineHover || m_pd.m_mousePress || m_pd.m_mouseDrag;

        viewRect.adjust(0, 0, -1, -1);

        const QBrush fillBrush (QColor(255, 255, 255, tiny ? 255 : (mouseActive ? 180 : 100)));

        QPen outlinePen (Qt::SolidLine);
        outlinePen.setWidth(1);
        outlinePen.setCapStyle(Qt::SquareCap);
        outlinePen.setJoinStyle(Qt::MiterJoin);

        painter.fillRect(viewRect, fillBrush);

        outlinePen.setColor(Qt::darkGray);
        painter.setPen(outlinePen);
        painter.drawRect(QRectF(viewRect.topLeft() + QPointF(1, 1), viewRect.size() - QSizeF(2, 2)));

        outlinePen.setColor(mouseActive ? Qt::green : Qt::red);
        painter.setPen(outlinePen);
        painter.drawRect(viewRect);
    }
}

void ImageView::MiniMap::showEvent (QShowEvent *const evt)
{
    QWidget::showEvent(evt);
    m_pd.localUpdate();
}

void ImageView::MiniMap::resizeEvent (QResizeEvent *const evt)
{
    QWidget::resizeEvent(evt);
    m_pd.localUpdate();
}

void ImageView::MiniMap::enterEvent (QEvent *const evt)
{
    m_pd.cursorUpdate();

    update();
    QWidget::leaveEvent(evt);
}
void ImageView::MiniMap::leaveEvent (QEvent *const evt)
{
    m_pd.m_mouseOutlineHover = false;

    update();
    QWidget::leaveEvent(evt);
}

void ImageView::MiniMap::mousePressEvent (QMouseEvent *const evt)
{
    const auto viewPos = evt->pos() / m_pd.m_localScaling;

    if (evt->button() == Qt::LeftButton)
    {
        if (m_pd.m_viewRect.contains(viewPos))
        {
            m_pd.m_mouseDrag = true;
            m_pd.m_offsetInit = offset();
            m_pd.m_mousePressPos = evt->globalPos();
        }
        else
        {
            m_pd.m_mousePress = true;
            moveOffsetOver(viewPos);
            emit offsetChanged(offset() * m_pd.m_scaling);
        }

        evt->accept();
    }

    m_pd.cursorUpdate(evt->globalPos());
}
void ImageView::MiniMap::mouseMoveEvent (QMouseEvent *const evt)
{
    if (m_pd.m_mouseDrag)
    {
        const QPointF localDiff = evt->globalPos() - m_pd.m_mousePressPos;
        if (localDiff.manhattanLength() >= 3)
        {
            const QPointF offsetDiff = (QPointF(localDiff) / m_pd.m_localScaling).toPoint();
            setOffset(m_pd.m_offsetInit + offsetDiff);
            emit offsetChanged(offset() * m_pd.m_scaling);
        }
    }
    else if (m_pd.m_mousePress)
    {
        moveOffsetOver(evt->pos() / m_pd.m_localScaling);
        emit offsetChanged(offset() * m_pd.m_scaling);

        evt->accept();
    }

    m_pd.cursorUpdate(evt->globalPos());
}
void ImageView::MiniMap::mouseReleaseEvent (QMouseEvent *const evt)
{
    if (evt->button() == Qt::LeftButton)
    {
        m_pd.m_mouseDrag = false;
        m_pd.m_mousePress = false;

        evt->accept();
    }

    m_pd.cursorUpdate(evt->globalPos());
}

void ImageView::MiniMap::wheelEvent (QWheelEvent *const evt)
{
    const int deltaY = evt->angleDelta().y();
    if (deltaY > 0)
        emit scrollIn();
    else if (deltaY < 0)
        emit scrollOut();
}
