/* FILE:        ImageViewNative.cpp
 * DATE:        7/15/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Tools and definitions for integration with the current operating system.
 */

#include "ImageViewNative_Private.h"
#include "ImageViewStringHelpers.h"
#include "ImageViewVersionNo.h"

#include <iostream>

#if defined(_WIN32)

#include <cstdint>

#include <sstream>

#include <map>
#include <set>
#include <vector>

#include <mutex>

static std::map<DWORD, HHOOK> g_wndRetProcHooks;
static std::map<IMAGE_VIEW_NATIVE_WINDOW, std::set<IMAGE_VIEW_NATIVE_WINDOW>> g_parentChildWindows;
static std::mutex g_parentChildWindowsMutex;

// Win32 specific methods.
static LRESULT WINAPI ImageViewApplicationCallWndProc (int nCode, WPARAM wParam, LPARAM lParam)
{
    HHOOK hHook = NULL;
    PCWPSTRUCT paramStruct = (PCWPSTRUCT)lParam;

    HWND parentWindow = NULL;
    std::vector<HWND> childWindows;

    // Find all windows that need to be moved, and retrieve the hook handle.
    {
        const std::lock_guard<std::mutex> parentChildWindowsLock (g_parentChildWindowsMutex);

        const auto wdIter = g_parentChildWindows.find(paramStruct->hwnd);;
        if (g_parentChildWindows.end() != wdIter)
        {
            parentWindow = wdIter->first;

            if (WM_WINDOWPOSCHANGED == paramStruct->message)
            {
                // If we have a parent window, make sure we're above it.
                HWND child = parentWindow;
                while (child = GetNextWindow(child, GW_HWNDNEXT))
                {
                    if (wdIter->second.find(child) != wdIter->second.end())
                        childWindows.push_back(child);
                }
            }
        }

        // Retrieve the hook handle.
        const auto hookIter = g_wndRetProcHooks.find(GetCurrentThreadId());
        if (g_wndRetProcHooks.end() != hookIter)
            hHook = hookIter->second;
    }

    // Perform the shuffle.
    if (childWindows.size())
    {
        if (WM_WINDOWPOSCHANGED == paramStruct->message)
        {
            // Move the children up to the parent.
            HWND moveUnder = parentWindow;
            for (auto const& child : childWindows)
            {
                SetWindowPos(child, moveUnder, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
                moveUnder = child;
            }

            // Place the parent behind the lowest child.
            SetWindowPos(parentWindow, moveUnder, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
        }
    }

    // If we have a hook handle, forward the hook.
    if (hHook)
        return CallNextHookEx(hHook, nCode, wParam, lParam);
    else
        return NULL;
}

void ImageView::Native::Win32::SetParentChildWindow (IMAGE_VIEW_NATIVE_WINDOW parent, IMAGE_VIEW_NATIVE_WINDOW child)
{
    const std::lock_guard<std::mutex> parentChildWindowsLock (g_parentChildWindowsMutex);

    // Make sure we have an appropriate hook.
    const DWORD windowThreadId = GetWindowThreadProcessId(parent, NULL);
    const auto wndProcHook = g_wndRetProcHooks.find(windowThreadId);
    if (g_wndRetProcHooks.end() == wndProcHook)
    {
        const HHOOK hook = SetWindowsHookEx(WH_CALLWNDPROC, ImageViewApplicationCallWndProc, GetModuleHandle(NULL), windowThreadId);
        if (hook)
            g_wndRetProcHooks.insert(std::make_pair(windowThreadId, hook));
    }

    // Setup the parent-child pair.
    g_parentChildWindows[parent].insert(child);
}
void ImageView::Native::Win32::ClearParentChildWindow (IMAGE_VIEW_NATIVE_WINDOW parent, IMAGE_VIEW_NATIVE_WINDOW child)
{
    const std::lock_guard<std::mutex> parentChildWindowsLock (g_parentChildWindowsMutex);

    const auto childSetIter = g_parentChildWindows.find(parent);
    if (g_parentChildWindows.end() != childSetIter)
    {
        childSetIter->second.erase(child);
        if (childSetIter->second.empty())
            g_parentChildWindows.erase(childSetIter);
    }
}

// General platform-specific methods.
bool ImageView::Native::LibraryVersion (int &major, int &minor, bool &debug)
{
    major = IMAGE_VIEW_SDK_VERSION_MAJOR & ~IMAGE_VIEW_SDK_VERSION_BUG;
    minor = IMAGE_VIEW_SDK_VERSION_MINOR;
    debug = !!(IMAGE_VIEW_SDK_VERSION_MAJOR & IMAGE_VIEW_SDK_VERSION_BUG);
    return true;
}

bool ImageView::Native::DriverVersion (int &major, int &minor, bool &debug)
{
    bool success = false;
    if (HMODULE bfdDll = LoadLibrary(TEXT("BFD.dll")))
    {
        if (auto BFDriverVersion = (std::uint32_t(*)(std::uint32_t*, std::uint32_t*))GetProcAddress(bfdDll, "BFDriverVersion"))
        {
            std::uint32_t majorVer, minorVer;
            BFDriverVersion(&majorVer, &minorVer);

            major = majorVer & ~IMAGE_VIEW_SDK_VERSION_BUG;
            minor = minorVer;
            debug = !!(majorVer & IMAGE_VIEW_SDK_VERSION_BUG);
            success = true;
        }
        FreeLibrary(bfdDll);
    }
    return success;
}

void ImageView::Native::Cleanup (void)
{
    const std::lock_guard<std::mutex> parentChildWindowsLock (g_parentChildWindowsMutex);

    for (auto const& hookPair : g_wndRetProcHooks)
        UnhookWindowsHookEx(hookPair.second);
    g_wndRetProcHooks.clear();

    g_parentChildWindows.clear();
}

bool ImageView::Native::LibraryVersionStr (char *const strBuf, size_t *const strBufSize)
{
    int major, minor;
    bool debug;
    if (!LibraryVersion(major, minor, debug))
        return false;

    std::stringstream verStrStrm;
    verStrStrm << std::dec << major << "." << std::hex << minor;
    if (debug)
        verStrStrm << " [debug]";

    return ImageView::String::set(verStrStrm.str(), strBuf, strBufSize);
}

bool ImageView::Native::DriverVersionStr (char *const strBuf, size_t *const strBufSize)
{
    int major, minor;
    bool debug;
    if (!DriverVersion(major, minor, debug))
        return false;

    std::stringstream verStrStrm;
    verStrStrm << std::dec << major << "." << std::hex << minor;
    if (debug)
        verStrStrm << " [debug]";

    return ImageView::String::set(verStrStrm.str(), strBuf, strBufSize);
}

#elif defined(__GNUC__)

#include <BFciLib.h>

#include <sstream>

// General platform-specific methods.
bool ImageView::Native::LibraryVersion (int &major, int &minor, bool &debug)
{
    tCIU32 libVer, drvVer;
    if (kCIEnoErr == CiSysGetVersions(&libVer, &drvVer))
    {
        major = (libVer >> 24) & 0x000000FF;
        minor = libVer & 0x00FFFFFF;
        debug = false;
        return true;
    }

    return false;
}

bool ImageView::Native::DriverVersion (int &major, int &minor, bool &debug)
{
    tCIU32 libVer, drvVer;
    if (kCIEnoErr == CiSysGetVersions(&libVer, &drvVer))
    {
        major = (drvVer >> 24) & 0x000000FF;
        minor = drvVer & 0x00FFFFFF;
        debug = false;
        return true;
    }

    return false;
}

void ImageView::Native::Cleanup(void)
{
}

bool ImageView::Native::LibraryVersionStr (char *const strBuf, size_t *const strBufSize)
{
    int major, minor;
    bool debug;
    if (!LibraryVersion(major, minor, debug))
        return false;

    std::stringstream verStrStrm;
    verStrStrm << std::dec << major << ".";
    verStrStrm.width(2);
    verStrStrm.fill('0');
    verStrStrm << std::hex << ((minor >> 16) & 0xFF);
    verStrStrm << ", Build " << std::hex << ((minor >> 8) & 0xFFFF);
    if (debug)
        verStrStrm << " [debug]";

    return ImageView::String::set(verStrStrm.str(), strBuf, strBufSize);
}
bool ImageView::Native::DriverVersionStr (char *const strBuf, size_t *const strBufSize)
{
    int major, minor;
    bool debug;
    if (!DriverVersion(major, minor, debug))
        return false;

    std::stringstream verStrStrm;
    verStrStrm << std::dec << major << ".";
    verStrStrm.width(2);
    verStrStrm.fill('0');
    verStrStrm << std::hex << ((minor >> 16) & 0xFF);
    verStrStrm << ", Build " << std::hex << ((minor >> 8) & 0xFFFF);
    if (debug)
        verStrStrm << " [debug]";

    return ImageView::String::set(verStrStrm.str(), strBuf, strBufSize);
}

#else
#   error Native library support not implemented on this platform.
#endif

