/* FILE:        ImageViewPopover.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::PopoverWindow.
 */

#include "ImageViewPopover_Private.h"

#include <QApplication>
#include <QStyle>
#include <QShortcut>

#include <QDesktopWidget>

#include <QFocusEvent>

ImageView::PopoverWindow::PrivateData::PrivateData (PopoverWindow &window)
    : QObject (&window)
    , m_window (window)
    , m_affinityWidget (nullptr)
    , m_affinity (BottomLeftAffinity)
    , m_centralWidget (nullptr)
    , m_popoverLayout (nullptr)
{ }

ImageView::PopoverWindow::PopoverWindow (QWidget *const parent)
    : QFrame (parent)
    , m_pd (*new PrivateData(*this))
{
    setFrameShape(QFrame::StyledPanel);
    setWindowFlags(Qt::Popup | Qt::CustomizeWindowHint | Qt::FramelessWindowHint | Qt::WindowStaysOnTopHint);

    m_pd.m_popoverLabel = new QLabel (this);
    m_pd.m_popoverLabel->setVisible(false);

    m_pd.m_rejectButton = new QPushButton (this);
    m_pd.m_rejectButton->setAutoDefault(false);
    m_pd.m_rejectButton->setFocusPolicy(Qt::NoFocus);
    m_pd.m_rejectButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogDiscardButton));
    m_pd.m_rejectButton->setVisible(false);
    
    m_pd.m_acceptButton = new QPushButton (this);
    m_pd.m_acceptButton->setAutoDefault(true);
    m_pd.m_acceptButton->setFocusPolicy(Qt::NoFocus);
    m_pd.m_acceptButton->setIcon(QApplication::style()->standardIcon(QStyle::SP_DialogApplyButton));

    m_pd.reflow(Qt::Horizontal);

    auto *const startFixedEnter = new QShortcut(Qt::Key_Enter, this);
    auto *const startFixedReturn = new QShortcut(Qt::Key_Return, this);
    auto *const startFixedEscape = new QShortcut(Qt::Key_Escape, this);

    connect(startFixedEnter, SIGNAL(activated()), this, SLOT(accept()));
    connect(startFixedReturn, SIGNAL(activated()), this, SLOT(accept()));
    connect(m_pd.m_acceptButton, SIGNAL(clicked()), this, SLOT(accept()));

    connect(startFixedEscape, SIGNAL(activated()), this, SLOT(reject()));
    connect(m_pd.m_rejectButton, SIGNAL(clicked()), this, SLOT(reject()));
}

ImageView::PopoverWindow::~PopoverWindow (void)
{
    delete &m_pd;
}

QString ImageView::PopoverWindow::labelText (void) const
{
    return m_pd.m_popoverLabel->text();
}
void ImageView::PopoverWindow::setLabelText (QString const& text)
{
    m_pd.m_popoverLabel->setText(text);
    m_pd.m_popoverLabel->setVisible(!text.isEmpty() || (m_pd.m_popoverLabel->pixmap() && !m_pd.m_popoverLabel->pixmap()->isNull()));
    m_pd.updateMargin();
}

const QPixmap* ImageView::PopoverWindow::labelPixmap (void) const
{
    return m_pd.m_popoverLabel->pixmap();
}
void ImageView::PopoverWindow::setLabelPixmap (QPixmap const& pixmap)
{
    m_pd.m_popoverLabel->setPixmap(pixmap);
    m_pd.m_popoverLabel->setVisible(!m_pd.m_popoverLabel->text().isEmpty() || !pixmap.isNull());
    m_pd.updateMargin();
}

QWidget* ImageView::PopoverWindow::affinityWidget (void) const
{
    return m_pd.m_affinityWidget;
}
void ImageView::PopoverWindow::setAffinityWidget (QWidget *const widget, AffinityEnum affinity)
{
    if (m_pd.m_affinityWidget)
        m_pd.m_affinityWidget->removeEventFilter(&m_pd);

    m_pd.m_affinityWidget = widget;
    m_pd.m_affinity = affinity;

    if (m_pd.m_affinityWidget)
        m_pd.m_affinityWidget->installEventFilter(&m_pd);
}

QWidget* ImageView::PopoverWindow::centralWidget (void) const
{
    return m_pd.m_centralWidget;
}

void ImageView::PopoverWindow::setCentralWidget (QWidget *const widget)
{
    if (m_pd.m_centralWidget)
        m_pd.m_centralWidget->setParent(nullptr);

    m_pd.m_centralWidget = widget;

    m_pd.reflow(layoutOrientation());
}

void ImageView::PopoverWindow::accept (void)
{
    if (isVisible())
    {
        emit accepted();
        close();
    }
}

void ImageView::PopoverWindow::reject (void)
{
    if (isVisible())
    {
        emit rejected();
        close();
    }
}

void ImageView::PopoverWindow::open (void)
{
    if (!isVisible())
    {
        if (m_pd.m_centralWidget)
            m_pd.m_centralWidget->setFocus();

        show();
        QApplication::setActiveWindow(this);
        m_pd.reallign();

        emit opened();
    }
}

// Helpers.
void ImageView::PopoverWindow::PrivateData::reallign (void)
{
    if (m_affinityWidget)
    {
        // Get our current values.
        const QRect thisRect = m_window.rect();
        const QRect affinityRect = m_affinityWidget->rect();
        const QSize desktopSize = QApplication::desktop()->size();

        // Naive alignment to the affinity.
        QPoint pos;
        switch (m_affinity)
        {
        case TopLeftAffinity:
            pos = affinityRect.topLeft();
            pos.ry() -= thisRect.height();
            break;
        case TopRightAffinity:
            pos = affinityRect.topRight();
            pos.ry() -= thisRect.height();
            pos.rx() -= thisRect.width();
            break;
        case BottomLeftAffinity:
            pos = affinityRect.bottomLeft();
            break;
        case BottomRightAffinity:
            pos = affinityRect.bottomRight();
            pos.rx() -= thisRect.width();
            break;
        case TopLeftOverAffinity:
            pos = affinityRect.topLeft();
            break;
        case TopRightOverAffinity:
            pos = affinityRect.topRight();
            pos.rx() -= thisRect.width();
            break;
        case BottomLeftOverAffinity:
            pos = affinityRect.bottomLeft();
            pos.ry() -= thisRect.height();
            break;
        case BottomRightOverAffinity:
            pos = affinityRect.bottomRight();
            pos.rx() -= thisRect.width();
            pos.ry() -= thisRect.height();
            break;
        }

        // Move to the global scope.
        const QPoint globalPos = m_affinityWidget->mapToGlobal(pos);

        // Fit within the desktop as best as possible.
        QRect newRect (globalPos, thisRect.size());
        
        if (0 > newRect.left())
            newRect.setLeft(0);
        else if (desktopSize.width() < newRect.right())
            newRect.setRight(desktopSize.width());

        if (0 > newRect.top())
            newRect.setTop(0);
        else if (desktopSize.height() < newRect.bottom())
            newRect.setBottom(desktopSize.height());

        // Apply the move.
        m_window.move(newRect.topLeft());
    }
}

void ImageView::PopoverWindow::PrivateData::reflow (const Qt::Orientation orientation)
{
    delete m_popoverLayout;

    m_popoverLayout = Qt::Horizontal == orientation ? static_cast<QBoxLayout*>(new QHBoxLayout (&m_window)) : new QVBoxLayout (&m_window);

    m_popoverLayout->addWidget(m_popoverLabel);

    if (m_centralWidget)
    {
        m_popoverLayout->addWidget(m_centralWidget);
        m_acceptButton->setFocusProxy(m_centralWidget);
        m_centralWidget->setParent(&m_window);
    }

    m_popoverLayout->addWidget(m_rejectButton);
    m_popoverLayout->addWidget(m_acceptButton);

    updateMargin();
}
void ImageView::PopoverWindow::PrivateData::updateMargin (void)
{
    const auto orientation = m_window.layoutOrientation();

    const bool requiresMargin = m_popoverLabel->isVisibleTo(&m_window) || m_window.rejectButtonVisible() || m_window.acceptButtonVisible();

    const int fMargin = requiresMargin ? 3 : 1;
    const int eMargin = requiresMargin ? 1 : 1;

    const int hMargin = Qt::Horizontal == orientation ? fMargin : eMargin;
    const int vMargin = Qt::Vertical == orientation ? fMargin : eMargin;

    m_popoverLayout->setContentsMargins(hMargin, vMargin, hMargin, vMargin);
}

// Properties.
Qt::Orientation ImageView::PopoverWindow::layoutOrientation (void) const
{
    if (qobject_cast<const QVBoxLayout*>(m_pd.m_popoverLayout))
        return Qt::Vertical;
    return Qt::Horizontal;
}
void ImageView::PopoverWindow::setLayoutOrientation (const Qt::Orientation orientation)
{
    if (layoutOrientation() != orientation)
        m_pd.reflow(orientation);
}

bool ImageView::PopoverWindow::rejectButtonVisible (void) const
{
    return m_pd.m_rejectButton->isVisibleTo(this);
}
void ImageView::PopoverWindow::setRejectButtonVisible (const bool visible)
{
    m_pd.m_rejectButton->setVisible(visible);
    m_pd.updateMargin();
}

bool ImageView::PopoverWindow::acceptButtonVisible (void) const
{
    return m_pd.m_acceptButton->isVisibleTo(this);
}
void ImageView::PopoverWindow::setAcceptButtonVisible (const bool visible)
{
    m_pd.m_acceptButton->setVisible(visible);
    m_pd.updateMargin();
}

bool ImageView::PopoverWindow::rejectOnFocusOut (void) const
{
    return m_pd.m_rejectOnFocusOut;
}
void ImageView::PopoverWindow::setRejectOnFocusOut (const bool enabled)
{
    m_pd.m_rejectOnFocusOut = enabled;
}

// Event overrides.
bool ImageView::PopoverWindow::PrivateData::eventFilter (QObject *const watched, QEvent *const evt)
{
    switch (evt->type())
    {
    case QEvent::Destroy:
        if (watched == m_affinityWidget)
            m_window.setAffinityWidget(nullptr);
        break;
    case QEvent::Move:
    case QEvent::Resize:
        if (watched == m_affinityWidget)
            reallign();
        break;
    }

    return false;
}

void ImageView::PopoverWindow::showEvent (QShowEvent *evt)
{
    QFrame::showEvent(evt);
    activateWindow();
}

void ImageView::PopoverWindow::focusOutEvent (QFocusEvent *const evt)
{
    if (m_pd.m_rejectOnFocusOut)
    {
        evt->accept();
        reject();
    }

    QWidget::focusOutEvent(evt);
}

void ImageView::PopoverWindow::closeEvent (QCloseEvent *const evt)
{
    emit closed();
    QWidget::closeEvent(evt);
}

