#ifndef INCLUDED__BIT_FLOW__IMAGE__VIEW__STRING__HELPERS__H
#define INCLUDED__BIT_FLOW__IMAGE__VIEW__STRING__HELPERS__H

/* FILE:        ImageViewStringHelpers.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Methods to simplify common string conversions and other operations.
 */

#include <cstring>

#include <vector>
#include <string>
#include <algorithm>
#include <functional>

namespace ImageView
{
    struct String
    {
        // Set the input string on the output string. pOutStr is optional, and may be nullptr. pOutStrSize should
        // be the buffer size of pOutStr, as an input variable. As an output variable, returns the required output
        // buffer byte size to contain the full string, including null-terminator; pOutStr will be filled as much
        // as possible, and always null-terminated. Returns true, unless pOutStrSize is nullptr.
        inline static bool set (std::string const& inStr, char *const pOutStr, size_t *const pOutStrSize)
        {
            if (!pOutStrSize)
                return false;

            const size_t outStrSize = *pOutStrSize;
            const size_t inStrSize = inStr.size() + 1;

            if (pOutStr)
            {
                const size_t copySize = std::min(outStrSize, inStrSize);
                memcpy(pOutStr, inStr.c_str(), copySize);
                if (outStrSize > 0)
                    pOutStr[outStrSize - 1] = 0;
            }

            *pOutStrSize = inStrSize;

            return true;
        }

        // Read a string from a function, converting it to a std::string.
        inline static std::string read (std::function<bool(char *const pOutStr, size_t *const pOutStrSize)> const& strFunc)
        {
            size_t iSize;
            if (!strFunc(nullptr, &iSize))
                return std::string();

            std::vector<char> outStr (iSize);
            if (!strFunc(outStr.data(), &iSize))
                return std::string();

            return outStr.data();
        }

        template <typename FuncT, typename Arg1T>
        inline static std::string read (FuncT &&func, Arg1T &&arg1)
        {
            return read(std::bind(std::forward<FuncT>(func), std::forward<Arg1T>(arg1), std::placeholders::_1, std::placeholders::_2));
        }

        template <typename FuncT, typename Arg1T, typename Arg2T>
        inline static std::string read (FuncT &&func, Arg1T &&arg1, Arg2T &&arg2)
        {
            return read(std::bind(std::forward<FuncT>(func), std::forward<Arg1T>(arg1), std::forward<Arg2T>(arg2), std::placeholders::_1, std::placeholders::_2));
        }

        template <typename FuncT, typename Arg1T, typename Arg2T, typename Arg3T>
        inline static std::string read (FuncT &&func, Arg1T &&arg1, Arg2T &&arg2, Arg3T &&arg3)
        {
            return read(std::bind(std::forward<FuncT>(func), std::forward<Arg1T>(arg1), std::forward<Arg2T>(arg2), std::forward<Arg3T>(arg3), std::placeholders::_1, std::placeholders::_2));
        }

        template <typename FuncT, typename Arg1T, typename Arg2T, typename Arg3T, typename Arg4T>
        inline static std::string read (FuncT &&func, Arg1T &&arg1, Arg2T &&arg2, Arg3T &&arg3, Arg4T &&arg4)
        {
            return read(std::bind(std::forward<FuncT>(func), std::forward<Arg1T>(arg1), std::forward<Arg2T>(arg2), std::forward<Arg3T>(arg3), std::forward<Arg4T>(arg4), std::placeholders::_1, std::placeholders::_2));
        }
    };
}

#endif // INCLUDED__BIT_FLOW__IMAGE__VIEW__STRING__HELPERS__H
