/* FILE:        ImageViewSurface.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::Surface.
 */

#include "ImageViewSurface_Private.h"

#include <QPainter>

#include <QPaintEvent>
#include <QWheelEvent>

#include <QTimer>

#include <chrono>
#include <cmath>

//// Public member functions.

// Constructor/Desctructor.
ImageView::Surface::Surface (QWidget *const parent)
    : IMAGE_VIEW_SURFACE_BASE (parent)
    , m_pd(*new PrivateData(*this))
{
    setMouseTracking(true);

    setMinimumSize(1, 1);
    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred);
}

ImageView::Surface::~Surface (void)
{
    delete &m_pd;
}

// The current pixmap.
QPixmap ImageView::Surface::pixmap (void) const
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);
    return m_pd.m_pixmap;
}
void ImageView::Surface::setPixmap (QPixmap const& pxmp)
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);

    const bool sizeChanged = m_pd.m_pixmap.size() != pxmp.size();
    m_pd.m_pixmap = pxmp;

#if IMAGE_VIEW_SURFACE_USE_QGL
    m_pd.m_textureExpired.store(true);
#endif

    if (sizeChanged)
    {
        if (!m_pd.unlocked_checkScaleToFit(pixmapLock))
            emit pixelRangeChanged();
    }

    update();
}

// Set the raw data for the current image, and its AcquisitionInterface. This may be cleared by setting nullptr.
void ImageView::Surface::setImageSrc (AbstractAcquisitionInterface *const acqIf, RawImage const& rawImage)
{
    const std::lock_guard<std::mutex> imageBufferLock (m_pd.m_imageBufferMutex);

    if (m_pd.m_imageBufferSizeChange.load() || m_pd.m_acqIf != acqIf || m_pd.m_rawImage.m_format != rawImage.m_format || m_pd.m_rawImage.m_width != rawImage.m_width || m_pd.m_rawImage.m_height != rawImage.m_height)
        m_pd.m_imageBufferSizeChange.store(true);

    m_pd.m_acqIf = acqIf;
    m_pd.m_rawImage = rawImage;

    m_pd.m_image = QImage();

    update();
}

void ImageView::Surface::setImageSrc (QImage const& image)
{
    const std::lock_guard<std::mutex> imageBufferLock (m_pd.m_imageBufferMutex);

    if (m_pd.m_imageBufferSizeChange.load()  || m_pd.m_image.format() != image.format() || m_pd.m_image.size() != image.size())
        m_pd.m_imageBufferSizeChange.store(true);

    m_pd.m_acqIf = nullptr;
    m_pd.m_rawImage.m_buffer = nullptr;

    m_pd.m_image = image;

    update();
}

void ImageView::Surface::clearImageSrc (void)
{
    const std::lock_guard<std::mutex> imageBufferLock (m_pd.m_imageBufferMutex);

    m_pd.m_acqIf = nullptr;
    m_pd.m_rawImage.m_buffer = nullptr;

    m_pd.m_image = QImage();

    update();
}

// Get the image source pixel value, if any.
bool ImageView::Surface::imageSrcPixVal (PixPos const& pos, PixVal &val)
{
    const std::lock_guard<std::mutex> imageBufferLock (m_pd.m_imageBufferMutex);

    if (m_pd.m_acqIf && m_pd.m_rawImage.m_buffer)
    {
        if (const auto bufferCriticalSection = m_pd.m_acqIf->getFrameLock(m_pd.m_rawImage))
            return m_pd.m_imageAccessor.BufferPixelValue(m_pd.m_rawImage, pos, val);
    }
    else if (!m_pd.m_image.isNull())
    {
        if (m_pd.m_image.rect().contains(pos.x, pos.y))
        {
            const QRgb pixRgbVal = m_pd.m_image.pixel(pos.x, pos.y);

            val.isGrayscale = false;
            val.isColor = true;
            val.r = qRed(pixRgbVal);
            val.g = qGreen(pixRgbVal);
            val.b = qBlue(pixRgbVal);

            return true;
        }
    }

    return false;
}

// Property accessors.
double ImageView::Surface::scalingRatio (void)
{
    return 1.125;
}
double ImageView::Surface::scalingMin (void)
{
    return pow(scalingRatio(), -40);
}
double ImageView::Surface::scalingMax (void)
{
    return pow(scalingRatio(), 40);
}
double ImageView::Surface::scaling (void) const
{
    return m_pd.m_scaling;
}
QSizeF ImageView::Surface::scaledSize (void) const
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);
    return m_pd.unlocked_scaledSize(pixmapLock);
}
QRectF ImageView::Surface::scaledRect (void) const
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);
    return m_pd.unlocked_scaledRect(pixmapLock);
}
QPointF ImageView::Surface::scaledOffset (void) const
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);
    return m_pd.unlocked_scaledOffset(pixmapLock);
}
QRect ImageView::Surface::visibleImageRect (void) const
{
    const auto visibleRect = visibleImageRectF();

    const auto topLeftAdj = visibleRect.topLeft() - QPointF(0.5, 0.5);
    const auto bottomRightRnd = visibleRect.bottomRight() + QPointF(0.5, 0.5);

    return QRect(topLeftAdj.toPoint(), bottomRightRnd.toPoint());
}
QRectF ImageView::Surface::visibleImageRectF (void) const
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);
    return m_pd.unlocked_visibleImageRectF(pixmapLock);
}

double ImageView::Surface::enforceScalingRatio (const double scaling)
{
    const double fittedScaling = std::max(scalingMin(), std::min(scalingMax(), scaling));
    const double scalingPower = log(fittedScaling) / log(scalingRatio());
    const double correctedScaling = pow(scalingRatio(), floor(scalingPower + 0.5));
    return correctedScaling;
}

// Property slots.
void ImageView::Surface::setScaling (const double scaling)
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);

    const bool oldScaleToFit = m_pd.m_scaleToFit;
    const double oldScaling = m_pd.m_scaling;
    const double fittedScaling = std::max(scalingMin(), std::min(scalingMax(), scaling));

    m_pd.m_scaling = fittedScaling;
    m_pd.m_scaleToFit = false;

    const QPointF newOffset = (m_pd.m_offset / oldScaling) * fittedScaling;
    m_pd.unlocked_setFittedOffset(pixmapLock, newOffset);

    const bool hasNewScaling = std::fabs(oldScaling - m_pd.m_scaling) >= std::numeric_limits<double>::epsilon();
    if (hasNewScaling)
    {
        emit pixelRangeChanged();
        emit scalingChanged(m_pd.m_scaleToFit, m_pd.m_scaling);
    }
    else if (oldScaleToFit)
        emit scalingChanged(m_pd.m_scaleToFit, m_pd.m_scaling);

    update();
}

void ImageView::Surface::setScaledOffset (QPointF const& offset)
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);
    m_pd.unlocked_setFittedOffset(pixmapLock, offset);
    update();
}
void ImageView::Surface::setScaledOffsetX (const double xOffset)
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);

    const auto oldOffset = m_pd.unlocked_scaledOffset(pixmapLock);
    const QPointF newOffset (xOffset, oldOffset.y());
    m_pd.unlocked_setFittedOffset(pixmapLock, newOffset);

    update();
}
void ImageView::Surface::setScaledOffsetY (const double yOffset)
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);

    const auto oldOffset = m_pd.unlocked_scaledOffset(pixmapLock);
    const QPointF newOffset (oldOffset.x(), yOffset);
    m_pd.unlocked_setFittedOffset(pixmapLock, newOffset);

    update();
}
void ImageView::Surface::setScaledOffsetX (const int xOffset)
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);

    const auto oldOffset = m_pd.unlocked_scaledOffset(pixmapLock);
    const QPointF newOffset (xOffset, oldOffset.y());
    m_pd.unlocked_setFittedOffset(pixmapLock, newOffset);

    update();
}
void ImageView::Surface::setScaledOffsetY (const int yOffset)
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);

    const auto oldOffset = m_pd.unlocked_scaledOffset(pixmapLock);
    const QPointF newOffset (oldOffset.x(), yOffset);
    m_pd.unlocked_setFittedOffset(pixmapLock, newOffset);

    update();
}

bool ImageView::Surface::scaleToFit (void) const
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);
    return m_pd.m_scaleToFit;
}
void ImageView::Surface::setScaleToFit (const bool enabled)
{
    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);

    m_pd.m_scaleToFit = enabled;
    if (enabled)
        m_pd.unlocked_checkScaleToFit(pixmapLock, true);
}

// Whether or not to render the hovered pixel crosshair.
bool ImageView::Surface::crosshairVisible (void) const
{
    return m_pd.m_crosshairVisible;
}

void ImageView::Surface::setCrosshairVisible (const bool visible)
{
    m_pd.m_crosshairVisible = visible;
    update();
}

// The crosshair location in the image, in pixels.
QPoint ImageView::Surface::activePixel (void) const
{
    return m_pd.m_activePixel;
}

// Whether or not the pixel value overlay will be rendered at high enough zoom.
bool ImageView::Surface::valueOverlayEnabled (void) const
{
    return m_pd.m_valueOverlayEnabled;
}
void ImageView::Surface::setValueOverlayEnabled (const bool enabled)
{
    m_pd.m_valueOverlayEnabled = enabled;
    update();
}

void ImageView::Surface::setActivePixel (QPoint const& pixPoint)
{
    m_pd.m_activePixel = pixPoint;
    if (m_pd.m_crosshairVisible)
        update();
}
void ImageView::Surface::setActivePixel (const int x, const int y)
{
    m_pd.m_activePixel = QPoint(x, y);
    if (m_pd.m_crosshairVisible)
        update();
}

// Reimplemented methods.
void ImageView::Surface::paintEvent (QPaintEvent *evt)
{
    // Get the latest pixmap.
    const QRectF targetRect = evt->rect();;

    // Start painting.
    QPainter painter (this);
    
    // Draw the image.
    painter.save();
    m_pd.renderPixmap(painter, targetRect);
    painter.restore();
    
    // Optionally draw the overlay.
    if (m_pd.m_valueOverlayEnabled)
    {
        painter.save();
        m_pd.renderPixelOverlay(painter, targetRect);
        painter.restore();
    }

    // Optionally draw the pixel crosshair lines.
    if (m_pd.m_crosshairVisible)
    {
        painter.save();
        m_pd.renderCrosshair(painter, targetRect);
        painter.restore();
    }
}

void ImageView::Surface::resizeEvent (QResizeEvent *evt)
{
    IMAGE_VIEW_SURFACE_BASE::resizeEvent(evt);

    const std::lock_guard<std::mutex> pixmapLock (m_pd.m_pixmapMutex);

    if (!m_pd.unlocked_checkScaleToFit(pixmapLock))
        emit pixelRangeChanged();

    const auto offsetDiff = QSizeF(evt->size() - evt->oldSize()) / 2.;
    const QPointF newOffset (m_pd.m_offset.x() - offsetDiff.width(), m_pd.m_offset.y() - offsetDiff.height());
    m_pd.unlocked_setFittedOffset(pixmapLock, newOffset);
}
