#ifndef INCLUDED__BIT_FLOW__IMAGE__VIEW__SURFACE__H
#define INCLUDED__BIT_FLOW__IMAGE__VIEW__SURFACE__H

/* FILE:        AcquisitionSurface.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Image display surface widget for the ImageView library.
 */

#define IMAGE_VIEW_SURFACE_USE_QGL (0)

#include "ImageView.h"

#if IMAGE_VIEW_SURFACE_USE_QGL
#   include <QGLWidget>
#   define IMAGE_VIEW_SURFACE_BASE QGLWidget
#else
#   include <QWidget>
#   define IMAGE_VIEW_SURFACE_BASE QWidget
#endif

#include <QPixmap>

#include <cstdint>
#include <mutex>

namespace ImageView
{
    class IMAGEVIEW_EXPORT Surface : public IMAGE_VIEW_SURFACE_BASE
    {
        Q_OBJECT

        Q_PROPERTY(double   scalingMin          READ    scalingMin);
        Q_PROPERTY(double   scalingMax          READ    scalingMax);
        Q_PROPERTY(double   scaling             READ    scaling             WRITE   setScaling);

        Q_PROPERTY(QSizeF   scaledSize          READ    scaledSize);
        Q_PROPERTY(QRectF   scaledRect          READ    scaledRect);
        Q_PROPERTY(QPointF  scaledOffset        READ    scaledOffset        WRITE   setScaledOffset);
        Q_PROPERTY(bool     scaleToFit          READ    scaleToFit          WRITE   setScaleToFit);

        Q_PROPERTY(QRect    visibleImageRect    READ    visibleImageRect);
        Q_PROPERTY(QRectF   visibleImageRectF   READ    visibleImageRectF);

        Q_PROPERTY(bool     crosshairVisible    READ    crosshairVisible    WRITE   setCrosshairVisible);
        Q_PROPERTY(QPoint   activePixel         READ    activePixel         WRITE   setActivePixel);

        Q_PROPERTY(bool     valueOverlayEnabled READ    valueOverlayEnabled WRITE   setValueOverlayEnabled);

    private:
        struct PrivateData;
        PrivateData &m_pd;

        // Illegal.
        Surface (Surface const&);
        Surface& operator= (Surface const&);

    public:
        //// Public member functions.

        // Constructor/Desctructor.
        Surface (QWidget *const parent = nullptr);
        ~Surface (void);

        // The current pixmap.
        QPixmap pixmap (void) const;
        void setPixmap (QPixmap const& pxmp);

        // Set the raw data for the current image, and its AcquisitionInterface.
        void setImageSrc (AbstractAcquisitionInterface *const acqIf, RawImage const& rawImage);
        void setImageSrc (QImage const& image);
        void clearImageSrc (void);

        // Get the image source pixel value, if any.
        bool imageSrcPixVal (PixPos const& pos, PixVal &val);

        // Property accessors.
        static double scalingRatio (void);
        static double scalingMax (void);
        static double scalingMin (void);
        static double enforceScalingRatio (const double scaling);

        double scaling (void) const;

        QSizeF scaledSize (void) const;
        QRectF scaledRect (void) const;
        QPointF scaledOffset (void) const;
        bool scaleToFit (void) const;

        QRect visibleImageRect (void) const;
        QRectF visibleImageRectF (void) const;

        // Whether or not to render the hovered pixel crosshair.
        bool crosshairVisible (void) const;

        // The crosshair location in the image, in pixels.
        QPoint activePixel (void) const;

        // Whether or not the pixel value overlay will be rendered at high enough zoom.
        bool valueOverlayEnabled (void) const;

    public slots:
            // Property slots
        void setScaling (const double scaling);
        void setScaledOffset (QPointF const& offset);

        void setScaledOffsetX (const double xOffset);
        void setScaledOffsetY (const double yOffset);
        void setScaledOffsetX (const int xOffset);
        void setScaledOffsetY (const int yOffset);

        void setScaleToFit (const bool enabled = true);

        // Whether or not to render the hovered pixel crosshair.
        void setCrosshairVisible (const bool visible);

        void setActivePixel (QPoint const& pixPoint);
        void setActivePixel (const int x, const int y);

        // Enable or disable the pixel value overlay.
        void setValueOverlayEnabled (const bool enabled);

    signals:
        void pixelRangeChanged (void);
        void pixelOffsetChanged (QPointF const& offset);

        void scalingChanged (const bool autoFit, const double scaling);

    private:
        // Reimplemented methods.
        virtual void paintEvent (QPaintEvent *evt);
        virtual void resizeEvent(QResizeEvent *evt);
    };
}

#endif // INCLUDED__BIT_FLOW__IMAGE__VIEW__SURFACE__H
