/* FILE:        ImageViewSurfaceScrollArea.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::SurfaceScrollArea.
 */

#include "ImageViewSurfaceScrollArea_Private.h"

#include <QGridLayout>

#include <cmath>

ImageView::SurfaceScrollArea::SurfaceScrollArea (QWidget *const parent)
    : QWidget (parent)
    , m_pd (*new PrivateData(*this))
{
    // Spawn our primary widgets.
    m_pd.m_imageSurface = new ImageView::Surface(this);
    m_pd.m_horizontalScrollBar = new QScrollBar(Qt::Horizontal, this);
    m_pd.m_verticalScrollBar = new QScrollBar(Qt::Vertical, this);

    // Layout all widgets.
    m_pd.m_gridLayout = new QGridLayout (this);
    m_pd.m_gridLayout->setMargin(0);
    m_pd.m_gridLayout->setSpacing(0);

    m_pd.m_gridLayout->addWidget(m_pd.m_imageSurface, 0, 0);
    m_pd.m_gridLayout->addWidget(m_pd.m_horizontalScrollBar, 1, 0);
    m_pd.m_gridLayout->addWidget(m_pd.m_verticalScrollBar, 0, 1);

    setLayout(m_pd.m_gridLayout);

    // Configure our appearance.
    m_pd.m_imageSurface->setAutoFillBackground(true);
    auto imageBackgroundPalette = palette();
    imageBackgroundPalette.setColor(QPalette::Window, Qt::darkGray);
    m_pd.m_imageSurface->setPalette(imageBackgroundPalette);

    // Generate keyboard shortcuts.
    m_pd.m_leftShortcut = new QShortcut (Qt::Key_Left, this);
    m_pd.m_rightShortcut = new QShortcut (Qt::Key_Right, this);
    m_pd.m_upShortcut = new QShortcut (Qt::Key_Up, this);
    m_pd.m_downShortcut = new QShortcut (Qt::Key_Down, this);

    m_pd.m_ctrlLeftShortcut = new QShortcut (Qt::CTRL + Qt::Key_Left, this);
    m_pd.m_ctrlRightShortcut = new QShortcut (Qt::CTRL + Qt::Key_Right, this);
    m_pd.m_ctrlUpShortcut = new QShortcut (Qt::CTRL + Qt::Key_Up, this);
    m_pd.m_ctrlDownShortcut = new QShortcut (Qt::CTRL + Qt::Key_Down, this);

    connect(m_pd.m_leftShortcut, SIGNAL(activated()), this, SLOT(stepLeft()));
    connect(m_pd.m_rightShortcut, SIGNAL(activated()), this, SLOT(stepRight()));
    connect(m_pd.m_upShortcut, SIGNAL(activated()), this, SLOT(stepUp()));
    connect(m_pd.m_downShortcut, SIGNAL(activated()), this, SLOT(stepDown()));

    connect(m_pd.m_ctrlLeftShortcut, SIGNAL(activated()), this, SLOT(pageLeft()));
    connect(m_pd.m_ctrlRightShortcut, SIGNAL(activated()), this, SLOT(pageRight()));
    connect(m_pd.m_ctrlUpShortcut, SIGNAL(activated()), this, SLOT(pageUp()));
    connect(m_pd.m_ctrlDownShortcut, SIGNAL(activated()), this, SLOT(pageDown()));

    // Setup event filters.
    new ViewportEventFilter(this, m_pd.m_imageSurface);
    new ImageSurfaceEventFilter(this, m_pd.m_imageSurface);

    connect(m_pd.m_imageSurface, SIGNAL(pixelRangeChanged()), &m_pd, SLOT(updateScrollRange()), Qt::QueuedConnection);
    connect(m_pd.m_imageSurface, SIGNAL(pixelOffsetChanged(QPointF)), &m_pd, SLOT(updateScrollPos(QPointF)), Qt::QueuedConnection);

    connect(m_pd.m_horizontalScrollBar, SIGNAL(valueChanged(int)), m_pd.m_imageSurface, SLOT(setScaledOffsetX(int)));
    connect(m_pd.m_verticalScrollBar, SIGNAL(valueChanged(int)), m_pd.m_imageSurface, SLOT(setScaledOffsetY(int)));

    m_pd.m_horizontalScrollBar->setVisible(false);
    m_pd.m_verticalScrollBar->setVisible(false);
}
ImageView::SurfaceScrollArea::~SurfaceScrollArea (void)
{
    delete &m_pd;
}

// ImageView::Surface accessor.
ImageView::Surface* ImageView::SurfaceScrollArea::surface (void)
{
    return m_pd.m_imageSurface;
}
const ImageView::Surface* ImageView::SurfaceScrollArea::surface (void) const
{
    return m_pd.m_imageSurface;
}

// Scrollbar accessors.
QScrollBar* ImageView::SurfaceScrollArea::horizontalScrollBar (void)
{
    return m_pd.m_horizontalScrollBar;
}
const QScrollBar* ImageView::SurfaceScrollArea::horizontalScrollBar (void) const
{
    return m_pd.m_horizontalScrollBar;
}

QScrollBar* ImageView::SurfaceScrollArea::verticalScrollBar (void)
{
    return m_pd.m_verticalScrollBar;
}
const QScrollBar* ImageView::SurfaceScrollArea::verticalScrollBar (void) const
{
    return m_pd.m_verticalScrollBar;
}
// Corner widget. The ImageView::Area takes ownership of the corner widget. If ownership is now
// reassigned explicitly, any current corner widget will be deleted upon call of
QWidget* ImageView::SurfaceScrollArea::cornerWidget (void)
{
    return m_pd.m_cornerWidget;
}
const QWidget* ImageView::SurfaceScrollArea::cornerWidget (void) const
{
    return m_pd.m_cornerWidget;
}
void ImageView::SurfaceScrollArea::setCornerWidget (QWidget *const widget)
{
    // Remove and possibly delete any extant corner widget.
    if (m_pd.m_cornerWidget)
    {
        m_pd.m_gridLayout->removeWidget(m_pd.m_cornerWidget);
        if (m_pd.m_cornerWidget->parent() == this && m_pd.m_cornerWidget != widget)
            delete m_pd.m_cornerWidget;
    }

    // Set the new corner widget.
    m_pd.m_cornerWidget = widget;

    // If not nullptr, setup the new corner widget.
    if (m_pd.m_cornerWidget)
    {
        m_pd.m_cornerWidget->setParent(this);
        m_pd.m_gridLayout->addWidget(m_pd.m_cornerWidget, 1, 1);
    }

    m_pd.updateScrollRange();
}

double ImageView::SurfaceScrollArea::zoomScaling (void) const
{
    return m_pd.m_imageSurface->scaling();
}
bool ImageView::SurfaceScrollArea::zoomToFit (void) const
{
    return m_pd.m_imageSurface->scaleToFit();
}

// Helper slots.
void ImageView::SurfaceScrollArea::setZoomTo (const double scale, QPoint const& centerRef)
{
    if (m_pd.m_imageSurface->pixmap().isNull())
        return;

    // Record the actual initial scaling value.
    const double scalingBefore = m_pd.m_imageSurface->scaling();
    const QPointF offsetBefore = m_pd.m_imageSurface->scaledOffset();

    // Record information necessary to reorient the offset position.
    const QRectF frameRect = m_pd.m_imageSurface->rect();

    // Attempt to apply the new scaling value.
    m_pd.m_imageSurface->setScaling(scale);

    // Record the actual after scaling value.
    const double scalingAfter = m_pd.m_imageSurface->scaling();

    // If the actual scaling meaningfully changed, reorient the scroll position.
    if (std::fabs(scalingBefore - scalingAfter) > Surface::scalingMin())
    {
        const QPointF frameCenter = frameRect.contains(centerRef) ? centerRef : frameRect.center();
        const QPointF visibleCenter = offsetBefore + frameCenter;

        const QPointF scaledCenter = visibleCenter * scalingAfter / scalingBefore;

        const QPointF scaledTopLeft = scaledCenter - frameCenter;

        m_pd.m_imageSurface->setScaledOffset(scaledTopLeft);
    }
}
void ImageView::SurfaceScrollArea::doZoomIn (QPoint const& centerRef)
{
    setZoomTo(m_pd.m_imageSurface->scaling() * Surface::scalingRatio(), centerRef);
}
void ImageView::SurfaceScrollArea::doZoomOut (QPoint const& centerRef)
{
    setZoomTo(m_pd.m_imageSurface->scaling() / Surface::scalingRatio(), centerRef);
}
void ImageView::SurfaceScrollArea::setZoomToFit (const bool enabled)
{
    m_pd.m_imageSurface->setScaleToFit(enabled);
}

void ImageView::SurfaceScrollArea::doZoomTo25 (void)
{
    setZoomTo(0.25);
}
void ImageView::SurfaceScrollArea::doZoomTo50 (void)
{
    setZoomTo(0.5);
}
void ImageView::SurfaceScrollArea::doZoomTo100 (void)
{
    setZoomTo(1.);
}
void ImageView::SurfaceScrollArea::doZoomTo200 (void)
{
    setZoomTo(2.);
}
void ImageView::SurfaceScrollArea::doZoomTo400 (void)
{
    setZoomTo(4.);
}
void ImageView::SurfaceScrollArea::doZoomTo800 (void)
{
    setZoomTo(8.);
}

void ImageView::SurfaceScrollArea::stepLeft (void)
{
    m_pd.m_horizontalScrollBar->triggerAction(QScrollBar::SliderSingleStepSub);
}
void ImageView::SurfaceScrollArea::stepRight (void)
{
    m_pd.m_horizontalScrollBar->triggerAction(QScrollBar::SliderSingleStepAdd);
}
void ImageView::SurfaceScrollArea::stepUp (void)
{
    m_pd.m_verticalScrollBar->triggerAction(QScrollBar::SliderSingleStepSub);
}
void ImageView::SurfaceScrollArea::stepDown (void)
{
    m_pd.m_verticalScrollBar->triggerAction(QScrollBar::SliderSingleStepAdd);
}

void ImageView::SurfaceScrollArea::pageLeft (void)
{
    m_pd.m_horizontalScrollBar->triggerAction(QScrollBar::SliderPageStepSub);
}
void ImageView::SurfaceScrollArea::pageRight (void)
{
    m_pd.m_horizontalScrollBar->triggerAction(QScrollBar::SliderPageStepAdd);
}
void ImageView::SurfaceScrollArea::pageUp (void)
{
    m_pd.m_verticalScrollBar->triggerAction(QScrollBar::SliderPageStepSub);
}
void ImageView::SurfaceScrollArea::pageDown (void)
{
    m_pd.m_verticalScrollBar->triggerAction(QScrollBar::SliderPageStepAdd);
}

QPoint ImageView::SurfaceScrollArea::imageOffset (void) const
{
    return QPoint (m_pd.m_horizontalScrollBar->value(), m_pd.m_verticalScrollBar->value());
}
void ImageView::SurfaceScrollArea::setImageOffset (QPointF const& offset)
{
    const bool hsbBlocked = m_pd.m_horizontalScrollBar->blockSignals(true);
    m_pd.m_horizontalScrollBar->setValue(offset.x());
    m_pd.m_horizontalScrollBar->blockSignals(hsbBlocked);

    const bool vsbBlocked = m_pd.m_verticalScrollBar->blockSignals(true);
    m_pd.m_verticalScrollBar->setValue(offset.y());
    m_pd.m_verticalScrollBar->blockSignals(vsbBlocked);

    m_pd.m_imageSurface->setScaledOffset(offset);
}

bool ImageView::SurfaceScrollArea::hBarScrollable (void) const
{
    return m_pd.m_horizontalScrollBar->isEnabled() && m_pd.m_horizontalScrollBar->isVisible();
}
bool ImageView::SurfaceScrollArea::vBarScrollable (void) const
{
    return m_pd.m_verticalScrollBar->isEnabled() && m_pd.m_verticalScrollBar->isVisible();
}
