/* FILE:        ImageViewSurfaceScrollArea_Private.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::SurfaceScrollArea::PrivateData.
 */

#include "ImageViewSurfaceScrollArea_Private.h"

ImageView::SurfaceScrollArea::PrivateData::PrivateData (SurfaceScrollArea &scrollArea)
    : QObject (&scrollArea)
    , m_scrollArea (scrollArea)
    , m_cornerWidget (nullptr)
{ }

void ImageView::SurfaceScrollArea::PrivateData::updateScrollRange (void)
{
    // Determine what we need to do.
    const bool hBarWasScrollable = m_scrollArea.hBarScrollable();
    const bool vBarWasScrollable = m_scrollArea.vBarScrollable();

    const auto scaledSize = m_imageSurface->scaledSize().toSize();
    const bool barsDesirable = !m_imageSurface->scaleToFit() && !scaledSize.isEmpty();
    const int pixelSize = (int)std::max(1., m_imageSurface->scaling());
    const auto viewerSize = m_imageSurface->size();

    const int maxH = std::max(0, scaledSize.width() - viewerSize.width());
    const bool hBarScrollable = barsDesirable && maxH > 0;

    // Set the horizontal scroll bar parameters.
    const bool hBarSignalsBlocked = m_horizontalScrollBar->blockSignals(true);

    m_horizontalScrollBar->setRange(0, maxH);
    m_horizontalScrollBar->setVisible(hBarScrollable);
    m_horizontalScrollBar->setEnabled(hBarScrollable);

    m_horizontalScrollBar->setSingleStep(pixelSize);
    m_horizontalScrollBar->setPageStep(viewerSize.width());

    m_horizontalScrollBar->blockSignals(hBarSignalsBlocked);

    const int maxV = std::max(0, scaledSize.height() - viewerSize.height());
    const bool vBarScrollable = barsDesirable && maxV > 0;

    // Set the vertical scroll bar parameters.
    const bool vBarSignalsBlocked = m_verticalScrollBar->blockSignals(true);

    m_verticalScrollBar->setRange(0, maxV);
    m_verticalScrollBar->setVisible(vBarScrollable);
    m_verticalScrollBar->setEnabled(vBarScrollable);

    m_verticalScrollBar->setSingleStep(pixelSize);
    m_verticalScrollBar->setPageStep(viewerSize.height());

    m_verticalScrollBar->blockSignals(vBarSignalsBlocked);

    // Enforce the desired corner widget size.
    if (m_cornerWidget)
    {
        const int smallerSize = std::min(m_verticalScrollBar->width(), m_horizontalScrollBar->height());
        m_cornerWidget->setFixedSize(smallerSize, smallerSize);
        m_cornerWidget->setVisible(hBarScrollable || vBarScrollable);
    }

    // Update any externals depending upon our scrollbar visibility.
    if (hBarWasScrollable != hBarScrollable || vBarWasScrollable != vBarScrollable)
    {
        m_gridLayout->removeWidget(m_imageSurface);
        m_gridLayout->addWidget(m_imageSurface, 0, 0, hBarScrollable ? 1 : 2, vBarScrollable ? 1 : 2);

        emit m_scrollArea.scrollabilityChanged(hBarScrollable, vBarScrollable);
    }
}
void ImageView::SurfaceScrollArea::PrivateData::updateScrollPos (QPointF const& offset)
{
    updateScrollRange();

    const bool hsbBlocked = m_horizontalScrollBar->blockSignals(true);
    m_horizontalScrollBar->setValue(offset.x());
    m_horizontalScrollBar->blockSignals(hsbBlocked);

    const bool vsbBlocked = m_verticalScrollBar->blockSignals(true);
    m_verticalScrollBar->setValue(offset.y());
    m_verticalScrollBar->blockSignals(vsbBlocked);
}

// eventFilter QObject for viewport events.
ViewportEventFilter::ViewportEventFilter (ImageView::SurfaceScrollArea *const area, ImageView::Surface *const surface)
    : QObject (surface)
    , m_surface (surface)
    , m_area (area)
{
    m_surface->installEventFilter(this);
}

bool ViewportEventFilter::eventFilter (QObject *o, QEvent *e)
{
    switch (e->type())
    {
    case QEvent::Wheel:
        return wheelEvent(o, static_cast<QWheelEvent*>(e));
    }
    return false;
}

bool ViewportEventFilter::wheelEvent (QObject *o, QWheelEvent *evt)
{
    const auto zoomCenter = m_surface->mapFromGlobal(evt->globalPos());

    const auto delta = evt->angleDelta();
    if (delta.y() > 0)
        m_area->doZoomIn(zoomCenter);
    else if (delta.y() < 0)
        m_area->doZoomOut(zoomCenter);
    else
        return false;

    evt->accept();
    return true;
}

ImageSurfaceEventFilter::ImageSurfaceEventFilter (ImageView::SurfaceScrollArea *const area, ImageView::Surface *const parent)
    : QObject (parent)
    , m_area (area)
    , m_surface (parent)
    , m_dragActive (false)
{
    m_surface->installEventFilter(this);
    m_surface->setMouseTracking(true);
    m_surface->setCursor(Qt::OpenHandCursor);
}

bool ImageSurfaceEventFilter::eventFilter (QObject *o, QEvent *e)
{
    switch (e->type())
    {
    case QEvent::MouseButtonPress:
        return mousePressEvent(o, static_cast<QMouseEvent*>(e));
    case QEvent::MouseMove:
        return mouseMoveEvent(o, static_cast<QMouseEvent*>(e));
    case QEvent::MouseButtonRelease:
        return mouseReleaseEvent(o, static_cast<QMouseEvent*>(e));
    case QEvent::Enter:
        return enterEvent(o, static_cast<QEnterEvent*>(e));
    case QEvent::Leave:
        return leaveEvent(o, e);
    }

    return false;
}
bool ImageSurfaceEventFilter::mousePressEvent (QObject *o, QMouseEvent *evt)
{
    if (evt->buttons() == Qt::LeftButton)
    {
        m_surface->setCursor(Qt::ClosedHandCursor);
        m_pressPos = evt->globalPos();
    }

    return false;
}
bool ImageSurfaceEventFilter::mouseMoveEvent (QObject *o, QMouseEvent *evt)
{
    // Determine if we should enter the drag active state.
    if (evt->buttons() == Qt::LeftButton)
    {
        if (!m_dragActive && !m_surface->pixmap().isNull() && (m_pressPos - evt->globalPos()).manhattanLength() >= 3)
        {
            m_dragActive = true;
            m_dragPointInit = m_area->imageOffset() + m_pressPos;
        }
    }

    // Handle any active drag.
    if (m_dragActive)
    {
        m_area->setImageOffset(m_dragPointInit - evt->globalPos());
        evt->accept();
        return true;
    }

    // Update the crosshair position for the mouse location.
    const auto surfacePos = m_surface->mapFromGlobal(evt->globalPos());

    QPoint imagePos (-1, -1);
    if (surfacePos.x() >= 0 && surfacePos.y() >= 0)
    {
        const auto imageRect = m_surface->scaledRect();
        imagePos = ((surfacePos - imageRect.topLeft()) / m_surface->scaling() - QPointF(0.5, 0.5)).toPoint();
    }

    m_surface->setActivePixel(imagePos);

    return false;
}
bool ImageSurfaceEventFilter::mouseReleaseEvent (QObject *o, QMouseEvent *evt)
{
    if (evt->buttons() | Qt::LeftButton)
    {
        m_surface->setCursor(Qt::OpenHandCursor);

        if (m_dragActive)
        {
            m_dragActive = false;
            evt->accept();
            return true;
        }
    }

    return false;
}
bool ImageSurfaceEventFilter::enterEvent (QObject *o, QEnterEvent *evt)
{
    const auto surfacePos = m_surface->mapFromGlobal(evt->globalPos());

    QPoint imagePos (-1, -1);
    if (surfacePos.x() >= 0 && surfacePos.y() >= 0)
    {
        const auto imageRect = m_surface->scaledRect();
        imagePos = ((surfacePos - imageRect.topLeft()) / m_surface->scaling() - QPointF(0.5, 0.5)).toPoint();
    }

    m_surface->setActivePixel(imagePos);

    return false;
}
bool ImageSurfaceEventFilter::leaveEvent (QObject *o, QEvent *evt)
{
    m_surface->setActivePixel(-1, -1);

    m_surface->setCursor(Qt::OpenHandCursor);
    m_dragActive = false;

    return false;
}
