/* FILE:        ImageViewSurfaceStream.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::SurfaceStream.
 */

#include "ImageViewSurfaceStream_Private.h"

#include "AcquisitionSignalHandler.h"

// Construct a SurfaceStream for the given Surface (set as our parent), streaming from the
// given AbstractAcquisitionInterface.
ImageView::SurfaceStream::SurfaceStream (Surface *const surface, AbstractAcquisitionInterface *const acqIf)
    : QObject (surface)
    , m_pd (*new PrivateData(*this, *surface, *acqIf))
{
    // Setup the getNextFrame timer.
    m_pd.m_refreshTimer = new QTimer (this);
    connect(m_pd.m_refreshTimer, SIGNAL(timeout()), &m_pd, SLOT(getNextFrame()));

    m_pd.m_refreshTimer->setSingleShot(false);
    m_pd.m_refreshTimer->setInterval(1000 / 60);
    setRefreshRateAuto(true);

    // Setup the PrivateData acquisition routine.
    m_pd.startAcquisitionThread();

    // Connect the acquisition interface signal handler.
    connect(m_pd.m_acqIf.signalHandler(), SIGNAL(acquisitionStateChanged()), this, SLOT(updateAcquisitionState()), Qt::QueuedConnection);
    connect(m_pd.m_acqIf.signalHandler(), SIGNAL(getFrameNow()), &m_pd, SLOT(getNextFrame()), Qt::QueuedConnection);
}

// Destructor.
ImageView::SurfaceStream::~SurfaceStream (void)
{
    if (m_pd.m_refreshTimer->isActive())
        m_pd.m_refreshTimer->stop();

    m_pd.stopAcquisitionThread();
}

ImageView::Surface* ImageView::SurfaceStream::surface (void)
{
    return &m_pd.m_imageSurface;
}
const ImageView::Surface* ImageView::SurfaceStream::surface (void) const
{
    return &m_pd.m_imageSurface;
}

ImageView::AbstractAcquisitionInterface* ImageView::SurfaceStream::acquisitionInterface (void)
{
    return &m_pd.m_acqIf;
}
const ImageView::AbstractAcquisitionInterface* ImageView::SurfaceStream::acquisitionInterface (void) const
{
    return &m_pd.m_acqIf;
}

// The rate in frames-per-second at which the ImageView::Window attempts to update the display.
bool ImageView::SurfaceStream::refreshEnabled (void) const
{
    return m_pd.m_refreshTimer->isActive();
}
void ImageView::SurfaceStream::setRefreshEnabled (const bool enabled)
{
    if (enabled)
    {
        if (!m_pd.m_refreshTimer->isActive())
            m_pd.m_refreshTimer->start();
    }
    else
    {
        if (m_pd.m_refreshTimer->isActive())
            m_pd.m_refreshTimer->stop();
    }
}

bool ImageView::SurfaceStream::refreshRateAuto (void) const
{
    return m_pd.m_refreshRateAuto;
}
void ImageView::SurfaceStream::setRefreshRateAuto (const bool enabled)
{
    const bool oldRefreshRateAuto = m_pd.m_refreshRateAuto;

    const auto *const primaryScreen = QApplication::primaryScreen();

    m_pd.m_refreshRateAuto = enabled && primaryScreen;
    if (m_pd.m_refreshRateAuto)
    {
        m_pd.applyRefreshRateHz(primaryScreen->refreshRate());

        if (!oldRefreshRateAuto)
            connect(primaryScreen, SIGNAL(refreshRateChanged(qreal)), &m_pd, SLOT(applyRefreshRateHz(qreal)));
    }
    else if (primaryScreen)
        disconnect(primaryScreen, SIGNAL(refreshRateChanged(qreal)), &m_pd, SLOT(applyRefreshRateHz(qreal)));
}

int ImageView::SurfaceStream::refreshRate (void) const
{
    return 1000 / m_pd.m_refreshTimer->interval();
}
void ImageView::SurfaceStream::setRefreshRate (const int rate)
{
    setRefreshRateAuto(false);
    m_pd.applyRefreshRateHz(rate);
}

bool ImageView::SurfaceStream::bayerDemosaicEnabled (void) const
{
    return m_pd.m_decodeBayerImages;
}
void ImageView::SurfaceStream::setBayerDemosaicEnabled (const bool enabled)
{
    m_pd.setBayerDecodingEnabled(enabled);
}

ImageView::Format ImageView::SurfaceStream::bayerDemosaicFilter (void) const
{
    return m_pd.m_bayerDemosaicFilter;
}
void ImageView::SurfaceStream::setBayerDemosaicFilter (const Format filter)
{
    m_pd.setBayerDemosaicFilter(filter);
}

bool ImageView::SurfaceStream::rgbOrderReversed (void) const
{
    return m_pd.m_reverseRgbOrder;
}
void ImageView::SurfaceStream::setRgbOrderReversed (const bool reversed)
{
    m_pd.setRgbOrderReversed(reversed);
}

// GUI update slots.
void ImageView::SurfaceStream::updateAcquisitionState (void)
{
    m_pd.updateStreamStates();
    m_pd.updateImage();
}

// Properties.
QImage ImageView::SurfaceStream::cloneLatestImage (void) const
{
    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);

    if (m_pd.m_activeImage->m_image.isNull())
        return QImage();

    QImage imageClone = m_pd.m_activeImage->m_image.copy();
    imageClone.bits();

    return imageClone;
}
QImage ImageView::SurfaceStream::cloneLatestImageForRender (void) const
{
    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);

    if (m_pd.m_activeImage->m_image.isNull())
        return QImage();

    QImage imageClone = m_pd.m_activeImage->m_image.convertToFormat(QImage::Format_RGB32);
    imageClone.bits();

    return imageClone;
}
bool ImageView::SurfaceStream::latestImage (RawImage *const rawImage, ScopedLock *const criticalSection) const
{
    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);

    auto cs = m_pd.m_acqIf.getFrameLock(m_pd.m_activeImage->m_rawImage);
    if (!cs)
        return false;

    *rawImage = m_pd.m_activeImage->m_rawImage;
    *criticalSection = std::move(cs);

    return true;
}

QPixmap ImageView::SurfaceStream::peekPixmap (void) const
{
    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);
    return m_pd.m_activeImage->m_pixmap;
}

QVector<QRgb> ImageView::SurfaceStream::colorPalette (void) const
{
    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);
    return m_pd.m_colorTable;
}
void ImageView::SurfaceStream::setColorPalette (QVector<QRgb> const& palette)
{
    {
        const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);
        m_pd.m_colorTable = palette;
    }
    m_pd.applyColorTable();
}

QSize ImageView::SurfaceStream::latestImageSize (void) const
{
    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);
    return m_pd.m_activeImage->m_image.size();
}
ImageView::Format ImageView::SurfaceStream::latestImageFormat (void) const
{
    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);
    if (!m_pd.m_activeImage->m_rawImage.m_buffer)
        return UnknownFormat;
    return m_pd.m_activeImage->m_rawImage.m_format;
}
QImage::Format ImageView::SurfaceStream::latestQImageFormat (void) const
{
    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);
    if (m_pd.m_activeImage->m_image.isNull())
        return QImage::Format_Invalid;
    return m_pd.m_activeImage->m_image.format();
}

// Methods to enforce a constant bitmap. Format must be 8-, 24- or 32-bit.
bool ImageView::SurfaceStream::bitmapLocked (void) const
{
    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);
    return m_pd.m_bitmapLocked;
}
bool ImageView::SurfaceStream::setBitmapLocked (const int width, const int height, const Format format)
{
    switch (format)
    {
    default:
        return false;
    case Gray8: case Color24: case Color32:
        break;
    }

    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);

    m_pd.m_bitmapLocked = m_pd.m_imageAccessor.AllocateQImage(m_pd.m_activeImage->m_image, *m_pd.m_activeImage->m_data, width, height, format);
    emit bitmapLockChanged(m_pd.m_bitmapLocked);

    return m_pd.m_bitmapLocked;
}
void ImageView::SurfaceStream::unlockBitmap (void)
{
    m_pd.m_bitmapLocked = false;
    emit bitmapLockChanged(m_pd.m_bitmapLocked);
}

// Retrieve the locked bitmap pointer, if any. nullptr, if not bitmapLocked.
bool ImageView::SurfaceStream::getBitmap (RawImage *const rawImage) const
{
    const std::lock_guard<std::mutex> imagesLock (m_pd.m_imagesMutex);
    if (!m_pd.m_bitmapLocked)
        return false;

    switch (m_pd.m_activeImage->m_image.format())
    {
    default:
        return false;
    case QImage::Format_Indexed8:
        rawImage->m_format = ImageView::Gray8;
        break;
    case QImage::Format_RGB888:
        rawImage->m_format = ImageView::Color24;
        break;
    case QImage::Format_RGBX8888:
        rawImage->m_format = ImageView::Color32;
        break;
    }

    rawImage->m_buffer = m_pd.m_activeImage->m_data->data();
    rawImage->m_width = m_pd.m_activeImage->m_image.width();
    rawImage->m_height = m_pd.m_activeImage->m_image.height();

    return true;
}
