#ifndef INCLUDED__BIT_FLOW__IMAGE__VIEW__SURFACE__STREAM__H
#define INCLUDED__BIT_FLOW__IMAGE__VIEW__SURFACE__STREAM__H

/* FILE:        ImageViewSurfaceStream.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: SurfaceStream for ImageView, continuously streaming images from
 *              and AbstractAcquisitionInterface to an ImageView::Surface, for
 *              live conversion and display.
 */

#include "ImageView.h"

#include <QObject>
#include <QImage>

namespace ImageView
{
    class IMAGEVIEW_EXPORT SurfaceStream : public QObject
    {
        Q_OBJECT

        Q_PROPERTY(bool     refreshEnabled          READ refreshEnabled         WRITE setRefreshEnabled);
        Q_PROPERTY(bool     refreshRateAuto         READ refreshRateAuto        WRITE setRefreshRateAuto);
        Q_PROPERTY(int      refreshRate             READ refreshRate            WRITE setRefreshRate);

        Q_PROPERTY(bool     bayerDemosaicEnabled    READ bayerDemosaicEnabled   WRITE setBayerDemosaicEnabled);
        Q_PROPERTY(bool     rgbOrderReversed        READ rgbOrderReversed       WRITE setRgbOrderReversed);

    private:
        struct PrivateData;
        PrivateData &m_pd;

    public:
        // Construct a SurfaceStream for the given Surface (set as our parent), streaming from the
        // given AbstractAcquisitionInterface, which we will deallocate.
        SurfaceStream (Surface *const surface, AbstractAcquisitionInterface *const acqIf);

        // Destructor.
        ~SurfaceStream (void);

        Surface* surface (void);
        const Surface* surface (void) const;

        AbstractAcquisitionInterface* acquisitionInterface (void);
        const AbstractAcquisitionInterface* acquisitionInterface (void) const;

        // Stream control.
        bool isStreaming (void) const;
        void start (void);
        void stop (void);

        // Properties
        bool refreshEnabled (void) const;
        bool refreshRateAuto (void) const;
        int refreshRate (void) const;

        bool bayerDemosaicEnabled (void) const;

        Format bayerDemosaicFilter (void) const;
        void setBayerDemosaicFilter (const Format filter);

        bool rgbOrderReversed (void) const;

        QVector<QRgb> colorPalette (void) const;
        void setColorPalette (QVector<QRgb> const& palette);

        QImage cloneLatestImage (void) const;
        QImage cloneLatestImageForRender (void) const;

        bool latestImage (RawImage *const rawImage, ScopedLock *const criticalSection) const;
        QSize latestImageSize (void) const;
        Format latestImageFormat (void) const;

        QPixmap peekPixmap (void) const;

        QImage::Format latestQImageFormat (void) const;

        // Methods to enforce a constant bitmap. Format must be 8-, 24- or 32-bit.
        bool bitmapLocked (void) const;
        bool setBitmapLocked (const int width, const int height, const Format format);
        void unlockBitmap (void);

        // Retrieve the locked bitmap pointer, if any. nullptr, if not bitmapLocked.
        bool getBitmap (RawImage *const rawImage) const;

    public slots:
        // GUI update slots.
        void updateAcquisitionState (void);

        // Properties.
        void setRefreshEnabled (const bool enabled);
        void setRefreshRateAuto (const bool enabled);
        void setRefreshRate (const int rate);

        void setBayerDemosaicEnabled (const bool enabled);
        void setRgbOrderReversed (const bool reversed);

    signals:
        void imageUpdated (void);
        void bitmapLockChanged (const bool locked);
    };
}

#endif // INCLUDED__BIT_FLOW__IMAGE__VIEW__SURFACE__STREAM__H
