/* FILE:        ImageViewSurfaceStream_Private.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::SurfaceStream::PrivateData.
 */

#include "ImageViewSurfaceStream_Private.h"

#include "ImageViewThread.h"

ImageView::SurfaceStream::PrivateData::PrivateData (SurfaceStream &stream, Surface &surface, AbstractAcquisitionInterface &acqIf)
    : QObject (&stream)
    , m_surfaceStream (stream)
    , m_imageSurface (surface)
    , m_acqIf (acqIf)

    , m_refreshRateAuto (false)
    , m_decodeBayerImages (true)
    , m_bayerDemosaicFilter (UnknownFormat)
    , m_reverseRgbOrder (false)

    , m_bitmapLocked (false)

    , m_convertFrameThreadKeepRunning (false)
{
    m_activeImage = &m_images[0];
    m_inactiveImage = &m_images[1];
}

ImageView::SurfaceStream::PrivateData::~PrivateData (void)
{
    stopAcquisitionThread();
}

void ImageView::SurfaceStream::PrivateData::startAcquisitionThread (void)
{
    // Start the image retrieval/conversion thread.
    std::unique_lock<std::mutex> convertFrameLock (m_convertFrameMutex);
    if (!m_convertFrameThreadKeepRunning.exchange(true))
    {
        m_convertFrameThread = ImageView::thread(&PrivateData::getLatestFrameThreadFunc, this);
        m_convertFrameCondition.wait(convertFrameLock);
    }
}

void ImageView::SurfaceStream::PrivateData::stopAcquisitionThread (void)
{
    // Stop the frame acquisition and conversion thread. This should be done BEFORE stopping the acquisition.
    ImageView::thread convertFrameThread;
    {
        const std::lock_guard<std::mutex> convertFrameLock (m_convertFrameMutex);
        m_convertFrameThreadKeepRunning.store(false);
        m_convertFrameCondition.notify_all();
        convertFrameThread = std::move(m_convertFrameThread);
    }

    if (convertFrameThread.joinable())
        convertFrameThread.join();
}

void ImageView::SurfaceStream::PrivateData::setBayerDecodingEnabled (const bool enabled)
{
    bool doUpdate = false;
    {
        const std::unique_lock<std::mutex> imagesLock (m_imagesMutex);

        const bool checkReRender = m_decodeBayerImages != enabled;
        m_decodeBayerImages = enabled;

        if (checkReRender)
        {
            const bool hasFilter = (m_activeImage->m_rawImage.m_format & BayerFormatMask) || (m_bayerDemosaicFilter & BayerFormatMask);
            if (hasFilter && isGrayscale(m_activeImage->m_rawImage.m_format))
                doUpdate = unlocked_reRenderDisplayImage(imagesLock);
        }
    }

    if (doUpdate)
        updateImage();
}
void ImageView::SurfaceStream::PrivateData::setBayerDemosaicFilter (const Format filter)
{
    bool doUpdate = false;
    {
        const std::unique_lock<std::mutex> imagesLock (m_imagesMutex);

        const Format oldFilter = m_bayerDemosaicFilter;
        m_bayerDemosaicFilter = (Format)(filter & BayerFormatMask);

        if (oldFilter != m_bayerDemosaicFilter && isGrayscale(m_activeImage->m_rawImage.m_format))
            doUpdate = unlocked_reRenderDisplayImage(imagesLock);
    }

    if (doUpdate)
        updateImage();
}
void ImageView::SurfaceStream::PrivateData::setRgbOrderReversed (const bool reversed)
{
    bool doUpdate = false;
    {
        const std::unique_lock<std::mutex> imagesLock (m_imagesMutex);

        m_reverseRgbOrder = reversed;

        if (isColor(m_activeImage->m_rawImage.m_format))
            doUpdate = unlocked_reRenderDisplayImage(imagesLock);
    }

    if (doUpdate)
        updateImage();
}

void ImageView::SurfaceStream::PrivateData::applyColorTable (void)
{
    const std::lock_guard<std::mutex> imagesLock (m_imagesMutex);
    if (!m_activeImage->m_pixmap.isNull() && !m_activeImage->m_image.isNull())
    {
        if (m_activeImage->m_image.format() == QImage::Format_Indexed8)
        {
            m_activeImage->updatePixmap(m_colorTable);
            m_imageSurface.setPixmap(m_activeImage->m_pixmap);
        }
    }
}

void ImageView::SurfaceStream::PrivateData::applyRefreshRateHz (const qreal rateHz)
{
    if (rateHz > 0.)
        m_refreshTimer->setInterval(1000. / rateHz);
}

void ImageView::SurfaceStream::PrivateData::updateStreamStates (void)
{
    const bool isSetup = m_acqIf.isSetup();

    if (isSetup)
    {
        if (!m_refreshTimer->isActive())
            m_refreshTimer->start();
    }
    else
    {
        if (m_refreshTimer->isActive())
            m_refreshTimer->stop();

        {
            const std::lock_guard<std::mutex> imagesLock (m_imagesMutex);
            m_activeImage->m_data->clear();
            m_activeImage->m_image = QImage();
            m_activeImage->m_pixmap = QPixmap();
            m_activeImage->m_rawImage.m_buffer = nullptr;
            m_activeImage->m_rawImageImposed.m_buffer = nullptr;

            m_inactiveImage->m_data->clear();
            m_inactiveImage->m_image = QImage();
            m_inactiveImage->m_pixmap = QPixmap();
            m_inactiveImage->m_rawImage.m_buffer = nullptr;
            m_inactiveImage->m_rawImageImposed.m_buffer = nullptr;
        }

        updateImage();
    }
}

void ImageView::SurfaceStream::PrivateData::updateImage (void)
{
    {
        const std::lock_guard<std::mutex> imagesLock (m_imagesMutex);

        QPixmap dispPix = m_imageSurface.pixmap();
        const bool wasNull = dispPix.isNull();
        const bool isNull = m_activeImage->m_pixmap.isNull();
        if (!wasNull && isNull)
        {
            m_imageSurface.setPixmap(QPixmap());
            m_imageSurface.clearImageSrc();
        }
        else if (!isNull)
        {
            m_imageSurface.setPixmap(m_activeImage->m_pixmap);

            if (m_decodeBayerImages && (m_activeImage->m_rawImageImposed.m_format & BayerFormatMask))
                m_imageSurface.setImageSrc(m_activeImage->m_image);
            else
                m_imageSurface.setImageSrc(&m_acqIf, m_activeImage->m_rawImageImposed);
        }
    }

    emit m_surfaceStream.imageUpdated();
}

void ImageView::SurfaceStream::PrivateData::getNextFrame (const bool getNew)
{
    size_t latestFrameNum;
    if (getNew)
    {
        std::unique_lock<std::mutex> convertFrameLock (m_convertFrameMutex);

        m_latestFrameNum = m_convertedFrameCnt.load();
        m_latestDispFrameNum = m_latestFrameNum;
        m_convertFrameCondition.notify_one();
        m_convertFrameCondition.wait(convertFrameLock);

        latestFrameNum = m_convertedFrameCnt.load();
    }
    else
    {
        const std::lock_guard<std::mutex> convertFrameLock (m_convertFrameMutex);
        m_convertFrameCondition.notify_one();

        latestFrameNum = m_convertedFrameCnt.load();
    }

    if (latestFrameNum != m_latestFrameNum)
    {
        m_latestFrameNum = latestFrameNum;
        updateImage();
    }
}

void ImageView::SurfaceStream::PrivateData::getLatestFrameThreadFunc (void)
{
    std::unique_lock<std::mutex> convertFrameLock (m_convertFrameMutex);
    m_convertedFrameCnt.store(0);
    m_latestFrameNum = 0;
    m_latestDispFrameNum = 0;
    m_convertFrameCondition.notify_one();

    while (m_convertFrameThreadKeepRunning.load())
    {
        m_convertFrameCondition.wait(convertFrameLock);
        if (!m_convertFrameThreadKeepRunning.load())
            break;

        const std::unique_lock<std::mutex> imagesLock (m_imagesMutex);

        RawImage rawImage;
        ScopedLock bufferCriticalSection;
        if (m_acqIf.getNextFrame(&rawImage, &bufferCriticalSection))
        {
            const bool bitmapLocked = m_bitmapLocked;

            // Swap the active image buffer set.
            WindowImage::swap(m_activeImage, m_inactiveImage, bitmapLocked);

            // Store the old image data, and apply the new image data.
            const auto rawImageOld = m_activeImage->m_rawImage;
            const auto rawImageImposedOld = m_activeImage->m_rawImageImposed;
            m_activeImage->m_rawImage = rawImage;
            m_activeImage->m_rawImageImposed = rawImage;

            // Attempt the conversion
            if (unlocked_renderDisplayImage(imagesLock, bufferCriticalSection))
                m_convertedFrameCnt.fetch_add(1);
            else
            {
                // On failure, restore the old image data and the previously active image.
                m_activeImage->m_rawImage = rawImageOld;
                m_activeImage->m_rawImageImposed = rawImageImposedOld;

                WindowImage::swap(m_activeImage, m_inactiveImage, bitmapLocked);
            }
        }

        m_convertFrameCondition.notify_one();
    }
}

bool ImageView::SurfaceStream::PrivateData::unlocked_renderDisplayImage (std::unique_lock<std::mutex> const& imagesLock, ScopedLock const& bufferCriticalSection)
{
#if defined(_DEBUG)
    if (!imagesLock.owns_lock() || imagesLock.mutex() != &m_imagesMutex)
        return false;
    if (!bufferCriticalSection)
        return false;
#endif

    // Check explicitly for null images. Simply clear the display.
    if (!m_activeImage->m_rawImage.m_buffer)
    {
        m_activeImage->m_pixmap = QPixmap();
        return true;
    }

    // Setup the source image format.
    RawImage rawImageImposed = m_activeImage->m_rawImage;

    int imposedFormat = rawImageImposed.m_format & BaseFormatMask;

    if (m_decodeBayerImages && !m_bitmapLocked)
    {
        if (m_bayerDemosaicFilter & BayerFormatMask)
            imposedFormat |= m_bayerDemosaicFilter & BayerFormatMask;
        else
            imposedFormat |= rawImageImposed.m_format & BayerFormatMask;
    }

    if (m_reverseRgbOrder)
    {
        if ((rawImageImposed.m_format & ColorOrderFormatMask) == ColorOrderBGR)
            imposedFormat |= ColorOrderRGB;
        else
            imposedFormat |= ColorOrderBGR;
    }
    else
        imposedFormat |= rawImageImposed.m_format & ColorOrderFormatMask;

    switch (rawImageImposed.m_format & PackedFormatMask)
    {
    case Packed:
        imposedFormat |= Packed;
        break;
    case PackedNoSwap:
        imposedFormat |= PackedNoSwap;
        break;
    }

    rawImageImposed.m_format = (Format)imposedFormat;

    // Perform the conversion from our raw buffer to a QImage.
    bool qImageRgbSwap = false;
    if (m_bitmapLocked)
    {
        // If the data buffer is the QImage buffer, conversion may not be necessary.
        const bool externalBuffer = m_activeImage->m_data->data() != rawImageImposed.m_buffer;

        if (externalBuffer)
        {
            // The bitmap is locked, so do not reallocate the current image data.
            if (!m_imageAccessor.QImageFromBufferConst(m_activeImage->m_image, *m_activeImage->m_data, rawImageImposed))
                return false;
        }
        else
            // If this is color, and isn't BGR, we need to swap RGB explicitly. So this
            // won't collide with possibly ongoing external processing, use QImage::rgbSwapped.
            qImageRgbSwap = isColor(rawImageImposed.m_format) && (rawImageImposed.m_format & ColorOrderFormatMask) != ColorOrderBGR;
    }
    else
    {
        // Perform whatever conversion and allocation might be necessary.
        if (!m_imageAccessor.QImageFromBuffer(m_activeImage->m_image, *m_activeImage->m_data, rawImageImposed))
            return false;
    }

    // Save the imposed image parameters.
    m_activeImage->m_rawImageImposed = rawImageImposed;

    // Convert the QImage to a QPixmap.
    m_activeImage->updatePixmap(m_colorTable, qImageRgbSwap);

    return true;
}

bool ImageView::SurfaceStream::PrivateData::unlocked_reRenderDisplayImage (std::unique_lock<std::mutex> const& imagesLock)
{
#if defined(_DEBUG)
    if (!imagesLock.owns_lock() || imagesLock.mutex() != &m_imagesMutex)
        return false;
#endif

    // Do nothing, if we lack an active image.
    if (m_activeImage->m_pixmap.isNull() || m_activeImage->m_image.isNull())
        return false;

    // Lock the raw buffer with a critical section.
    const auto bufferCriticalSection = m_acqIf.getFrameLock(m_activeImage->m_rawImage);
    if (!bufferCriticalSection)
        return false;

    return unlocked_renderDisplayImage(imagesLock, bufferCriticalSection);
}
