#ifndef INCLUDED__BIT_FLOW__IMAGE__VIEW__SURFACE__STREAM_____PRIVATE__H
#define INCLUDED__BIT_FLOW__IMAGE__VIEW__SURFACE__STREAM_____PRIVATE__H

/* FILE:        ImageViewSurfaceStream_Private.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: PrivateData for ImageView::SurfaceStream.
 */

#include "ImageViewSurfaceStream.h"
#include "ImageViewSurface.h"
#include "AbstractAcquisitionInterface.h"
#include "ImageViewImageAccessor.h"

#include "ImageViewThread.h"

#include <vector>

#include <atomic>
#include <condition_variable>

#include <QApplication>
#include <QScreen>

#include <QTimer>
#include <QVector>

#include <QImage>
#include <QPixmap>

struct WindowImage
{
private:
    std::vector<uchar> m_dataInst;

public:
    std::vector<uchar> *m_data;
    QImage m_image;
    QPixmap m_pixmap;
    ImageView::RawImage m_rawImage;
    ImageView::RawImage m_rawImageImposed;

    WindowImage (void)
        : m_data (&m_dataInst)
    { }

    inline QImage finalImage (QVector<uint> const& colorTable = QVector<uint>(), const bool rgbSwapped = false)
    {
        if (m_image.format() == QImage::Format_Indexed8)
        {
            if (colorTable.size() != 256)
            {
#if QT_VERSION >= QT_VERSION_CHECK(5, 5, 0)
                return QImage(m_data->data(), m_image.width(), m_image.height(), m_image.bytesPerLine(), QImage::Format_Grayscale8);
#else
                QVector<uint> defaultColorTable (256);
                for (uint i = 0; i < 256; i++)
                    defaultColorTable[i] = qRgb(i, i, i);
                m_image.setColorTable(defaultColorTable);
#endif
            }
            else
                m_image.setColorTable(colorTable);
        }
        else if (rgbSwapped)
            return m_image.rgbSwapped();

        return m_image;
    }

    inline void updatePixmap (QVector<uint> const& colorTable = QVector<uint>(), const bool rgbSwapped = false)
    {
        m_pixmap.convertFromImage(finalImage(colorTable, rgbSwapped));
    }

    inline static void swap (WindowImage *& left, WindowImage *& right, const bool bitmapLocked)
    {
        std::swap(left, right);

        if (bitmapLocked)
        {
            std::swap(left->m_data, right->m_data);
            std::swap(left->m_image, right->m_image);
        }
    }
};

struct ImageView::SurfaceStream::PrivateData : public QObject
{
    Q_OBJECT

public:
    PrivateData (SurfaceStream &stream, Surface &surface, AbstractAcquisitionInterface &acqIf);
    ~PrivateData (void);

    void startAcquisitionThread (void);
    void stopAcquisitionThread (void);

    void setBayerDecodingEnabled (const bool enabled);
    void setBayerDemosaicFilter (const Format filter);
    void setRgbOrderReversed (const bool reversed);

    void applyColorTable (void);

    // Members.
    QTimer *m_refreshTimer;

    bool m_refreshRateAuto;
    bool m_decodeBayerImages;
    Format m_bayerDemosaicFilter;
    bool m_reverseRgbOrder;

    SurfaceStream &m_surfaceStream;
    Surface &m_imageSurface;
    AbstractAcquisitionInterface &m_acqIf;

    ImageAccessor m_imageAccessor;

    QVector<QRgb> m_colorTable;

    WindowImage m_images[2];
    WindowImage *m_activeImage, *m_inactiveImage;
    bool m_bitmapLocked;
    mutable std::mutex m_imagesMutex;

    ImageView::thread m_convertFrameThread;
    std::mutex m_convertFrameMutex;
    std::condition_variable m_convertFrameCondition;
    std::atomic<bool> m_convertFrameThreadKeepRunning;

    std::atomic<size_t> m_convertedFrameCnt;
    size_t m_latestFrameNum;
    size_t m_latestDispFrameNum;

public slots:
    void applyRefreshRateHz (const qreal rateHz);

    void updateStreamStates (void);
    void updateImage (void);

    void getNextFrame (const bool getNew = false);

private:
    void getLatestFrameThreadFunc (void);
    bool unlocked_renderDisplayImage (std::unique_lock<std::mutex> const& imagesLock, ScopedLock const& bufferCriticalSection);
    bool unlocked_reRenderDisplayImage (std::unique_lock<std::mutex> const& imagesLock);
};

#endif // INCLUDED__BIT_FLOW__IMAGE__VIEW__SURFACE__STREAM_____PRIVATE__H
