/* FILE:        ImageViewSurface_Private.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::Surface::PrivateData.
 */

#include "ImageViewSurface_Private.h"

#include "BayerDemosaic.h"

#include <atomic>
#include <vector>
#include <cmath>

#include <QPainter>

#include <QFontDatabase>

const double ImageView::Surface::PrivateData::s_paddingRatio = 9. / 8.;

ImageView::Surface::PrivateData::PrivateData (ImageView::Surface &parent)
    : m_parent (parent)

#if IMAGE_VIEW_SURFACE_USE_QGL
    , m_pixmapTexture (0)
    , m_textureExpired (false)
#endif

    , m_acqIf (nullptr)

    , m_scaling (1.)
    , m_offset (0, 0)
    , m_scaleToFit (true)

    , m_crosshairVisible (false)
    , m_activePixel (-1, -1)

    , m_valueOverlayEnabled (true)

    , m_imageBufferSizeChange (true)
    , m_imagePixPosSizeMax (0, 0)
    , m_imagePixValSizeMax (0, 0)
    , m_overallPixSizeMax (0, 0)

    , m_painterFont (QFontDatabase::systemFont(QFontDatabase::SmallestReadableFont))
    , m_painterFontPointSizeDefault (QFontDatabase::systemFont(QFontDatabase::SmallestReadableFont).pointSize())
{
    m_rawImage.m_buffer = nullptr;
}

ImageView::Surface::PrivateData::~PrivateData (void)
{
#if IMAGE_VIEW_SURFACE_USE_QGL
    if (m_pixmapTexture)
    {
        m_parent.deleteTexture(m_pixmapTexture);
        m_pixmapTexture = 0;
    }
#endif
}

void ImageView::Surface::PrivateData::unlocked_setFittedOffset (std::lock_guard<std::mutex> const& lock, QPointF const& offset)
{
    const auto oldOffset = m_offset;

    if (m_pixmap.isNull())
    {
        m_offset.setX(std::max(offset.x(), 0.));
        m_offset.setY(std::max(offset.y(), 0.));
    }
    else
    {
        const QSizeF scaledPixmapSize = QSizeF(m_pixmap.size()) * m_scaling;
        const QSizeF surfaceSize = m_parent.size();
        const QSizeF sizeDiff = scaledPixmapSize - surfaceSize;

        const double maxXOffset = std::max(0., sizeDiff.width());
        const double maxYOffset = std::max(0., sizeDiff.height());

        m_offset.setX(std::min(std::max(offset.x(), 0.), maxXOffset));
        m_offset.setY(std::min(std::max(offset.y(), 0.), maxYOffset));
    }

    if (oldOffset != m_offset)
        emit m_parent.pixelOffsetChanged(m_offset);
}
bool ImageView::Surface::PrivateData::unlocked_checkScaleToFit (std::lock_guard<std::mutex> const& lock, const bool setEnabled)
{
    if (!m_scaleToFit || m_pixmap.isNull())
        return false;

    const QSizeF viewerSize = m_parent.size();
    const QSizeF imageSize = m_pixmap.size();

    double newScaling;
    if (imageSize.isEmpty())
        newScaling = 1.;
    else if (imageSize.width() / imageSize.height() > viewerSize.width() / viewerSize.height())
        newScaling = imageSize.width() == 0. ? 1. : viewerSize.width() / imageSize.width();
    else
        newScaling = imageSize.height() == 0. ? 1. : viewerSize.height() / imageSize.height();

    const double fittedScaling = std::max(scalingMin(), std::min(scalingMax(), newScaling));
    const double oldScaling = m_scaling;
    m_scaling = fittedScaling;

    unlocked_setFittedOffset(lock, m_offset);

    const bool hasNewScaling = std::fabs(oldScaling - m_scaling) >= std::numeric_limits<double>::epsilon();
    if (hasNewScaling)
    {
        emit m_parent.pixelRangeChanged();
        emit m_parent.scalingChanged(m_scaleToFit, m_scaling);
    }
    else if (setEnabled)
        emit m_parent.scalingChanged(m_scaleToFit, m_scaling);

    m_parent.update();

    return hasNewScaling;
}

QSizeF ImageView::Surface::PrivateData::unlocked_scaledSize (std::lock_guard<std::mutex> const& lock) const
{
    return QSizeF(m_pixmap.size()) * m_scaling;
}
QRectF ImageView::Surface::PrivateData::unlocked_scaledRect (std::lock_guard<std::mutex> const& lock) const
{
    return QRectF(-unlocked_scaledOffset(lock), unlocked_scaledSize(lock));
}
QPointF ImageView::Surface::PrivateData::unlocked_scaledOffset (std::lock_guard<std::mutex> const& lock) const
{
    if (m_pixmap.isNull())
        return m_offset;

    const QSizeF widgetSize = m_parent.size();
    const QSizeF scaledPixmapSize = QSizeF(m_pixmap.size()) * m_scaling;
    const QSizeF sizeDiffHalf = (widgetSize - scaledPixmapSize) / 2.;

    const QPointF realOffset
        (
            sizeDiffHalf.width() > 0 ? -sizeDiffHalf.width() : m_offset.x(),
            sizeDiffHalf.height() > 0 ? -sizeDiffHalf.height() : m_offset.y()
        );
    return realOffset;
}
QRectF ImageView::Surface::PrivateData::unlocked_visibleImageRectF (std::lock_guard<std::mutex> const& lock) const
{
    const double scaling = m_scaling;
    const QSizeF widgetSize = m_parent.size();
    const QRectF pixmapRect = m_pixmap.rect();
    const QSizeF scaledPixmapSize = pixmapRect.size() * scaling;

    const QPointF scaledOffset
        (
            widgetSize.width() > scaledPixmapSize.width() ? 0. : m_offset.x(),
            widgetSize.height() > scaledPixmapSize.height() ? 0. : m_offset.y()
        );

    const QSizeF visibleSize
        (
            std::min(widgetSize.width(), scaledPixmapSize.width()),
            std::min(widgetSize.height(), scaledPixmapSize.height())
        );

    const QRectF visibleImgRect (scaledOffset / scaling, visibleSize / scaling);

    return pixmapRect.intersected(visibleImgRect);
}

bool ImageView::Surface::PrivateData::renderPixmap (QPainter &painter, QRectF const& targetRect)
{
#if IMAGE_VIEW_SURFACE_USE_QGL

    if (m_textureExpired.exchange(false))
    {
        if (m_pixmapTexture)
        {
            m_parent.deleteTexture(m_pixmapTexture);
            m_pixmapTexture = 0;
        }
    
        const std::lock_guard<std::mutex> pixmapLock (m_pixmapMutex);
        if (!m_pixmap.isNull())
            m_pixmapTexture = m_parent.bindTexture(m_pixmap, GL_TEXTURE_2D, GL_RGB, QGLContext::LinearFilteringBindOption);
    }

    if (m_pixmapTexture)
    {
        const auto textureRect = m_parent.scaledRect();
        const QRectF textureRectSwapped (textureRect.bottomLeft(), textureRect.topRight());
        m_parent.drawTexture(textureRectSwapped, m_pixmapTexture);
    }

#else

    const std::lock_guard<std::mutex> pixmapLock (m_pixmapMutex);

    const auto pixmapRect = unlocked_visibleImageRectF(pixmapLock);
    const auto scaledRect = unlocked_scaledRect(pixmapLock);
    const QSizeF widgetSize = m_parent.size();

    const QPointF scaledTopLeft = scaledRect.topLeft();

    const QPointF renderPoint (std::max(0., scaledTopLeft.x()), std::max(0., scaledTopLeft.y()));
    const QSizeF renderSize (std::min(scaledRect.width(), widgetSize.width()), std::min(scaledRect.height(), widgetSize.height()));
    const QRectF renderRect (renderPoint, renderSize);

    if (!renderRect.isEmpty() && !pixmapRect.isEmpty())
        painter.drawPixmap(renderRect, m_pixmap, pixmapRect);

#endif

    return true;
}
bool ImageView::Surface::PrivateData::renderCrosshair (QPainter &painter, QRectF const& targetRect)
{
    const auto visibleImageRect = m_parent.visibleImageRect();
    const double scaling = m_scaling;
    const auto location = m_activePixel;

    QVector<QLineF> lines;
    if (!visibleImageRect.contains(location))
        return false;

    const auto offsetInverse = -m_parent.scaledOffset();

    const double pixTop = double(location.y()) * scaling + offsetInverse.y();
    const double pixLeft = double(location.x()) * scaling + offsetInverse.x();

    const QLineF topLine (targetRect.left(), pixTop, targetRect.right(), pixTop);
    lines.push_back(topLine);

    const QLineF leftLine (pixLeft, targetRect.top(), pixLeft, targetRect.bottom());
    lines.push_back(leftLine);

    if (scaling >= 1.5)
    {
        const QLineF bottomLine (pixLeft, pixTop + scaling, pixLeft + scaling, pixTop + scaling);
        lines.push_back(bottomLine);

        const QLineF rightLine (pixLeft + scaling, pixTop, pixLeft + scaling, pixTop + scaling);
        lines.push_back(rightLine);
    }

    QPen linePen (Qt::white);
    QVector<qreal> dashPattern;
    dashPattern << 4 << 4;
    linePen.setDashPattern(dashPattern);

    painter.setPen(linePen);
    painter.setBackground(Qt::black);
    painter.setBackgroundMode(Qt::OpaqueMode);
    painter.drawLines(lines);

    return true;
}
bool ImageView::Surface::PrivateData::renderPixelOverlay (QPainter &painter, QRectF const& targetRect)
{
    // Extract pixel values.
    PixPairVector pixelPairs;
    QSizeF posSizeMax, valSizeMax;
    {
        const std::lock_guard<std::mutex> imageBufferLock (m_imageBufferMutex);
        if (!unlocked_overlayPixelsFromBuffer(imageBufferLock, targetRect, pixelPairs, posSizeMax, valSizeMax))
        {
            if (!unlocked_overlayPixelsFromImage(imageBufferLock, targetRect, pixelPairs, posSizeMax, valSizeMax))
                return false;
        }
    }

    const auto imageRect = m_parent.scaledRect();
    const double scaling = m_scaling;

    // Do the painting.
    const QSize targetSize = targetRect.size().toSize();
    if (targetSize.isEmpty())
        return false;

    painter.setFont(m_painterFont);
    painter.setBrush(Qt::white);
    painter.setBackground(Qt::white);
    painter.setBackgroundMode(Qt::OpaqueMode);

    painter.setRenderHint(QPainter::HighQualityAntialiasing);
    painter.setRenderHint(QPainter::TextAntialiasing);
    
    const QTextOption textOptionCenter (Qt::AlignCenter);
    const QTextOption textOptionJustify (Qt::AlignJustify);

    const QPointF pixBtmRgt (scaling, scaling);

    const double biggerWidth = std::max(posSizeMax.width(), valSizeMax.width());
    QRectF unitedOutlineRect (QPointF(), QSizeF(biggerWidth, posSizeMax.height() + valSizeMax.height()));
    QRectF posOutlineRect (QPointF(), posSizeMax);
    QRectF valOutlineRect (QPointF(), valSizeMax);

    const QPointF unitedOutlineOffset = (pixBtmRgt - unitedOutlineRect.bottomRight()) / 2.;
    const QPointF posOutlineOffset ((pixBtmRgt.x() - posOutlineRect.width()) / 2., unitedOutlineOffset.y());
    const QPointF valOutlineOffset ((pixBtmRgt.x() - valOutlineRect.width()) / 2., posOutlineOffset.y() + posOutlineRect.height());
    
    for (auto const& pixPair : pixelPairs)
    {
        const QPointF pixPos = pixPair.first * scaling + imageRect.topLeft();

        posOutlineRect.moveTopLeft(pixPos + posOutlineOffset);
        valOutlineRect.moveTopLeft(pixPos + valOutlineOffset);
        unitedOutlineRect.moveTopLeft(pixPos + unitedOutlineOffset);

        const QString stringPos = " " + QString::number(pixPair.first.x()) + ", " + QString::number(pixPair.first.y()) + " ";

        QString stringVal;
        QTextOption strValOpt;
        if (pixPair.second.isColor)
        {
            stringVal = " R: " + QString::number(pixPair.second.r) + " \n G: " + QString::number(pixPair.second.g) + " \n B: " + QString::number(pixPair.second.b) + " ";
            strValOpt = textOptionJustify;
        }
        else if (pixPair.second.isGrayscale)
        {
            stringVal = " " + QString::number(pixPair.second.gray) + " ";
            strValOpt = textOptionCenter;
        }

        painter.setPen(Qt::black);
        painter.drawRect(unitedOutlineRect);

        painter.drawText(posOutlineRect, stringPos, textOptionCenter);
        painter.drawText(valOutlineRect, stringVal, strValOpt);

        painter.setPen(Qt::DotLine);
        painter.drawLine(posOutlineRect.bottomLeft() + QPointF(5, 0), posOutlineRect.bottomRight() - QPointF(5, 0));
    }

    return true;
}
bool ImageView::Surface::PrivateData::unlocked_overlayPixelsFromBuffer (std::lock_guard<std::mutex> const& lock, QRectF const& targetRect, PixPairVector &pixPairs, QSizeF &posSizeMax, QSizeF &valSizeMax)
{
    if (!m_acqIf || !m_rawImage.m_buffer)
        return false;

    const auto imageRect = m_parent.scaledRect();
    const double scaling = m_scaling;

    // Determine the maximum string size in pixels.
    if (m_imageBufferSizeChange.exchange(false))
    {
        std::uint16_t maxVal = ~0;
        bool gray = false;
        switch (m_rawImage.m_format & BaseFormatMask)
        {
        case Gray8: case Gray10: case Gray12: case Gray14: case Gray16:
            maxVal = (1u << (m_rawImage.m_format & BaseFormatMask)) - 1;
            gray = true;
            break;
        case Color24: case Color32:
            maxVal = (1u << 8) - 1;
            gray = false;
            break;
        case Color30:
            maxVal = (1u << 10) - 1;
            gray = false;
            break;
        case HiColor36: case HiColor42: case HiColor48:
            maxVal = (1u << ((m_rawImage.m_format & BaseFormatMask) / 3)) - 1;
            gray = false;
            break;
        }

        const QString maxPosStr = " " + QString::number(m_rawImage.m_width - 1) + ", " + QString::number(m_rawImage.m_height - 1) + " ";

        QString maxValStr;
        if (gray)
            maxValStr = " " + QString::number(maxVal) + " ";
        else
            maxValStr = " R: " + QString::number(maxVal) + " \n G: " + QString::number(maxVal) + " \n B: " + QString::number(maxVal) + " ";

        unlocked_findOptimalOverlaySizes(lock, maxPosStr, maxValStr);
    }

    posSizeMax = m_imagePixPosSizeMax;
    valSizeMax = m_imagePixValSizeMax;

    // Ensure the values fit.
    if (m_overallPixSizeMax.width() * s_paddingRatio >= scaling || m_overallPixSizeMax.height() * s_paddingRatio >= scaling)
    {
        m_imageBufferSizeChange.store(true);
        return false;
    }

    // If we can get a critical section, extract all visible pixel values and locations.
    const auto rawImage = m_rawImage;
    if (const auto criticalSection = m_acqIf->getFrameLock(rawImage))
    {
        const size_t xBase = floor((targetRect.left() - imageRect.left()) / scaling);
        const size_t yBase = floor((targetRect.top() - imageRect.top()) / scaling);
        const size_t xEndPadded = xBase + ceil(targetRect.width() / scaling) + 1.;
        const size_t yEndPadded = yBase + ceil(targetRect.height() / scaling) + 1.;
        const size_t xEnd = std::min(xEndPadded, rawImage.m_width);
        const size_t yEnd = std::min(yEndPadded, rawImage.m_height);

        PixVal pixVal;
        for (size_t xPos = xBase; xPos < xEnd; xPos++)
        {
            for (size_t yPos = yBase; yPos < yEnd; yPos++)
            {
                const PixPos pixPos (xPos, yPos);
                if (m_imageAccessor.BufferPixelValue(rawImage, pixPos, pixVal))
                    pixPairs.push_back(qMakePair(QPointF(xPos, yPos), pixVal));
            }
        }
    }
    else
        return false;

    return true;
}
bool ImageView::Surface::PrivateData::unlocked_overlayPixelsFromImage (std::lock_guard<std::mutex> const& lock, QRectF const& targetRect, PixPairVector &pixPairs, QSizeF &posSizeMax, QSizeF &valSizeMax)
{
    if (m_image.isNull())
        return false;

    const auto imageRect = m_parent.scaledRect();
    const double scaling = m_scaling;

    // Determine the maximum string size in pixels.
    if (m_imageBufferSizeChange.exchange(false))
    {
        const std::uint16_t maxVal = 255;

        const QString maxPosStr = " " + QString::number(m_image.width() - 1) + ", " + QString::number(m_image.height() - 1) + " ";

        QString maxValStr;
        if (m_image.format() == QImage::Format_Indexed8)
            maxValStr = " " + QString::number(maxVal) + " ";
        else
            maxValStr = " R: " + QString::number(maxVal) + " \n G: " + QString::number(maxVal) + " \n B: " + QString::number(maxVal) + " ";

        unlocked_findOptimalOverlaySizes(lock, maxPosStr, maxValStr);
    }

    posSizeMax = m_imagePixPosSizeMax;
    valSizeMax = m_imagePixValSizeMax;

    // Ensure the values fit.
    if (m_overallPixSizeMax.width() * s_paddingRatio >= scaling || m_overallPixSizeMax.height() * s_paddingRatio >= scaling)
    {
        m_imageBufferSizeChange.store(true);
        return false;
    }

    // Extract all visible pixel values and locations.
    const size_t xBase = floor((targetRect.left() - imageRect.left()) / scaling);
    const size_t yBase = floor((targetRect.top() - imageRect.top()) / scaling);
    const size_t xEndPadded = xBase + ceil(targetRect.width() / scaling) + 1.;
    const size_t yEndPadded = yBase + ceil(targetRect.height() / scaling) + 1.;
    const size_t xEnd = std::min<size_t>(xEndPadded, m_image.width());
    const size_t yEnd = std::min<size_t>(yEndPadded, m_image.height());

    PixVal pixVal;
    pixVal.isGrayscale = m_image.format() == QImage::Format_Indexed8;
    pixVal.isColor = !pixVal.isGrayscale;
    for (size_t xPos = xBase; xPos < xEnd; xPos++)
    {
        for (size_t yPos = yBase; yPos < yEnd; yPos++)
        {
            const QRgb pixel = m_image.pixel(xPos, yPos);
            pixVal.r = qRed(pixel);
            pixVal.g = qGreen(pixel);
            pixVal.b = qBlue(pixel);
            pixVal.gray = pixVal.r;
            pixPairs.push_back(qMakePair(QPointF(xPos, yPos), pixVal));
        }
    }

    return true;
}
void ImageView::Surface::PrivateData::unlocked_findOptimalOverlaySizes (std::lock_guard<std::mutex> const& lock, QString const& posStr, QString const& valStr)
{
    const double scalingMax = Surface::scalingMax();

    if (scalingMax > 10. * m_scaling)
    {
        m_imagePixPosSizeMax = QSizeF(scalingMax, scalingMax);
        m_imagePixValSizeMax = QSizeF(scalingMax, scalingMax);
        m_overallPixSizeMax = QSizeF(scalingMax, scalingMax);
    }
    else
    {
        bool badSize;
        int nextPtSize = m_painterFontPointSizeDefault;

        do
        {
            m_painterFont.setPointSize(nextPtSize);
            QFontMetricsF metrics (m_painterFont);

            m_imagePixPosSizeMax = metrics.size(0, posStr) + QSizeF(4, 4);
            m_imagePixValSizeMax = metrics.size(0, valStr) + QSizeF(4, 4);
            m_overallPixSizeMax = QSizeF(std::max(m_imagePixPosSizeMax.width(), m_imagePixValSizeMax.width()), m_imagePixPosSizeMax.height() * 2 + m_imagePixValSizeMax.height());

            badSize = m_overallPixSizeMax.width() * s_paddingRatio >= scalingMax || m_overallPixSizeMax.height() * s_paddingRatio >= scalingMax;

            --nextPtSize;
        }
        while (badSize && nextPtSize > 0);
    }
}
