/* FILE:        ImageViewWindow.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::Window.
 */

#include <QtCore>
#include <QtGui>
#include <QWindow>

#if defined(_WIN32)
#   include <QtWin>
#endif

#include "ImageViewWindow_Private.h"
#include "ImageViewNative_Private.h"

#include "ImageViewAction.h"

#include "ImageViewStringHelpers.h"

#include "ImageViewSurface.h"
#include "ImageViewSurfaceStream.h"
#include "ImageViewSurfaceScrollArea.h"

#include <string>
#include <iostream>

ImageView::Window::PrivateData::PrivateData (Window &parent)
    : m_parent (parent)
    , m_appAboutDialog (nullptr)
    , m_nativeWindow (0)
    , m_parentWindow (0)
{ }

// Constructor. Window takes permanent ownership of the acquisition interface.
ImageView::Window::Window (ScopedAcquisitionInterface &&acqIf)
    : m_pd (*new PrivateData(*this))
{
    try
    {
        m_pd.m_windowQt = new WindowQt(std::move(acqIf));
    }
    catch (...)
    {
        delete &m_pd;
        throw;
    }

    QObject::connect(m_pd.m_windowQt, SIGNAL(destroyed()), &m_pd, SLOT(windowQtDestroyed()));
}

// Constructor. Window takes permanent ownership of the acquisition and file interfaces.
ImageView::Window::Window (ScopedAcquisitionInterface &&acqIf, ScopedFileInterface &&fileIf)
    : m_pd (*new PrivateData(*this))
{
    try
    {
        m_pd.m_windowQt = new WindowQt(std::move(acqIf), std::move(fileIf));
    }
    catch (...)
    {
        delete &m_pd;
        throw;
    }

    QObject::connect(m_pd.m_windowQt, SIGNAL(destroyed()), &m_pd, SLOT(windowQtDestroyed()));
}

// Destructor
ImageView::Window::~Window (void)
{
#if defined(_WIN32)
    if (m_pd.m_parentWindow && m_pd.m_nativeWindow)
        ImageView::Native::Win32::ClearParentChildWindow(m_pd.m_parentWindow, m_pd.m_nativeWindow);
#endif

    m_pd.m_nativeWindow = NULL;
    m_pd.m_parentWindow = NULL;

    if (m_pd.m_windowQt)
        delete m_pd.m_windowQt;
    delete &m_pd;
}

// Indicates whether or not the underlying window is still extant.
bool ImageView::Window::good (void) const
{
    return !!m_pd.m_windowQt;
}

// Retrieve the Qt widget handle for the underlying window.
QWidget* ImageView::Window::qt_handle (void)
{
    return m_pd.m_windowQt;
}
const QWidget* ImageView::Window::qt_handle (void) const
{
    return m_pd.m_windowQt;
}

// Retrieve the native window handle for this Window.
IMAGE_VIEW_NATIVE_WINDOW ImageView::Window::native_handle (void) const
{
    return reinterpret_cast<IMAGE_VIEW_NATIVE_WINDOW>(m_pd.m_windowQt->effectiveWinId());
}

// Manage the transient parent for this window.
IMAGE_VIEW_NATIVE_WINDOW ImageView::Window::transientParent (void) const
{
    return m_pd.m_parentWindow;
}

void ImageView::Window::setTransientParent (IMAGE_VIEW_NATIVE_WINDOW a_parentWindow)
{
#if defined(_WIN32)

    if (m_pd.m_parentWindow && m_pd.m_nativeWindow)
        ImageView::Native::Win32::ClearParentChildWindow(m_pd.m_parentWindow, m_pd.m_nativeWindow);
    m_pd.m_nativeWindow = NULL;
    m_pd.m_parentWindow = NULL;

    if (a_parentWindow)
    {
        m_pd.m_nativeWindow = native_handle();
        m_pd.m_parentWindow = a_parentWindow;
        ImageView::Native::Win32::SetParentChildWindow(m_pd.m_parentWindow, m_pd.m_nativeWindow);
    }

#else

    m_pd.m_parentWindow = 0;
    if (m_pd.m_windowQt->isTopLevel())
    {
        m_pd.m_windowQt->winId();
        auto *const lclWindow = m_pd.m_windowQt->windowHandle();
        if (lclWindow)
        {
            auto *const parentWindow = QWindow::fromWinId(reinterpret_cast<WId>(a_parentWindow));
            if (parentWindow)
            {
                lclWindow->setTransientParent(parentWindow);
                m_pd.m_parentWindow = a_parentWindow;
            }
        }
    }

#endif
}

void ImageView::Window::PrivateData::windowQtDestroyed (void)
{
    m_windowQt = nullptr;
}

// Retrieve the acquisition interface.
ImageView::AbstractAcquisitionInterface* ImageView::Window::acquisitionInterface (void)
{
    return m_pd.m_windowQt->acquisitionInterface();
}
const ImageView::AbstractAcquisitionInterface* ImageView::Window::acquisitionInterface (void) const
{
    return m_pd.m_windowQt->acquisitionInterface();
}

// Retrieve the file interface, if any.
ImageView::AbstractFileInterface* ImageView::Window::fileInterface (void)
{
    return m_pd.m_windowQt->fileInterface();
}
const ImageView::AbstractFileInterface* ImageView::Window::fileInterface (void) const
{
    return m_pd.m_windowQt->fileInterface();
}

// Retrieve the surface scroll area.
ImageView::SurfaceScrollArea* ImageView::Window::surfaceScrollArea (void)
{
    return m_pd.m_windowQt->surfaceScrollArea();
}
const ImageView::SurfaceScrollArea* ImageView::Window::surfaceScrollArea (void) const
{
    return m_pd.m_windowQt->surfaceScrollArea();
}

// Retrieve the surface stream.
ImageView::SurfaceStream* ImageView::Window::surfaceStream (void)
{
    return m_pd.m_windowQt->surfaceStream();
}
const ImageView::SurfaceStream* ImageView::Window::surfaceStream (void) const
{
    return m_pd.m_windowQt->surfaceStream();
}

// SurfaceStream properties.
bool ImageView::Window::refreshEnabled (void) const
{
    return surfaceStream()->refreshEnabled();
}
void ImageView::Window::setRefreshEnabled (const bool enabled)
{
    surfaceStream()->setRefreshEnabled(enabled);
}

bool ImageView::Window::refreshRateAuto (void) const
{
    return surfaceStream()->refreshRateAuto();
}
void ImageView::Window::setRefreshRateAuto (const bool enabled)
{
    surfaceStream()->setRefreshRateAuto(enabled);
}

int ImageView::Window::refreshRate (void) const
{
    return surfaceStream()->refreshRate();
}
void ImageView::Window::setRefreshRate (const int rate)
{
    surfaceStream()->setRefreshRate(rate);
}

// Window properties.
bool ImageView::Window::title (char *const pTitle, size_t *const pSize)
{
    const QString windowTitle = m_pd.m_windowQt->windowTitle();
    return String::set(windowTitle.toLocal8Bit().data(), pTitle, pSize);
}
void ImageView::Window::setTitle (const char *const title)
{
    m_pd.m_windowQt->setWindowTitle (title);
}

bool ImageView::Window::isVisible (void) const
{
    return m_pd.m_windowQt->isVisible();
}
void ImageView::Window::setVisible (const bool visible)
{
    m_pd.m_windowQt->setVisible(visible);
}

void ImageView::Window::show (void)
{
    m_pd.m_windowQt->show();
}
void ImageView::Window::hide (void)
{
    m_pd.m_windowQt->hide();
}
void ImageView::Window::raise (void)
{
    m_pd.m_windowQt->raise();
}

bool ImageView::Window::closeable (void) const
{
    return m_pd.m_windowQt->closeable();
}
void ImageView::Window::setCloseable (const bool enabled)
{
    m_pd.m_windowQt->setCloseable(enabled);
}

int ImageView::Window::x (void) const
{
    return m_pd.m_windowQt->x();
}
int ImageView::Window::y (void) const
{
    return m_pd.m_windowQt->y();
}
int ImageView::Window::width (void) const
{
    return m_pd.m_windowQt->width();
}
int ImageView::Window::height (void) const
{
    return m_pd.m_windowQt->height();
}

void ImageView::Window::move (const int x, const int y)
{
    m_pd.m_windowQt->move(x, y);
}
void ImageView::Window::resize (const int w, const int h)
{
    m_pd.m_windowQt->resize(w, h);
}

void ImageView::Window::getGeometry (int &x, int &y, int &width, int &height)
{
    const auto wg = m_pd.m_windowQt->geometry();
    x = wg.x();
    y = wg.y();
    width = wg.width();
    height = wg.height();
}
void ImageView::Window::setGeometry (const int x, const int y, const int width, const int height)
{
    m_pd.m_windowQt->setGeometry(x, y, width, height);
}

void ImageView::Window::getFrameGeometry (int &x, int &y, int &width, int &height)
{
    const auto fg = m_pd.m_windowQt->frameGeometry();
    x = fg.x();
    y = fg.y();
    width = fg.width();
    height = fg.height();
}
void ImageView::Window::setFrameGeometry (const int x, const int y, const int width, const int height)
{
    const auto fg = m_pd.m_windowQt->frameGeometry();
    const auto wg = m_pd.m_windowQt->geometry();

    m_pd.m_windowQt->move(x, y);
    m_pd.m_windowQt->resize(width - fg.width() + wg.width(), height - fg.height() + wg.height());
}

bool ImageView::Window::saveGeometry (char *const geometryBuf, size_t *const pSize) const
{
    const QByteArray byteArray = m_pd.m_windowQt->saveGeometry();
    const QByteArray hexArray = byteArray.toHex();
    return ImageView::String::set(hexArray.data(), geometryBuf, pSize);
}
bool ImageView::Window::restoreGeometry (const char *const geometryBuf)
{
    const QByteArray byteArray = QByteArray::fromHex(geometryBuf);
    return m_pd.m_windowQt->restoreGeometry(byteArray);
}

// Zoom level in percent.
double ImageView::Window::zoomLevel (void) const
{
    return surfaceStream()->surface()->scaling() * 100.;
}
void ImageView::Window::setZoomLevel (const double percent)
{
    surfaceStream()->surface()->setScaling(percent / 100.);
}

bool ImageView::Window::zoomToFit (void) const
{
    return surfaceStream()->surface()->scaleToFit();
}
void ImageView::Window::setZoomToFit (const bool enabled)
{
    surfaceStream()->surface()->setScaleToFit(enabled);
}

int ImageView::Window::imageOffsetX (void) const
{
    return surfaceStream()->surface()->scaledOffset().x();
}
int ImageView::Window::imageOffsetY (void) const
{
    return surfaceStream()->surface()->scaledOffset().y();
}
void ImageView::Window::setImageOffset (const int x, const int y)
{
    surfaceStream()->surface()->setScaledOffset(QPointF(x, y));
}

// Window display.
bool ImageView::Window::isFullScreen (void) const
{
    return m_pd.m_windowQt->isFullScreen();
}
void ImageView::Window::setFullScreen (const bool enabled)
{
    m_pd.m_windowQt->setFullScreen(enabled);
}

bool ImageView::Window::isMaximized (void) const
{
    return m_pd.m_windowQt->isMaximized();
}
bool ImageView::Window::isMinimized (void) const
{
    return m_pd.m_windowQt->isMinimized();
}

void ImageView::Window::maximize (void)
{
    m_pd.m_windowQt->setWindowState((m_pd.m_windowQt->windowState() & ~Qt::WindowMinimized) | Qt::WindowMaximized | Qt::WindowActive);
}
void ImageView::Window::minimize (void)
{
    m_pd.m_windowQt->setWindowState(m_pd.m_windowQt->windowState() | Qt::WindowMinimized);
}
void ImageView::Window::restore (void)
{
    m_pd.m_windowQt->setWindowState((m_pd.m_windowQt->windowState() & ~Qt::WindowMinimized & ~Qt::WindowMaximized) | Qt::WindowActive);
}

// Various properties
bool ImageView::Window::previewZoomLevel (void) const
{
    return m_pd.m_windowQt->previewZoomLevel();
}
void ImageView::Window::setPreviewZoomLevel (const bool enabled)
{
    m_pd.m_windowQt->setPreviewZoomLevel(enabled);
}

bool ImageView::Window::previewImageDimensions (void) const
{
    return m_pd.m_windowQt->previewImageDimensions();
}
void ImageView::Window::setPreviewImageDimensions (const bool enabled)
{
    m_pd.m_windowQt->setPreviewImageDimensions(enabled);
}

bool ImageView::Window::previewImageFormat (void) const
{
    return m_pd.m_windowQt->previewImageFormat();
}
void ImageView::Window::setPreviewImageFormat (const bool enabled)
{
    m_pd.m_windowQt->setPreviewImageFormat(enabled);
}

bool ImageView::Window::previewPixelValue (void) const
{
    return m_pd.m_windowQt->previewPixelValue();
}
void ImageView::Window::setPreviewPixelValue (const bool enabled)
{
    m_pd.m_windowQt->setPreviewPixelValue(enabled);
}

bool ImageView::Window::previewFrameRate (void) const
{
    return m_pd.m_windowQt->previewFrameRate();
}
void ImageView::Window::setPreviewFrameRate (const bool enabled)
{
    m_pd.m_windowQt->setPreviewFrameRate(enabled);
}

bool ImageView::Window::previewBuffersQueued (void) const
{
    return m_pd.m_windowQt->previewBuffersQueued();
}
void ImageView::Window::setPreviewBuffersQueued (const bool enabled)
{
    m_pd.m_windowQt->setPreviewBuffersQueued(enabled);
}

bool ImageView::Window::previewBuffersFilled (void) const
{
    return m_pd.m_windowQt->previewBuffersFilled();
}
void ImageView::Window::setPreviewBuffersFilled (const bool enabled)
{
    m_pd.m_windowQt->setPreviewBuffersFilled(enabled);
}

bool ImageView::Window::previewBuffersDropped (void) const
{
    return m_pd.m_windowQt->previewBuffersDropped();
}
void ImageView::Window::setPreviewBuffersDropped (const bool enabled)
{
    m_pd.m_windowQt->setPreviewBuffersDropped(enabled);
}

bool ImageView::Window::previewBuffersDelivered (void) const
{
    return m_pd.m_windowQt->previewBuffersDelivered();
}
void ImageView::Window::setPreviewBuffersDelivered (const bool enabled)
{
    m_pd.m_windowQt->setPreviewBuffersDelivered(enabled);
}

bool ImageView::Window::keepImageInfoVisible (void) const
{
    return m_pd.m_windowQt->keepImageInfoVisible();
}
void ImageView::Window::setKeepImageInfoVisible (const bool visible)
{
    m_pd.m_windowQt->setKeepImageInfoVisible(visible);
}

bool ImageView::Window::previewSetupMode (void) const
{
    return m_pd.m_windowQt->previewSetupMode();
}
void ImageView::Window::setPreviewSetupMode (const bool visible)
{
    m_pd.m_windowQt->setPreviewSetupMode(visible);
}

bool ImageView::Window::bayerDemosaicEnabled (void) const
{
    return m_pd.m_windowQt->bayerDemosaicEnabled();
}
void ImageView::Window::setBayerDemosaicEnabled (const bool enabled)
{
    m_pd.m_windowQt->setBayerDemosaicEnabled(enabled);
}

bool ImageView::Window::rgbOrderReversed (void) const
{
    return m_pd.m_windowQt->rgbOrderReversed();
}
void ImageView::Window::setRgbOrderReversed (const bool enabled)
{
    m_pd.m_windowQt->setRgbOrderReversed(enabled);
}

bool ImageView::Window::pixelCrosshairVisible (void) const
{
    return m_pd.m_windowQt->pixelCrosshairVisible();
}
void ImageView::Window::setPixelCrosshairVisible (const bool enabled)
{
    m_pd.m_windowQt->setPixelCrosshairVisible(enabled);
}

// Acquisition control. Asynchronous, if no other operation is ongoing. Use the
// AbstractAcquisitionInterface directly for more nuanced control.
void ImageView::Window::setRunning (void)
{
    m_pd.m_windowQt->setRunning();
}
void ImageView::Window::setConfigured (void)
{
    m_pd.m_windowQt->setConfigured();
}
void ImageView::Window::setUnconfigured (void)
{
    m_pd.m_windowQt->setUnconfigured();
}

// The Bayer mosaic override patther.
ImageView::Format ImageView::Window::bayerFilterOverride (void) const
{
    return surfaceStream()->bayerDemosaicFilter();
}
void ImageView::Window::setBayerFilterOverride (const Format bayerFormat)
{
    switch (bayerFormat)
    {
    case Unencoded:
        m_pd.m_windowQt->setBayerDemosaicFilterAuto();
        break;
    case BayerRGGB:
        m_pd.m_windowQt->setBayerDemosaicFilterRGGB();
        break;
    case BayerGRBG:
        m_pd.m_windowQt->setBayerDemosaicFilterGRBG();
        break;
    case BayerBGGR:
        m_pd.m_windowQt->setBayerDemosaicFilterBGGR();
        break;
    case BayerGBRG:
        m_pd.m_windowQt->setBayerDemosaicFilterGBRG();
        break;
    default:
        std::cerr << __FUNCTION__ << ": Unknown Bayer filter with index " << int(bayerFormat) << ".\n";
    }
}

// Color table for grayscale image rendering.
ImageView::ColorTable ImageView::Window::colorTable (void) const
{
    return m_pd.m_windowQt->colorTable();
}
void ImageView::Window::setColorTable (const ColorTable ct)
{
    m_pd.m_windowQt->setColorTable(ct);
}

// Give the Window an application dialog with the title, description (may
// contain simple HTML), and icon.
bool ImageView::Window::applicationAboutTitle (char *const title, size_t *const pSize) const
{
    if (m_pd.m_appAboutDialog)
        return ImageView::String::set(m_pd.m_appAboutDialog->windowTitle().toLatin1().constData(), title, pSize);
    else
        return ImageView::String::set("", title, pSize);
}
bool ImageView::Window::applicationAboutDescription(char *const desc, size_t *const pSize) const
{
    if (m_pd.m_appAboutDialog)
        return ImageView::String::set(m_pd.m_appAboutDialogUi.imageViewDetailsLabel->text().toLatin1().constData(), desc, pSize);
    else
        return ImageView::String::set("", desc, pSize);
}

#if defined(_WIN32)
void ImageView::Window::setApplicationAbout (const char *const title, const char *const desc, const IMAGE_VIEW_NATIVE_ICON iconBig, const IMAGE_VIEW_NATIVE_ICON iconMed, const IMAGE_VIEW_NATIVE_ICON iconSmall)
{
    const QString titleStr (title);
    const QString descStr (desc);

    if (!m_pd.m_appAboutDialog)
    {
        m_pd.m_appAboutDialog = new QDialog (m_pd.m_windowQt);
        m_pd.m_appAboutDialogUi.setupUi(m_pd.m_appAboutDialog);
    }

    m_pd.m_appAboutDialog->setWindowTitle(titleStr);

    m_pd.m_appAboutDialogUi.imageViewMainLabel->setText("<big><b>" + titleStr.toHtmlEscaped() + "</b></big>");

    QIcon newWindowIcon;

    if (iconSmall)
    {
        const QPixmap winPix = QtWin::fromHICON(iconSmall);
        m_pd.m_appAboutDialog->setWindowIcon((QIcon)winPix);
        newWindowIcon.addPixmap(winPix);
    }

    if (iconMed)
    {
        const QPixmap winPix = QtWin::fromHICON(iconMed);
        newWindowIcon.addPixmap(winPix);
    }

    if (iconBig)
    {
        const QPixmap winPix = QtWin::fromHICON(iconBig);
        m_pd.m_appAboutDialogUi.imageViewMainLabel->setPixmap(winPix.scaled(QSize(128, 128), Qt::KeepAspectRatio, Qt::SmoothTransformation));
        newWindowIcon.addPixmap(winPix);
    }

    if (!newWindowIcon.isNull() && QApplication::windowIcon().isNull())
        QApplication::setWindowIcon(newWindowIcon);

    m_pd.m_appAboutDialogUi.imageViewDetailsLabel->setText(descStr);

    if (titleStr.length() || descStr.length() || iconBig)
    {
        m_pd.m_appAboutDialog->setMinimumSize(QSize(0, 0));
        m_pd.m_appAboutDialog->setMaximumSize(QSize(QWIDGETSIZE_MAX, QWIDGETSIZE_MAX));
        m_pd.m_appAboutDialog->adjustSize();
        m_pd.m_appAboutDialog->setFixedSize(m_pd.m_appAboutDialog->size());

        m_pd.m_windowQt->setApplicationAbout(m_pd.m_appAboutDialog);
    }
    else
        m_pd.m_windowQt->setApplicationAbout(nullptr);
}
#endif

void ImageView::Window::setApplicationAbout (const char *const title, const char *const desc, const char *const xpmIconBig[], const char *const xpmIconMed[], const char *const xpmIconSmall[])
{
    const QString titleStr (title);
    const QString descStr (desc);

    if (!m_pd.m_appAboutDialog)
    {
        m_pd.m_appAboutDialog = new QDialog (m_pd.m_windowQt);
        m_pd.m_appAboutDialogUi.setupUi(m_pd.m_appAboutDialog);
    }

    m_pd.m_appAboutDialog->setWindowTitle(titleStr);

    m_pd.m_appAboutDialogUi.imageViewMainLabel->setText("<big><b>" + titleStr.toHtmlEscaped() + "</b></big>");

    QIcon newWindowIcon;

    if (xpmIconSmall)
    {
        const QPixmap pix (xpmIconSmall);
        m_pd.m_appAboutDialog->setWindowIcon((QIcon)pix);
        newWindowIcon.addPixmap(pix);
    }

    if (xpmIconMed)
    {
        const QPixmap pix (xpmIconMed);
        newWindowIcon.addPixmap(pix);
    }

    if (xpmIconBig)
    {
        const QPixmap pix (xpmIconBig);
        m_pd.m_appAboutDialogUi.imageViewMainLabel->setPixmap(pix.scaled(QSize(128, 128), Qt::KeepAspectRatio, Qt::SmoothTransformation));
        newWindowIcon.addPixmap(pix);
    }

    if (!newWindowIcon.isNull() && QApplication::windowIcon().isNull())
        QApplication::setWindowIcon(newWindowIcon);

    m_pd.m_appAboutDialogUi.imageViewDetailsLabel->setText(descStr);

    if (titleStr.length() || descStr.length() || xpmIconBig)
    {
        m_pd.m_appAboutDialog->setMinimumSize(QSize(0, 0));
        m_pd.m_appAboutDialog->setMaximumSize(QSize(QWIDGETSIZE_MAX, QWIDGETSIZE_MAX));
        m_pd.m_appAboutDialog->adjustSize();
        m_pd.m_appAboutDialog->setFixedSize(m_pd.m_appAboutDialog->size());

        m_pd.m_windowQt->setApplicationAbout(m_pd.m_appAboutDialog);
    }
    else
        m_pd.m_windowQt->setApplicationAbout(nullptr);
}

// Methods to add custom actions to the Window GUI.
ImageView::Action* ImageView::Window::addMainMenuAction (ImageView::ScopedActionInterface &&a_actionIf, const char *const parentMenuName, const char *const text)
{
    auto *const qtAction = m_pd.m_windowQt->addMainMenuAction(parentMenuName, text ? text : "");
    return new ImageView::Action(this, qtAction, std::move(a_actionIf));
}
ImageView::Action* ImageView::Window::addToolBarAction (ImageView::ScopedActionInterface &&a_actionIf, const char *const text)
{
    auto *const qtAction = m_pd.m_windowQt->addToolBarAction(text ? text : "");
    return new ImageView::Action(this, qtAction, std::move(a_actionIf));
}
ImageView::Action* ImageView::Window::addImageStatsAction (ImageView::ScopedActionInterface &&a_actionIf, const char *const text)
{
    auto *const qtAction = m_pd.m_windowQt->addImageStatsAction(text ? text : "");
    return new ImageView::Action(this, qtAction, std::move(a_actionIf));
}

// Methods to enforce a constant bitmap. Format must be 8-, 24- or 32-bit.
bool ImageView::Window::bitmapLocked (void) const
{
    return surfaceStream()->bitmapLocked();
}
bool ImageView::Window::setBitmapLocked (const int width, const int height, const Format format)
{
    return surfaceStream()->setBitmapLocked(width, height, format);
}
void ImageView::Window::unlockBitmap (void)
{
    surfaceStream()->unlockBitmap();
}

// Retrieve the locked bitmap pointer, if any. nullptr, if not bitmapLocked.
bool ImageView::Window::getBitmap (RawImage *const bitmap) const
{
    return surfaceStream()->getBitmap(bitmap);
}
