#ifndef INCLUDED__BIT_FLOW__IMAGE__VIEW__WINDOW__H
#define INCLUDED__BIT_FLOW__IMAGE__VIEW__WINDOW__H

/* FILE:        ImageViewWindow.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Public display Window class for the ImageView library. May only be
 *              used after construction of the ImageView::Application global object.
 *              Use WindowQt, for greater control.
 */

#include "ImageView.h"
#include "ImageViewBaseClass.h"

#include "ImageViewNative.h"

namespace ImageView
{
    class IMAGEVIEW_EXPORT Window : public BaseClass
    {
    private:
        // Private data.
        struct PrivateData;
        PrivateData &m_pd;

        // Illegal.
        Window (void);
        Window (Window const&);
        Window& operator= (Window const&);

    public:
        // Constructor. Window takes permanent ownership of the acquisition interface.
        Window (ScopedAcquisitionInterface &&acqIf);

        // Constructor. Window takes permanent ownership of the acquisition and file interfaces.
        Window (ScopedAcquisitionInterface &&acqIf, ScopedFileInterface &&fileIf);

        // Destructor
        ~Window (void);

        // Indicates whether or not the underlying window is still extant. The Window cannot be
        // used any longer, if not good, and should be disposed of at first convenience.
        bool good (void) const;

        // Retrieve the Qt widget handle for the underlying window.
        QWidget* qt_handle (void);
        const QWidget* qt_handle (void) const;

        // Retrieve the native handle for this Window.
        IMAGE_VIEW_NATIVE_WINDOW native_handle (void) const;

        // Manage the transient parent for this window.
        IMAGE_VIEW_NATIVE_WINDOW transientParent (void) const;
        void setTransientParent (IMAGE_VIEW_NATIVE_WINDOW a_parentWindow);

        // Retrieve the acquisition interface.
        AbstractAcquisitionInterface* acquisitionInterface (void);
        const AbstractAcquisitionInterface* acquisitionInterface (void) const;

        // Retrieve the file interface, if any.
        AbstractFileInterface* fileInterface (void);
        const AbstractFileInterface* fileInterface (void) const;

        // Retrieve the surface scroll area.
        SurfaceScrollArea* surfaceScrollArea (void);
        const SurfaceScrollArea* surfaceScrollArea (void) const;

        // Retrieve the surface stream.
        SurfaceStream* surfaceStream (void);
        const SurfaceStream* surfaceStream (void) const;

        // SurfaceStream properties.
        bool refreshEnabled (void) const;
        void setRefreshEnabled (const bool enabled);

        bool refreshRateAuto (void) const;
        void setRefreshRateAuto (const bool enabled);

        int refreshRate (void) const;
        void setRefreshRate (const int rate);

        // Window properties.
        bool title (char *const pTitle, size_t *const pSize);
        void setTitle (const char *const title);

        bool isVisible (void) const;
        void setVisible (const bool visible);

        void show (void);
        void hide (void);
        void raise (void);

        bool closeable (void) const;
        void setCloseable (const bool enabled);

        int x (void) const;
        int y (void) const;
        int width (void) const;
        int height (void) const;

        void move (const int x, const int y);
        void resize (const int w, const int h);

        void getGeometry (int &x, int &y, int &width, int &height);
        void setGeometry (const int x, const int y, const int width, const int height);

        void getFrameGeometry (int &x, int &y, int &width, int &height);
        void setFrameGeometry (const int x, const int y, const int width, const int height);

        bool saveGeometry (char *const geometryBuf, size_t *const pSize) const;
        bool restoreGeometry (const char *const geometryBuf);

        // Zoom level and offset controls.
        double zoomLevel (void) const;
        void setZoomLevel (const double percent);

        bool zoomToFit (void) const;
        void setZoomToFit (const bool enabled = true);

        int imageOffsetX (void) const;
        int imageOffsetY (void) const;
        void setImageOffset (const int x, const int y);

        // Window display.
        bool isFullScreen (void) const;
        void setFullScreen (const bool enabled);

        bool isMaximized (void) const;
        bool isMinimized (void) const;

        void maximize (void);
        void minimize (void);
        void restore (void);

        // Various Window properties
        bool previewZoomLevel (void) const;
        void setPreviewZoomLevel (const bool enabled);

        bool previewImageDimensions (void) const;
        void setPreviewImageDimensions (const bool enabled);

        bool previewImageFormat (void) const;
        void setPreviewImageFormat (const bool enabled);

        bool previewPixelValue (void) const;
        void setPreviewPixelValue (const bool enabled);

        bool previewFrameRate (void) const;
        void setPreviewFrameRate (const bool enabled);

        bool previewBuffersQueued (void) const;
        void setPreviewBuffersQueued (const bool enabled);

        bool previewBuffersFilled (void) const;
        void setPreviewBuffersFilled (const bool enabled);

        bool previewBuffersDropped (void) const;
        void setPreviewBuffersDropped (const bool enabled);

        bool previewBuffersDelivered (void) const;
        void setPreviewBuffersDelivered (const bool enabled);

        bool keepImageInfoVisible (void) const;
        void setKeepImageInfoVisible (const bool visible);

        bool previewSetupMode (void) const;
        void setPreviewSetupMode (const bool visible);

        bool bayerDemosaicEnabled (void) const;
        void setBayerDemosaicEnabled (const bool enabled);

        bool rgbOrderReversed (void) const;
        void setRgbOrderReversed (const bool enabled);

        bool pixelCrosshairVisible (void) const;
        void setPixelCrosshairVisible (const bool enabled);

        // Acquisition control. Asynchronous, if no other operation is ongoing. Use the
        // AbstractAcquisitionInterface directly for more nuanced control.
        void setRunning (void);
        void setConfigured (void);
        void setUnconfigured (void);

        // The Bayer mosaic override patther.
        Format bayerFilterOverride (void) const;
        void setBayerFilterOverride (const Format bayerFormat);

        // Color table for grayscale image rendering.
        ColorTable colorTable (void) const;
        void setColorTable (const ColorTable ct);

        // Give the Window an application dialog with the title, description (may
        // contain simple HTML), and icon.
        bool applicationAboutTitle (char *const title, size_t *const pSize) const;
        bool applicationAboutDescription(char *const desc, size_t *const pSize) const;

#if defined(_WIN32)
        void setApplicationAbout (const char *const title, const char *const desc, const IMAGE_VIEW_NATIVE_ICON iconBig = 0, const IMAGE_VIEW_NATIVE_ICON iconMed = 0, const IMAGE_VIEW_NATIVE_ICON iconSmall = 0);
#endif

        void setApplicationAbout (const char *const title, const char *const desc, const char *const xpmIconBig[] = 0, const char *const xpmIconMed[] = 0, const char *const xpmIconSmall[] = 0);

        // Methods to add custom actions to the Window GUI.
        ImageView::Action* addMainMenuAction (ImageView::ScopedActionInterface &&a_actionIf, const char *const parentMenuName, const char *const text = nullptr);
        ImageView::Action* addToolBarAction (ImageView::ScopedActionInterface &&a_actionIf, const char *const text = nullptr);
        ImageView::Action* addImageStatsAction (ImageView::ScopedActionInterface &&a_actionIf, const char *const text = nullptr);

        // Methods to enforce a constant bitmap. Format must be 8-, 24- or 32-bit.
        bool bitmapLocked (void) const;
        bool setBitmapLocked (const int width, const int height, const Format format);
        void unlockBitmap (void);

        // Retrieve the locked bitmap pointer, if any. nullptr, if not bitmapLocked.
        bool getBitmap (RawImage *const bitmap) const;
    };
}

#endif // INCLUDED__BIT_FLOW__IMAGE__VIEW__WINDOW__H
