/* FILE:        ImageViewWindowQt.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::WindowQt.
 */

#include "ImageViewWindowQt_Private.h"

#include "ImageViewStringHelpers.h"

#include <iostream>

#include <vector>

#define IMAGE_VIEW_PI (3.1415926535897932384626433832795)

// Constructor. ImageView takes permanent ownership of the acquisition interface.
ImageView::WindowQt::WindowQt (ScopedAcquisitionInterface &&acqIf, QWidget *const parent)
    : QMainWindow (parent)
    , m_pd (*new PrivateData(*this, std::move(acqIf)))
{
    try
    {
        m_pd.initialize();
    }
    catch (...)
    {
        delete &m_pd;
        throw;
    }
}

// Constructor. ImageView takes permanent ownership of the acquisition and file interfaces.
ImageView::WindowQt::WindowQt (ScopedAcquisitionInterface &&acqIf, ScopedFileInterface &&fileIf, QWidget *const parent)
    : QMainWindow (parent)
    , m_pd (*new PrivateData(*this, std::move(acqIf), std::move(fileIf)))
{
    try
    {
        m_pd.initialize();
    }
    catch (...)
    {
        delete &m_pd;
        throw;
    }
}

ImageView::WindowQt::~WindowQt (void)
{
    delete &m_pd;
}

void ImageView::WindowQt::closeEvent (QCloseEvent *evt)
{
    if (!m_pd.m_ui.actionClose->isEnabled())
    {
        evt->ignore();
        return;
    }

    // Execute the acquisition interface close handler.
    m_pd.m_acqIf->uxClosed();

    // Stop any ongoing acquisition.
    m_pd.flushBusyWork();

    // Accept the event.
    evt->accept();

    // Delete this WindowQt.
    deleteLater();
}

void ImageView::WindowQt::changeEvent (QEvent *evt)
{
    switch (evt->type())
    {
    case QEvent::WindowStateChange:
        if (!!(windowState() & Qt::WindowFullScreen) != m_pd.m_ui.actionFull_Screen->isChecked())
            setFullScreen(windowState() & Qt::WindowFullScreen);
        break;
    }
}

ImageView::AbstractAcquisitionInterface* ImageView::WindowQt::acquisitionInterface (void)
{
    return m_pd.m_acqIf.get();
}
const ImageView::AbstractAcquisitionInterface* ImageView::WindowQt::acquisitionInterface (void) const
{
    return m_pd.m_acqIf.get();
}

ImageView::AbstractFileInterface* ImageView::WindowQt::fileInterface (void)
{
    return m_pd.m_fileIf.get();
}
const ImageView::AbstractFileInterface* ImageView::WindowQt::fileInterface (void) const
{
    return m_pd.m_fileIf.get();
}

ImageView::SurfaceScrollArea* ImageView::WindowQt::surfaceScrollArea (void)
{
    return m_pd.m_imageArea;
}
const ImageView::SurfaceScrollArea* ImageView::WindowQt::surfaceScrollArea (void) const
{
    return m_pd.m_imageArea;
}

ImageView::SurfaceStream* ImageView::WindowQt::surfaceStream (void)
{
    return m_pd.m_surfaceStream;
}
const ImageView::SurfaceStream* ImageView::WindowQt::surfaceStream (void) const
{
    return m_pd.m_surfaceStream;
}

// Various properties.
bool ImageView::WindowQt::closeable (void) const
{
    return m_pd.m_ui.actionClose->isEnabled();
}
void ImageView::WindowQt::setCloseable (const bool enabled)
{
    m_pd.m_ui.actionClose->setEnabled(enabled);

    const bool wasVisible = isVisible();

    const auto state = windowState();

    const auto flags = windowFlags();
    const auto windowType = flags & Qt::WindowType_Mask;
    if (enabled)
        setWindowFlags(windowType);
    else
        setWindowFlags(windowType | Qt::CustomizeWindowHint | Qt::WindowTitleHint | Qt::WindowMinMaxButtonsHint);

    setWindowState(Qt::WindowNoState);
    setWindowState(state);

    if (wasVisible)
        show();
    else
        hide();
}

bool ImageView::WindowQt::previewZoomLevel (void) const
{
    return m_pd.m_ui.actionZoom_Level->isChecked();
}
void ImageView::WindowQt::setPreviewZoomLevel (const bool visible)
{
    m_pd.m_ui.actionZoom_Level->setChecked(visible);
}

bool ImageView::WindowQt::previewImageDimensions (void) const
{
    return m_pd.m_ui.actionImageDimensions->isChecked();
}
void ImageView::WindowQt::setPreviewImageDimensions (const bool visible)
{
    return m_pd.m_ui.actionImageDimensions->setChecked(visible);
}

bool ImageView::WindowQt::previewImageFormat (void) const
{
    return m_pd.m_ui.actionImage_Format->isChecked();
}
void ImageView::WindowQt::setPreviewImageFormat (const bool visible)
{
    return m_pd.m_ui.actionImage_Format->setChecked(visible);
}

bool ImageView::WindowQt::previewPixelValue (void) const
{
    return m_pd.m_ui.actionPixel_Value->isChecked();
}
void ImageView::WindowQt::setPreviewPixelValue (const bool visible)
{
    return m_pd.m_ui.actionPixel_Value->setChecked(visible);
}

bool ImageView::WindowQt::previewFrameRate (void) const
{
    return m_pd.m_ui.actionFrame_Rate->isChecked();
}
void ImageView::WindowQt::setPreviewFrameRate (const bool visible)
{
    return m_pd.m_ui.actionFrame_Rate->setChecked(visible);
}

bool ImageView::WindowQt::previewBuffersQueued (void) const
{
    return m_pd.m_ui.actionBuffers_Queued->isChecked();
}
void ImageView::WindowQt::setPreviewBuffersQueued (const bool visible)
{
    return m_pd.m_ui.actionBuffers_Queued->setChecked(visible);
}

bool ImageView::WindowQt::previewBuffersFilled (void) const
{
    return m_pd.m_ui.actionBuffers_Filled->isChecked();
}
void ImageView::WindowQt::setPreviewBuffersFilled (const bool visible)
{
    return m_pd.m_ui.actionBuffers_Filled->setChecked(visible);
}

bool ImageView::WindowQt::previewBuffersDropped (void) const
{
    return m_pd.m_ui.actionBuffers_Dropped->isChecked();
}
void ImageView::WindowQt::setPreviewBuffersDropped (const bool visible)
{
    return m_pd.m_ui.actionBuffers_Dropped->setChecked(visible);
}

bool ImageView::WindowQt::previewBuffersDelivered (void) const
{
    return m_pd.m_ui.actionBuffers_Delivered->isChecked();
}
void ImageView::WindowQt::setPreviewBuffersDelivered (const bool visible)
{
    return m_pd.m_ui.actionBuffers_Delivered->setChecked(visible);
}

bool ImageView::WindowQt::previewSetupMode (void) const
{
    return m_pd.m_ui.actionActive_Setup_Mode->isChecked();
}
void ImageView::WindowQt::setPreviewSetupMode (const bool visible)
{
    m_pd.m_ui.actionActive_Setup_Mode->setChecked(visible);
}

bool ImageView::WindowQt::keepImageInfoVisible (void) const
{
    return m_pd.m_ui.actionKeep_Stats_Visible->isChecked();
}
void ImageView::WindowQt::setKeepImageInfoVisible (const bool visible)
{
    const bool change = m_pd.m_ui.actionKeep_Stats_Visible->isChecked() != visible;

    m_pd.m_ui.actionKeep_Stats_Visible->setChecked(visible);

    if (change)
    {
        m_pd.showWindowBars();
        m_pd.m_fullScreenBarHiderTimer->start();
    }
}

bool ImageView::WindowQt::bayerDemosaicEnabled (void) const
{
    return m_pd.m_ui.actionDecode_Bayer_Images->isChecked();
}
void ImageView::WindowQt::setBayerDemosaicEnabled (const bool enabled)
{
    m_pd.m_surfaceStream->setBayerDemosaicEnabled(enabled);
    m_pd.m_ui.actionDecode_Bayer_Images->setChecked(enabled);
}

bool ImageView::WindowQt::rgbOrderReversed (void) const
{
    return m_pd.m_ui.actionReverse_RGB_Order->isChecked();
}
void ImageView::WindowQt::setRgbOrderReversed (const bool reversed)
{
    m_pd.m_surfaceStream->setRgbOrderReversed(reversed);
    m_pd.m_ui.actionReverse_RGB_Order->setChecked(reversed);
}

bool ImageView::WindowQt::pixelCrosshairVisible (void) const
{
    return m_pd.m_ui.actionPixel_Crosshair_Visible->isChecked();
}
void ImageView::WindowQt::setPixelCrosshairVisible (const bool visible)
{
    m_pd.m_imageArea->surface()->setCrosshairVisible(visible);
    m_pd.m_ui.actionPixel_Crosshair_Visible->setChecked(visible);
}

ImageView::ColorTable ImageView::WindowQt::colorTable (void) const
{
    return m_pd.m_colorTableCurrent;
}
void ImageView::WindowQt::setColorTable (const ColorTable ct)
{
    switch (ct)
    {
    case ColorTable::Gray:
        setColorTableGray();
        break;
    case ColorTable::Gray2x:
        setColorTableGray2x();
        break;
    case ColorTable::GrayInverse:
        setColorTableGrayInverse();
        break;
    case ColorTable::Sine:
        setColorTableSine();
        break;
    case ColorTable::Cosine:
        setColorTableCosine();
        break;
    case ColorTable::Rainbow:
        setColorTableRainbow();
        break;
    case ColorTable::Random:
        setColorTableRandom();
        break;
    case ColorTable::Contours:
        setColorTableContours();
        break;
    default:
        std::cerr << __FUNCTION__ << ": Unknown ColorTable with index " << int(ct) << ".\n";
    }
}

// Set an external QDialog to use as an application about dialog. Window ownership is not changed.
QDialog* ImageView::WindowQt::applicationAbout (void) const
{
    return m_pd.m_appAboutDialog;
}
void ImageView::WindowQt::setApplicationAbout (QDialog *const abtDlg)
{
    if (m_pd.m_appAboutDialog)
    {
        disconnect(m_pd.m_ui.actionAbout_Application, 0, m_pd.m_appAboutDialog, 0);

        disconnect(m_pd.m_appAboutDialog, &QWidget::windowTitleChanged, nullptr, nullptr);
        disconnect(m_pd.m_appAboutDialog, &QWidget::windowIconChanged, nullptr, nullptr);
    }

    m_pd.m_appAboutDialog = abtDlg;

    if (m_pd.m_appAboutDialog)
    {
        m_pd.m_ui.actionAbout_Application->setText(m_pd.m_appAboutDialog->windowTitle() + "...");
        m_pd.m_ui.actionAbout_Application->setIcon(m_pd.m_appAboutDialog->windowIcon());

        connect(m_pd.m_appAboutDialog, &QWidget::windowTitleChanged, [this](QString const& t) { m_pd.m_ui.actionAbout_Application->setText(t + "..."); });
        connect(m_pd.m_appAboutDialog, &QWidget::windowIconChanged, [this](QIcon const& i) { m_pd.m_ui.actionAbout_Application->setIcon(i); });

        connect(m_pd.m_ui.actionAbout_Application, SIGNAL(triggered()), m_pd.m_appAboutDialog, SLOT(exec()));
    }

    m_pd.m_ui.actionAbout_Application->setVisible(!!m_pd.m_appAboutDialog);
}

// Methods to a custom actions to the Window GUI.
QAction* ImageView::WindowQt::addMainMenuAction (QString const& parentMenuName, QString const& text)
{
    const QString parentMenuObjectName = "ImageViewUserMenu_" + parentMenuName;

    QMenu *parentMenu = m_pd.m_ui.menuBar->findChild<QMenu*>(parentMenuObjectName);
    if (!parentMenu)
    {
        parentMenu = new QMenu (parentMenuName, m_pd.m_ui.menuBar);
        parentMenu->setObjectName(parentMenuObjectName);
        parentMenu->setSeparatorsCollapsible(true);
        parentMenu->setToolTipsVisible(true);
        m_pd.m_ui.menuBar->insertMenu(m_pd.m_ui.menuHelp->menuAction(), parentMenu);
    }

    auto *const newAction = parentMenu->addAction(text);

    connect(newAction, SIGNAL(destroyed()), &m_pd, SLOT(userMenuActionDestroyed()));

    return newAction;
}
QAction* ImageView::WindowQt::addToolBarAction (QString const& text)
{
    if (m_pd.m_userToolBarActions.isEmpty())
        m_pd.m_ui.acquisitionToolBar->addSeparator();

    auto *const newAction = m_pd.m_ui.acquisitionToolBar->addAction(text);
    m_pd.m_userToolBarActions.push_back(newAction);

    connect(newAction, SIGNAL(destroyed()), &m_pd, SLOT(userToolBarActionDestroyed()));

    return newAction;
}
QAction* ImageView::WindowQt::addImageStatsAction (QString const& text)
{
    auto *const newAction = new QAction (m_pd.m_imageStatsMenu);
    newAction->setText(text);

    m_pd.m_imageStatsMenu->insertAction(m_pd.m_imageStatsMenuUserActionSeparator, newAction);

    return newAction;
}

// Acquisition control. Asynchronous, if no other operation is ongoing.
void ImageView::WindowQt::setRunning (void)
{
    if (!m_pd.m_acqIf->isRunning())
        m_pd.setupStartAsync();
}
void ImageView::WindowQt::setConfigured (void)
{
    if (m_pd.m_acqIf->isRunning())
        m_pd.stopAsync();
    else
        m_pd.setupAsync();
}
void ImageView::WindowQt::setUnconfigured (void)
{
    if (m_pd.m_acqIf->isSetup())
        m_pd.cleanupAsync();
}

// GUI update slots.
void ImageView::WindowQt::updateFeatureSet (void)
{
    // Marshal supported features.
    const bool hasSetup = m_pd.m_acqIf->hasSetup();
    const bool hasSetupMode = m_pd.m_acqIf->hasSetupMode();
    const bool hasCleanup = m_pd.m_acqIf->hasCleanup();
    const bool hasSetupCleanup = hasSetup || hasCleanup;

    const bool hasTrigger = m_pd.m_acqIf->hasTrigger();

    const bool hasStartSingleReverse = m_pd.m_acqIf->hasStartSingleReverse();
    const bool hasStartReverse = m_pd.m_acqIf->hasStartReverse();
    const bool hasStart = m_pd.m_acqIf->hasStart();
    const bool hasStartFixed = m_pd.m_acqIf->hasStartFixed();
    const bool hasStartSingle = m_pd.m_acqIf->hasStartSingle();
    const bool hasPause = m_pd.m_acqIf->hasPause();
    const bool hasStop = m_pd.m_acqIf->hasStop();
    const bool hasPlayback = hasStartSingleReverse || hasStartReverse || hasStart || hasStartFixed || hasStartSingle || hasPause || hasStop;

    const bool hasFrameRate = m_pd.m_acqIf->hasFrameRate();

    const bool hasFramesQueued = m_pd.m_acqIf->hasFramesQueued();
    const bool hasFramesFilled = m_pd.m_acqIf->hasFramesFilled();
    const bool hasFramesDropped = m_pd.m_acqIf->hasFramesDropped();
    const bool hasFramesDelivered = m_pd.m_acqIf->hasFramesDelivered();

    QStringList setupModeNames, setupModeDescs;
    if (hasSetupMode)
    {
        for (size_t i = 0, cnt = m_pd.m_acqIf->setupModeCount(); i < cnt; i++)
        {
            const auto name = ImageView::String::read(&AbstractAcquisitionInterface::setupModeName, m_pd.m_acqIf.get(), i);
            const auto desc = ImageView::String::read(&AbstractAcquisitionInterface::setupModeDescription, m_pd.m_acqIf.get(), i);
            if (name.size())
            {
                setupModeNames.push_back(name.c_str());
                setupModeDescs.push_back(desc.c_str());
            }
        }
    }

    //// Setup feature visibility, refilling the acquisition toolbar from scratch.
    m_pd.m_ui.acquisitionToolBar->clear();

    // Setup/cleanup.
    m_pd.m_ui.actionSetup->setVisible(hasSetup);
    m_pd.m_ui.actionSetup->blockSignals(!hasSetup);

    m_pd.m_ui.actionCleanup->setVisible(hasCleanup);
    m_pd.m_ui.actionCleanup->blockSignals(!hasCleanup);

    m_pd.m_setupCleanupButton->setVisible(hasSetupCleanup);
    m_pd.m_ui.actionSetup_Cleanup->setVisible(hasSetupCleanup);
    m_pd.m_ui.actionSetup_Cleanup->blockSignals(!hasSetupCleanup);

    if (hasSetupCleanup)
    {
        m_pd.m_ui.acquisitionToolBar->addWidget(m_pd.m_setupCleanupButton);
        m_pd.m_ui.acquisitionToolBar->addSeparator();
    }

    // Setup mode support.
    m_pd.m_setupModeMenu->clear();
    m_pd.m_setupModeMenu->setEnabled(hasSetupMode);
    m_pd.m_ui.actionSetup_Mode->setVisible(hasSetupMode);

    m_pd.m_setupCleanupButton->setMenu(hasSetupMode ? m_pd.m_setupModeMenu : nullptr);
    m_pd.m_setupCleanupButton->setPopupMode(hasSetupMode ? QToolButton::MenuButtonPopup : QToolButton::DelayedPopup);

    for (int i = 0, cnt = setupModeNames.size(); i < cnt; i++)
    {
        auto *const modeAction = m_pd.m_setupModeMenu->addAction(setupModeNames[i], &m_pd, SLOT(setupModeAsync()));
        modeAction->setWhatsThis(setupModeNames[i]);
        modeAction->setToolTip(setupModeDescs[i]);
        modeAction->setCheckable(true);
    }

    // Playback.
    m_pd.m_ui.actionStart_Single_Reverse->setVisible(hasStartSingleReverse);
    m_pd.m_ui.actionStart_Single_Reverse->blockSignals(!hasStartSingleReverse);
    if (hasStartSingleReverse)
        m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionStart_Single_Reverse);

    m_pd.m_ui.actionStart_Reverse->setVisible(hasStartReverse);
    m_pd.m_ui.actionStart_Reverse->blockSignals(!hasStartReverse);
    if (hasStartReverse)
        m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionStart_Reverse);

    m_pd.m_ui.actionStart->setVisible(hasStart);
    m_pd.m_ui.actionStart->blockSignals(!hasStart);
    if (hasStart)
        m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionStart);

    m_pd.m_ui.actionStart_Fixed->setVisible(hasStartFixed);
    m_pd.m_ui.actionStart_Fixed->blockSignals(!hasStartFixed);
    if (hasStartFixed)
        m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionStart_Fixed);

    m_pd.m_ui.actionStart_Single->setVisible(hasStartSingle);
    m_pd.m_ui.actionStart_Single->blockSignals(!hasStartSingle);
    if (hasStartSingle)
        m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionStart_Single);

    m_pd.m_ui.actionPause->setVisible(hasPause);
    m_pd.m_ui.actionPause->blockSignals(!hasPause);
    if (hasPause)
        m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionPause);

    m_pd.m_ui.actionStop->setVisible(hasStop);
    m_pd.m_ui.actionStop->blockSignals(!hasStop);
    if (hasStop)
        m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionStop);

    if (hasPlayback)
        m_pd.m_ui.acquisitionToolBar->addSeparator();

    // Restore the window menu.
    m_pd.m_ui.actionNo_Controls_Available->setVisible(!hasPlayback);

    // Triggering.
    m_pd.m_ui.actionTrigger->setVisible(hasTrigger);
    m_pd.m_ui.actionTrigger->blockSignals(!hasTrigger);
    if (hasTrigger)
    {
        m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionTrigger);
        m_pd.m_ui.acquisitionToolBar->addSeparator();
    }

    // Zoom control.
    m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionZoom_Editor);
    m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionZoom_Fit);
    m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_ui.actionZoom_100);

    // Color palette control.
    m_pd.m_ui.acquisitionToolBar->addSeparator();
    m_pd.m_ui.acquisitionToolBar->addAction(m_pd.m_colorPaletteButtonAction);

    // Re-add any user actions.
    if (m_pd.m_userToolBarActions.size())
    {
        m_pd.m_ui.acquisitionToolBar->addSeparator();
        for (auto *const userAction : m_pd.m_userToolBarActions)
            m_pd.m_ui.acquisitionToolBar->addAction(userAction);
    }

    // Image stats.
    m_pd.m_ui.actionFrame_Rate->setVisible(hasFrameRate);
    m_pd.m_ui.actionFrame_Rate->blockSignals(!hasFrameRate);

    m_pd.m_ui.actionBuffers_Queued->setVisible(hasFramesQueued);
    m_pd.m_ui.actionBuffers_Queued->blockSignals(!hasFramesQueued);

    m_pd.m_ui.actionBuffers_Filled->setVisible(hasFramesFilled);
    m_pd.m_ui.actionBuffers_Filled->blockSignals(!hasFramesFilled);

    m_pd.m_ui.actionBuffers_Dropped->setVisible(hasFramesDropped);
    m_pd.m_ui.actionBuffers_Dropped->blockSignals(!hasFramesDropped);

    m_pd.m_ui.actionBuffers_Delivered->setVisible(hasFramesDelivered);
    m_pd.m_ui.actionBuffers_Delivered->blockSignals(!hasFramesDelivered);

    m_pd.m_ui.actionActive_Setup_Mode->setVisible(hasSetupMode);
    m_pd.m_ui.actionActive_Setup_Mode->blockSignals(!hasSetupMode);
    
    // Start-fixed range.
    m_pd.m_startFixedSpinBox->setRange(m_pd.m_acqIf->startFixedMin(), m_pd.m_acqIf->startFixedMax());
    m_pd.m_startFixedSpinBox->setSingleStep(m_pd.m_acqIf->startFixedInc());
}
void ImageView::WindowQt::updateAcquisitionState (void)
{
    m_pd.updateActionStates();
    m_pd.updateImageStats();
}

// Window display.
void ImageView::WindowQt::setFullScreen (const bool enabled)
{
    const auto oldWindowState = windowState();
    const bool windowStateChange = !!(oldWindowState & Qt::WindowFullScreen) != enabled;

    m_pd.m_fullScreenBarHiderTimer->blockSignals(!enabled);
    m_pd.m_fullScreenBarHiderTimer->setInterval(enabled ? 500 : 0);

    m_pd.m_ui.actionFull_Screen->setIcon(QIcon(enabled ? ":/Icons/LeaveFullScreen.png" : ":/Icons/EnterFullScreen.png"));
    m_pd.m_ui.actionFull_Screen->setToolTip(enabled ? "Leave Full Screen" : "Enter Full Screen");
    m_pd.m_ui.actionFull_Screen->setChecked(enabled);

    m_pd.showWindowBars();

    if (windowStateChange)
    {
        const auto newWindowState = enabled ? (oldWindowState | Qt::WindowFullScreen) : m_pd.m_nonFullScreenWindowState;
        setWindowState(newWindowState);
    }

    if (enabled)
    {
        m_pd.m_nonFullScreenWindowState = oldWindowState & ~Qt::WindowFullScreen;

        raise();
        m_pd.m_fullScreenBarHiderTimer->start();
    }
}
void ImageView::WindowQt::toggleFullScreen (void)
{
    setFullScreen(!isFullScreen());
}
void ImageView::WindowQt::enterFullScreen (void)
{
    setFullScreen(true);
}
void ImageView::WindowQt::leaveFullScreen (void)
{
    setFullScreen(false);
}

// External image access.
void ImageView::WindowQt::saveCurrentImage (void)
{
    // Ask the user where we should save the image.
    QString fileFormats;

    const bool noFileIf = !m_pd.m_fileIf;
    const bool hasFileIf = !noFileIf;

    if (noFileIf || !m_pd.m_fileIf->hasFormat(".png"))
        fileFormats.append(fileFormats.size() ? ";;" : "").append("Portable Network Graphics Image (*.png)");

    if (noFileIf || (!m_pd.m_fileIf->hasFormat(".jpg") && !m_pd.m_fileIf->hasFormat(".jpeg")))
        fileFormats.append(fileFormats.size() ? ";;" : "").append("Joint Photographic Expert Group Image (*.jpg *.jpeg)");

    if (noFileIf || !m_pd.m_fileIf->hasFormat(".bmp"))
        fileFormats.append(fileFormats.size() ? ";;" : "").append("Bitmap Image (*.bmp)");

    // Retrieve the list of AcquisitionInterface formats.
    if (hasFileIf)
    {
        size_t formatCount;
        const QString ifFormats = String::read(&AbstractFileInterface::queryFormats, m_pd.m_fileIf.get(), &formatCount).c_str();

        // Extract the file save pattern for each supported AcquisitionInterface format.
        const QStringList acqIfFileFormats = ifFormats.split(' ');
        for (auto const& acqIfFormat : acqIfFileFormats)
        {
            const auto primaryExtension = acqIfFormat.toLocal8Bit();
            const QString ifFormatPattern = String::read(&AbstractFileInterface::queryFormatPattern, m_pd.m_fileIf.get(), primaryExtension.data()).c_str();

            if (ifFormatPattern.length())
                fileFormats += (fileFormats.size() ? ";;" : "") + ifFormatPattern;
        }
    }

    // Query the user for a save path.
#if defined(_WIN32)
    const QString filePath = QFileDialog::getSaveFileName(this, "Save Current Image", m_pd.m_prevSavePath, fileFormats, &m_pd.m_prevSaveFileFormat);
#elif defined(__GNUC__)
    const QString filePath = QFileDialog::getSaveFileName(this, "Save Current Image", m_pd.m_prevSavePath, fileFormats, &m_pd.m_prevSaveFileFormat, QFileDialog::DontUseNativeDialog);
#else
#   error File dialog behavior undefined for this platform.
#endif

    QString warningMessage;
    if (!filePath.isEmpty())
    {
        // Attempt to save the image.
        if (hasFileIf && m_pd.m_fileIf->hasFormat(QFileInfo(filePath).suffix().toLocal8Bit()))
        {
            // Save using the AcquisitionInterface method, saving the RAW image data buffer.
            ScopedLock criticalSection;
            RawImage rawImage;
            if (!m_pd.m_surfaceStream->latestImage(&rawImage, &criticalSection))
                warningMessage = "The image data buffer cannot be accessed at this time.";
            else if (m_pd.m_fileIf->save(rawImage, filePath.toLocal8Bit()))
                m_pd.m_prevSavePath = filePath;
            else
                warningMessage = "File save of an image data buffer failed. Please verify that the current user has access to the selected location.";
        }
        else
        {
            // Save the display buffer using the default Qt method.
            const auto saveImage = m_pd.m_surfaceStream->cloneLatestImageForRender();
            if (saveImage.isNull())
                warningMessage = "The display image cannot be accessed at this time.";
            else if (saveImage.save(filePath))
                m_pd.m_prevSavePath = filePath;
            else
                warningMessage = "File save of a display image failed. Please verify that the current user has access to the selected location.";
        }
    }

    if (warningMessage.size())
        QMessageBox::warning(this, "File Save Error", warningMessage);

    // Update our action states.
    m_pd.updateActionStates();
}
void ImageView::WindowQt::copyCurrentImage (void)
{
    const QImage copyImage = m_pd.m_surfaceStream->cloneLatestImageForRender();

    if (!copyImage.isNull())
        QApplication::clipboard()->setImage(copyImage);

    m_pd.updateActionStates();
}
void ImageView::WindowQt::copyCurrentPixelValue (void)
{
    QVariant pixelValueData = m_pd.m_ui.actionCopy_Pixel_Value->data();
    if (pixelValueData.isNull())
        pixelValueData = m_pd.m_ui.actionPixel_Value->data();

    QApplication::clipboard()->setText(pixelValueData.toString());
}
void ImageView::WindowQt::copyCurrentImageStats (void)
{
    const auto preLoadData = m_pd.m_ui.actionCopy_Image_Stats->data();
    const QString statsText = preLoadData.isNull() ? m_pd.imageStatsText() : preLoadData.toString();
    QApplication::clipboard()->setText(statsText);
}

// Helper methods.
void ImageView::WindowQt::setColorTableGray (void)
{
    m_pd.m_colorTableCurrent = ColorTable::Gray;
    m_pd.m_colorTable.resize(256);

    uint i = 0;
    for (auto &e : m_pd.m_colorTable)
    {
        e = qRgb(i, i, i);
        i++;
    }

    m_pd.applyColorTable(m_pd.m_ui.actionColor_Palette_Gray);
}
void ImageView::WindowQt::setColorTableGray2x (void)
{
    m_pd.m_colorTableCurrent = ColorTable::Gray2x;
    m_pd.m_colorTable.resize(256);

    uint i = 0;
    for (auto &e : m_pd.m_colorTable)
    {
        const uint val = std::min<uint>(i, 255);
        e = qRgb(val, val, val);
        i += 2;
    }

    m_pd.applyColorTable(m_pd.m_ui.actionColor_Palette_Gray_2x);
}
void ImageView::WindowQt::setColorTableGrayInverse (void)
{
    m_pd.m_colorTableCurrent = ColorTable::GrayInverse;
    m_pd.m_colorTable.resize(256);

    uint i = 255;
    for (auto &e : m_pd.m_colorTable)
    {
        e = qRgb(i, i, i);
        --i;
    }

    m_pd.applyColorTable(m_pd.m_ui.actionColor_Palette_Gray_Inverse);
}
void ImageView::WindowQt::setColorTableSine (void)
{
    m_pd.m_colorTableCurrent = ColorTable::Sine;
    m_pd.m_colorTable.resize(256);

    for (uint i = 0; i < 256; i++)
    {
        const uint iSin = 127 * sin(IMAGE_VIEW_PI * double(i) / 127.) + 128;
        m_pd.m_colorTable[i] = qRgb(iSin, iSin, iSin);
    }

    m_pd.applyColorTable(m_pd.m_ui.actionColor_Palette_Sine);
}
void ImageView::WindowQt::setColorTableCosine (void)
{
    m_pd.m_colorTableCurrent = ColorTable::Cosine;
    m_pd.m_colorTable.resize(256);

    for (uint i = 0; i < 256; i++)
    {
        const uint iCos = 127 * cos(IMAGE_VIEW_PI * double(i) / 127.) + 128;
        m_pd.m_colorTable[i] = qRgb(iCos, iCos, iCos);
    }

    m_pd.applyColorTable(m_pd.m_ui.actionColor_Palette_Cosine);
}
void ImageView::WindowQt::setColorTableContours (void)
{
    m_pd.m_colorTableCurrent = ColorTable::Contours;
    m_pd.m_colorTable.resize(256);

    for (uint i = 0; i < 256; ++i)
    {
        const std::uint8_t c = (std::uint8_t)((256L * (2 * i + 1)) / (2 * 256));
        m_pd.m_colorTable[i] = qRgb(c, (c % 32) << 3, (c % 128) << 1);
    }

    m_pd.applyColorTable(m_pd.m_ui.actionColor_Palette_Contours);
}
void ImageView::WindowQt::setColorTableRainbow (void)
{
    m_pd.m_colorTableCurrent = ColorTable::Rainbow;
    const double freq = 2. * IMAGE_VIEW_PI / m_pd.m_ui.rainbow_periodSpinBox->value();
    const double rPhase = m_pd.m_ui.rainbow_rPhaseSpinBox->value();
    const double gPhase = m_pd.m_ui.rainbow_gPhaseSpinBox->value();
    const double bPhase = m_pd.m_ui.rainbow_bPhaseSpinBox->value();
    const double amplitude = m_pd.m_ui.rainbow_amplitudeSpinBox->value();
    const double center = m_pd.m_ui.rainbow_centerSpinBox->value();

    PrivateData::generateRainbowColorTable(m_pd.m_colorTable, freq, rPhase, gPhase, bPhase, amplitude, center);

    m_pd.applyColorTable(m_pd.m_ui.actionColor_Palette_Rainbow);
}
void ImageView::WindowQt::setColorTableRandom (void)
{
    m_pd.m_colorTableCurrent = ColorTable::Random;
    m_pd.m_colorTable.resize(256);

    const auto randomSeed = std::chrono::system_clock::now().time_since_epoch().count();
    std::default_random_engine generator (randomSeed);
    std::uniform_int_distribution<int> uidRand (0, 255);

    for (auto &e : m_pd.m_colorTable)
        e = qRgb(uidRand(generator), uidRand(generator), uidRand(generator));

    m_pd.applyColorTable(m_pd.m_ui.actionColor_Palette_Random);
}

void ImageView::WindowQt::setBayerDemosaicFilterAuto (void)
{
    m_pd.m_surfaceStream->setBayerDemosaicFilter(UnknownFormat);
    m_pd.applyBayerFilter(m_pd.m_ui.actionBayer_Filter_Auto);
}
void ImageView::WindowQt::setBayerDemosaicFilterRGGB (void)
{
    m_pd.m_surfaceStream->setBayerDemosaicFilter(BayerRGGB);
    m_pd.applyBayerFilter(m_pd.m_ui.actionBayerRG_Filter);
}
void ImageView::WindowQt::setBayerDemosaicFilterGRBG (void)
{
    m_pd.m_surfaceStream->setBayerDemosaicFilter(BayerGRBG);
    m_pd.applyBayerFilter(m_pd.m_ui.actionBayerGR_Filter);
}
void ImageView::WindowQt::setBayerDemosaicFilterGBRG (void)
{
    m_pd.m_surfaceStream->setBayerDemosaicFilter(BayerGBRG);
    m_pd.applyBayerFilter(m_pd.m_ui.actionBayerGB_Filter);
}
void ImageView::WindowQt::setBayerDemosaicFilterBGGR (void)
{
    m_pd.m_surfaceStream->setBayerDemosaicFilter(BayerBGGR);
    m_pd.applyBayerFilter(m_pd.m_ui.actionBayerBG_Filter);
}
