#ifndef INCLUDED__IMAGE__VIEW__WINDOW__QT__H
#define INCLUDED__IMAGE__VIEW__WINDOW__QT__H

/* FILE:        ImageViewWindowQt.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Qt display Window class for the ImageView library.
 */

#include "ImageView.h"

#include <QMainWindow>

namespace ImageView
{
    // The ImageView window class.
    class IMAGEVIEW_EXPORT WindowQt : public QMainWindow
    {
        Q_OBJECT

        Q_PROPERTY(bool     closeable                   READ closeable                  WRITE setCloseable);

        Q_PROPERTY(bool     previewZoomLevel            READ previewZoomLevel           WRITE setPreviewZoomLevel);
        Q_PROPERTY(bool     previewImageDimensions      READ previewImageDimensions     WRITE setPreviewImageDimensions);
        Q_PROPERTY(bool     previewImageFormat          READ previewImageFormat         WRITE setPreviewImageFormat);
        Q_PROPERTY(bool     previewPixelValue           READ previewPixelValue          WRITE setPreviewPixelValue);
        Q_PROPERTY(bool     previewFrameRate            READ previewFrameRate           WRITE setPreviewFrameRate);
        Q_PROPERTY(bool     previewBuffersQueued        READ previewBuffersQueued       WRITE setPreviewBuffersQueued);
        Q_PROPERTY(bool     previewBuffersFilled        READ previewBuffersFilled       WRITE setPreviewBuffersFilled);
        Q_PROPERTY(bool     previewBuffersDropped       READ previewBuffersDropped      WRITE setPreviewBuffersDropped);
        Q_PROPERTY(bool     previewBuffersDelivered     READ previewBuffersDelivered    WRITE setPreviewBuffersDelivered);
        Q_PROPERTY(bool     previewSetupMode            READ previewSetupMode           WRITE setPreviewSetupMode);

        Q_PROPERTY(bool     keepImageInfoVisible     READ keepImageInfoVisible    WRITE setKeepImageInfoVisible);

        Q_PROPERTY(bool     bayerDemosaicEnabled        READ bayerDemosaicEnabled       WRITE setBayerDemosaicEnabled);
        Q_PROPERTY(bool     rgbOrderReversed            READ rgbOrderReversed           WRITE setRgbOrderReversed);
        Q_PROPERTY(bool     pixelCrosshairVisible       READ pixelCrosshairVisible      WRITE setPixelCrosshairVisible);

    private:
        struct PrivateData;
        PrivateData &m_pd;

        // Illegal.
        WindowQt (void);
        WindowQt (WindowQt const&);
        WindowQt& operator= (WindowQt const&);

    public:
        // Constructor. WindowQt takes permanent ownership of the acquisition interface.
        WindowQt (ScopedAcquisitionInterface &&acqIf, QWidget *const parent = nullptr);

        // Constructor. WindowQt takes permanent ownership of the acquisition and file interfaces.
        WindowQt (ScopedAcquisitionInterface &&acqIf, ScopedFileInterface &&fileIf, QWidget *const parent = nullptr);

        // Destructor.
        ~WindowQt (void);

        // Retrieve the acquisition interface.
        AbstractAcquisitionInterface* acquisitionInterface (void);
        const AbstractAcquisitionInterface* acquisitionInterface (void) const;

        // Retrieve the file interface, if any.
        AbstractFileInterface* fileInterface (void);
        const AbstractFileInterface* fileInterface (void) const;

        // Retrieve the surface scroll area.
        SurfaceScrollArea* surfaceScrollArea (void);
        const SurfaceScrollArea* surfaceScrollArea (void) const;

        // Retrieve the surface stream.
        SurfaceStream* surfaceStream (void);
        const SurfaceStream* surfaceStream (void) const;

        // Various properties.
        bool closeable (void) const;
        void setCloseable (const bool enabled);

        bool previewZoomLevel (void) const;
        void setPreviewZoomLevel (const bool visible);

        bool previewImageDimensions (void) const;
        void setPreviewImageDimensions (const bool visible);

        bool previewImageFormat (void) const;
        void setPreviewImageFormat (const bool visible);

        bool previewPixelValue (void) const;
        void setPreviewPixelValue (const bool visible);

        bool previewFrameRate (void) const;
        void setPreviewFrameRate (const bool visible);

        bool previewBuffersQueued (void) const;
        void setPreviewBuffersQueued (const bool visible);

        bool previewBuffersFilled (void) const;
        void setPreviewBuffersFilled (const bool visible);

        bool previewBuffersDropped (void) const;
        void setPreviewBuffersDropped (const bool visible);

        bool previewBuffersDelivered (void) const;
        void setPreviewBuffersDelivered (const bool visible);

        bool previewSetupMode (void) const;
        void setPreviewSetupMode (const bool visible);

        bool keepImageInfoVisible (void) const;
        void setKeepImageInfoVisible (const bool visible);

        bool bayerDemosaicEnabled (void) const;

        bool rgbOrderReversed (void) const;

        bool pixelCrosshairVisible (void) const;

        // The color table used to render grayscale images.
        ColorTable colorTable (void) const;
        void setColorTable (const ColorTable ct);

        // Set an external QDialog to use as an application about dialog. Window ownership is not changed.
        QDialog* applicationAbout (void) const;
        void setApplicationAbout (QDialog *const abtDlg);

        // Methods to add custom actions to the Window GUI.
        QAction* addMainMenuAction (QString const& parentMenuName, QString const& text = QString());
        QAction* addToolBarAction (QString const& text = QString());
        QAction* addImageStatsAction (QString const& text = QString());

    public slots:
        // Acquisition control. Asynchronous, if no other operation is ongoing. Use the
        // AbstractAcquisitionInterface directly for more nuanced control.
        void setRunning (void);
        void setConfigured (void);
        void setUnconfigured (void);

        // GUI update slots.
        void updateFeatureSet (void);
        void updateAcquisitionState (void);

        // Window display.
        void setFullScreen (const bool enabled);
        void toggleFullScreen (void);
        void enterFullScreen (void);
        void leaveFullScreen (void);

        // System image access.
        void saveCurrentImage (void);
        void copyCurrentImage (void);
        void copyCurrentPixelValue (void);
        void copyCurrentImageStats (void);

        // Properties.
        void setBayerDemosaicEnabled (const bool enabled);
        void setRgbOrderReversed (const bool reversed);
        void setPixelCrosshairVisible (const bool visible);

        // Helper slots.
        void setColorTableGray (void);
        void setColorTableGray2x (void);
        void setColorTableGrayInverse (void);
        void setColorTableSine (void);
        void setColorTableCosine (void);
        void setColorTableContours (void);
        void setColorTableRainbow (void);
        void setColorTableRandom (void);

        void setBayerDemosaicFilterAuto (void);
        void setBayerDemosaicFilterRGGB (void);
        void setBayerDemosaicFilterGRBG (void);
        void setBayerDemosaicFilterGBRG (void);
        void setBayerDemosaicFilterBGGR (void);

    private:
        // Event overrides.
        virtual void closeEvent (QCloseEvent *evt) Q_DECL_OVERRIDE;
        virtual void changeEvent (QEvent *evt) Q_DECL_OVERRIDE;
    };
}

#endif // INCLUDED__IMAGE__VIEW__WINDOW__QT__H
