/* FILE:        ImageViewWindowQt_Private.cpp
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: Implementation of ImageView::WindowQt::PrivateData.
 */

#include <QShortcut>
#include <QKeySequence>

#include <QList>

#include <QDoubleSpinBox>
#include <QSpinBox>
#include <QMessageBox>

#include <QDir>
#include <QRgb>

#include <QDateTime>

#include <QPainter>
#include <QIcon>
#include <QPixmap>

#include <cmath>

#include "ImageViewNative_Private.h"

#include "ImageViewWindowQt_Private.h"

#include "ImageViewStringHelpers.h"

#include "ImageViewVersionNo.h"

#include <BFQtLgplDialog.h>

bool ImageView::WindowQt::PrivateData::register_meta_types (void)
{
    qRegisterMetaType<BusyWorkResult>("BusyWorkResult");
    return true;
}
const bool ImageView::WindowQt::PrivateData::meta_types_registered = ImageView::WindowQt::PrivateData::register_meta_types();

ImageView::WindowQt::PrivateData::PrivateData (WindowQt &imageView, ScopedAcquisitionInterface &&acqIf, ScopedFileInterface &&fileIf)
    : m_window (imageView)
    , m_appAboutDialog (nullptr)
    , m_acqIf (std::move(acqIf))
    , m_fileIf (std::move(fileIf))
    , m_prevSavePath (QDir::homePath())
    , m_busyWorkCount (0)
{ }

// Destructor.
ImageView::WindowQt::PrivateData::~PrivateData (void)
{
    // Stop any ongoing acquisition.
    flushBusyWork();

    // Delete the surface stream before trying to delete anything it might depend upon.
    delete m_surfaceStream;
}

// Singleton initializer (call immediately prior to construction).
void ImageView::WindowQt::PrivateData::initialize (void)
{
    // Setup the Window UI.
    m_ui.setupUi(&m_window);

    // Benign workaround for systems (cough-cough, Ubuntu) that interfere with the menu action shortcuts.
    m_window.addActions(m_ui.menuBar->actions());

    // Hide the rainbow-rules dock widget for now.
    m_ui.rainbowRulesDockWidget->setVisible(false);

    // Nullify the About Application setup.
    m_window.setApplicationAbout(nullptr);

    // Setup the About ImageView dialog.
    setupAboutImageView();

    // Setup the About Qt and Qt LGPL Licensing menu entries.
    setupAboutQt();

    // Setup the busyWork cleanup slot.
    connect(this, SIGNAL(busyWorkComplete(BusyWorkResult)), this, SLOT(cleanupBusyWork(BusyWorkResult)));

    // Setup the Area.
    m_imageArea = new ImageView::SurfaceScrollArea(&m_window);
    m_ui.centralwidget->layout()->addWidget(m_imageArea);

    // Initialize the Surface and SurfaceStream.
    auto *const imageSurface = m_imageArea->surface();
    connect(imageSurface, SIGNAL(scalingChanged(bool, double)), this, SLOT(updateZoomStates()), Qt::QueuedConnection);

    m_surfaceStream = new ImageView::SurfaceStream(imageSurface, m_acqIf.get());

    connect(m_surfaceStream, SIGNAL(imageUpdated()), this, SLOT(updateActionStates()), Qt::QueuedConnection);
    connect(m_surfaceStream, SIGNAL(bitmapLockChanged(bool)), this, SLOT(updateBitmapLockedStates(bool)), Qt::QueuedConnection);

    imageSurface->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(imageSurface, SIGNAL(customContextMenuRequested(QPoint)), this, SLOT(showSurfaceContextMenu(QPoint)));

    // Retrieve the current application style.
    auto *const style = QApplication::style();

    // Setup the color palette menu.
    setupColorPaletteMenu();

    // Setup the Decode Bayer Images action.
    connect(m_ui.actionDecode_Bayer_Images, SIGNAL(triggered(bool)), &m_window, SLOT(setBayerDemosaicEnabled(bool)));

    connect(m_ui.actionBayer_Filter_Auto, SIGNAL(triggered()), &m_window, SLOT(setBayerDemosaicFilterAuto()));
    connect(m_ui.actionBayerRG_Filter, SIGNAL(triggered()), &m_window, SLOT(setBayerDemosaicFilterRGGB()));
    connect(m_ui.actionBayerGR_Filter, SIGNAL(triggered()), &m_window, SLOT(setBayerDemosaicFilterGRBG()));
    connect(m_ui.actionBayerGB_Filter, SIGNAL(triggered()), &m_window, SLOT(setBayerDemosaicFilterGBRG()));
    connect(m_ui.actionBayerBG_Filter, SIGNAL(triggered()), &m_window, SLOT(setBayerDemosaicFilterBGGR()));

    m_window.setBayerDemosaicFilterAuto();
    applyBayerFilter(m_ui.actionBayer_Filter_Auto);

    // Setup the Reverse RGB Order action
    connect(m_ui.actionReverse_RGB_Order, SIGNAL(triggered(bool)), &m_window, SLOT(setRgbOrderReversed(bool)));

    // Setup the Pixel Crosshair Visible action.
    connect(m_ui.actionPixel_Crosshair_Visible, SIGNAL(triggered(bool)), &m_window, SLOT(setPixelCrosshairVisible(bool)));

    // Setup the image stats menu.
    setupImageStatsMenu();

    // Setup the system control buttons.
    connect(m_ui.actionSave_Image, SIGNAL(triggered()), &m_window, SLOT(saveCurrentImage()));
    connect(m_ui.actionCopy_Image, SIGNAL(triggered()), &m_window, SLOT(copyCurrentImage()));
    connect(m_ui.actionCopy_Pixel_Value, SIGNAL(triggered()), &m_window, SLOT(copyCurrentPixelValue()));

    m_ui.actionSave_Image->setIcon(style->standardIcon(QStyle::SP_DialogSaveButton));

    m_window.addAction(m_ui.actionCopy_Pixel_Value);

    // Setup mode controls.
    m_setupCleanupButton = new QToolButton(m_ui.acquisitionToolBar);
    m_setupCleanupButton->setDefaultAction(m_ui.actionSetup_Cleanup);
    m_setupCleanupButton->setPopupMode(QToolButton::MenuButtonPopup);

    m_setupModeMenu = new QMenu(m_setupCleanupButton);
    m_setupModeMenu->setToolTipsVisible(true);

    m_ui.actionSetup_Mode->setMenu(m_setupModeMenu);

    // Setup the control buttons.
    m_ui.actionSetup->setShortcut(Qt::Key_F5);
    m_ui.actionCleanup->setShortcut(Qt::SHIFT + Qt::Key_F5);
    connect(m_ui.actionSetup, SIGNAL(triggered()), this, SLOT(setupAsync()));
    connect(m_ui.actionCleanup, SIGNAL(triggered()), this, SLOT(cleanupAsync()));

    connect(m_ui.actionTrigger, SIGNAL(triggered()), this, SLOT(triggerAsync()));

    connect(m_ui.actionStart_Single_Reverse, SIGNAL(triggered()), this, SLOT(startSingleReverseAsync()));
    connect(m_ui.actionStart_Reverse, SIGNAL(triggered()), this, SLOT(startReverseAsync()));
    connect(m_ui.actionStart, SIGNAL(triggered()), this, SLOT(startAsync()));
    connect(m_ui.actionStart_Fixed, SIGNAL(triggered()), this, SLOT(startFixedSelect()));
    connect(m_ui.actionStart_Single, SIGNAL(triggered()), this, SLOT(startSingleAsync()));
    connect(m_ui.actionPause, SIGNAL(triggered()), this, SLOT(pauseAsync()));
    connect(m_ui.actionStop, SIGNAL(triggered()), this, SLOT(stopAsync()));

    // Setup the zoom buttons.
    setupZoomEditorPopover();
    m_ui.actionZoom_In->setShortcut(QKeySequence::ZoomIn);
    m_ui.actionZoom_Out->setShortcut(QKeySequence::ZoomOut);

    connect(m_ui.actionZoom_In, SIGNAL(triggered()), m_imageArea, SLOT(doZoomIn()));
    connect(m_ui.actionZoom_In, SIGNAL(triggered()), this, SLOT(zoomEditorOpen()));

    connect(m_ui.actionZoom_Out, SIGNAL(triggered()), m_imageArea, SLOT(doZoomOut()));
    connect(m_ui.actionZoom_Out, SIGNAL(triggered()), this, SLOT(zoomEditorOpen()));

    connect(m_ui.actionZoom_Fit, SIGNAL(triggered(bool)), m_imageArea, SLOT(setZoomToFit(bool)));

    connect(m_ui.actionZoom_25, SIGNAL(triggered()), m_imageArea, SLOT(doZoomTo25()));
    connect(m_ui.actionZoom_50, SIGNAL(triggered()), m_imageArea, SLOT(doZoomTo50()));
    connect(m_ui.actionZoom_100, SIGNAL(triggered()), m_imageArea, SLOT(doZoomTo100()));
    connect(m_ui.actionZoom_200, SIGNAL(triggered()), m_imageArea, SLOT(doZoomTo200()));
    connect(m_ui.actionZoom_400, SIGNAL(triggered()), m_imageArea, SLOT(doZoomTo400()));
    connect(m_ui.actionZoom_800, SIGNAL(triggered()), m_imageArea, SLOT(doZoomTo800()));

    // Update the GUI for the interface state.
    auto *const acqIfSignalHandler = m_acqIf->signalHandler();
    connect(acqIfSignalHandler, SIGNAL(featureSetChanged()), &m_window, SLOT(updateFeatureSet()), Qt::QueuedConnection);
    connect(acqIfSignalHandler, SIGNAL(acquisitionStateChanged()), &m_window, SLOT(updateAcquisitionState()), Qt::QueuedConnection);

    // Setup the Full Screen actions.
    connect(m_ui.actionFull_Screen, SIGNAL(triggered()), &m_window, SLOT(toggleFullScreen()));
    new QShortcut(Qt::Key_Escape, &m_window, SLOT(leaveFullScreen()));

    m_ui.actionFull_Screen->setIcon(QIcon(":/Icons/EnterFullScreen.png"));

    m_fullScreenButton = new QToolButton(m_ui.menuBar);
    m_fullScreenButton->setDefaultAction(m_ui.actionFull_Screen);
    m_fullScreenButton->setAutoRaise(true);

    m_ui.menuBar->setCornerWidget(m_fullScreenButton, Qt::TopRightCorner);

    // Setup the Full Screen window bar hiding timer and animation.
    m_fullScreenBarHiderTimer = new QTimer(this);
    m_fullScreenBarHiderTimer->setInterval(500);
    m_fullScreenBarHiderTimer->setSingleShot(true);

    connect(m_fullScreenBarHiderTimer, SIGNAL(timeout()), this, SLOT(hideWindowBars()));

    for (auto *const childMenu : m_ui.menuBar->findChildren<QMenu*>())
    {
        connect(childMenu, SIGNAL(aboutToHide()), m_fullScreenBarHiderTimer, SLOT(start()));
        connect(childMenu, SIGNAL(aboutToShow()), this, SLOT(showWindowBars()));
    }
    connect(m_colorPaletteMenu, SIGNAL(aboutToHide()), m_fullScreenBarHiderTimer, SLOT(start()));
    connect(m_imageStatsMenu, SIGNAL(aboutToHide()), m_fullScreenBarHiderTimer, SLOT(start()));

    connect(m_ui.actionKeep_Stats_Visible, SIGNAL(triggered()), this, SLOT(showWindowBars()));
    connect(m_ui.actionKeep_Stats_Visible, SIGNAL(triggered()), m_fullScreenBarHiderTimer, SLOT(start()));

    m_fullScreenBarHiderAnimationGroup = new QParallelAnimationGroup(this);
    m_fullScreenBarHiderAnimation_menuBar = addAnimatedWidgetHide(m_fullScreenBarHiderAnimationGroup, m_ui.menuBar);
    m_fullScreenBarHiderAnimation_acquisitionToolBar = addAnimatedWidgetHide(m_fullScreenBarHiderAnimationGroup, m_ui.acquisitionToolBar);
    m_fullScreenBarHiderAnimation_statusBar = addAnimatedWidgetHide(m_fullScreenBarHiderAnimationGroup, m_ui.statusBar);
    m_fullScreenBarHiderAnimationGroup->setDirection(QAbstractAnimation::Backward);

    // Setup the start-fixed popover.
    setupStartFixedPopover();

    // Setup the quick-pan minimap popover.
    setupQuickPanPopover();

    // Install this PrivateData class as an event filter for image surface.
    m_ui.menuBar->installEventFilter(this);
    m_ui.acquisitionToolBar->installEventFilter(this);
    m_ui.statusBar->installEventFilter(this);

    m_window.updateFeatureSet();
    m_window.updateAcquisitionState();
}

void ImageView::WindowQt::PrivateData::setupAboutImageView (void)
{
    // Setup the dialog itself.
    auto *const aboutImageViewDialog = new QDialog(&m_window);
    m_aboutDialog.setupUi(aboutImageViewDialog);

    // Determine the BitFlow SDK version installed.
    QString sdkVerStr = String::read(&Native::LibraryVersionStr).c_str();
    if (sdkVerStr.isEmpty())
        sdkVerStr = "Unknown";

    // Determine the BitFlow driver version installed.
    QString driverVerStr = String::read(&Native::DriverVersionStr).c_str();
    if (driverVerStr.isEmpty())
        driverVerStr = "Unknown";

    // Generate the dynamic strings.
    const QString versionLabelText =
        "<html align=\"center\">"
        "<p><span style=\"font-size:16pt;font-weight:600;\">ImageView</span></p>"
        "<p><span style=\"font-size:12pt;\">Library for image render and display.</span></p>"
        "<p><table>"
        "<tr><td align=right><b>ImageView Version&nbsp;</b></td><td>" IMAGE_VIEW_APPVERSION ", " IMAGE_VIEW_BUILDVERSION "</td></tr>"
        "<tr><td align=right><b>SDK Version&nbsp;</b></td><td>" + sdkVerStr + "</td></tr>"
        "<tr><td align=right><b>Driver Version&nbsp;</b></td><td>" + driverVerStr + "</td></tr>"
        "<tr><td align=right><b>Copyright&nbsp;</b></td><td>" IMAGE_VIEW_COPYRIGHT_HTML "</td></tr>"
        "</table></p>"
        "</html>";

    const QString versionLabelHelp =
        "<html>"
        "<center><table>"
        "<tr><td align=right><b>ImageView Version&nbsp;</b></td><td>The release version of the ImageView library.</td></tr>"
        "<tr><td align=right><b>SDK Version&nbsp;</b></td><td>The version of the " + QString(IMAGE_VIEW_CORPORATE_NAME_ABRV) + " SDK that this build of ImageView was built with.</td></tr>"
        "<tr><td align=right><b>Driver Version&nbsp;</b></td><td>The version of the " + QString(IMAGE_VIEW_CORPORATE_NAME_ABRV) + " driver that is currently running. This need not match the SDK version.</td></tr>"
        "<tr><td align=right><b>Copyright&nbsp;</b></td><td>The ImageView copyright notice.</a></td></tr>"
        "</table></center>"
        "</html>";

    // Set strings.
    m_aboutDialog.imageViewMainLabel->setPixmap(QIcon(":/Icons/ImageViewIcon.png").pixmap(128, 128));

    m_aboutDialog.imageViewDetailsLabel->setText(versionLabelText);
    m_aboutDialog.imageViewDetailsLabel->setToolTip(versionLabelHelp);
    m_aboutDialog.imageViewDetailsLabel->setWhatsThis(versionLabelHelp);

    // Finalize the dialog.
    aboutImageViewDialog->adjustSize();
    aboutImageViewDialog->setFixedSize(aboutImageViewDialog->size());

    connect(m_ui.actionAbout_ImageView, SIGNAL(triggered()), aboutImageViewDialog, SLOT(exec()));
}

void ImageView::WindowQt::PrivateData::setupAboutQt (void)
{
    // About Qt... shows the standard Qt about screen.
    connect(m_ui.actionAbout_Qt, SIGNAL(triggered()), this, SLOT(showAboutQt()));

    // Create the Qt LGPL Licensing dialog.
    auto *const qtLgplLicensingDialog = new BFQtLgplDialog(&m_window);
    connect(m_ui.actionQt_LGPL_Licensing, SIGNAL(triggered()), qtLgplLicensingDialog, SLOT(exec()));
}

QPropertyAnimation* ImageView::WindowQt::PrivateData::addAnimatedWidgetHide (QParallelAnimationGroup *const group, QWidget *const toHide) const
{
    auto *const animation = new QPropertyAnimation(toHide, "maximumHeight");

    animation->setStartValue(QWIDGETSIZE_MAX);
    const int originalHeight = toHide->height();
    animation->setKeyValueAt(0.01, originalHeight);
    animation->setEndValue(1);
    animation->setDuration(200);

    group->addAnimation(animation);

    return animation;
}

void ImageView::WindowQt::PrivateData::setupColorPaletteMenu (void)
{
    m_colorPaletteButton = new QToolButton(&m_window);
    m_ui.acquisitionToolBar->addSeparator();
    m_colorPaletteButtonAction = m_ui.acquisitionToolBar->addWidget(m_colorPaletteButton);

    m_colorPaletteMenu = new QMenu(&m_window);
    m_colorPaletteButton->setMenu(m_colorPaletteMenu);
    m_colorPaletteButton->setPopupMode(QToolButton::InstantPopup);
    m_colorPaletteButton->setFocusPolicy(Qt::NoFocus);

    m_colorPaletteMenu->addAction(m_ui.actionColor_Palette_Gray);
    m_colorPaletteMenu->addAction(m_ui.actionColor_Palette_Gray_2x);
    m_colorPaletteMenu->addAction(m_ui.actionColor_Palette_Gray_Inverse);
    m_colorPaletteMenu->addAction(m_ui.actionColor_Palette_Sine);
    m_colorPaletteMenu->addAction(m_ui.actionColor_Palette_Cosine);
    m_colorPaletteMenu->addAction(m_ui.actionColor_Palette_Contours);
    m_colorPaletteMenu->addAction(m_ui.actionColor_Palette_Rainbow);
    m_colorPaletteMenu->addAction(m_ui.actionColor_Palette_Random);

    connect(m_ui.actionColor_Palette_Gray, SIGNAL(triggered()), &m_window, SLOT(setColorTableGray()));
    connect(m_ui.actionColor_Palette_Gray_2x, SIGNAL(triggered()), &m_window, SLOT(setColorTableGray2x()));
    connect(m_ui.actionColor_Palette_Gray_Inverse, SIGNAL(triggered()), &m_window, SLOT(setColorTableGrayInverse()));
    connect(m_ui.actionColor_Palette_Sine, SIGNAL(triggered()), &m_window, SLOT(setColorTableSine()));
    connect(m_ui.actionColor_Palette_Cosine, SIGNAL(triggered()), &m_window, SLOT(setColorTableCosine()));
    connect(m_ui.actionColor_Palette_Contours, SIGNAL(triggered()), &m_window, SLOT(setColorTableContours()));
    connect(m_ui.actionColor_Palette_Rainbow, SIGNAL(triggered()), &m_window, SLOT(setColorTableRainbow()));
    connect(m_ui.actionColor_Palette_Random, SIGNAL(triggered()), &m_window, SLOT(setColorTableRandom()));

    connect(m_ui.rainbow_generatePushButton, SIGNAL(clicked()), &m_window, SLOT(setColorTableRainbow()));

    auto *const enterRainbowModeAction = new QAction(m_ui.rainbowRulesDockWidget);
    enterRainbowModeAction->setShortcuts(QList<QKeySequence>() << Qt::Key_Enter << Qt::Key_Return);
    enterRainbowModeAction->setShortcutContext(Qt::WidgetShortcut);
    connect(enterRainbowModeAction, SIGNAL(triggered()), &m_window, SLOT(setColorTableRainbow()));

    m_ui.rainbow_periodSpinBox->addAction(enterRainbowModeAction);
    m_ui.rainbow_rPhaseSpinBox->addAction(enterRainbowModeAction);
    m_ui.rainbow_gPhaseSpinBox->addAction(enterRainbowModeAction);
    m_ui.rainbow_bPhaseSpinBox->addAction(enterRainbowModeAction);
    m_ui.rainbow_amplitudeSpinBox->addAction(enterRainbowModeAction);
    m_ui.rainbow_centerSpinBox->addAction(enterRainbowModeAction);

    m_window.setColorTableGray();
    applyColorTable(m_ui.actionColor_Palette_Gray);
}

void ImageView::WindowQt::PrivateData::setupImageStatsMenu (void)
{
    m_imageStatsButton = new QToolButton (&m_window);
    m_ui.statusBar->addPermanentWidget(m_imageStatsButton);

    m_imageStatsButton->setToolButtonStyle(Qt::ToolButtonTextOnly);
    m_imageStatsButton->setText("Image Info");
    m_imageStatsButton->setToolTip("Image Info");
    m_imageStatsButton->setAutoRaise(true);
    m_imageStatsButton->setFocusPolicy(Qt::NoFocus);

    m_imageStatsMenu = new QMenu(&m_window);
    m_imageStatsMenu->setSeparatorsCollapsible(true);
    m_imageStatsButton->setMenu(m_imageStatsMenu);
    m_imageStatsButton->setPopupMode(QToolButton::InstantPopup);

    m_imageStatsMenu->setWindowTitle("Image Info");
    m_imageStatsMenu->setTitle("Image Info");
    m_imageStatsMenu->setTearOffEnabled(true);
    m_imageStatsMenu->setToolTipsVisible(true);

    m_imageStatsMenu->addAction(m_ui.actionPixel_Value);
    m_imageStatsMenu->addAction(m_ui.actionZoom_Level);
    m_imageStatsMenu->addAction(m_ui.actionActive_Setup_Mode);
    m_imageStatsMenu->addAction(m_ui.actionImageDimensions);
    m_imageStatsMenu->addAction(m_ui.actionImage_Format);
    m_imageStatsMenu->addAction(m_ui.actionFrame_Rate);
    m_imageStatsMenu->addAction(m_ui.actionBuffers_Queued);
    m_imageStatsMenu->addAction(m_ui.actionBuffers_Filled);
    m_imageStatsMenu->addAction(m_ui.actionBuffers_Dropped);
    m_imageStatsMenu->addAction(m_ui.actionBuffers_Delivered);
    m_imageStatsMenu->addSeparator();
    m_imageStatsMenuUserActionSeparator = m_imageStatsMenu->addSeparator();
    m_imageStatsMenu->addAction(m_ui.actionCopy_Image_Stats);
    m_imageStatsMenu->addSeparator();
    m_imageStatsMenu->addAction(m_ui.actionKeep_Stats_Visible);

    connect(m_ui.actionCopy_Image_Stats, SIGNAL(triggered()), &m_window, SLOT(copyCurrentImageStats()));
    connect(m_ui.actionImage_Stats, SIGNAL(triggered()), m_imageStatsButton, SLOT(click()));

    m_imageStatsTimer = new QTimer(this);
    connect(m_imageStatsTimer, SIGNAL(timeout()), this, SLOT(updateImageStats()));

    m_imageStatsTimer->setInterval(1000 / 10);
    m_imageStatsTimer->setSingleShot(false);
}

void ImageView::WindowQt::PrivateData::setupZoomEditorPopover (void)
{
    m_zoomEditorPopover = new ImageView::PopoverWindow (&m_window);

    auto *const editorLayoutWidget = new QWidget (m_zoomEditorPopover);
    auto *const editorLayout = new QVBoxLayout (editorLayoutWidget);
    editorLayout->setMargin(0);
    editorLayout->setSpacing(0);

    auto *const editorTopLayoutWidget = new QWidget (editorLayoutWidget);
    auto *const editorTopLayout = new QHBoxLayout (editorTopLayoutWidget);
    editorTopLayout->setMargin(0);
    editorTopLayout->setSpacing(0);

    auto *const editorBtmLayoutWidget = new QWidget (editorLayoutWidget);
    auto *const editorBtmLayout = new QHBoxLayout (editorBtmLayoutWidget);
    editorBtmLayout->setMargin(0);
    editorBtmLayout->setSpacing(0);

    editorLayout->addWidget(editorTopLayoutWidget);
    editorLayout->addWidget(editorBtmLayoutWidget);

    // Top

    m_zoomEditorSpinBox = new QDoubleSpinBox (editorTopLayoutWidget);
    m_zoomEditorSpinBox->setSuffix("%");
    const int widgetHeight = m_zoomEditorSpinBox->height();

    auto *const spacer0 = new QWidget (editorTopLayoutWidget);
    spacer0->setFixedWidth(3);

    m_zoomEditorSlider = new QSlider (editorTopLayoutWidget);
    m_zoomEditorSlider->setOrientation(Qt::Horizontal);
    m_zoomEditorSlider->setMinimumWidth(124);

    auto *const zoomOutButton = new QToolButton (editorTopLayoutWidget);
    zoomOutButton->setDefaultAction (m_ui.actionZoom_Out);
    zoomOutButton->setAutoRaise(true);
    zoomOutButton->setAutoRepeat(true);

    auto *const zoomInButton = new QToolButton (editorTopLayoutWidget);
    zoomInButton->setDefaultAction (m_ui.actionZoom_In);
    zoomInButton->setAutoRaise(true);
    zoomInButton->setAutoRepeat(true);

    // Bottom

    auto *const zoomFitButton = new QToolButton (editorBtmLayoutWidget);
    zoomFitButton->setDefaultAction (m_ui.actionZoom_Fit);
    zoomFitButton->setAutoRaise(true);
    zoomFitButton->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    auto *const zoom25Button = new QToolButton (editorBtmLayoutWidget);
    zoom25Button->setDefaultAction (m_ui.actionZoom_25);
    zoom25Button->setAutoRaise(true);
    zoom25Button->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    auto *const zoom50Button = new QToolButton (editorBtmLayoutWidget);
    zoom50Button->setDefaultAction (m_ui.actionZoom_50);
    zoom50Button->setAutoRaise(true);
    zoom50Button->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    auto *const zoom100Button = new QToolButton (editorBtmLayoutWidget);
    zoom100Button->setDefaultAction (m_ui.actionZoom_100);
    zoom100Button->setAutoRaise(true);
    zoom100Button->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    auto *const zoom200Button = new QToolButton (editorBtmLayoutWidget);
    zoom200Button->setDefaultAction (m_ui.actionZoom_200);
    zoom200Button->setAutoRaise(true);
    zoom200Button->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    auto *const zoom400Button = new QToolButton (editorBtmLayoutWidget);
    zoom400Button->setDefaultAction (m_ui.actionZoom_400);
    zoom400Button->setAutoRaise(true);
    zoom400Button->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    auto *const zoom800Button = new QToolButton (editorBtmLayoutWidget);
    zoom800Button->setDefaultAction (m_ui.actionZoom_800);
    zoom800Button->setAutoRaise(true);
    zoom800Button->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    editorTopLayout->addWidget(m_zoomEditorSpinBox);
    editorTopLayout->addWidget(spacer0);
    editorTopLayout->addWidget(m_zoomEditorSlider);
    editorTopLayout->addWidget(zoomOutButton);
    editorTopLayout->addWidget(zoomInButton);

    editorBtmLayout->addWidget(zoomFitButton);

    editorBtmLayout->addWidget(zoom25Button);
    editorBtmLayout->addWidget(zoom50Button);
    editorBtmLayout->addWidget(zoom100Button);
    editorBtmLayout->addWidget(zoom200Button);
    editorBtmLayout->addWidget(zoom400Button);
    editorBtmLayout->addWidget(zoom800Button);

    m_zoomEditorPopover->setCentralWidget(editorLayoutWidget);
    m_zoomEditorPopover->setAffinityWidget(m_ui.acquisitionToolBar);
    m_zoomEditorPopover->setAcceptButtonVisible(false);
    m_zoomEditorPopover->setRejectOnFocusOut(true);

    m_zoomEditorPopover->setToolTip("<html>Modify the image zoom level.</html>");
    m_zoomEditorPopover->setLabelPixmap(QPixmap(":/Icons/Zoom.png").scaled(widgetHeight, widgetHeight, Qt::IgnoreAspectRatio, Qt::SmoothTransformation));

    zoomEditorUpdate();

    connect(m_surfaceStream->surface(), SIGNAL(scalingChanged(bool, double)), this, SLOT(zoomEditorUpdate()), Qt::QueuedConnection);

    connect(m_zoomEditorSpinBox, SIGNAL(editingFinished()), this, SLOT(zoomSpinBoxChanged()));
    connect(m_zoomEditorSlider, SIGNAL(valueChanged(int)), this, SLOT(zoomSliderChanged()));

    connect(m_zoomEditorPopover, SIGNAL(closed()), this, SLOT(zoomEditorClosed()));

    connect(m_ui.actionZoom_Editor, SIGNAL(triggered()), this, SLOT(zoomEditorOpen()));
}

void ImageView::WindowQt::PrivateData::setupStartFixedPopover (void)
{
    m_startFixedPopover = new ImageView::PopoverWindow (&m_window);

    m_startFixedSpinBox = new QSpinBox (m_startFixedPopover);
    m_startFixedSpinBox->setValue(10);
    const int widgetHeight = m_startFixedSpinBox->height();

    m_startFixedPopover->setCentralWidget(m_startFixedSpinBox);
    m_startFixedPopover->setAffinityWidget(m_ui.acquisitionToolBar);
    m_startFixedPopover->setRejectOnFocusOut(true);

    m_startFixedPopover->setToolTip("<html>Enter the number of frames to acquire. Enter/Return to start, Esc to cancel.</html>");
    m_startFixedPopover->setLabelPixmap(QPixmap(":/Icons/PlayFixed.png").scaled(widgetHeight, widgetHeight, Qt::IgnoreAspectRatio, Qt::SmoothTransformation));

    connect(m_startFixedSpinBox, SIGNAL(valueChanged(int)), this, SLOT(startFixedSet(int)));

    connect(m_startFixedSpinBox, SIGNAL(editingFinished()), m_startFixedPopover, SLOT(reject()));

    connect(m_startFixedPopover, SIGNAL(accepted()), this, SLOT(startFixedAccept()));
    connect(m_startFixedPopover, SIGNAL(closed()), this, SLOT(startFixedClose()));
}

void ImageView::WindowQt::PrivateData::setupQuickPanPopover (void)
{
    m_quickViewPopoverButton = new QToolButton (&m_window);
    m_quickViewPopover = new ImageView::PopoverWindow (&m_window);
    m_quickViewMiniMap = new ImageView::MiniMap (&m_window);

    m_quickViewPopoverButton->setDefaultAction(m_ui.actionQuick_View);
    m_quickViewPopoverButton->setAutoRaise(true);

    m_imageArea->setCornerWidget(m_quickViewPopoverButton);

    m_quickViewPopover->setCentralWidget(m_quickViewMiniMap);
    m_quickViewPopover->setAffinityWidget(m_imageArea->surface(), ImageView::PopoverWindow::BottomRightOverAffinity);

    m_quickViewPopover->setAcceptButtonVisible(false);
    m_quickViewPopover->setRejectOnFocusOut(true);

    m_quickViewPopover->setLabelText("");

    connect(m_ui.actionQuick_View, SIGNAL(triggered()), this, SLOT(quickViewPopoverOpen()));

    connect(m_surfaceStream, SIGNAL(imageUpdated()), this, SLOT(quickViewPopoverUpdatePixmap()), Qt::QueuedConnection);

    auto *const surface = m_surfaceStream->surface();
    connect(surface, SIGNAL(scalingChanged(bool, double)), this, SLOT(quickViewPopoverUpdateView()), Qt::QueuedConnection);
    connect(surface, SIGNAL(pixelRangeChanged()), this, SLOT(quickViewPopoverUpdateView()), Qt::QueuedConnection);
    connect(surface, SIGNAL(pixelOffsetChanged(QPointF)), this, SLOT(quickViewPopoverUpdateView()), Qt::QueuedConnection);

    connect(m_quickViewMiniMap, SIGNAL(offsetChanged(QPointF)), m_imageArea, SLOT(setImageOffset(QPointF)));
    connect(m_quickViewMiniMap, SIGNAL(scrollIn()), m_imageArea, SLOT(doZoomIn()));
    connect(m_quickViewMiniMap, SIGNAL(scrollOut()), m_imageArea, SLOT(doZoomOut()));

    connect(m_quickViewPopover, SIGNAL(closed()), this, SLOT(quickViewPopoverClose()));
}

// Helper methods.
void ImageView::WindowQt::PrivateData::applyColorTable (QObject *const sender)
{
    for (auto *const palette : m_ui.menuColor_Palette->actions())
    {
        const bool isSender = palette == sender;
        palette->setChecked(isSender);
        if (isSender)
            m_colorPaletteButton->setText("Palette: " + palette->text());
    }

    if (ImageView::ColorTable::Gray == m_colorTableCurrent)
        m_surfaceStream->setColorPalette(QVector<uint>());
    else
        m_surfaceStream->setColorPalette(m_colorTable);
}
void ImageView::WindowQt::PrivateData::applyBayerFilter (QObject *const sender)
{
    m_ui.actionBayer_Filter_Auto->setChecked(m_ui.actionBayer_Filter_Auto == sender);
    m_ui.actionBayerRG_Filter->setChecked(m_ui.actionBayerRG_Filter == sender);
    m_ui.actionBayerGR_Filter->setChecked(m_ui.actionBayerGR_Filter == sender);
    m_ui.actionBayerGB_Filter->setChecked(m_ui.actionBayerGB_Filter == sender);
    m_ui.actionBayerBG_Filter->setChecked(m_ui.actionBayerBG_Filter == sender);
}
void ImageView::WindowQt::PrivateData::generateRainbowColorTable (QVector<uint> &table, const double freq, const double rPhase, const double gPhase, const double bPhase, const double width, const double center)
{
    table.resize(256);

    for (uint i = 0; i < 256; i++)
    {
        const uint r = (uint)std::min(255., std::max(0., (sin(freq * i + rPhase) * width + center)));
        const uint g = (uint)std::min(255., std::max(0., (sin(freq * i + gPhase) * width + center)));
        const uint b = (uint)std::min(255., std::max(0., (sin(freq * i + bPhase) * width + center)));
        table[i] = qRgb(r, g, b);
    }
}

QString ImageView::WindowQt::PrivateData::imageStatsText (void) const
{
    QString statsText;
    for (auto *const action : m_imageStatsButton->menu()->actions())
    {
        if (action->isVisible() && action->data().isValid())
        {
            if (statsText.length())
                statsText.append('\n');
            statsText.append(action->text());
        }
    }

    if (statsText.length())
        statsText.prepend('\n').prepend("Image Info @ " + QDateTime::currentDateTime().toString(Qt::ISODate));

    return statsText;
}

void ImageView::WindowQt::PrivateData::updateActionStates (void)
{
    const bool canDoAnything = 0 >= m_busyWorkCount;

    const bool canSetup                 = canDoAnything && m_acqIf->canSetup();
    const bool canSetupMode             = canDoAnything && m_acqIf->canSetupMode();
    const bool canCleanup               = canDoAnything && m_acqIf->canCleanup();
    const bool canTrigger               = canDoAnything && m_acqIf->canTrigger();
    const bool canStartSingleReverse    = canDoAnything && m_acqIf->canStartSingleReverse();
    const bool canStartReverse          = canDoAnything && m_acqIf->canStartReverse();
    const bool canStart                 = canDoAnything && m_acqIf->canStart();
    const bool canStartFixed            = canDoAnything && m_acqIf->canStartFixed();
    const bool canStartSingle           = canDoAnything && m_acqIf->canStartSingle();
    const bool canPause                 = canDoAnything && m_acqIf->canPause();
    const bool canStop                  = canDoAnything && m_acqIf->canStop();

    const bool isSetup = m_acqIf->isSetup();
    const bool isStopped = m_acqIf->isStopped();
    const bool isRunningSingleReverse = m_acqIf->isRunningSingleReverse();
    const bool isRunningReverse = m_acqIf->isRunningReverse();
    const bool isRunning = m_acqIf->isRunning();
    const bool isRunningFixed = m_acqIf->isRunningFixed();
    const bool isRunningSingle = m_acqIf->isRunningSingle();
    const bool isPaused = m_acqIf->isPaused();

    const QString activeSetupMode = String::read(&AbstractAcquisitionInterface::activeSetupMode, m_acqIf.get()).c_str();

    const QSize latestImageSize = m_surfaceStream->latestImageSize();
    const QImage::Format latestImageFormat = m_surfaceStream->latestQImageFormat();

    const bool hasImage = isSetup && !latestImageSize.isNull();
    const bool hasIndexedImage = hasImage && latestImageFormat == QImage::Format_Indexed8;

    m_surfaceStream->setRefreshEnabled(isSetup);
    if (isSetup)
    {
        if (!m_imageStatsTimer->isActive())
            m_imageStatsTimer->start();
    }
    else
    {
        if (m_imageStatsTimer->isActive())
            m_imageStatsTimer->stop();

        updateImageStats();
    }

    m_ui.actionSetup->setEnabled(canSetup);
    m_ui.actionCleanup->setEnabled(canCleanup);
    m_ui.actionSetup_Cleanup->setEnabled(canSetup || canCleanup);
    if (canCleanup)
    {
        m_ui.actionSetup_Cleanup->setIcon(QIcon(":/Icons/Setup.png"));
        m_ui.actionSetup_Cleanup->setText(m_ui.actionCleanup->text());
        m_ui.actionSetup_Cleanup->setIconText(m_ui.actionCleanup->iconText());
        m_ui.actionSetup_Cleanup->setToolTip("Depress to " + m_ui.actionCleanup->toolTip());
        m_ui.actionSetup_Cleanup->setChecked(true);

        disconnect(m_ui.actionSetup_Cleanup, SIGNAL(triggered()), nullptr, nullptr);
        connect(m_ui.actionSetup_Cleanup, SIGNAL(triggered()), this, SLOT(cleanupAsync()));
    }
    else if (canSetup)
    {
        m_ui.actionSetup_Cleanup->setIcon(QIcon(":/Icons/Cleanup.png"));
        m_ui.actionSetup_Cleanup->setText(m_ui.actionSetup->text());
        m_ui.actionSetup_Cleanup->setIconText(m_ui.actionSetup->iconText());
        m_ui.actionSetup_Cleanup->setToolTip("Press to " + m_ui.actionSetup->toolTip());
        m_ui.actionSetup_Cleanup->setChecked(false);

        disconnect(m_ui.actionSetup_Cleanup, SIGNAL(triggered()), nullptr, nullptr);
        connect(m_ui.actionSetup_Cleanup, SIGNAL(triggered()), this, SLOT(setupAsync()));
    }

    // Setup mode handling.
    for (auto *const modeAction : m_setupModeMenu->findChildren<QAction*>())
    {
        modeAction->setEnabled(canSetupMode);
        modeAction->setChecked(modeAction->whatsThis() == activeSetupMode);
    }

    // Determine the most appropriate window icon.
    QPixmap windowStatePixmap;
    if (!isSetup)
        windowStatePixmap = QPixmap(":/Icons/Cleanup.png");
    else if (isStopped)
    {
        if (hasImage)
            windowStatePixmap = QPixmap(":/Icons/Stop.png");
        else
            windowStatePixmap = QPixmap(":/Icons/Setup.png");
    }
    else if (isPaused)
        windowStatePixmap = QPixmap(":/Icons/Pause.png");
    else if (isRunningSingle)
        windowStatePixmap = QPixmap(":/Icons/PlayOne.png");
    else if (isRunningSingleReverse)
        windowStatePixmap = QPixmap(":/Icons/PlayOneReverse.png");
    else if (isRunningReverse)
        windowStatePixmap = QPixmap(":/Icons/PlayReverse.png");
    else if (isRunningFixed)
        windowStatePixmap = QPixmap(":/Icons/PlayFixed.png");
    else if (isRunning)
        windowStatePixmap = QPixmap(":/Icons/Play.png");
    else
        windowStatePixmap = QPixmap(":/Icons/AmbiguousState.png");

    const auto windowIcon = QApplication::windowIcon();

    auto pixmapSizes = windowIcon.availableSizes();
    if (pixmapSizes.isEmpty())
        pixmapSizes << QSize(48, 48);

    // Generate the new window icon.
    QIcon newWindowIcon;
    for (auto pixmapSize : pixmapSizes)
    {
        QPixmap newPixmap;
        if (windowIcon.isNull())
        {
            newPixmap = QPixmap(pixmapSize);
            newPixmap.fill(Qt::transparent);
        }
        else
        {
            newPixmap = windowIcon.pixmap(pixmapSize);
            pixmapSize = newPixmap.size();
        }

        if (!windowStatePixmap.isNull())
        {
            const auto overlayRect = windowIcon.isNull() ? newPixmap.rect()
                : QRect(pixmapSize.width() / 2, 0, pixmapSize.width() / 2, pixmapSize.height() / 2);

            QPainter painter (&newPixmap);

            painter.setBrush(Qt::white);
            if (pixmapSize.width() >= 24 && pixmapSize.height() >= 24)
            {
                painter.setPen(Qt::black);
                painter.setRenderHint(QPainter::Antialiasing);
                painter.drawRoundedRect(overlayRect, pixmapSize.width() / 8, pixmapSize.height() / 8);
            }
            else
            {
                painter.setPen(Qt::lightGray);
                painter.drawRect(overlayRect);
            }

            painter.drawPixmap(overlayRect, windowStatePixmap);
        }

        newWindowIcon.addPixmap(newPixmap);
    }

    m_window.setWindowIcon(newWindowIcon);

    // Enable/disable window actions.
    m_ui.actionTrigger->setEnabled(canTrigger);

    m_ui.actionStart_Single_Reverse->setEnabled(canStartSingleReverse);
    m_ui.actionStart_Reverse->setEnabled(canStartReverse);
    m_ui.actionStart->setEnabled(canStart);
    m_ui.actionStart_Fixed->setEnabled(canStartFixed);
    m_ui.actionStart_Single->setEnabled(canStartSingle);
    m_ui.actionPause->setEnabled(canPause);
    m_ui.actionStop->setEnabled(canStop);

    m_ui.actionSave_Image->setEnabled(hasImage);
    m_ui.actionCopy_Image->setEnabled(hasImage);

    m_ui.actionCopy_Image_Stats->setEnabled(hasImage);

    m_ui.menuColor_Palette->setEnabled(hasIndexedImage);
    m_colorPaletteButton->setEnabled(hasIndexedImage);
    m_ui.rainbowRulesContentsWidget->setEnabled(hasIndexedImage);

    m_ui.actionQuick_View->setEnabled(hasImage);
    if (!hasImage)
        quickViewPopoverClose();

    updateZoomStates();
}

void ImageView::WindowQt::PrivateData::updateZoomStates (void)
{
    const bool isSetup = m_acqIf->isSetup();
    const QSize latestImageSize = m_surfaceStream->latestImageSize();
    const bool hasImage = isSetup && !latestImageSize.isNull();

    // Enable/disable scaling actions.
    const double scaling = m_imageArea->zoomScaling();
    const bool canScaleIn = scaling + std::numeric_limits<double>::epsilon() < Surface::scalingMax();
    const bool canScaleOut = scaling - std::numeric_limits<double>::epsilon() > Surface::scalingMin();

    const bool canScale25 = std::fabs(scaling - 0.25) > std::numeric_limits<double>::epsilon();
    const bool canScale50 = std::fabs(scaling - 0.5) > std::numeric_limits<double>::epsilon();
    const bool canScale100 = std::fabs(scaling - 1.) > std::numeric_limits<double>::epsilon();
    const bool canScale200 = std::fabs(scaling - 2.) > std::numeric_limits<double>::epsilon();
    const bool canScale400 = std::fabs(scaling - 4.) > std::numeric_limits<double>::epsilon();
    const bool canScale800 = std::fabs(scaling - 8.) > std::numeric_limits<double>::epsilon();

    m_ui.actionZoom_Editor->setEnabled(hasImage);
    m_zoomEditorSpinBox->setEnabled(hasImage);
    m_zoomEditorSlider->setEnabled(hasImage);
    m_ui.actionZoom_In->setEnabled(hasImage && canScaleIn);
    m_ui.actionZoom_Out->setEnabled(hasImage && canScaleOut);
    m_ui.actionZoom_Fit->setEnabled(hasImage);
    m_ui.actionZoom_Fit->setChecked(m_imageArea->zoomToFit());

    m_ui.actionZoom_25->setEnabled(hasImage && canScale25);
    m_ui.actionZoom_50->setEnabled(hasImage && canScale50);
    m_ui.actionZoom_100->setEnabled(hasImage && canScale100);
    m_ui.actionZoom_200->setEnabled(hasImage && canScale200);
    m_ui.actionZoom_400->setEnabled(hasImage && canScale400);
    m_ui.actionZoom_800->setEnabled(hasImage && canScale800);
}

void ImageView::WindowQt::PrivateData::updateImageStats (void)
{
    const auto *const surface = m_imageArea->surface();

    const QPoint imagePos = surface->activePixel();
    const PixPos pixPos (imagePos.x(), imagePos.y());
    PixVal pixVal;
    const bool validImagePos = m_imageArea->surface()->imageSrcPixVal(pixPos, pixVal);

    const QSize latestImageSize = m_surfaceStream->latestImageSize();
    const Format latestImageFormat = m_surfaceStream->latestImageFormat();
    const bool nullImage = latestImageSize.isNull();

    const auto zoomStr = QString::number(100. * m_imageArea->zoomScaling(), 'f', 2) + "%";

    const auto dimensionsStr = nullImage ? "N/A"
        : QString::number(latestImageSize.width())
        + "x" + QString::number(latestImageSize.height())
        + "@" + QString::number(latestImageFormat & BaseFormatMask) + "-bit";

    QString encodingStr;
    switch (latestImageFormat & BayerFormatMask)
    {
    case BayerRGGB: encodingStr = "BayerRG"; break;
    case BayerGBRG: encodingStr = "BayerGB"; break;
    case BayerBGGR: encodingStr = "BayerBG"; break;
    case BayerGRBG: encodingStr = "BayerGR"; break;
    }

    QString colorOrderStr;
    switch (latestImageFormat & ColorOrderFormatMask)
    {
    default:
    case ColorOrderRGB: colorOrderStr = "RGB"; break;
    case ColorOrderBGR: colorOrderStr = "BGR"; break;
    }

    QString packedStr;
    switch (latestImageFormat & PackedFormatMask)
    {
    case Packed: packedStr = "p"; break;
    case PackedNoSwap: packedStr = "p_NoSwap"; break;
    }

    QString formatStr;
    switch (latestImageFormat & BaseFormatMask)
    {
    default:        formatStr = nullImage ? "N/A" : "Unknown"; break;
    case Gray8:     formatStr = (encodingStr.length() ? encodingStr : "Mono") + "8"; break;
    case Gray10:    formatStr = (encodingStr.length() ? encodingStr : "Mono") + "10" + packedStr; break;
    case Gray12:    formatStr = (encodingStr.length() ? encodingStr : "Mono") + "12"; break;
    case Gray14:    formatStr = (encodingStr.length() ? encodingStr : "Mono") + "14"; break;
    case Gray16:    formatStr = (encodingStr.length() ? encodingStr : "Mono") + "16"; break;
    case Color24:   formatStr = colorOrderStr + "8"; break;
    case Color30:   formatStr = colorOrderStr + "10p32"; break;
    case Color32:   formatStr = colorOrderStr + "a8"; break;
    case HiColor36: formatStr = colorOrderStr + "12"; break;
    case HiColor42: formatStr = colorOrderStr + "14"; break;
    case HiColor48: formatStr = colorOrderStr + "16"; break;
    }

    const auto imagePosStr = (pixVal.isColor ? "RGB(" : "P(") + QString::number(imagePos.x()) + ", " + QString::number(imagePos.y()) + ")";

    QString imagePosValStr;
    if (pixVal.isGrayscale)
        imagePosValStr = QString::number(pixVal.gray);
    else if (pixVal.isColor)
        imagePosValStr = "{" + QString::number(pixVal.r) + ", " + QString::number(pixVal.g) + ", " + QString::number(pixVal.b) + "}";

    const auto pixelValueStr = validImagePos ? imagePosStr + "=" + imagePosValStr : QString();

    if (m_acqIf->hasFrameRate())
    {
        const QString str = QString::number(m_acqIf->frameRate(), 'f', 2) + " FPS";
        m_ui.actionFrame_Rate->setText("Frame Rate: " + str);
        m_ui.actionFrame_Rate->setData(str);
    }

    if (m_acqIf->hasFramesQueued())
    {
        const QString str = QString::number(m_acqIf->framesQueued());
        m_ui.actionBuffers_Queued->setText("Frames Queued: " + str);
        m_ui.actionBuffers_Queued->setData(str + " Queued");
    }
    if (m_acqIf->hasFramesFilled())
    {
        const QString str = QString::number(m_acqIf->framesFilled());
        m_ui.actionBuffers_Filled->setText("Frames Filled: " + str);
        m_ui.actionBuffers_Filled->setData(str + " Filled");
    }

    if (m_acqIf->hasFramesDropped())
    {
        const QString str = QString::number(m_acqIf->framesDropped());
        m_ui.actionBuffers_Dropped->setText("Frames Dropped: " + str);
        m_ui.actionBuffers_Dropped->setData(str + " Dropped");
    }

    if (m_acqIf->hasFramesDelivered())
    {
        const QString str = QString::number(m_acqIf->framesDelivered());
        m_ui.actionBuffers_Delivered->setText("Frames Delivered: " + str);
        m_ui.actionBuffers_Delivered->setData(str + " Delivered");
    }

    if (m_acqIf->hasSetupMode())
    {
        const QString activeSetupMode = String::read(&AbstractAcquisitionInterface::activeSetupMode, m_acqIf.get()).c_str();
        m_ui.actionActive_Setup_Mode->setText("Playback Mode: " + activeSetupMode);
        m_ui.actionActive_Setup_Mode->setData(activeSetupMode);
    }

    m_ui.actionZoom_Level->setText("Zoom Level: " + zoomStr);
    m_ui.actionZoom_Level->setData(zoomStr);

    m_ui.actionImageDimensions->setText("Dimensions: " + dimensionsStr);
    m_ui.actionImageDimensions->setData(dimensionsStr);

    m_ui.actionImage_Format->setText("Format: " + formatStr);
    m_ui.actionImage_Format->setData(formatStr);

    m_ui.actionPixel_Value->setText("Pixel Value: " + pixelValueStr);
    m_ui.actionPixel_Value->setData(pixelValueStr);

    QStringList statsLabelStrs;
    for (auto *const action : m_imageStatsButton->menu()->actions())
    {
        if (action->isVisible() && action->isChecked())
        {
            const QString labelStr = action->data().toString();
            if (labelStr.size())
                statsLabelStrs.append(labelStr);
        }
    }

    if (statsLabelStrs.size() > 0)
    {
        m_imageStatsButton->setText(statsLabelStrs.join(" | "));
        m_imageStatsButton->setToolTip("<html>" + statsLabelStrs.join("<br/>") + "</html>");
    }
    else
    {
        m_imageStatsButton->setText("Image Info");
        m_imageStatsButton->setToolTip("Image Info");
    }
}

void ImageView::WindowQt::PrivateData::updateBitmapLockedStates (const bool bitmapLocked)
{
    // Enable/disable features that require a dynamic bitmap.
    m_ui.actionDecode_Bayer_Images->setDisabled(bitmapLocked);

    m_ui.actionBayer_Filter_Auto->setDisabled(bitmapLocked);
    m_ui.actionBayerRG_Filter->setDisabled(bitmapLocked);
    m_ui.actionBayerGR_Filter->setDisabled(bitmapLocked);
    m_ui.actionBayerGB_Filter->setDisabled(bitmapLocked);
    m_ui.actionBayerBG_Filter->setDisabled(bitmapLocked);

    m_ui.menuBayer_Filter->setDisabled(bitmapLocked);
}

void ImageView::WindowQt::PrivateData::performBusyWork (QString const& tskMsg, std::function<bool(void)> const& workFunc)
{
    flushBusyWork();

    m_busyWorkCount++;

    const auto perfFunc = [this, workFunc](void)
    {
        BusyWorkResult result;
        try
        {
            result = workFunc() ? BusyWorkSuccess : BusyWorkFailure;
        }
        catch (...)
        {
            result = BusyWorkException;
        }

        emit busyWorkComplete(result);
    };

#if defined(_WIN32)
    m_busyWorkFuture = boost::async(boost::launch::async, perfFunc);
    const bool busyWorkTimeout = boost::future_status::timeout == m_busyWorkFuture.wait_for(boost::chrono::milliseconds(1000 / 60));
#elif defined(__GNUC__)
    m_busyWorkFuture = std::async(std::launch::async, perfFunc);
    const bool busyWorkTimeout = std::future_status::timeout == m_busyWorkFuture.wait_for(std::chrono::milliseconds(1000 / 60));
#else
#   error Future support undefined for this platform.
#endif

    if (busyWorkTimeout)
    {
        updateActionStates();
        m_ui.statusBar->showMessage(tskMsg);
    }
}
void ImageView::WindowQt::PrivateData::flushBusyWork (void)
{
#if defined(_WIN32)
    while (m_busyWorkFuture.valid() && boost::future_status::timeout == m_busyWorkFuture.wait_for(boost::chrono::milliseconds(1000 / 60)))
        QApplication::processEvents();
#elif defined(__GNUC__)
    while (m_busyWorkFuture.valid() && std::future_status::timeout == m_busyWorkFuture.wait_for(std::chrono::milliseconds(1000 / 60)))
        QApplication::processEvents();
#else
#   error Future support undefined for this platform.
#endif
}
void ImageView::WindowQt::PrivateData::cleanupBusyWork (const BusyWorkResult result)
{
    if (m_busyWorkCount > 0)
        m_busyWorkCount--;

    updateActionStates();

    m_ui.statusBar->clearMessage();

    switch (result)
    {
    case BusyWorkFailure:
        m_ui.statusBar->showMessage("Operation failed!", 2000);
        break;
    case BusyWorkException:
        m_ui.statusBar->showMessage("Operation aborted!", 2000);
        break;
    }
}

void ImageView::WindowQt::PrivateData::setupAsync (void)
{
    performBusyWork("Acquisition setup ongoing", std::bind(&AbstractAcquisitionInterface::setup, m_acqIf.get()));
}
void ImageView::WindowQt::PrivateData::setupModeAsync (void)
{
    auto *const modeAction = qobject_cast<QAction*>(sender());
    if (modeAction)
    {
        const std::string modeName = modeAction->whatsThis().toLocal8Bit().data();
        performBusyWork("Acquisition setup (" + modeAction->text() + ") ongoing", [this, modeName](void)
            {
                return m_acqIf->setupMode(modeName.c_str());
            });
    }
    else
        m_ui.statusBar->showMessage("Setup mode unidentified!", 2000);
}
void ImageView::WindowQt::PrivateData::setupStartAsync (void)
{
    performBusyWork("Acquisition setup and start ongoing", [this](void)
    {
        if (m_acqIf->isSetup() || m_acqIf->setup())
            return m_acqIf->start();
        return false;
    });
}
void ImageView::WindowQt::PrivateData::triggerAsync (void)
{
    performBusyWork("Acquisition trigger ongoing", std::bind(&AbstractAcquisitionInterface::trigger, m_acqIf.get()));
}
void ImageView::WindowQt::PrivateData::startSingleReverseAsync (void)
{
    performBusyWork("Acquisition reverse start-single ongoing", std::bind(&AbstractAcquisitionInterface::startSingleReverse, m_acqIf.get()));
}
void ImageView::WindowQt::PrivateData::startReverseAsync (void)
{
    performBusyWork("Acquisition reverse start ongoing", std::bind(&AbstractAcquisitionInterface::startReverse, m_acqIf.get()));
}
void ImageView::WindowQt::PrivateData::startAsync (void)
{
    performBusyWork("Acquisition start ongoing", std::bind(&AbstractAcquisitionInterface::start, m_acqIf.get()));
}
void ImageView::WindowQt::PrivateData::startFixedAsync (const size_t cnt)
{
    performBusyWork("Acquisition start-fixed ongoing", std::bind(&AbstractAcquisitionInterface::startFixed, m_acqIf.get(), cnt));
}
void ImageView::WindowQt::PrivateData::startSingleAsync (void)
{
    performBusyWork("Acquisition start-single ongoing", std::bind(&AbstractAcquisitionInterface::startSingle, m_acqIf.get()));
}
void ImageView::WindowQt::PrivateData::pauseAsync (void)
{
    performBusyWork("Acquisition pause ongoing", std::bind(&AbstractAcquisitionInterface::pause, m_acqIf.get()));
}
void ImageView::WindowQt::PrivateData::stopAsync (void)
{
    performBusyWork("Acquisition stop ongoing", std::bind(&AbstractAcquisitionInterface::stop, m_acqIf.get()));
}
void ImageView::WindowQt::PrivateData::cleanupAsync (void)
{
    performBusyWork("Acquisition cleanup ongoing", std::bind(&AbstractAcquisitionInterface::cleanup, m_acqIf.get()));
}

void ImageView::WindowQt::PrivateData::showSurfaceContextMenu (QPoint const& pos)
{
    // Pre-load the pixel value.
    const auto pixelValueStr = m_ui.actionPixel_Value->data().toString();
    m_ui.actionCopy_Pixel_Value->setData(pixelValueStr);
    m_ui.actionCopy_Pixel_Value->setEnabled(pixelValueStr.length());

    // Pre-load the image stats text.
    m_ui.actionCopy_Image_Stats->setData(imageStatsText());

    // Setup the context menu.
    QMenu contextMenu;
    contextMenu.addAction(m_ui.actionSave_Image);
    contextMenu.addAction(m_ui.actionCopy_Image);
    contextMenu.addSeparator();
    contextMenu.addAction(m_ui.actionCopy_Image_Stats);
    contextMenu.addSeparator();
    contextMenu.addAction(m_ui.actionCopy_Pixel_Value);

    // Execute the menu.
    contextMenu.exec(m_surfaceStream->surface()->mapToGlobal(pos));

    // Clear the pre-loaded data.
    m_ui.actionCopy_Pixel_Value->setData(QVariant());
    m_ui.actionCopy_Pixel_Value->setEnabled(true);

    m_ui.actionCopy_Image_Stats->setData(QVariant());
}

void ImageView::WindowQt::PrivateData::zoomEditorOpen (void)
{
    if (m_ui.acquisitionToolBar->isVisible())
        m_zoomEditorPopover->setAffinityWidget(m_ui.acquisitionToolBar);
    else
        m_zoomEditorPopover->setAffinityWidget(m_ui.menuBar);

    const bool editorWasVisible = m_zoomEditorPopover->isVisible();
    if (!editorWasVisible)
        m_zoomEditorPopover->open();

    m_zoomEditorSpinBox->selectAll();
    if (!editorWasVisible)
        m_zoomEditorSpinBox->setFocus();

    zoomEditorUpdate();
}
void ImageView::WindowQt::PrivateData::zoomEditorUpdate (void)
{
    const auto *const surface = m_surfaceStream->surface();

    const double min = surface->scalingMin();
    const double max = surface->scalingMax();
    m_zoomEditorSliderRatio = log(max) / std::max(1, m_zoomEditorSlider->width());

    m_zoomEditorSpinBox->setRange(100. * min, 100. * max);
    m_zoomEditorSlider->setRange(log(min) / m_zoomEditorSliderRatio, log(max) / m_zoomEditorSliderRatio);

    const double scaling = surface->scaling();

    const bool spinBoxSignalsBlocked = m_zoomEditorSpinBox->blockSignals(true);
    m_zoomEditorSpinBox->setValue(100. * scaling);
    m_zoomEditorSpinBox->blockSignals(spinBoxSignalsBlocked);

    const bool sliderSignalsBlocked = m_zoomEditorSlider->blockSignals(true);
    m_zoomEditorSlider->setValue(log(scaling) / m_zoomEditorSliderRatio);
    m_zoomEditorSlider->blockSignals(sliderSignalsBlocked);

    m_ui.actionZoom_Editor->setChecked(m_zoomEditorPopover->isVisible());
}
void ImageView::WindowQt::PrivateData::zoomSpinBoxChanged (void)
{
    const double newScaling = m_zoomEditorSpinBox->value() / 100.;
    const double oldScaling = m_imageArea->zoomScaling();

    const double practicalEpsilon = pow(10., -m_zoomEditorSpinBox->decimals()) / 100. / 2.;

    if (std::fabs(newScaling - oldScaling) > practicalEpsilon)
        m_imageArea->setZoomTo(newScaling);
}
void ImageView::WindowQt::PrivateData::zoomSliderChanged (void)
{
    const int zoomValue = m_zoomEditorSlider->value();

    if (m_zoomEditorSlider->minimum() == zoomValue)
        m_imageArea->setZoomTo(Surface::scalingMin());
    else if (m_zoomEditorSlider->maximum() == zoomValue)
        m_imageArea->setZoomTo(Surface::scalingMax());
    else
    {
        const auto newScaling = exp(zoomValue * m_zoomEditorSliderRatio);
        m_imageArea->setZoomTo(newScaling);
    }
}
void ImageView::WindowQt::PrivateData::zoomEditorClosed (void)
{
    m_ui.actionZoom_Editor->setChecked(false);
    m_fullScreenBarHiderTimer->start();
}

void ImageView::WindowQt::PrivateData::startFixedSelect (void)
{
    if (m_acqIf->canStartFixed())
    {
        m_startFixedSpinBox->selectAll();

        if (m_ui.acquisitionToolBar->isVisible())
            m_startFixedPopover->setAffinityWidget(m_ui.acquisitionToolBar);
        else
            m_startFixedPopover->setAffinityWidget(m_ui.menuBar);

        m_startFixedPopover->open();
    }
    else
        m_startFixedPopover->close();
}
void ImageView::WindowQt::PrivateData::startFixedAccept (void)
{
    startFixedAsync(m_startFixedSpinBox->value());
    startFixedClose();
}
void ImageView::WindowQt::PrivateData::startFixedClose (void)
{
    if (m_startFixedPopover->isVisible())
        m_startFixedPopover->reject();

    m_ui.actionStart_Fixed->setChecked(false);
    m_fullScreenBarHiderTimer->start();
}
void ImageView::WindowQt::PrivateData::startFixedSet (const int val)
{
    const std::int32_t min = m_startFixedSpinBox->minimum();
    const std::int32_t inc = m_startFixedSpinBox->singleStep();

    if ((val - min) % inc)
    {
        m_startFixedSpinBox->setValue(min + ((val - min) / inc) * inc);
        m_startFixedSpinBox->selectAll();
    }
}

void ImageView::WindowQt::PrivateData::quickViewPopoverOpen (void)
{
    const auto *const surface = m_surfaceStream->surface();

    m_quickViewMiniMap->setPixmap(surface->pixmap());
    m_quickViewMiniMap->setScaling(surface->scaling());
    m_quickViewMiniMap->setViewRect(surface->visibleImageRectF());

    m_quickViewPopover->open();

    m_ui.actionQuick_View->setChecked(true);
}
void ImageView::WindowQt::PrivateData::quickViewPopoverClose (void)
{
    if (m_quickViewPopover->isVisible())
        m_quickViewPopover->reject();

    m_ui.actionQuick_View->setChecked(false);
}
void ImageView::WindowQt::PrivateData::quickViewPopoverUpdatePixmap (void)
{
    if (m_quickViewPopover->isVisible())
    {
        const auto pixmap = m_surfaceStream->peekPixmap();
        if (pixmap.isNull())
            quickViewPopoverClose();
        else
            m_quickViewMiniMap->setPixmap(pixmap);
    }
}
void ImageView::WindowQt::PrivateData::quickViewPopoverUpdateView (void)
{
    if (m_quickViewPopover->isVisible())
    {
        const auto *const surface = m_surfaceStream->surface();
        m_quickViewMiniMap->setScaling(surface->scaling());
        m_quickViewMiniMap->setViewRect(surface->visibleImageRectF());
    }
}

void ImageView::WindowQt::PrivateData::hideWindowBars (void)
{
    const bool isFullScreen = m_window.isFullScreen();
    m_ui.statusBar->setMouseTracking(isFullScreen);

    if (isFullScreen)
    {
        const bool keepStatsVisible = m_ui.actionKeep_Stats_Visible->isChecked();

        if (sender() == m_fullScreenBarHiderTimer)
        {
            for (auto *const childMenu : m_ui.menuBar->findChildren<QMenu*>())
            {
                if (childMenu->isVisible())
                    return;
            }

            if (m_setupModeMenu->isVisible())
                return;

            if (m_colorPaletteButton->menu()->isVisible())
                return;

            if (m_ui.menuBar->underMouse() || m_ui.acquisitionToolBar->underMouse())
                return;

            if (m_startFixedPopover->isVisible() || m_zoomEditorPopover->isVisible())
                return;

            if (!keepStatsVisible)
            {
                if (m_imageStatsButton->menu()->isVisible() || m_ui.statusBar->underMouse())
                    return;
            }
        }

        // Determine if the statusbar should be open, and make certain that it is.
        const int statusBarAnimationIndex = m_fullScreenBarHiderAnimationGroup->indexOfAnimation(m_fullScreenBarHiderAnimation_statusBar);
        if (keepStatsVisible)
        {
            if (statusBarAnimationIndex >= 0)
                m_fullScreenBarHiderAnimationGroup->removeAnimation(m_fullScreenBarHiderAnimation_statusBar);

            if (m_fullScreenBarHiderAnimation_statusBar->direction() != QAbstractAnimation::Backward)
            {
                m_fullScreenBarHiderAnimation_statusBar->setDirection(QAbstractAnimation::Backward);
                m_fullScreenBarHiderAnimation_statusBar->start();
            }
        }
        else if (statusBarAnimationIndex < 0)
            m_fullScreenBarHiderAnimationGroup->addAnimation(m_fullScreenBarHiderAnimation_statusBar);

        // Run the hide animation.
        if (m_fullScreenBarHiderAnimationGroup->direction() != QAbstractAnimation::Forward)
        {
            m_fullScreenBarHiderAnimationGroup->setDirection(QAbstractAnimation::Forward);
            m_fullScreenBarHiderAnimationGroup->start();
        }
    }
    else
        showWindowBars();
}
void ImageView::WindowQt::PrivateData::showWindowBars (void)
{
    if (m_fullScreenBarHiderAnimationGroup->direction() != QAbstractAnimation::Backward)
    {
        m_fullScreenBarHiderAnimationGroup->setDirection(QAbstractAnimation::Backward);
        m_fullScreenBarHiderAnimationGroup->start();
    }
}

// Filter events from the window.
bool ImageView::WindowQt::PrivateData::eventFilter (QObject *const watched, QEvent *const evt)
{
    switch (evt->type())
    {
    case QEvent::Enter:
        if (m_window.isFullScreen())
        {
            if (!m_ui.actionKeep_Stats_Visible->isChecked() || watched != m_ui.statusBar)
                showWindowBars();
        }
        break;

    case QEvent::Leave:
        if (m_window.isFullScreen())
            m_fullScreenBarHiderTimer->start();
        break;
    }
    return false;
}

void ImageView::WindowQt::PrivateData::showAboutQt (void)
{
    QMessageBox::aboutQt(&m_window, "About Qt");
}

void ImageView::WindowQt::PrivateData::userMenuActionDestroyed (void)
{
    if (auto *const sndr = sender())
    {
        if (auto *const menu = qobject_cast<QMenu*>(sndr->parent()))
        {
            const auto menuActions = menu->actions();
            if (menuActions.isEmpty() || (menuActions.size() == 1 && menuActions[0] == sndr))
                delete menu;
        }
    }
}

void ImageView::WindowQt::PrivateData::userToolBarActionDestroyed (void)
{
    m_userToolBarActions.removeAll(static_cast<QAction*>(sender()));
}
