#ifndef INCLUDED__IMAGE__VIEW__WINDOW__QT_____PRIVATE__H
#define INCLUDED__IMAGE__VIEW__WINDOW__QT_____PRIVATE__H

/* FILE:        ImageViewWindowQt_Private.h
 * DATE:        1/7/2016
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2016, BitFlow, Inc.
 * DESCRIPTION: PrivateData for ImageView::WindowQt.
 */

#include "ImageViewWindowQt.h"
#include "ImageViewSurface.h"
#include "ImageViewSurfaceStream.h"
#include "ImageViewSurfaceScrollArea.h"

#include "ImageViewPopover.h"
#include "ImageViewMiniMap.h"

#include "AbstractAcquisitionInterface.h"
#include "AcquisitionSignalHandler.h"

#include "AbstractFileInterface.h"

#include "ImageViewThread.h"

#include <atomic>
#include <mutex>
#include <condition_variable>

#include <cstdint>
#include <chrono>
#include <random>

#if defined(_WIN32)
#   include <boost/thread/future.hpp>
#   include <boost/chrono.hpp>
#elif defined(__GNUC__)
#   include <future>
#   include <chrono>
#else
#   error Future support undefined for this platform.
#endif

#include "ui_ImageViewWindow.h"
#include "ui_AboutImageView.h"

#include <QTimer>

#include <QCloseEvent>
#include <QResizeEvent>
#include <QWheelEvent>

#include <QImage>

#include <QClipboard>
#include <QFileDialog>
#include <QMessageBox>

#include <QMenu>
#include <QToolButton>
#include <QScrollBar>
#include <QSlider>
#include <QSpinBox>

#include <QParallelAnimationGroup>
#include <QPropertyAnimation>

Q_ENUMS(BusyWorkResult);

enum BusyWorkResult
{
    BusyWorkSuccess = 0,
    BusyWorkFailure,
    BusyWorkException
};

struct ImageView::WindowQt::PrivateData : public QObject
{
    Q_OBJECT

private:
    static bool register_meta_types (void);
    static const bool meta_types_registered;

signals:
    void busyWorkComplete (BusyWorkResult result);

private:
    // Illegal.
    PrivateData (void);
    PrivateData (PrivateData const&);
    PrivateData& operator = (PrivateData const&);

    // Helpers.
    void setupAboutImageView (void);
    void setupAboutQt (void);

    QPropertyAnimation* addAnimatedWidgetHide (QParallelAnimationGroup *const group, QWidget *const toHide) const;

    void setupColorPaletteMenu (void);
    void setupImageStatsMenu (void);
    void setupZoomEditorPopover (void);
    void setupStartFixedPopover (void);
    void setupQuickPanPopover (void);

public:
    // Constructor
    PrivateData (WindowQt &imageView, ScopedAcquisitionInterface &&acqIf, ScopedFileInterface &&fileIf = ScopedFileInterface());

    // Singleton initializer (call immediately prior to construction).
    void initialize (void);

    // Destructor.
    ~PrivateData (void);

    // Helper methods.
    void applyColorTable (QObject *const sender);
    void applyBayerFilter (QObject *const sender);
    static void generateRainbowColorTable (QVector<uint> &table, const double freq, const double rPhase, const double gPhase, const double bPhase, const double width, const double center);

    QString imageStatsText (void) const;

    void performBusyWork (QString const& tskMsg, std::function<bool(void)> const& workFunc);
    void flushBusyWork (void);

    // Members.
    WindowQt &m_window;
    QThread *m_viewerThread;

    Ui::imageDisplayWindow m_ui;
    Ui::aboutImageViewDialog m_aboutDialog;

    QDialog *m_appAboutDialog;

    const ScopedAcquisitionInterface m_acqIf;
    const ScopedFileInterface m_fileIf;

    SurfaceScrollArea *m_imageArea;
    SurfaceStream *m_surfaceStream;

    QMenu *m_colorPaletteMenu;
    QMenu *m_imageStatsMenu;
    QAction *m_imageStatsMenuUserActionSeparator;

    ImageView::PopoverWindow *m_zoomEditorPopover;
    QDoubleSpinBox *m_zoomEditorSpinBox;
    QSlider *m_zoomEditorSlider;
    double m_zoomEditorSliderRatio;

    ImageView::PopoverWindow *m_startFixedPopover;
    QSpinBox *m_startFixedSpinBox;

    ImageView::MiniMap *m_quickViewMiniMap;
    ImageView::PopoverWindow *m_quickViewPopover;
    QToolButton *m_quickViewPopoverButton;

    QToolButton *m_setupCleanupButton;
    QMenu *m_setupModeMenu;

    QToolButton *m_colorPaletteButton;
    QAction *m_colorPaletteButtonAction;
    QVector<QRgb> m_colorTable;
    ColorTable m_colorTableCurrent;

    QToolButton *m_imageStatsButton;
    QTimer *m_imageStatsTimer;

    QToolButton *m_fullScreenButton;

    QList<QAction*> m_userToolBarActions;

    QString m_prevSavePath;
    QString m_prevSaveFileFormat;

    QTimer *m_fullScreenBarHiderTimer;
    QParallelAnimationGroup *m_fullScreenBarHiderAnimationGroup;
    QPropertyAnimation *m_fullScreenBarHiderAnimation_menuBar;
    QPropertyAnimation *m_fullScreenBarHiderAnimation_acquisitionToolBar;
    QPropertyAnimation *m_fullScreenBarHiderAnimation_statusBar;

    Qt::WindowStates m_nonFullScreenWindowState;

#if defined(_WIN32)
    boost::unique_future<void> m_busyWorkFuture;
#elif defined(__GNUC__)
    std::future<void> m_busyWorkFuture;
#else
#   error Future support undefined for this platform.
#endif

    size_t m_busyWorkCount;

    // Filter events from the window.
    bool eventFilter (QObject *const watched, QEvent *const evt) override;

public slots:
    void updateActionStates (void);
    void updateZoomStates (void);
    void updateImageStats (void);

    void updateBitmapLockedStates (const bool bitmapLocked);

    void cleanupBusyWork (const BusyWorkResult result);

    void setupAsync (void);
    void setupModeAsync (void);
    void setupStartAsync (void);

    void triggerAsync (void);

    void startSingleReverseAsync (void);
    void startReverseAsync (void);
    void startAsync (void);
    void startFixedAsync (const size_t cnt);
    void startSingleAsync (void);

    void pauseAsync (void);
    void stopAsync (void);
    void cleanupAsync (void);

    void showSurfaceContextMenu (QPoint const& pos);

    void zoomEditorOpen (void);
    void zoomEditorUpdate (void);
    void zoomSpinBoxChanged (void);
    void zoomSliderChanged (void);
    void zoomEditorClosed (void);

    void startFixedSelect (void);
    void startFixedAccept (void);
    void startFixedClose (void);
    void startFixedSet (const int val);

    void quickViewPopoverOpen (void);
    void quickViewPopoverClose (void);
    void quickViewPopoverUpdatePixmap (void);
    void quickViewPopoverUpdateView (void);

    void hideWindowBars (void);
    void showWindowBars (void);

    void showAboutQt (void);

    void userMenuActionDestroyed (void);
    void userToolBarActionDestroyed (void);
};

#endif // INCLUDED__IMAGE__VIEW__WINDOW__QT_____PRIVATE__H
