/* FILE:        ImperativeAcquisitionInterface.cpp
 * DATE:        12/10/2015
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2015, BitFlow, Inc.
 * DESCRIPTION: Implementation of the ImageView AbstractAcquisitionInterface
 *              providing an imperative method to set frames for display. Simply
 *              call setNextFrame to display an image. By default,
 */

#include "ImperativeAcquisitionInterface_Private.h"

#include <memory>

ImageView::ImperativeAcquisitionInterface::ImperativeAcquisitionInterface (void)
    : m_pd (*new PrivateData)
{ }

ImageView::ImperativeAcquisitionInterface::~ImperativeAcquisitionInterface (void)
{
    delete &m_pd;
}

//// Imperative acquisition methods and properties.

// Set the next image frame to display.
void ImageView::ImperativeAcquisitionInterface::setNextFrame (RawImage const& rawImage)
{
    m_pd.setNextFrame(rawImage);
}

// Convenience method to setNextFrame frome a frame's constituent parts.
void ImageView::ImperativeAcquisitionInterface::setNextFrame (const std::uint8_t *const buffer, const size_t width, const size_t height, const Format format, void *const userData)
{
    const RawImage rawImage =
    {
        buffer,
        width,
        height,
        format,
        userData
    };

    setNextFrame(rawImage);
}

// Clear any pending next frame.
void ImageView::ImperativeAcquisitionInterface::clearNextFrame (void)
{
    m_pd.clearNextFrame();
}

// Whether or not the frames set to setNextFrame are implicitly locked.
// If an implicitly locked frame is in use, this first waits for it to
// be unlocked before applying the new value.
void ImageView::ImperativeAcquisitionInterface::setFramesLocked (const bool locked)
{
    m_pd.setFramesLocked(locked);
}

bool ImageView::ImperativeAcquisitionInterface::framesLocked (void) const
{
    return m_pd.framesLocked();
}

//// Non-imperative local helpers.

// Method called by the Imperative getNextFrame method before attempting to return
// the next frame, allowing clients to modify states declaratively. The default
// implementation does nothing.
void ImageView::ImperativeAcquisitionInterface::readyNextFrame (void)
{
    // NoOp.
}

//// Acquisition interface feature implementation.

// Return the last image set by set next frame. After a successful return, there is next frame.
bool ImageView::ImperativeAcquisitionInterface::getNextFrame (RawImage *const rawImage, ScopedLock *const imageLock)
{
    // Let the client prepare, if necessary.
    readyNextFrame();

    // Get the next frame, if any.
    const auto nextFrame = m_pd.takeNextFrame();
    if (!nextFrame)
        return false;

    // If necessary, return the image lock.
    if (imageLock)
    {
        // The frame lock may come locally, or be implemented by a sub-class.
        auto frameLock = getFrameLock(*nextFrame);
        if (!frameLock)
            return false;

        *imageLock = std::move(frameLock);
    }

    // Return the next frame.
    memcpy(rawImage, nextFrame.get(), sizeof(RawImage));

    return true;
}

// Return a scoped NoOpImageLock.
ImageView::ScopedLock ImageView::ImperativeAcquisitionInterface::getFrameLock (RawImage const& rawImage)
{
    return PrivateData::FramesLockedImageLock(m_pd).toScoped();
}
