#ifndef INCLUDED_BIT_FLOW__IMPERATIVE__ACQUISITION__INTERFACE__H
#define INCLUDED_BIT_FLOW__IMPERATIVE__ACQUISITION__INTERFACE__H

/* FILE:        ImperativeAcquisitionInterface.h
 * DATE:        12/10/2015
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2015, BitFlow, Inc.
 * DESCRIPTION: Abstract implementation of the ImageView AbstractAcquisitionInterface
 *              providing an imperative method to set frames for display. Simply
 *              call setNextFrame to display an image. isSetup and isRunning must
 *              still be implemented.
 */

#include "ImageView.h"
#include "AbstractAcquisitionInterface.h"

namespace ImageView
{
    class IMAGEVIEW_EXPORT ImperativeAcquisitionInterface : public AbstractAcquisitionInterface
    {
    private:
        struct PrivateData;
        PrivateData &m_pd;

        // Illegal.
        ImperativeAcquisitionInterface (ImperativeAcquisitionInterface const&);
        ImperativeAcquisitionInterface& operator= (ImperativeAcquisitionInterface const&);

    public:
        ImperativeAcquisitionInterface (void);
        virtual ~ImperativeAcquisitionInterface (void);

        //// Imperative acquisition methods and properties.

        // Set the next image frame to display.
        void setNextFrame (RawImage const& rawImage);

        // Convenience method to setNextFrame frome a frame's constituent parts.
        void setNextFrame (const std::uint8_t *const buffer, const size_t width, const size_t height, const Format format, void *const userData = nullptr);

        // Clear any pending next frame.
        void clearNextFrame (void);

        // Whether or not the frames set to setNextFrame are implicitly locked.
        // If an implicitly locked frame is in use, this first waits for it to
        // be unlocked before applying the new value.
        void setFramesLocked (const bool locked);
        bool framesLocked (void) const;

        //// Non-imperative local helpers.

        // Method called by the Imperative getNextFrame method before attempting to return
        // the next frame, allowing clients to modify states declaratively. The default
        // implementation does nothing.
        virtual void readyNextFrame (void);

        //// Acquisition interface feature implementation.

        // Return the last image set by set next frame. An imageLock is only returned if
        // framesLocked is true, or if an inheriting class getFrameLock override returns
        // a valid ScopedLock.
        bool getNextFrame (RawImage *const rawImage, ScopedLock *const imageLock) final;

        // Return a scoped NoOpImageLock.
        ScopedLock getFrameLock (RawImage const& rawImage) override;
    };
}

#endif // INCLUDED_BIT_FLOW__IMPERATIVE__ACQUISITION__INTERFACE__H
