#ifndef INCLUDED__BIT_FLOW__IMPERATIVE__ACQUISITION__INTERFACE_____PRIVATE__H
#define INCLUDED__BIT_FLOW__IMPERATIVE__ACQUISITION__INTERFACE_____PRIVATE__H

/* FILE:        ImperativeAcquisitionInterface_Private.h
 * DATE:        12/10/2015
 * AUTHOR:      Jeremy Greene
 * COMPANY:     BitFlow, Inc.
 * COPYRIGHT:   Copyright (C) 2015, BitFlow, Inc.
 * DESCRIPTION: Private data definition for the ImperativeAcquisitionInterface class.
 */

#include "ImperativeAcquisitionInterface.h"

#include <memory>
#include <mutex>
#include <atomic>

struct ImageView::ImperativeAcquisitionInterface::PrivateData
{
private:
    // Members.
    std::atomic<RawImage*> m_nextFrame;
    mutable std::mutex m_framesLockedMutex;
    bool m_framesLocked;

public:
    PrivateData (void)
        : m_framesLocked (false)
        , m_nextFrame (nullptr)
    {
    }

    ~PrivateData (void)
    {
        clearNextFrame();
    }

    // Frame helpers.
    inline void setNextFrame (ImageView::RawImage const& a_image)
    {
        delete m_nextFrame.exchange(new ImageView::RawImage(a_image));
    }

    inline std::unique_ptr<ImageView::RawImage> takeNextFrame (void)
    {
        return std::unique_ptr<ImageView::RawImage>(m_nextFrame.exchange(nullptr));
    }

    inline void clearNextFrame (void)
    {
        delete m_nextFrame.exchange(nullptr);
    }

    inline bool framesLocked (void) const
    {
        const std::unique_lock<std::mutex> framesLockedLock (m_framesLockedMutex);
        return m_framesLocked;
    }

    inline void setFramesLocked (const bool a_framesLocked)
    {
        const std::unique_lock<std::mutex> framesLockedLock (m_framesLockedMutex);
        m_framesLocked = a_framesLocked;
    }

    // AbstractLock implementation.
    class FramesLockedImageLock : public AbstractLock
    {
    private:
        std::unique_lock<std::mutex> m_framesLockedLock;

        // Illegal.
        FramesLockedImageLock (void);
        FramesLockedImageLock (FramesLockedImageLock const&);
        FramesLockedImageLock& operator= (FramesLockedImageLock const&);

    public:
        FramesLockedImageLock (PrivateData const& toLock)
        {
            std::unique_lock<std::mutex> lock (toLock.m_framesLockedMutex);
            if (toLock.m_framesLocked)
                m_framesLockedLock = std::move(lock);
        }

        FramesLockedImageLock (FramesLockedImageLock &&to_take)
            : m_framesLockedLock (std::move(to_take.m_framesLockedLock))
        { }

        inline bool locked (void) const
        {
            return m_framesLockedLock.owns_lock();
        }

        inline ScopedLock toScoped (void)
        {
            return locked() ? ScopedLock(new FramesLockedImageLock(std::move(*this))) : ScopedLock();
        }
    };
};

#endif // INCLUDED__BIT_FLOW__IMPERATIVE__ACQUISITION__INTERFACE_____PRIVATE__H
